/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1842 $
 * $Date: 2006-04-18 15:34:13 -0700 (Tue, 18 Apr 2006) $
 * $Author: hedstrom $
 * $Id: ENDF_file.cpp 1842 2006-04-18 22:34:13Z hedstrom $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

#include "ENDF_file.hpp"
#include "record_types.hpp"

// the length of an ENDF line
const int Line_len = 80;

// ------------------- skip ---------------
void ENDF_file::skip(int N)
// Skip the next N entries in the file
{
  char linje[Line_len];
  string linebuff;
  string strbuff;
  double dummy;

  for( int iN=0; iN < N; iN++ ) //Loop over N numbers
  {
    int iC=iN % 6;  //There are 6 items per line
    if ( iC == 0 )
    {
      getline( linje, Line_len-1 ); //read in new line
      linebuff = linje;
    }
    strbuff = linebuff.substr(iC*11, 11); //Grab the right entry
    read_d(&strbuff, &dummy );
  }
}
// ------------------- mfx_line1 ---------------
void ENDF_file::mfx_line1(int *ZA, double *AWR, int *I1, int *I2,
  int *I3, int *I4)
// read the first header line for an mf? file
{
  char linje[Line_len];

  getline(linje, Line_len-1 ); //read first line
  string linebuff(linje); // the input string

  read_idiiii(&linebuff, 
	      ZA,              // target Z & A 
	      AWR,             // target mass
              I1, I2, I3, I4); // depends on the mf number
}
// ------------------- mfx_line2 ---------------
void ENDF_file::mfx_line2(double *D1, double *D2, int *I1, int *I2,
  int *I3, int *I4)
// read the second header line for an mf? file
{
  char linje[Line_len];

  getline(linje, Line_len-1 ); //read first line
  string linebuff(linje); // the input string

  read_ddiiii(&linebuff,
	      D1, D2,            // depends on the mf number
              I1, I2, I3, I4);   // depends on the mf number
}
// ------------------- get_regions ---------------
void ENDF_file::get_regions(int NR, vector<int>& NBT, vector<int>& INT)
// read the interpolation information
{
  char linje[Line_len];

  string linebuff; // the input string
  string strbuff; // the current data

  int NBT_value, INT_value;

  // read the data
  for ( int iNR = 0; iNR < NR; iNR++ ) //Loop over NR NBT,INT pairs
  {
    int iR = iNR % 3;  //There are 3 pairs per line
    if ( iR == 0 )
    {
      getline(linje, Line_len-1); //read the line
      linebuff = linje;
    }
    strbuff = linebuff.substr(iR*22, 22); //Grab the right pair
    read_ii( &strbuff, &NBT_value, &INT_value );
    NBT.push_back( NBT_value );
    INT.push_back( INT_value );
  }
}

// ------------------- mf1_file::read_line1 ---------------
void mf1_file::read_line1(int *ZA, double *AWR, int *LNU)
// read the first header line for an mf1 file
{
  int iDum;

  mfx_line1(ZA,             // target Z & A 
	    AWR,            // target mass
	    &iDum,
            LNU,            // flag for spontaneous fission
	    &iDum, &iDum);  // zeros
}
// ------------------- mf1_file::delay_head ---------------
void mf1_file::delay_head(int *NC)
// read the header line for time constants in an mf1 file
{
  int iDum;
  double zero;

  mfx_line2(&zero,
            &zero,
            &iDum, &iDum,   // zero
            NC,             // number of INT regions 
            &iDum);         // zero
}
// ------------------- mf1_file::mult_head ---------------
void mf1_file::mult_head(int *NR, int *NP)
// read the header line for multiplicities in an mf1 file
{
  int iDum;
  double zero;

  mfx_line2(&zero,
            &zero,
            &iDum, &iDum,   // not used
            NR,             // number of INT regions 
            NP);            // number of data points
}
// ------------------- mf3_file::read_line1 ---------------
void mf3_file::read_line1(int *ZA, double *AWR)
// read the first header line for an mf3 file
{
  int iDum;

  mfx_line1(ZA,             // target Z & A 
	    AWR,            // target mass
	    &iDum, &iDum,       // zeros
	    &iDum, &iDum);      // zeros
}
// ------------------- mf3_file::read_line2 ---------------
void mf3_file::read_line2(double *QM, double *QI, int *LR, int *NR, int *NP)
// read the second header line for an mf3 file
{
  int iDum;

  mfx_line2(QM,             // mass difference Q value
            QI,             // reaction Q value for the lowest state
            &iDum,          // zero
            LR,             // complex or "breakup" flag
            NR,             // number of INT regions 
            NP);            // number of data points
}
// ------------------- mf4_file::read_line1 ---------------
void mf4_file::read_line1(int *ZA, double *AWR, int *LVT, int *LTT)
// read the first header line for an mf4 file
{
  int iDum;

  mfx_line1(ZA,             // target Z & A 
	    AWR,            // target mass
	    LVT,            // tranfsformation matrix flag
	    LTT,            // representation flag
        &iDum, &iDum);  // zeros
}
// ------------------- mf4_file::read_line2 ---------------
void mf4_file::read_line2(int *LI, int *LCT, int *NK, int *NM)
// read the second header line for an mf4 file
{
  double ZERO;
  double AWR;

  mfx_line2(&ZERO,
            &AWR,           // target mass
            LI,             // flag: isotropic or not
            LCT,            // 1 -> lab frame, 2 -> CM system
            NK,             // number of elements in transformation matrix 
            NM );           // max order of Legendre (should be even)
}
// ------------------- mf4_file::read_line3 ---------------
void mf4_file::read_line3(int *NR, int *NE)
// read the third header line for an mf4 file
{
  double ZERO;
  int iDum;

  mfx_line2(&ZERO,            // zero
	    &ZERO,            // zero
	    &iDum,            // zero
	    &iDum,            // zero
	    NR,               // number of interpolation regions
	    NE );             // number of incident energy points
}
// ------------------- mf5_file::read_line1 ---------------
void mf5_file::read_line1(int *ZA, double *AWR, int *NK)
// read the first header line for an mf5 file
{
  int iDum;

  mfx_line1(ZA,             // target Z & A 
	    AWR,            // target mass
	    &iDum, &iDum,   // zeros
            NK,             // the number of partial regions
            &iDum);         // zero
}
// ------------------- mf5_file::read_line2 ---------------
void mf5_file::read_line2(double *U, int *LF, int *NR, int *NP)
// read the second header line for an mf5 file
{
  double ZERO;
  int iZERO;

  mfx_line2(U,               // upper energy limit of secondary
            &ZERO,           // zero
            &iZERO,          // zero
            LF,              // Energy law flag
            NR,              // number of interpolation regions
            NP );            // number of incident energy points
}
// ------------------- mf5_file::read_Madland ---------------
void mf5_file::read_Madland(double *EFL, double *EFH, int *NR, int *NE)
// read a Madland-Nix header line
{
  int iZERO;

  mfx_line2(EFL,             // average energy of light fission fragment
            EFH,             // average energy of heavy fission fragment
            &iZERO,          // zero
            &iZERO,          // Energy law flag
            NR,              // number of interpolation regions
            NE );            // number of incident energy points
}
// ------------------- mf6_file::read_line1 ---------------
void mf6_file::read_line1(int *ZA, double *AWR, int *LCT, int *NK)
// read the first header line for an mf6 file
{
  int iDum;

  mfx_line1(ZA,             // target Z & A 
	    AWR,            // target mass
	    &iDum,          // zero
            LCT,            // data reference frame
            NK,             // the number of subsections
            &iDum);         // zero
}
// ------------------- mf6_file::particle_line1 ---------------
void mf6_file::particle_line1(int *ZAP, double *AWP, int *LIP,
  int *LAW, int *NR, int *NP)
// read the header line for a particle in an mf6 file
{
  mfx_line1(ZAP,  // product Z & A
	    AWP,             // product mass
	    LIP,             // residual isomeric state
	    LAW,             // kinetics model
	    NR,              // number of interpolation regions
	    NP);             // number of multiplicity points
}
// ------------------- mf6_file::mf6_get_model ---------------
void mf6_file::get_model(int *LANG, int *LEP, int *NR, int *NE)
// get the model for a particle in an mf6 file
{
  double ZERO;

  mfx_line2(&ZERO, &ZERO, // 2 zeros
	    LANG,        // the model
	    LEP,         // interpolation for E_out
	    NR,          // number of regions
	    NE);         // number of incident energies
}
// ------------------- mf6_file::gamma_ein ---------------
void mf6_file::gamma_ein( double *e_in, int *ND, int *NEP )
// get the pertinant information for an mf6 gamma file
{
  double ZERO;
  int dummy;
  int NA;

  mfx_line2( &ZERO,      // zero
             e_in,       // energy of incident neutron
	     ND,         // number of discrete lines
	     &NA,        // 0
	     &dummy,     // number of regions
	     NEP );      // number of outgoing energies
  if( NA != 0 )
  {
    Unimplemented("mf6_file::gamma_ein","Implement anisotropic gammas.");
  }
}
// ------------------- mf6_file::phase_space ---------------
void mf6_file::phase_space( double *APSX, int *NPSX )
// get the total mass and number of particles for a phase-space file
{
  double Zero;
  int zero;
  mfx_line2(APSX,        // total mass
	    &Zero,       // zero
	    &zero, &zero, &zero, // 3 zeros
	    NPSX);       // number of particles
}
// ------------------- mf6_file::mf4_line1 ---------------
void mf6_file::mf4_line1(int *NR, int *NE)
// get the number of incident energies for a gen_mf4 file
{
  double ZERO;
  int zero;

  mfx_line2(&ZERO, &ZERO, // 2 zeros
	    &zero, &zero, // more zeros
	    NR,          // number of regions
	    NE);         // number of incident energies
}
// ------------------- mf6_file::mf4_line2 ---------------
void mf6_file::mf4_line2( double *e_in, int *LANG, int *NW, int *NL )
// get the model for a gen_mf4 incident energy
{
  double ZERO;
  int zero;

  mfx_line2(&ZERO,    //  zero
            e_in,     // incident energy
            LANG,     // the model
	    &zero,    // zero
	    NW,       // amount of data
	    NL);      // Legendre order or number of cosines
}
// ------------------- mf6_file::ENDL ---------------
void mf6_file::ENDL(int *NR, int *NE)
// get the number of incident energies for an ENDL_table
{
  mf4_line1( NR,         // number of interpolation regions
	     NE );       // number of incident energies
}
// ------------------- mf6_file::ENDL_head ---------------
void mf6_file::ENDL_head( double *e_in, int *NR, int *NE )
// get the incident energy or cosine for an ENDL_table
{
  double ZERO;
  int zero;

  mfx_line2(&ZERO,        // zero
            e_in,         // incident energy or cosine
	    &zero, &zero, // more zeros
	    NR,          // number of regions
	    NE);         // number of incident energies
}
// ------------------- mf6_file::Legendre_E_in ---------------
void mf6_file::Legendre_E_in( double *e_in, int *ND, int *order, int *NE )
// Gets the incident energy and the order for a Legendre table
{
  double ZERO;
  int dummy;

  mfx_line2(&ZERO,       // zero
            e_in,        // incident energy or cosine
	    ND,          // number of discrete lines
            order,       // Legendre order
	    &dummy,      // NE*(order + 2)
	    NE);         // number of incident energies
  if( dummy != (*NE)*(*order + 2) )
  {
    Warning( "mf6_file::Legendre_E_in", 
	     pastenum("order inconsistent for e_in: ", *e_in ) );
    *order = dummy/(*NE) - 2;
  }
}
// ------------------- mf6_file::read_spin ---------------
void mf6_file::read_spin( double *spin, int *NR, int *NE )
// Gets the spin, number of interpolation regions and number of energies
{
  int zero, dummy;
  double Zero;

  mfx_line2(spin,        // spin of incident particle
            &Zero,       // zero
	    &dummy,      // 1 if particle the same, otherwise 0
            &zero,       // zero
	    NR,          // number of interpolation regions
	    NE);         // number of incident energies
}
// ------------------- mf6_file::read_model ---------------
void mf6_file::read_model( double *e_in, int *LTP, int *NL )
// Gets the incident energy and the model for charged-particle scattering
{
  int zero, dummy;
  double Zero;

  mfx_line2(&Zero,       // zero
            e_in,        // incident energy
	    LTP,         // the model
            &zero,       // zero
	    &dummy,      // number of data entries
	    NL);         // depends on the model
}
// ------------------- mf12_file::read_line1 ---------------
void mf12_file::read_line1(int *ZA, double *AWR, int *LO, int *LG,
  int *NK)
// read the first header line for an mf12 file
{
  int iDum;

  mfx_line1(ZA,             // target Z & A 
            AWR,            // target mass
            LO,             // representation flag (=1 or 2) 
            LG,             // for LO = 2: flag for gamma source
            NK,             // number of discrete photons (inc continuum)
            &iDum);         // zero
}
// ------------------- mf12_file::read_line2 ---------------
void mf12_file::read_line2(int *NR, int *NP)
// read the second header line for an mf12 file
{
  double ZERO;
  int iZERO;

  mfx_line2(&ZERO, &ZERO,     // zeros
	    &iZERO, &iZERO,   // zeros
	    NR,               // number of interpolation regions
	    NP );             // number of data points
}
// ------------------- mf12_file::gamma_line1 ---------------
void mf12_file::gamma_line1(double *EG, double *ES, int *LP, int *LF,
  int *NR, int *NP)
// read the first line for particular gamma
{
  mfx_line2(EG,     //  gamma energy
            ES,     //  excitation level
	    LP,     // is this a primary gamma?
	    LF,     // discrete-continuum flag
	    NR,     // number of interpolation regions
	    NP);    // number of (E, multiplicity) pairs
}
// ------------------- mf13_file::read_line1 ---------------
void mf13_file::read_line1(int *ZA, double *AWR, int *NK)
// read the first header line for an mf13 file
{
  int iDum;

  mfx_line1(ZA,             // target Z & A 
            AWR,            // target mass
	    &iDum,          // zero
	    &iDum,          // zero
            NK,             // number of discrete photons (inc continuum)
            &iDum);         // zero
}
// ------------------- mf13_file::read_line2 ---------------
void mf13_file::read_line2(int *NR, int *NP)
// read the second header line for an mf13 file
{
  double ZERO;
  int iZERO;

  mfx_line2(&ZERO, &ZERO,     // zeros
	    &iZERO, &iZERO,   // zeros
	    NR,               // number of interpolation regions
	    NP );             // number of data points
}
// ------------------- mf13_file::gamma_line1 ---------------
void mf13_file::gamma_line1(double *EG, double *ES, int *LP, int *LF,
  int *NR, int *NP)
// read the first line for particular gamma
{
  mfx_line2(EG,     //  gamma energy
            ES,     //  excitation level
	    LP,     // is this a primary gamma?
	    LF,     // discrete-continuum flag
	    NR,     // number of interpolation regions
	    NP);    // number of (E, gamma production) pairs
}
// ------------------- mf14_file::first_line ---------------
void mf14_file::first_line(int *ZA, double *AWR, int *LI, int *LTT,
  int *NK, int *NI)
// read the first header line for an mf14 file
{
  mfx_line1(ZA,             // target Z & A 
            AWR,            // target mass
	    LI,             // isotropy flag
	    LTT,            // Legendre flag
	    NK,             // number of discreet photons (inc continuum)
	    NI);            // number of isotropic photon dists
}
// ------------------- mf14_file::read_EG ---------------
void mf14_file::read_EG( double *EG, int *NR, int *NE )
// read the gamma energy and the amount of data for one gamma
{
  double ES;
  int zero;

  mfx_line2(EG,              // gamma energy
            &ES,             // excitation energy
	    &zero,
	    &zero,
	    NR,             // number of interpolation regions
	    NE);            // number of incident energies
}
// ------------------- mf14_file::discrete ---------------
void mf14_file::discrete( double *EG )
// read the gamma energy for an isotropic distribution
{
  double ES;
  int zero;

  mfx_line2(EG,              // gamma energy
            &ES,             // excitation energy
	    &zero,
	    &zero,
	    &zero,
	    &zero);
}
// ------------------- mf14_file::Leg_head ---------------
void mf14_file::Leg_head( double *e_in, int *NL )
// read the incident energy and the Legendre order
{
  double Zero;
  int zero;

  mfx_line2(&Zero,
            e_in,             // energy of incident neutron
	    &zero,
	    &zero,
	    NL,               // Legendre order
	    &zero);
}
// ------------------- mf14_file::get_order ---------------
int mf14_file::get_order( int NK, int NI )
// return the maximum Legendre order for gamma distributions
// NK: total number of lines
// NI: number of isotropic lines
{
  int max_order = 0;
  int num_to_do = NK - NI;
  int count;
  double EG;

  // skip the isotropic lines
  skip_iso( NI );

  // the anisotropic data
  for( count = 0; count < num_to_do; ++count )
  {
    int NR, NE;
    read_EG( &EG, &NR, &NE );
    skip( 2*NR );  // skip the interpolation regions
    for( int iNE = 0; iNE < NE; ++iNE )
    {
      int num_coef;
      double e_in;
      Leg_head( &e_in, & num_coef );
      if( num_coef > max_order )
      {
        max_order = num_coef;
      }
      skip( num_coef );
    }
  }
  return max_order;
}
// ------------------- mf14_file::skip_iso ---------------
void mf14_file::skip_iso( int NI )
// skip NI lines of isotropic gammas
{
  // skip the isotropic lines
  for( int count = 0; count < NI; ++count )
  {
    double EG;
    discrete( &EG );
  }
}
// ------------------- mf15_file::read_line1 ---------------
void mf15_file::read_line1(int *ZA, double *AWR, int *NC)
// read the first header line for an mf15 file
{
  int iDum;

  mfx_line1(ZA,             // target Z & A 
            AWR,            // target mass
	    &iDum,          // zero
	    &iDum,          // zero
	    NC,             // number of partial distributions
	    &iDum);         // zero
}
// ------------------- mf15_file::weight_line ---------------
void mf15_file::weight_line(int *LF, int *NR, int *NP)
// read the header for the weights for this section
{
  int iDum;
  double zero;

  mfx_line1(&iDum,          // zero
            &zero,
	    &iDum,          // zero
	    LF,             // energy distribution model
	    NR,             // number of interpolation regions
            NP);            // number of incident energries
}
// ------------------- mf15_file::E_in_interp ---------------
void mf15_file::E_in_interp(int *NR, int *NP)
// get the number of incident neutron energies
{
  int iDum;
  double zero;

  mfx_line1(&iDum,          // zero
            &zero,
	    &iDum, &iDum,   // zeros
	    NR,             // number of interpolation regions
            NP);            // number of incident neutron energries
}
// ------------------- mf15_file::get_E_in ---------------
void mf15_file::get_E_in(double *E_in, int *NR, int *NP)
// get the number of secondary neutron energies
{
  int iDum;

  mfx_line1(&iDum,          // zero
            E_in,           // incident neutron energy
	    &iDum, &iDum,   // zeros
	    NR,             // number of interpolation regions
            NP);            // number of secondary neutron energries
}
// ------------------- mt458_file::read_line1 ---------------
void mt458_file::read_line1(int *ZA, double *AWR)
// read the first header line for an mf15 file
{
  int iDum;

  mfx_line1(ZA,             // target Z & A 
            AWR,            // target mass
	    &iDum,          // zero
	    &iDum,          // zero
	    &iDum,          // zero
	    &iDum);         // zero
}
// ------------------- mt458_file::read_line2 ---------------
void mt458_file::read_line2( int *num_E, int *num_data )
// read the number of energies
{
  int zero;
  double Zero;

  mfx_line2(&Zero, &Zero,
            &zero, &zero, 
	    num_E,     // number of (energy, d_energy) pairs
	    num_data); // amount of data
}
// ------------------- mt458_file::read_data ---------------
void mt458_file::read_data( double *E_1, double *dE_1, double *E_2,
    double *dE_2, double *E_3, double *dE_3 )
// read a data line for an mt458 file
{
  char linje[Line_len];

  getline(linje, Line_len-1 ); //read line
  string linebuff(linje); // the input string

  read_dddddd(&linebuff,
	      E_1, dE_1, E_2, dE_2, E_3, dE_3);
}
// ------------------- mt458_file::reduce_Q ---------------
double mt458_file::reduce_Q( double Q )
// return Q minus energy from delayed decay of fission fragments
{
  double new_Q = Q;  // initialize
  int ZA;
  double AWR;
  int num_E, num_data;

  read_line1(&ZA, &AWR);  // gobble the first line
  read_line2(&num_E, &num_data); // gobble the second line

  double EFR, d_EFR,  // kinetic energy of fission fragments
         ENP, d_ENP,  // kinetic energy of prompt fission neutrons
         END, d_END;  // kinetic energy of delayed fission neutrons
  read_data( &EFR, &d_EFR, &ENP, &d_ENP, &END, &d_END );
  new_Q -= END;  // take off energy of delayed fission neutrons

  double EGP, d_EGP,  // energy of prompt gammas
         EGD, d_EGD,  // energy of delayed gammas
         EB, d_EB;    // energy of delayed betas
  read_data( &EGP, &d_EGP, &EGD, &d_EGD, &EB, &d_EB );
  new_Q -= EGD + EB;  // take off energy of delayed gammas and betas

  return new_Q;
}
