/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1889 $
 * $Date: 2006-08-29 09:28:59 -0700 (Tue, 29 Aug 2006) $
 * $Author: dbrown $
 * $Id: ENDF_file.hpp 1889 2006-08-29 16:28:59Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// declarations for ENDF/B-VI input files

#ifndef ENDF_file_DEF
#define ENDF_file_DEF

#include <stdio.h>
#include <fstream>
#include <string>
#include <vector>

#include "endl_formats.hpp"
#include "global_params.hpp"
#include "messaging.hpp"

extern ENDLClass ENDL;
extern GlobalParameterClass Global;

using namespace std;

// ------------------- class ENDF_file ---------------
//! Base class for reading intermediate ENDF file sections
class ENDF_file : public ifstream
{
protected:
  //! Generic read of the first line
  void mfx_line1(int *ZA, double *AWR, int *I1, int *I2,
    int *I3, int *I4);

  //! Generic read of the second line
  void mfx_line2(double *D1, double *D2, int *I1, int *I2,
    int *I3, int *I4);

public:
  inline ENDF_file()
  {
  }
  inline ~ENDF_file()
  {
  }
  //! Opens the intermediate ENDF file
  inline void open( int F, int T, bool verbose=true )
  {
    if (verbose) Info("ENDF_file::open","Opening file "+endf_file_name(F, T)+"...");
    ifstream::open(endf_file_name(F, T).c_str());
  }

  //! Reads the interpolation information
  void get_regions(int NR, vector<int>& NBT, vector<int>& INT);

  //! Skips the next N entries in the file
  void skip(int N);
};

// ------------------- class mf1_file ---------------
//! Derived class used for reading intermediate MF=1 ENDF data
class mf1_file : public ENDF_file
{
public:
  //! Reads the first line
  void read_line1(int *ZA, double *AWR, int *LNU);

  //! Reads the time constants line
  void delay_head(int *NC);

  //! Reads the multiplicity line
  void mult_head(int *NR, int *NP);
};

// ------------------- class mf3_file ---------------
//! Derived class used for reading intermediate MF=3 ENDF data
class mf3_file : public ENDF_file
{
public:
  //! Reads the first line
  void read_line1(int *ZA, double *AWR);

  //! Reads the second line
  void read_line2(double *QM, double *QI, int *LR, int *NR, int *NP);
};

// ------------------- class mf4_file ---------------
//! Derived class used for reading intermediate MF=4 ENDF data
class mf4_file : public ENDF_file
{
public:
  //! Reads the first line
  void read_line1(int *ZA, double *AWR, int *LVT, int *LTT);

  //! Reads the second line
  void read_line2(int *LI, int *LCT, int *NK, int *NM);

  //! Reads the third line
  void read_line3(int *NR, int *NE);
};

// ------------------- class mf5_file ---------------
//! Derived class used for reading intermediate MF=5 ENDF data
class mf5_file : public ENDF_file
{
public:
  //! Reads the first line
  void read_line1(int *ZA, double *AWR, int *NK);

  //! Reads the second line
  void read_line2(double *U, int *LF, int *NR, int *NP);

  //! Reads the Madland-Nix header line
  void read_Madland(double *EFL, double *EFH, int *NR, int *NE);
};

// ------------------- class mf6_file ---------------
//! Derived class used for reading intermediate MF=6 ENDF data
class mf6_file : public ENDF_file
{
public:
  //! Reads the first line
  void read_line1(int *ZA, double *AWR, int *LCT, int *NK);

  //! Reads the header line for a particle in an mf6 file
  void particle_line1(int *ZAP, double *AWP, int *LIP,
    int *LAW, int *NR, int *NP);

  //! Gets the model for a particle in an mf6 file for LAW = 1
  void get_model( int *LANG, int *LEP, int *NR, int *NE );

  //! Gets the incident energy, number of discrete gammas, and number of E_out for isotropic gammas
  void gamma_ein( double *e_in, int *ND, int *NEP );

  //! Gets the total mass and number of particles for a phase-space file
  void phase_space( double *APSX, int *NPSX );

  //! Gets the number of incident energies for a gen_mf4 file
  void mf4_line1( int *NR, int *NE );

  //! Gets the model for a gen_mf4 incident energy
  void mf4_line2( double *e_in, int *LANG, int *NW, int *NL );

  //! Gets the number of incident energies for an ENDL_table
  void ENDL( int *NR, int *NE );

  //! Gets the incident energy or cosine for an ENDL_table
  void ENDL_head( double *e_in, int *NR, int *NE );

  //! Gets the incident energy and the order for a Legendre table
  void Legendre_E_in( double *e_in, int *ND, int *order, int *NE );

  //! Gets the spin for charged-particle scattering
  void read_spin( double *spin, int *NR, int *NE );

  //! Gets the model for charged-particle scattering
  void read_model( double *e_in, int *LTP, int *NL );
};

// ------------------- class mf12_file ---------------
//! Derived class used for reading intermediate MF=12 ENDF data
class mf12_file : public ENDF_file
{
public:
  //! Reads the first line
  void read_line1(int *ZA, double *AWR, int *LO, int *LG,
    int *NK);

  //! Reads the second line
  void read_line2(int *NR, int *NP);

  //! Reads the first line for particular gamma
  void gamma_line1(double *EG, double *ES, int *LP, int *LF,
    int *NR, int *NP);
};

// ------------------- class mf13_file ---------------
//! Derived class used for reading intermediate MF=13 ENDF data
class mf13_file : public ENDF_file
{
public:
  //! Reads the first line
  void read_line1(int *ZA, double *AWR, int *NK);

  //! Reads the second line
  void read_line2(int *NR, int *NP);

  //! Reads the first line for particular gamma
  void gamma_line1(double *EG, double *ES, int *LP, int *LF,
    int *NR, int *NP);
};

// ------------------- class mf14_file ---------------
//! Derived class used for reading intermediate MF=14 ENDF data
class mf14_file : public mf4_file
{
public:
  //! Reads the first line
  void first_line(int *ZA, double *AWR, int *LI, int *LTT,
    int *NK, int *NI);

  //! Reads the gamma energy and the amount of data
  void read_EG( double *EG, int *NR, int *NE );

  //! Reads the discrete energy for an isotropic distribution
  void discrete( double *EG );

  //! Reads the incident energy and the Legendre order
  void Leg_head( double *e_in, int *NL );

  //! Returns the maximum Legendre order for gamma distributions
  int get_order( int NK, int NI );

  //! Skips the NI isotropic lines
  void skip_iso( int NI );
};

// ------------------- class mf15_file ---------------
//! Derived class used for reading intermediate MF=15 ENDF data
class mf15_file : public ENDF_file
{
public:
  //! Reads the first line
  void read_line1(int *ZA, double *AWR, int *NC);

  //! Reads the line containing the weights
  void weight_line(int *LF, int *NR, int *NP);

  //! Get the number of incident neutron energies
  void E_in_interp(int *NR, int *NP);

  //! Get the number of secondary neutron energies
  void get_E_in(double *E_in, int *NR, int *NP);
};

// ------------------- class mt458_file ---------------
//! Derived class used for reading intermediate MF=458 ENDF data
class mt458_file : public ENDF_file
{
public:
  //! Reads the first line
  void read_line1(int *ZA, double *AWR);

  //! Reads the second line
  void read_line2(int *num_E, int *num_data);

  //! Reads a data line
  void read_data( double *E_1, double *dE_1, double *E_2,
    double *dE_2, double *E_3, double *dE_3 );

  //! Resets the Q value
  double reduce_Q( double Q );

};

#endif
