/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1752 $
 * $Date: 2006-03-16 11:24:52 -0800 (Thu, 16 Mar 2006) $
 * $Author: dbrown $
 * $Id: b6stream.hpp 1752 2006-03-16 19:24:52Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

#ifndef B6STREAM_CLASS
#define B6STREAM_CLASS

#include <stdio.h>  // standard I/O package
#include <fstream>  // standard file stream package
#include <iostream> // for file I/O functions
#include <string>   // for c++ strings
#include <cmath>
#include <list>
#include <vector>

using namespace std;

// ------------------------ class iMF_array_link ---------------
//! A vector of MF values for a given reaction (MT)
class iMF_array_link : public vector< int >
{
private:
  int MT_;

public:
  //!Common name used to access the MT value of this list.
  inline int& MT( )
  {
    return MT_;
  }

  //! Print the MF values for this MT
  void print_MF( );

  //! Scans the vector to determine whether angular data should be CM
  bool angle_cm( );
};

// ----------------------- class iMF_array_list -------------------
//!List of iMF_array_links.
class iMF_array_list : public list< iMF_array_link >
{
public:
  //! Function to get the link at "x".
  // Returns "false" if the entry is not found.
  bool at( int x, iMF_array_list::iterator& link );

  //! Check for the presence of MT data
  bool found( int MT );
};

// ------------------------ class b6stream ---------------
//! Reads in an ENDF file, creates single reaction files and gathers global info.

class b6stream
{ // This class reads the preprocessed ENDF file in, fills the global
  // ENDF variables and then sends the data into output files.
  // The output files are named by the MF and MT data they contain.
  // The MAT, MF, MT and NS numbers are removed from all the data
  // lines since they are either redundant or unused (NS). 

public:
  // Here are the global ENDF file variables as defined in the manual
  int NTAPE;
  int MAT, NS, yi, date;
  int LRP, LFI, NLIB, NMOD, LIS, LISO, NFOR;
  int ZA, NSUB, NVER, LDRV, NWD, NXC;
  double AWR, ELIS, STA, AWI, TEMP;

  // Here is the arrays of MF, MT, listed at end of comments.
  // The MT (reaction) numbers tag a linked list with
  // links containing the MF (data ID) numbers for each reaction.
  iMF_array_list MT_MF_list;

  // default Endf input file name
  string input_file;

  // Other useful (dummy) variables
  int iDummy;
  double dDummy;

  ifstream Efile; //instantiate the input file stream
  ofstream oFile;  //instantiate output file stream

  //! Default constructor.
  b6stream(string inputFile="endf.prepro"); 

  //! Default destructor.
  ~b6stream();

  //!Reads a single line from the ENDF file stream.
  void readENDF();

  //! Reads the MAT, MF, MT and NS entries at the end of the ENDF line
  //! and then strips them from the output stream.
  void get_MAT_MF_MT_NS();  //
  
  //! computes the yi given the sublibrary designator
  int NSUB_to_yi(int NSUB);

  void print_summary();

  //! Print all of the MT and MF values
  void print_MT_MF( );

private:
  // These variables used only internally
  int MF, MT;
  int prev_mf, prev_mt;
  string sBuff;
  bool lofile;
  // The "END" records are logicals in our code
  bool SEND, FEND, MEND, TEND, DONE;

  // The first 6 lines of the input file are special.
  // We can't use NS to count the lines in the file,
  // because it resets to 0 after 99999.
  bool file_top;

  // A consistency check for the data
  void check_mf_mt( );

protected:

};

#endif
