/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1747 $
 * $Date:$
 * $Author:$
 * $Id:$
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// Implementation of the classes used to translate the charged-particle
// scattering data for identical particles.

#include "charge.hpp"
#include "endl_formats.hpp"
#include "global_params.hpp"
#include "bdfls_tools.hpp"

extern ENDLClass ENDL;
extern GlobalParameterClass Global;
extern bdflsClass bdfls;  // for the data constants

// ******************* class one_d_charge ******************
// ------------ one_d_charge::f -----------------------
double one_d_charge::f( double mu )
{
  if( ( mu < -1.0 ) || ( mu >= 1.0 ) ||
      ( same_ && ( mu < 0.0 ) ) )
  {
    SevereError( "one_d_charge::f",
		 pastenum("mu out of range: ", mu ) );
  }

  // the different terms
  double coulomb;  // Coulomb portion
  double interfere;   // the electron interference term
  double nuke;     // the nuclear reaction term

  if( LTP_model == 1 )
  {
    // are the particles identical?
    if( same_ )
    {
      coulomb = coulomb_same( mu );
      interfere = interfere_same( mu );
      nuke = nuclear_same( mu );
    }
    else
    {
      coulomb = coulomb_diff( mu );
      interfere = interfere_diff( mu );
      nuke = nuclear_diff( mu );
    }
    return coulomb - interfere + nuke;
  }
  else
  {
    // are the particles identical?
    if( same_ )
    {
      coulomb = coulomb_same( mu );
    }
    else
    {
      coulomb = coulomb_diff( mu );
    }
    return coulomb;
  }
}
// ------------ one_d_charge::coulomb_same -----------------------
double one_d_charge::coulomb_same( double mu )
// 1/2 the first (Coulomb) term from (6.8) for identical particles
{
  double term = ( 2 * eta * eta /( k_sq * (1-mu)*(1-mu) ) ) *
    ( ( 1 + mu*mu ) / ( 1 - mu*mu ) +
      ( spin_factor / (2*spin_ + 1) ) *
      cos( eta * log( ( 1 + mu ) / ( 1 - mu ) ) ) );

  return 0.5*term;
}
// ------------ one_d_charge::interfere_same -----------------------
double one_d_charge::interfere_same( double mu )
// 1/2 the second (interference) term from (6.12) for identical particles
{
  complex< double > exponent( 0.0, eta * log( ( 1 - mu ) / 2 ) );
  complex< double > term_1 = ( 1 + mu ) * exp( exponent );

  complex< double > expon_2( 0.0, eta * log( ( 1 + mu ) / 2 ) );
  complex< double > term_2 = ( 1 - mu ) * exp( expon_2 );

  // the zeroth-order term
  complex< double > sum = 0.5* ( term_1 + term_2 ) * a_coef[0];

  // the first-order term
  if( order > 0 )
  {
    sum += 1.5 * ( term_1 - term_2 ) * a_coef[1] * mu;
  }

  // the remaining terms
  double sign = 1.0;    // int won't work
  double P_prev = 1.0;  // P_0(mu)
  double P_this = mu;   // P_1(mu)
  double P_next;
  for( int count = 2; count <= order; ++count )
  {
    P_next = ((2*count - 1)*mu*P_this - (count-1)*P_prev)/count;
    complex< double > this_coef = a_coef[count] *
      (count + 0.5)*( term_1 + sign*term_2 )*P_next;
    sum += this_coef;

    // shift the values
    P_prev = P_this;
    P_this = P_next;
    sign *= -1;
  }
    
  return eta / ( 1 - mu*mu ) * (sum.real( ));
}
// ------------ one_d_charge::nuclear_same -----------------------
double one_d_charge::nuclear_same( double mu )
// 1/2 the third (nuclear) term from (6.12) for identical particles
{

  // the zeroth-order term
  double sum =b_coef[0] / 2;

  // the remaining (even) terms
  double P_prev = 1.0;  // P_0(mu)
  double P_this = mu;   // P_1(mu)
  double P_next;
  for( int count = 2; count <= order; ++count )
  {
    P_next = ((2*count - 1)*mu*P_this - (count-1)*P_prev)/count;
    if( count % 2 == 0 )
    {
      sum += (count + 0.5)*P_next*b_coef[ count/2 ];
    }
    // shift the values
    P_prev = P_this;
    P_this = P_next;
  }
    
  return 0.5*sum;
}
// ------------ one_d_charge::coulomb_diff -----------------------
double one_d_charge::coulomb_diff( double mu )
// the first (Coulomb) term from (6.7) for different particles
{
  double term = eta * eta /( k_sq * (1-mu)*(1-mu) );

  return term;
}
// ------------ one_d_charge::interfere_diff -----------------------
double one_d_charge::interfere_diff( double mu )
// the second (interference) term from (6.11) for different particles
{
  // the zeroth-order term
  complex< double > sum = 0.5 * a_coef[0];

  // the first-order term
  if( order > 0 )
  {
    sum += 1.5 * a_coef[1] * mu;
  }

  // the remaining terms
  double P_prev = 1.0;  // P_0(mu)
  double P_this = mu;   // P_1(mu)
  double P_next;
  for( int count = 2; count <= order; ++count )
  {
    P_next = ((2*count - 1)*mu*P_this - (count-1)*P_prev)/count;
    complex< double > this_coef = a_coef[count] *
      (count + 0.5) * P_next;
    sum += this_coef;

    // shift the values
    P_prev = P_this;
    P_this = P_next;
  }
  
  complex< double > exponent( 0.0, eta * log( ( 1 - mu ) / 2 ) );
  sum *= exp( exponent );
  return 2 * eta / ( 1 - mu ) * sum.real( );
}
// ------------ one_d_charge::nuclear_diff -----------------------
double one_d_charge::nuclear_diff( double mu )
// the third (nuclear) term from (6.11) for different particles
{

  // the zeroth-order term
  double sum = b_coef[0] / 2;

  // the first-order term
  if( order > 0 )
  {
    sum += 1.5 * b_coef[1] * mu;
  }

  // the remaining (even) terms
  double P_prev = 1.0;  // P_0(mu)
  double P_this = mu;   // P_1(mu)
  double P_next;
  for( int count = 2; count <= 2*order; ++count )
  {
    P_next = ((2*count - 1)*mu*P_this - (count-1)*P_prev)/count;
    sum += (count + 0.5)*P_next*b_coef[ count ];

    // shift the values
    P_prev = P_this;
    P_this = P_next;
  }
    
  return sum;
}

// ------------ one_d_charge::initiate -----------------------
void one_d_charge::initiate( mf6_file& inFile, double e_in, double spin )
{
  // are the particles the same?
  int ZA_in = ENDL.yo_to_za( ENDL.incident_particle );
  same_ = ( ZA_in == ENDL.ZA );

  // Calculate the parameter eta
  E_in( ) = e_in;
  // the charge
  int Z_in = ENDL.get_Z( ENDL.incident_particle );
  int Z_targ = ENDL.ZA / 1000;

  // square of parameter k from (6.9)
  // But the correct formula is
  //  k = (A/(A+1)) * sqrt(2*m_1*E) / (h_bar*c)
  // We use m_ and E in MeV, h_bar in MeV sh, c in cm/sh
  // h_bar = bdfls.nuclear_constants[9] in MeV sh
  // c = bdfls.nuclear_constants[8] in cm/sh
  double denom = bdfls.nuclear_constants[8] *
    bdfls.nuclear_constants[9];
  double amu_to_MeV = bdfls.nuclear_constants[5];
  double mass_ratio = ENDL.target_mass / ENDL.projectile_mass;
  // the 10^{24} is to convert cm^2 to barns
  k_sq = ( mass_ratio / ( mass_ratio + 1 ) ) * ENDL.projectile_mass * 
    amu_to_MeV * e_in /
    ( denom * denom * 1.0e24 );

  // eta parameter from (6.10)
  // If h_bar = 1 and c = 1, then e^2 = alpha (fine-structure constant)
  // value from physics.nist.gov
  double alpha = 7.297352568e-3;
  eta = Z_in * Z_targ * alpha *
    sqrt( 0.5 * amu_to_MeV * ENDL.projectile_mass / e_in );

  // (-1)^{2*spin}
  double ds = ENDL_EPSILON( 1.0 );
  // is 2*spin even or odd?
  spin_factor = ( static_cast<int>( 2 * spin * (1+ds) ) % 2 == 0 ) ?
    1 : -1;

  // save the spin
  spin_ = spin;

  // the amount of data depends on the model
  if( LTP_model == 1 )
  {
    a_coef.reserve( order + 1 );
    a_coef.resize( order + 1 );
    if( same_ )
    {
      b_coef.reserve( order + 1 );
      b_coef.resize( order + 1 );
    }
    else
    {
      b_coef.reserve( 2*order + 1 );
      b_coef.resize( 2*order + 1 );
    }
  }
  else
  {
  }
}
// ----------- one_d_charge::read_data -----------
void one_d_charge::read_data( mf6_file& inFile )
{
  string linebuff;  // the input line
  string strbuff;   // a substring
  int num_data;
  int num_b;
  int count;
  int sub_count;

  if( LTP_model == 1 )
  {
    if( same_ )
    {
      num_b = order + 1;
      num_data = 3*order + 3;
    }
    else
    {
      num_b = 2*order + 1;
      num_data = 4*order + 3;
    }
    // read the b-coefficients
    for( count = 0; count < num_b; ++count )
    {
      double next_coef;
      sub_count = count % 6;  //  there are 6 entries per line
      if( sub_count == 0 )
      {
        getline( inFile, linebuff); //Read in new line
      }
      strbuff = linebuff.substr( sub_count*11, 11 );
      read_d( &strbuff, &next_coef );
      b_coef[ count ] = next_coef;
    }
    sub_count = ( sub_count + 1 ) % 6;  // we just read the last b
    // read the a-coefficients
    for(  ; count < num_data; count += 2 )
    {
      double real_part;
      double imag_part;

      if( sub_count == 0 )
      {
        getline( inFile, linebuff); //Read in new line
      }
      // the real and imaginary parts may be on different lines
      if( sub_count < 5 )
      {
	// same line
        strbuff = linebuff.substr( sub_count*11, 22 );
        read_dd( &strbuff, &real_part, &imag_part );
        a_coef[ count - num_b ] = complex< double >( real_part, imag_part );
      }
      else
      {
	// different lines
        strbuff = linebuff.substr( sub_count*11, 11 );
        read_d( &strbuff, &real_part );
        getline( inFile, linebuff); //Read in new line
        strbuff = linebuff.substr( 0, 11 );
        read_d( &strbuff, &imag_part );
        a_coef[ count - num_b ] = complex< double >( real_part, imag_part );
      }

      sub_count = ( sub_count + 2 ) % 6;  //  there are 6 entries per line
    }
  }
}
// ------------ one_d_charge::read_table -----------------------
void one_d_charge::read_table( mf6_file& inFile, double e_in, int num_pairs )
{
  string linebuff;  // the input line
  string strbuff;   // a substring

  dd_link link;  // for creating new links

  // read the (mu, probability) pairs---the deviation from Rutherford
  for( int count = 0; count < num_pairs; ++count )
  {
    int sub_count = count % 3;
    if( sub_count == 0 )
    {
      getline( inFile, linebuff); //Read in new line
    }
    strbuff = linebuff.substr( sub_count*22, 22 );
    read_dd( &strbuff, &link.x, &link.y );
    deviation.insert( deviation.end( ), link );
  }
}
// ----------- one_d_charge::expand -----------
void one_d_charge::expand( )
// Expand a charge model into a linked list.
{
  dd_link link;  // for creating new links

  // initialize the list with maximum and minimum mu
  double mu = Global.Value( "max_coulomb_mu" );
  if( same_ )
  {
    // initialize the list with 0 and maximum mu
    link = dd_link( 0.0, f( 0.0 ) );
  }
  else
  {
    // initialize the list with maximum and minimum mu
    link = dd_link( -1.0, f( -1.0 ) );
  }
  insert( end( ), link );

  link = dd_link( mu, f( mu ) );
  insert( end( ), link) ;

  thicken( );

  // For identical particles ENDF models only the forward emission;
  // the other particle gives the backward emission.  In order to
  // avoid the jump discontinuity, we use the average for both particles.
  one_d_charge::iterator this_link = begin( );
  for( ++this_link; this_link != end( ); ++this_link )
  {
    link.x = -this_link->x;
    link.y = this_link->y;
    push_front( link );
  }
}
// ----------- one_d_charge::expand_12 -----------
void one_d_charge::expand_12( double scale )
// Expand a charge model into a linked list.
{
  // the angular distribution is coulomb + xs*deviation
  deviation *= scale;
  *this += deviation;

  // check for positivity
  for( one_d_charge::iterator this_link = begin( );
       this_link != end( ); ++this_link )
  {
    if( this_link->y < 0.0 )
    {
      this_link->y = 0.0;
    }
  }
}

// ********* for class two_d_charge *************
// ----------- two_d_charge::master -----------------
// Reads in all the data and expands into linearly interpolable pointwise data
void two_d_charge::master( mf6_file& inFile )
{
  bool done = false;
  xs_read = false;  // cross section data not yet read

  double spin;
  int NR;  // number of interpolation regions
  int NE;  // number of incident energies

  inFile.read_spin( &spin, &NR, &NE );
  inFile.get_regions( NR, NBT, INT );

  // do we need to implement interpolation?
  bool bad_interp = false;
  int j;
  for( j = 0; j < INT.size( ); ++j )
  {
    if( ( INT[ j ] != 2 ) && ( INT[ j ] != 6 ) )
    {
      bad_interp = true;
      break;
    }
  }
  if ( bad_interp )
  {
    Warning( "two_d_charge::master",
	    pastenum( "strange interpolation: ", INT[ j ] ) );
  }

  // handle each incident neutron energy
  for ( int iNE = 0; iNE < NE; iNE++ )
  {
    if( !done )
    {
      one_E_in( inFile, spin, &done );
    }
  }

  two_d_charge::iterator this_link = begin( );
  for(  ; this_link != end( ); ++this_link )
  {
    // convert the formulas to tables
    this_link->expand( );
    if( this_link->LTP_model > 2 )
    {
      // add xs*deviation
      double scale = xs.evaluate( this_link->E_in( ) );
      this_link->expand_12( scale );
    }
  }
  // get the reaction cross section
  get_xs( );

  if( ENDL.write_file ) 
  {
    ENDL.set_yo( ENDL.incident_particle );
    write_endl( 1 ); //write the ENDL file
    // do we have charged-particle scattering for a light target?
    if ( ( ENDL.T == 2 ) && ( ENDL.ZA <= 2004 ) )
    {
      ENDL.outgoing_particle = 10 + ENDL.za_to_yo( ENDL.ZA );
      mirror( );
      write_endl( 1 );
    }
  }
}

// ----------- two_d_charge::one_E_in -----------------
// Reads in data for one incident energy
// Sets *done = true if the incident energy is at or above the maximum
void two_d_charge::one_E_in( mf6_file& inFile, double spin,
  bool *done )
{
  // make a new link for this incident energy
  one_d_charge new_link;

  // read the model and the incident energy
  double e_in;
  int LTP;  // the model
  int NL;   // depends on the model
  inFile.read_model( &e_in, &LTP, &NL );

  if( !(*done) )
  {
    // energy in MeV
    e_in *= ENDL.eV2MeV;

    // insert the 1-d list as a link
    insert( end( ), new_link );
    two_d_charge::iterator this_link = end( );
    --this_link;

    this_link->LTP_model = LTP;
    this_link->initiate( inFile, e_in, spin );

    if( LTP == 1 )
    {
      // for this model NL is the Legendre order
      this_link->order = NL;
      // read the Legendre coefficients
      this_link->read_data( inFile );
    }
    else if( LTP > 2 )
    {
      this_link->read_table( inFile, e_in, NL );
      if( LTP == 14 )
      {
        Unimplemented( "two_d_charge::one_E_in",
	  	 "implement lin-log interpolation" );
      }
      // did we already read the cross section file?
      if( !xs_read )
      {
	read_xs( );
      }
    }
    else
    {
      Unimplemented( "two_d_charge::one_E_in",
	  	 pastenum( "implement model: ", LTP ) );
    }
  }
  else
  {
    // just read the rest of the data
  }

  double dE = ENDL_EPSILON( ENDL.Max_E_in );
  if( e_in > ENDL.Max_E_in - dE )
  {
    *done = true;
  }
}
// ----------- two_d_charge::read_xs -----------------
// Reads in the cross section data
void two_d_charge::read_xs( )
{
  mf3_file xs_File;   //Instantiate the input file
  xs_File.open(3, ENDL.T);

  // get the target ZA and mass
  int ZA;
  double AWR;
  xs_File.read_line1(&ZA, &AWR);

  //read second line
  double QM, QI;
  int LR, NR, NP;
  xs_File.read_line2(
          &QM,             // mass difference Q value
          &QI,             // reaction Q value for the lowest state
          &LR,             // complex or "breakup" flag
          &NR,             // number of INT regions 
          &NP);            // number of data points
  xs_File.get_regions( NR, xs.NBT, xs.INT );
  xs.read_data( NP, xs_File, ENDL.Max_E_in );
  xs_File.close();

  xs_read = true;
}
// ----------- two_d_charge::get_xs -----------------
// Calculates the cross section
void two_d_charge::get_xs( )
{
  one_d_table new_xs;  // Instantiate the cross section object
  dd_link link;  // for creating new links

  two_d_charge::iterator this_link = begin( );
  for(  ; this_link != end( ); ++this_link )
  {
    link.x = this_link->E_in( );
    link.y = this_link->get_norm( );
    if( link.y > 0.0 )
    {
      this_link->renorm( link.y );
    }
    new_xs.insert( new_xs.end( ), link );
  }

  // print the cross sections
  ENDL.set_yo( 0 );
  new_xs.write_endl( 0 );
}
