/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1736 $
 * $Date:$
 * $Author:$
 * $Id:$
 * 
 * ******** fete: From ENDF To ENDL *********
 */

//Header for the classes used in charged-particle scattering

#ifndef CHARGE_CLASSES
#define CHARGE_CLASSES

#include <complex>

#include "distrib_base.hpp"
#include "list_2d.hpp"
#include "ENDF_file.hpp"

using namespace std;

// -------------------- class one_d_charge ---------------
//! Used for the angular distribution for charged particles at one incident energy.
//! Depending on the value of LTP, the data may be different types of parameters.
class one_d_charge : public distrib_base
{
private:
  double spin_;
  int spin_factor;  // (-1)^{2*spin}
  double k_sq;      // See equation (6.9)
  double eta;       // See equation (6.10)
  bool same_;       // are the target and incident particles identical?

  //! For LTP_model > 2 the data is a table of deviations from Rutherford
  one_d_table deviation;

  //! Evaluates the Coulomb effects for identical particles
  double coulomb_same( double mu );

  //! Evaluates the electron interference effects for identical particles
  double interfere_same( double mu );

  //! Evaluates the nuclear reaction effects for identical particles
  double nuclear_same( double mu );

  //! Evaluates the Coulomb effects for different particles
  double coulomb_diff( double mu );

  //! Evaluates the electron interference effects for different particles
  double interfere_diff( double mu );

  //! Evaluates the nuclear reaction effects for different particles
  double nuclear_diff( double mu );

public:
  //! The LTP identifier
  int LTP_model;

  // order of the Legendre expansions
  int order;

  // Legendre coefficients
  vector< complex< double > > a_coef;
  vector< double > b_coef;

  //! Sets up the parameters
  void initiate( mf6_file& inFile, double e_in, double spin );

  //! Evaluates the probability density
  double f( double mu );

  //! reads the ENDF data
  void read_data( mf6_file& inFile );

  //! reads tabular (mu, probability) data
  void read_table( mf6_file& inFile, double e_in, int num_data );

  //! Expands a one_d_charge model into a linked list
  void expand( );

  //! Expands an LTP-12 model into a linked list
  void expand_12( double scale );

};

// ----------- class two_d_charge -----------------
// derive this class from two_d_list
//! Class to handle charged-particle scattering data for identical particles
class two_d_charge : public two_d_list< one_d_charge >
{
private:
  //! We may need the cross section data
  one_d_table xs;

  bool xs_read;

  //! Reads the "cross section" data
  void read_xs( );

  //! Calculates the cross section
  void get_xs( );

public:
  //! Reads in all the data and expands into linearly interpolable pointwise data
  void master( mf6_file& inFile );

  //! Reads in data for one incident energy and expands.
  //! Sets *done true if the incident energy is at or above the maximum
  void one_E_in( mf6_file& inFile, double spin, bool *done );

};

#endif
