/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1781 $
 * $Date: 2006-03-22 09:55:41 -0800 (Wed, 22 Mar 2006) $
 * $Author: dbrown $
 * $Id: convert.cpp 1781 2006-03-22 17:55:41Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// function methods for conversion functions

#include "convert.hpp"
#include "messaging.hpp"

double stod( string instring )
{ 
  // First remove all the blanks from the string
  int nspace=instring.find(' ');
  while ( nspace < instring.size() )
    {
      instring.erase(nspace,1);
      nspace=instring.find(' ');
    }

  // Remove a possible single "e" from "properly" formatted numbers
  // so that they can be handled just like those that aren't
  instring=tolower(instring);
  int espace=instring.find('e');
  if ( espace >= 0 )
    {  
      //Need to make sure the next character is a "+" or "-",
      //if not, insert a "+" sign (i.e. 1.03e12 -> 1.03e+12)
      if ( instring.at(espace+1) != '+' &&
	   instring.at(espace+1) != '-' ) 
	instring.insert(espace+1,"+");

      instring.erase(espace,1);
    }

  // Must have only acceptable characters at this point
  if ( instring.find_first_not_of("0123456789+-.") != string::npos )
    {
      SevereError("stod","Unknown character in '"+instring+"'");
    }

  //Strip off leading + or -
  double fsign=1.0;  //either +1 or -1, depending on sign
  int lpsign = instring.find_first_of("+"); //Look for + sign
  if ( lpsign == 0 )
    {
      instring.erase(lpsign,1);
    }
  int lmsign = instring.find_first_of("-"); //Look for leading - sign
  if ( lmsign == 0 )
    {
      fsign=-1.0;
      instring.erase(lmsign,1);
    }

  // Find the first and last occurrence of signs
  int fpm = instring.find_first_of("+-");
  int lpm = instring.find_last_of("+-");
  if ( fpm !=  lpm )  //Should be the same
    {
      SevereError("stod","Too many signs!");
    }

  // The first and last occurrence of a decimal point should be the same
  // If not, we need to exit.  Program continues if no decimal is encountered.
  int fdp = instring.find_first_of(".");
  int ldp = instring.find_last_of(".");
  if ( fdp !=  ldp ) //Should be the same
    {
      SevereError("stod","Too many decimals!");
    }

  if ( fpm < 0 ) // There is no exponent
    {
      return fsign*atof(instring.c_str());
    }
  else if ( fpm == instring.size()-1 ) //Cannot end with a sign
    {
      SevereError("stod","Number cannot end with a sign!");
    }
  else if ( fdp > fpm )  //Cannot have decimal in exponent
    {
      SevereError("stod","No decimals allowed in exponent "+instring);
    }
  else
    {
      instring.insert(fpm,"e");      //Insert an "e" for easy conversion
      return fsign*atof(instring.c_str());
    }
  return 0.0; // should never get here
}

float stof( string instring )
{ 

  return static_cast<float>( stod( instring ) );

}

int stoi( string instring )
{ 
  
  return static_cast<int>( stod( instring ) );
}

//!Removes the leading blanks in a string.
string remove_leading_blanks( string instring )
{
  if( instring.size() <= 0 ) return( instring );

  int nspace=instring.find(' ');
  while ( nspace == 0 )
    {
      instring.erase(nspace,1);
      nspace=instring.find(' ');
    }
  return( instring );
}

//!Removes the trailing blanks in a string.
string remove_trailing_blanks( string instring )
{
  if ( instring.size() <= 0 ) return( instring );  //If there's nothing here, get out

  int nspace=instring.rfind(' ');
  while ( nspace == instring.size()-1 )
    {
      instring.erase(instring.size()-1,1);
      nspace=instring.rfind(' ');
    }
  return( instring );
}

//!Removes all extra blanks from a string.
string remove_extra_blanks( string instring )
{ 
  if( instring.size() == 0 ) return( instring );

  instring = remove_leading_blanks( instring );
  instring = remove_trailing_blanks( instring );

  if( instring.size() == 0 ) return( instring );

  int x = instring.find("  "); //Find a double space
  while( x < instring.size() && x >= 0 )
    {
      instring.erase(x,1);
      x = instring.find("  ");
    }
  return( instring );
}

//! Removes all blanks from a string.
string remove_all_blanks( string instring )
{
  if( instring.size() <= 0 ) return( instring );

  int nspace=instring.find(' ');
  while ( nspace < instring.size() )
    {
      instring.erase(nspace,1);
      nspace=instring.find(' ');
    }
  return( instring );
}

//! splits a string at all occurances of "pattern" and returns 
//! the substrings in a list 
vector< string > split(const string& s, const string pattern)
{
    vector< string > result;
    int begin_slice = 0;
    int end_slice = 0;
    int slice_length = 0;
    while ( s.find( pattern, begin_slice ) != string::npos )
    {
        end_slice = s.find( pattern, begin_slice )+1;
        slice_length = end_slice - begin_slice - 1;
        if ( slice_length > 0 )
            result.push_back( s.substr( begin_slice, slice_length ) );
        begin_slice = end_slice;
    }
    result.push_back( s.substr( begin_slice ) );
    return result;
}

//! joins a list of strings using "pattern" as glue
string join(const vector< string >& ls, const string pattern)
{
    string result="";
    for ( vector<string>::const_iterator itr = ls.begin(); itr != ls.end(); ++itr )
    {
        vector<string>::const_iterator next_itr = itr;
        ++next_itr;
        result += *itr;
        if ( next_itr != ls.end() ) result += pattern;
    }
    return result;
}

//! A function that lower-cases strings
string tolower(const string& s)
{
    string t = s;
    int length = t.length();
    for (int i = 0; i < length; i++) t[i] = tolower(t[i]);
    return t;
}

//! i copies of a string x pasted together
string operator*(int i, const string x){
    string xTimes("");
    for (int j=0;j<i;++j) xTimes+=x;
    return xTimes;
}

//! center a string in a field of a certain length
string center(const string stuff, int len){
    int numSpaces = len-stuff.size();
    numSpaces/=2;
    string aSpace(" ");
    string output(numSpaces*aSpace+stuff+numSpaces*aSpace);
    if (output.size()==len) return output; else return output+" ";
}

//! left justify a string in a field of a certain length
string ljust(const string stuff, int len){
    int numSpaces = len-stuff.size();
    string aSpace(" ");
    return stuff+numSpaces*aSpace;
}

//! right justify a string in a field of a certain length
string rjust(string stuff,int len){
    int numSpaces = len-stuff.size();
    string aSpace(" ");
    return numSpaces*aSpace+stuff;
}



