/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1878 $
 * $Date: 2006-05-23 11:11:18 -0700 (Tue, 23 May 2006) $
 * $Author: dbrown $
 * $Id: distrib_1d.cpp 1878 2006-05-23 18:11:18Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the one_d_distrib class

#include <cmath>   // math library

#include "distrib_1d.hpp"
#include "endl_formats.hpp"
#include "global_params.hpp"
#include "math_util.hpp"
#include "record_types.hpp"
#include "endl_precision.hpp"
#include "bdfls_tools.hpp"

extern GlobalParameterClass Global;
extern ENDLClass ENDL;
extern bdflsClass bdfls;  // for the data constants

// *********** one_d_table class ***********************
// ---------------- one_d_table::f --------------------
double one_d_table::f( double e_out )
// for lin-log or log-lin interpolation between links
{
  double ans;

  if( interp_ == 3 ) // log-lin
  {
    ans = y_0_ + b_*log( e_out / x_0_ );
  }
  else if( interp_ == 4 ) // lin-log
  {
    ans = y_0_ * exp( b_ * ( e_out - x_0_ ) );
  }
  else
  {
    SevereError("one_d_table::f",
      pastenum("attempt to interpolate with type ", interp_ ) );
  }
  return ans;
}
// ----------- one_d_table::expand_interp --------------------
void one_d_table::expand_interp( double Max_E )
// a version of expand_interp which allows lin-log and log-lin
// interpolation
{
  bool special_interp = false;

  // test for lin-log or log-lin interpolation
  for( int j = 0; j < INT.size( ); ++j)
  {
    if( ( INT[ j ] == 3 ) || ( INT[ j ] == 4 ) )
    {
      special_interp = true;
      break;
    }
  }

  if( !special_interp )
  {
    // only lin-lin, histogram, or log-log (or INT > 10)
    dd_list::expand_interp( Max_E );
  }
  else
  {
    if(INT[0] == -7)
    {
      SevereError("one_d_table::expand_interp","expand_interp called twice");
    }

    dd_list::iterator left_link = begin();

    for(int count = 1; count <= NBT.size(); ++count)
    {
      // pointer to the end of this segment
      dd_list::iterator last_link;
      if(count == NBT.size())
      {
        last_link = end();
        --last_link;
      }
      else
      {
        int j = NBT[count] - NBT[count-1];

        // get the corresponding link
        last_link = left_link;
        for(int k = 0; k < j; ++k)
        {
          ++last_link;
        }
      }

      // convert to lin-lin one segment at a time
      for( dd_list::iterator this_link = left_link;
           this_link != last_link; ++this_link )
      {
        x_0_ = this_link->x;
        y_0_ = this_link->y;
        dd_list::iterator next_link = this_link;
        ++next_link;

        if( INT[count-1] == 3 )
        {
          // log-lin
          interp_ = 3;
          double x_ratio = next_link->x / x_0_;

          // check for division by 0 and bad logarithm
          if( (x_ratio <= 0.0) || (x_ratio == 1.0) )
          {
            SevereError("one_d_table::expand_interp",
              pastenum("bad x_ratio:",x_ratio));
          }

          b_ = ( next_link->y - y_0_ ) / log( x_ratio );

          thicken( this_link, next_link );
        }
        else if( INT[count-1] == 4 )
        {
          // lin-log
          interp_ = 4;
          // check for proper inputs
          if( (y_0_ <= 0.0) || (next_link->y <= 0.0) ||
            (next_link->x == x_0_) )
          {
            SevereError("one_d_table::expand_interp","Bad inputs");
          }

          b_ = log( next_link->y / y_0_ ) / ( next_link->x - x_0_ );

          thicken( this_link, next_link );
        }
      }
      // save for the next set
      left_link = last_link;
    }
    // flag to ensure that we don't do this twice on the same data
    INT[0] = -7;
  }
}
// ----------- one_d_table::copy ---------------------------
void one_d_table::copy( one_d_table& list_2 )
// Copies a list
{
  dd_list::copy( list_2 );
  weight = list_2.weight;
  interp_type = list_2.interp_type;
}
// ----------- one_d_table::line_multiple ----------
double one_d_table::line_multiple( )
// This routine assumes that the list consists of pairs:
// (energy for a gamma line, its multiplicity).
// Returns the total multiplicity of the lines.
{
  double mult = 0;

  for( one_d_table::iterator link_ptr = begin( );
       link_ptr != end( ); ++link_ptr )
  {
    mult += link_ptr->y;
  }
  return mult;
}
// ----------- one_d_table::list_interp --------------------
void one_d_table::list_interp(double e_in, one_d_table& left_list,
  one_d_table& right_list)
// Interpolate this list between left_list and right_list.
{
  dd_list temp_list;

  if( !empty() )
  {
    SevereError("one_d_table::list_interp",
        "Trying to fill a one_d_table which has already been made");
  }

  if (( right_list.interp_type == 2 )
    ||( right_list.interp_type == 22 )) // lin-lin
  {
    double alpha = ( e_in - left_list.E_in( ) )/
      ( right_list.E_in( ) - left_list.E_in( ) );
    if( ( alpha <= 0.0 ) || ( alpha >= 1.0 ) )
    {
      SevereError("one_d_table::list_interp","Attempt to extrapolate in lin-lin");
    }

    // What we want is (1 - alpha)*left_list + alpha*right_list
    if( alpha >= 0.5 )
    {
      // start with right_list and add to it
      //  ( (1 - alpha)/alpha )*left_list
      copy( right_list );
      temp_list.copy( left_list );
      temp_list *= (1 - alpha)/alpha;
      *this += temp_list;
    }
    else
    {
      // start with left_list and add to it
      //  ( alpha/(1 - alpha) )*right_list
      copy( left_list );
      temp_list.copy( right_list );
      temp_list *= alpha/(1 - alpha);
      *this += temp_list;
    }
    E_in( ) = e_in;  // reset the incident energy tag
    renorm( );
    weight = ( 1 - alpha )*left_list.weight +
      alpha*right_list.weight;
  }
  else if ( right_list.interp_type == 1 ) // histogram
  {
    double alpha = ( e_in - left_list.E_in( ) )/
      ( right_list.E_in( ) - left_list.E_in( ) );
    if( ( alpha <= 0.0 ) || ( alpha >= 1.0 ) )
    {
      SevereError("one_d_table::list_interp"," Attempt to extrapolate a histogram");
    }

    // just make a copy of left_list
    copy( left_list );
    E_in( ) = e_in;  // reset the incident energy tag
    weight = ( 1 - alpha )*left_list.weight +
      alpha*right_list.weight;
  }
  else if ( right_list.interp_type == 3 ) // log-lin
  {
    double left_e_in = left_list.E_in( );
    double right_e_in = right_list.E_in( );
    // start with a copy of left_list
    copy( left_list );
    E_in( ) = e_in;  // reset the incident energy tag
    // insert links at E_out values for right_list
    fill_in_lists( *this, right_list );

    // loop through the links
    for( dd_list::iterator list_ptr = begin( );
        list_ptr != end( ); ++list_ptr )
    {
      double e_out = list_ptr->x;
      list_ptr->y = Log_Lin( e_in, left_e_in, left_list.evaluate( e_out ),
        right_e_in, right_list.evaluate( e_out ) );
    }
    thinit( );
    renorm( );
    if(bin_interp)
    {
      // We check interpolation using equiprobabile bins,
      // not unit-based interpolation.
      get_bins( );
    }
  }
  else
  {
    double left_e_in = left_list.E_in( );
    double right_e_in = right_list.E_in( );
    // start with a copy of left_list
    copy( left_list );
    E_in( ) = e_in;  // reset the incident energy tag
    // insert links at E_out values for right_list
    fill_in_lists( *this, right_list );

    // loop through the links
    for( dd_list::iterator list_ptr = begin( );
        list_ptr != end( ); ++list_ptr )
    {
      double e_out = list_ptr->x;
      list_ptr->y = Lin_Log( e_in, left_e_in, left_list.evaluate( e_out ),
        right_e_in, right_list.evaluate( e_out ) );
    }
    thinit( );
    renorm( );
    if(bin_interp)
    {
      // We check interpolation using equiprobabile bins,
      // not unit-based interpolation.
      get_bins( );
    }
  }
  interp_type = right_list.interp_type;
}
// ---------------- one_d_table::read_data --------------------------
void one_d_table::read_data( int NP, ENDF_file& inFile, double max_E )
// Read in 2-column data
{
  string linebuff;  // the input string
  string strbuff;   // a substring
  bool first_energy = true;
  bool done = false;
  double dE = ENDL_EPSILON( max_E );

  for( int iNP=0; iNP < NP; iNP++ ) //Loop over NP data pairs
  { 
    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 ) getline( inFile, linebuff); //Read in new line
    for( int i=0+iP*22; i<22+iP*22; i=i+22 ) //Grab the right pair
    { 
      strbuff = linebuff.substr(i,22);
    }
    double X,Y; //Real variables we will put the data pair
    read_dd(&strbuff, &X, &Y );
    if ( X==-999. || Y==-999. ) //read_PAIR sends -999. if problem
    {
      SevereError("one_d_table::read_data","read_PAIR failed!");
    }
    else 
    {
      dd_link XYdata;
      XYdata.E_in() = X*ENDL.eV2MeV; //convert to MeV
      XYdata.x_sec() = Y;
      if( !done )
      {
        insert(end(), XYdata);
        if( XYdata.E_in() > max_E - dE )
        {
          done = true;
        }

        if ( first_energy )  //Save the first energy for later use
        {
          first_energy =  false ;
          min_E_out = XYdata.E_in();  // for unit-base interpolation
          // is this a cross section table?
          if( ENDL.F == 3 )
          {
            ENDL.threshold = X*ENDL.eV2MeV; //convert to MeV
            if( ENDL.threshold > ENDL.Max_E_in )
            {
              ENDL.write_file = false;
              Info("one_d_table::read_data",
                 "Omit reaction: threshold above Max_energy");
            }
          }
        }
      }
    }
  }
  // There may be some double-valued cross sections so get rid of them
  widen_jumps();

  // This could be a probability table at one incident energy
  if( ENDL.F > 3 )
  {
    one_d_table::iterator last_link = end( );
    --last_link;
    double last_E = last_link->x;
    if( last_E > max_E )
    {
      chop( max_E );
      last_E = max_E;
    }
    max_E_out = max_E;  // for unit-base interpolation
  }
}

// *********** one_d_evap class ***********************
// ----------- one_d_evap constructor -----------------
// default constructor
one_d_evap::one_d_evap()
{
}
// ----------- one_d_evap::initiate --------------------
void one_d_evap::initiate(double e_in, double Theta, double U)
// set the parameters for the evaporation model
{
  // to make a delta function right at threshold
  double EPSILON = 1.0e-10;

  // are we below the threshold?
  if( e_in < ENDL.threshold )
  {
    Info("one_d_evap::initiate",
      pastenum("e_in = ", e_in)+pastenum(" below threshold.  Threshold is ",ENDL.threshold));
  }

  E_in() = e_in;
  // for unit-base interpolation
  min_E_out = 0.0;
  max_E_out = E_in() - U;
  if( max_E_out > ENDL.Max_E_out )
  {
    max_E_out = ENDL.Max_E_out;
  }

  Theta_ = Theta;
  U_ = U;

  if ( max_E_out <= 0.0 )
  {
    // below model threshold OK in multiple models if weight = 0
    // make a dummy list
    dd_link link( 0.0, 0.0 );  // for creating new links
    insert(end(), link);
    link.x = EPSILON;
    link.y = 1.0;
    insert(end(), link);
  }
  else
  {
    // this is a real list
    extrema();  // start the list by inserting the extrema
    thicken();
  }
  renorm();

  // If we are checking interpolation using the equiprobable
  // bins, get the bins.
  if(bin_interp)
  {
    get_bins();
  }
}
// ----------- one_d_evap::extrema -----------
void one_d_evap::extrema()
// Start a list for an evaporation model by inserting the
// maximum, minimum and inflection point
{
  dd_link link;  // for creating new links

  // the maximum secondary energy
  double E_max = max_E_out;

  // exit if we are below threshold
  if(E_max <= 0.0)
  {
    return;
  }

  // initialize the list with 0 secondary energy
  link = dd_link(0.0, f(0.0));
  insert(end(), link);

  // do we include the maximum at E = Theta?
  double next_E = Theta_;
  if(E_max > next_E)
  {
    link = dd_link(next_E, f(next_E));
    insert(end(), link);
  }

  // do we include the inflection point at E = 2*Theta?
  next_E = 2*Theta_;
  if(E_max > next_E)
  {
    link = dd_link(next_E, f(next_E));
    insert(end(), link);
  }

  // end the list with E_max
  link = dd_link(E_max, f(E_max));
  insert(end(), link);
}
// ----------- one_d_evap::copy ---------------------------
void one_d_evap::copy( one_d_evap& list_2 )
// Copies a list
{
  dd_list::copy( list_2 );
  weight = list_2.weight;
  interp_type = list_2.interp_type;
  U_ = list_2.U_;
  Theta_ = list_2.Theta_;
}
// -------------- one_d_evap::f -----------------
double one_d_evap::f(double E_out)
// f = E'*exp{-E'/Theta}, 0 <= E' <= E - U
{
  double E_max = E_in() - U_;
  const double eps = 1.0e-12;

  if((E_out < 0.0) || (E_out > E_max*(1 + eps)))
  {
    // we are outside the range
    Warning("one_d_evap::f",pastenum("E_out: ",E_out)+
        " outside range for evaporation function");
    return 0.0;
  }
  else
  {
    return E_out*exp(-E_out/Theta_);
  }
}
// ----------- one_d_evap::list_interp --------------------
void one_d_evap::list_interp(double e_in, one_d_evap& left_list,
  one_d_evap& right_list)
// Interpolate this list between left_list and right_list.
{
  if(!empty())
  {
    SevereError("one_d_evap::list_interp",
        "Trying to fill a one_d_evap which has already been made");
  }
  // set the parameters
  double Theta;
  double U;
  if ( right_list.interp_type == 2 ) // lin-lin
  {
    Theta = Lin_Lin( e_in, left_list.E_in(), left_list.Theta_,
      right_list.E_in(), right_list.Theta_);
    U = Lin_Lin( e_in, left_list.E_in(), left_list.U_,
      right_list.E_in(), right_list.U_);
  }
  else if ( right_list.interp_type == 3 ) // log-lin
  {
    Theta = Log_Lin( e_in, left_list.E_in(), left_list.Theta_,
      right_list.E_in(), right_list.Theta_);
    U = Log_Lin( e_in, left_list.E_in(), left_list.U_,
      right_list.E_in(), right_list.U_);
  }
  else if ( right_list.interp_type == 4 ) // lin-log
  {
    Theta = Lin_Log( e_in, left_list.E_in(), left_list.Theta_,
      right_list.E_in(), right_list.Theta_);
    U = Lin_Log( e_in, left_list.E_in(), left_list.U_,
      right_list.E_in(), right_list.U_);
  }
  else if ( right_list.interp_type == 5 ) // log-log
  {
    Theta = Log_Log( e_in, left_list.E_in(), left_list.Theta_,
      right_list.E_in(), right_list.Theta_);
    U = Log_Log( e_in, left_list.E_in(), left_list.U_,
      right_list.E_in(), right_list.U_);
  }
  else
  {
    SevereError("one_d_evap::list_interp", 
        pastenum("Implement for type: ",right_list.interp_type));
  }
  interp_type = right_list.interp_type;

  // fill out the list
  initiate(e_in, Theta, U);
}
// *********** one_d_Maxwell class ***********************
// ----------- one_d_Maxwell constructor -----------------
// default constructor
one_d_Maxwell::one_d_Maxwell()
{
}
// ----------- one_d_Maxwell::initiate --------------------
void one_d_Maxwell::initiate(double e_in, double Theta, double U)
// set the parameters for the Maxwell model
{
  // to increase the incident energy right at threshold
  double EPSILON = Global.Value( "mf5_shift" );

  // are we right at the threshold?
  if(e_in <= ENDL.threshold)
  {
    // go a little above it to avoid getting a delta function
    E_in() = e_in*( 1 + EPSILON );
  }
  else
  {
    E_in() = e_in;
  }
  // for unit-base interpolation
  min_E_out = 0.0;
  max_E_out = E_in() - U;
  if( max_E_out > ENDL.Max_E_out )
  {
    max_E_out = ENDL.Max_E_out;
  }

  Theta_ = Theta;
  U_ = U;

  if ( max_E_out <= 0.0 )
  {
    // below model threshold OK in multiple models if weight = 0
    // make a dummy list
    dd_link link( 0.0, 0.0 );  // for creating new links
    insert(end(), link);
    link.x = EPSILON;
    link.y = 1.0;
    insert(end(), link);
  }
  else
  {
    // this is a real list
    extrema();  // start the list by inserting the extrema
    thicken();
  }
  renorm();

  // If we are checking interpolation using the equiprobable
  // bins, get the bins.
  if(bin_interp)
  {
    get_bins();
  }
}
// ----------- one_d_Maxwell::extrema -----------
void one_d_Maxwell::extrema()
// Start a list for an Maxwell model by inserting the
// maximum, minimum and inflection point
{
  dd_link link;  // for creating new links

  // the maximum secondary energy
  double E_max = max_E_out;

  // exit if we are below threshold
  if(E_max <= 0.0)
  {
    return;
  }

  // initialize the list with 0 secondary energy
  link = dd_link(0.0, f(0.0));
  insert(end(), link);

  // do we include the maximum at E = Theta/2?
  double next_E = 0.5*Theta_;
  if(E_max > next_E)
  {
    link = dd_link(next_E, f(next_E));
    insert(end(), link);
  }

  // do we include the inflection point at E = Theta*(1 + sqrt(2))/2?
  next_E = Theta_*(1 + sqrt(2.0))/2;
  if(E_max > next_E)
  {
    link = dd_link(next_E, f(next_E));
    insert(end(), link);
  }

  // end the list with E_max
  link = dd_link(E_max, f(E_max));
  insert(end(), link);
}
// ----------- one_d_Maxwell::copy ---------------------------
void one_d_Maxwell::copy( one_d_Maxwell& list_2 )
// Copies a list
{
  dd_list::copy( list_2 );
  weight = list_2.weight;
  interp_type = list_2.interp_type;
  U_ = list_2.U_;
  Theta_ = list_2.Theta_;
}
// -------------- one_d_Maxwell::f -----------------
double one_d_Maxwell::f(double E_out)
// f = sqrt(E')*exp{-E'/Theta}, 0 <= E' <= E - U
{
  double E_max = E_in() - U_;
  const double eps = 1.0e-12;

  if((E_out < 0.0) || (E_out > E_max*(1 + eps)))
  {
    // we are outside the range
    Warning("one_d_Maxwell::f", 
        pastenum("E_out: ",E_out)+" outside range for Maxwell function");
    return 0.0;
  }
  else
  {
    return sqrt(E_out)*exp(-E_out/Theta_);
  }
}
// ----------- one_d_Maxwell::list_interp --------------------
void one_d_Maxwell::list_interp(double e_in, one_d_Maxwell& left_list,
  one_d_Maxwell& right_list)
// Interpolate this list between left_list and right_list.
{
  if(!empty())
  {
    SevereError("one_d_Maxwell::list_interp",
        "Trying to fill a one_d_Maxwell which has already been made");
  }
  // set the parameter
  double Theta;
  if ( right_list.interp_type == 2 ) // lin-lin
  {
    Theta = Lin_Lin( e_in, left_list.E_in(), left_list.Theta_,
      right_list.E_in(), right_list.Theta_);
  }
  else if ( right_list.interp_type == 3 ) // log-lin
  {
    Theta = Log_Lin( e_in, left_list.E_in(), left_list.Theta_,
      right_list.E_in(), right_list.Theta_);
  }
  else if ( right_list.interp_type == 4 ) // lin-log
  {
    Theta = Lin_Log( e_in, left_list.E_in(), left_list.Theta_,
      right_list.E_in(), right_list.Theta_);
  }
  else if ( right_list.interp_type == 5 ) // log-log
  {
    Theta = Log_Log( e_in, left_list.E_in(), left_list.Theta_,
      right_list.E_in(), right_list.Theta_);
  }
  else
  {
    Unimplemented("one_d_Maxwell::list_interp",
        pastenum("Implement for type: ",right_list.interp_type));
  }
  // U is independent of incident energy
  double U = right_list.U_;
  interp_type = right_list.interp_type;

  // fill out the list
  initiate(e_in, Theta, U);
}
// *********** one_d_Watt class ***********************
// ----------- one_d_Watt constructor -----------------
// default constructor
one_d_Watt::one_d_Watt()
{
}
// ------------- one_d_Watt::set_EUa ----------------------
void one_d_Watt::set_EUa( double e_in, double U, double a )
// set parameters E_in, U, and a
{
  // to increase the incident energy right at threshold
  double EPSILON = Global.Value( "mf5_shift" );

  // are we right at the threshold?
  if( e_in <= ENDL.threshold )
  {
    // go a little above it to avoid getting a delta function
    E_in() = e_in*( 1 + EPSILON );
  }
  else
  {
    E_in() = e_in;
  }
  U_ = U;
  a_ = a;

  // for possible mapping to unit base
  max_E_out = E_in() - U;
  if( max_E_out > ENDL.Max_E_out )
  {
    max_E_out = ENDL.Max_E_out;
  }
  min_E_out = 0.0;
}

// ----------- one_d_Watt::intermediate_b --------------------
void one_d_Watt::intermediate_b(double e_in, double U, double a,
  int a_interp, double b, int b_interp)
// set up a list for b between 2 incident energies for a
{
  // to increase the incident energy right at threshold
  double EPSILON = Global.Value( "mf5_shift" );

  // are we right at the threshold?
  if( e_in <= ENDL.threshold )
  {
    // go a little above it to avoid getting a delta function
    E_in() = e_in*( 1 + EPSILON );
  }
  else
  {
    E_in() = e_in;
  }
  U_ = U;
  a_ = a;
  b_ = b;
  interp_type = a_interp; // how to interpolate a
  interp_type_b = b_interp; // how to interpolate b
  // for possible mapping to unit base
  max_E_out = E_in() - U;
  if( max_E_out > ENDL.Max_E_out )
  {
    max_E_out = ENDL.Max_E_out;
  }
  min_E_out = 0.0;

  // expand the list
  initiate();
}
// ----------- one_d_Watt::same_b --------------------
void one_d_Watt::same_b( double b, int b_interp )
// set up a list for b at an incident energy for a
{
  // to increase the incident energy right at threshold
  b_ = b;
  interp_type_b = b_interp; // how to interpolate b
  // expand the list
  initiate();
}
// --------------------- one_d_Watt::initiate() -------------
void one_d_Watt::initiate()
// fill out the list
{
  // to increase the incident energy right at threshold
  double EPSILON = Global.Value( "mf5_shift" );

  if ( max_E_out <= 0.0 )
  {
    // below model threshold OK in multiple models if weight = 0
    // make a dummy list
    dd_link link( 0.0, 0.0 );  // for creating new links
    insert(end(), link);
    link.x = EPSILON;
    link.y = 1.0;
    insert(end(), link);
  }
  else
  {
    // this is a real list
    extrema();  // start the list by inserting the maximum
    thicken();
  }
  renorm();

  // If we are checking interpolation using the equiprobable
  // bins, get the bins.
  if(bin_interp)
  {
    get_bins();
  }
}
// ----------- one_d_Watt::extrema --------------------
void one_d_Watt::extrema()
// start the list with links a 0, the maximum probability,
// and the maximum energy
{
  dd_link link;  // for creating new links
  // initialize the list with 0 secondary energy
  link = dd_link(0.0, f(0.0));
  insert(end(), link);

  // end the list with E_max
  double E_max = max_E_out;
  link = dd_link(E_max, f(E_max));
  insert(end(), link);

  // to find the maximum
  double sigma = a_*b_/2;
  double beta = sqrt(b_*E_max);  // the largest possible beta
  // The zeroin function requires 2 dd_links on input
  dd_link B(0.0, 0.0);  // a link to the left of the maximum
  dd_link C(beta, Watt_max(beta, 0.0));  // a link to the right of the maximum
  if( C.y > sigma )
  {
    // There is an interior maximum
    beta = dd_list::zeroin(&Watt_max, sigma, B, C, 0.0, 1.0E-5);
    E_max = beta*beta/b_;
    if( E_max < max_E_out )
    {
      link = dd_link(E_max, f(E_max));
      dd_list::iterator list_ptr = end();
      --list_ptr;
      insert(list_ptr, link);
    }
  }
}
// ----------- one_d_Watt::copy ---------------------------
void one_d_Watt::copy( one_d_Watt& list_2 )
// Copies a list
{
  dd_list::copy( list_2 );
  weight = list_2.weight;
  interp_type = list_2.interp_type;
  U_ = list_2.U_;
  a_ = list_2.a_;
  b_ = list_2.b_;
}
// -------------- one_d_Watt::f -----------------
double one_d_Watt::f(double E_out)
// f = sinh{sqrt(b*E')}*exp{-E'/a}, 0 <= E' <= E - U
{
  double E_max = max_E_out;
  const double eps = 1.0e-12;

  if((E_out < 0.0) || (E_out > E_max*(1 + eps)))
  {
    // we are outside the range
    Warning("one_d_Watt::f",pastenum(" E_out: ",E_out)+
         " outside range for Watt function");
    return 0.0;
  }
  else
  {
    return sinh(sqrt(b_*E_out))*exp(-E_out/a_);
  }
}
// ----------- one_d_Watt::list_interp --------------------
void one_d_Watt::list_interp(double e_in, one_d_Watt& left_list,
  one_d_Watt& right_list)
// Interpolate this list between left_list and right_list.
{
  if( !empty() )
  {
    SevereError("one_d_Watt::list_interp",
        "Trying to fill a one_d_Watt which has already been made");
  }
  // set the parameters
  double a;
  double b;
  double U;
  if ( right_list.interp_type == 2 ) // lin-lin for a and U
  {
    a = Lin_Lin( e_in, left_list.E_in(), left_list.a_,
      right_list.E_in(), right_list.a_);
    U = Lin_Lin( e_in, left_list.E_in(), left_list.U_,
      right_list.E_in(), right_list.U_);
  }
  else if ( right_list.interp_type == 3 ) // log-lin
  {
    a = Log_Lin( e_in, left_list.E_in(), left_list.a_,
      right_list.E_in(), right_list.a_);
    U = Log_Lin( e_in, left_list.E_in(), left_list.U_,
      right_list.E_in(), right_list.U_);
  }
  else if ( right_list.interp_type == 4 ) // lin-log
  {
    a = Lin_Log( e_in, left_list.E_in(), left_list.a_,
      right_list.E_in(), right_list.a_);
    U = Lin_Log( e_in, left_list.E_in(), left_list.U_,
      right_list.E_in(), right_list.U_);
  }
  else if ( right_list.interp_type == 5 ) // log-log
  {
    a = Log_Log( e_in, left_list.E_in(), left_list.a_,
      right_list.E_in(), right_list.a_);
    U = Log_Log( e_in, left_list.E_in(), left_list.U_,
      right_list.E_in(), right_list.U_);
  }
  else
  {
    SevereError("one_d_Watt::list_interp",
        pastenum("Implement for type: ",right_list.interp_type));
  }

  if ( right_list.interp_type_b == 2 ) // lin-lin for b
  {
    b = Lin_Lin( e_in, left_list.E_in(), left_list.b_,
      right_list.E_in(), right_list.b_);
  }
  else if ( right_list.interp_type_b == 3 ) // log-lin
  {
    b = Log_Lin( e_in, left_list.E_in(), left_list.b_,
      right_list.E_in(), right_list.b_);
  }
  else if ( right_list.interp_type_b == 4 ) // lin-log
  {
    b = Lin_Log( e_in, left_list.E_in(), left_list.b_,
      right_list.E_in(), right_list.b_);
  }
  else if ( right_list.interp_type_b == 5 ) // log-log
  {
    b = Log_Log( e_in, left_list.E_in(), left_list.b_,
      right_list.E_in(), right_list.b_);
  }
  else
  {
    Unimplemented("one_d_Watt::list_interp b",
        pastenum("Implement for type: ",right_list.interp_type_b));
  }
  interp_type = right_list.interp_type;
  interp_type_b = right_list.interp_type_b;

  // set parameters
  set_EUa( e_in, U, a );

  // set b and fill out the list
  same_b( b, right_list.interp_type_b );
}
// --------------------- Watt_max --------------------
double Watt_max(double beta, double dummy)
// used to locate the maximum of the Watt function
{
  return beta*tanh(beta);
}

// *********** one_d_Madland class ***********************
// ----------- one_d_Madland constructor -----------------
// default constructor
one_d_Madland::one_d_Madland()
{
  E_max = ENDL.Max_E_in;
  Gamma_a = c_dgamma( 1.5 );
}
// ----------- one_d_Madland::initiate --------------------
void one_d_Madland::initiate(double e_in, double EFL, double EFH,
  double TM )
// set the parameters for the Madland model
{
  E_in() = e_in;
  EFL_ = EFL;
  EFH_ = EFH;
  TM_ = TM;

  // start with some number of links
  const int num_links = 20;
  double dE = E_max/num_links;
  for( int count = 0; count <= num_links; ++count)
  {
    double E_out = count * dE;
    dd_link new_link( E_out, f(E_out) );  // for creating new links
    insert(end(), new_link);
  }

  thicken();
  thinit();
  renorm();

  // If we are checking interpolation using the equiprobable
  // bins, get the bins.
  if(bin_interp)
  {
    get_bins();
  }
}
// -------------- one_d_Madland::g -----------------
double one_d_Madland::g( double E_out, double E_frag )
// function used by f
{
  double sqrt_E_out = sqrt( E_out );
  double sqrt_E_frag = sqrt ( E_frag );
  double u_1 = ( sqrt_E_out - sqrt_E_frag ) *
    ( sqrt_E_out - sqrt_E_frag ) / TM_;
  double u_2 = ( sqrt_E_out + sqrt_E_frag ) *
    ( sqrt_E_out + sqrt_E_frag ) / TM_;

  double gamma_1 = c_dgamit( 1.5, u_1 );
  gamma_1*= Gamma_a * u_1*sqrt(u_1);  // scale by Gamma(a) * x^a

  double gamma_2 = c_dgamit( 1.5, u_2 );
  gamma_2*= Gamma_a * u_2*sqrt(u_2);  // scale by Gamma(a) * x^a

  double ans = ( u_2 * sqrt( u_2 ) * expint( 1, u_2 ) -
                 u_1 * sqrt( u_1 ) * expint( 1, u_1 ) +
                 gamma_2 - gamma_1 ) / (3 * sqrt( E_frag * TM_ ));
  return ans;
}
// ----------- one_d_Madland::copy ---------------------------
void one_d_Madland::copy( one_d_Madland& list_2 )
// Copies a list
{
  dd_list::copy( list_2 );
  weight = list_2.weight;
  interp_type = list_2.interp_type;
  EFL_ = list_2.EFL_;
  EFH_ = list_2.EFH_;
  TM_ = list_2.TM_;
  E_max = list_2.E_max;
  Gamma_a = list_2.Gamma_a;
}
// ----------- one_d_Madland::list_interp --------------------
void one_d_Madland::list_interp(double e_in, one_d_Madland& left_list,
  one_d_Madland& right_list)
// Interpolate this list between left_list and right_list.
{
  if(!empty())
  {
    SevereError("one_d_Madland::list_interp",
        "Trying to fill a one_d_Madland which has already been made");
  }
  // set the parameters
  double TM;
  if ( right_list.interp_type == 2 ) // lin-lin
  {
    TM = Lin_Lin( e_in, left_list.E_in(), left_list.TM_,
      right_list.E_in(), right_list.TM_);
  }
  else if ( right_list.interp_type == 3 ) // log-lin
  {
    TM = Log_Lin( e_in, left_list.E_in(), left_list.TM_,
      right_list.E_in(), right_list.TM_);
  }
  else if ( right_list.interp_type == 4 ) // lin-log
  {
    TM = Lin_Log( e_in, left_list.E_in(), left_list.TM_,
      right_list.E_in(), right_list.TM_);
  }
  else if ( right_list.interp_type == 5 ) // log-log
  {
    TM = Log_Log( e_in, left_list.E_in(), left_list.TM_,
      right_list.E_in(), right_list.TM_);
  }
  else
  {
    Unimplemented("one_d_Madland::list_interp",
        pastenum("Implement for type: ",right_list.interp_type));
  }
  double EFL = right_list.EFL_;
  double EFH = right_list.EFH_;
  interp_type = right_list.interp_type;

  // fill out the list
  initiate(e_in, EFL, EFH, TM);
}
// ******************* class one_d_Legendre ******************
double one_d_Legendre::f(double mu)
{
  const double eps = 1.0e-12;  // a tolerance for the range

  if((mu < -1.0 - eps) || (mu > 1.0 + eps))
  {
    // we are outside the range
    Warning("one_d_Legendre::f",pastenum("cosine ",mu)+" outside its range");

    return 0.0;
  }
  else if(coef.size() < 1)
  {
    // failure
    SevereError("one_d_Legendre::f","Legendre_func called with no coefficients");
  }

  // First, handle the trivial cases
  double sum = 0.5*coef.at(0);
  if(coef.size() == 1)
  {
    return sum;
  }

  sum += 1.5*mu*coef.at(1);
  if(coef.size() == 2)
  {
    if(sum < 0.0)
    {
      Warning("one_d_Legendre::f",
        pastenum("Legendre sum < 0 for mu = ",mu));
    }
    return sum;
  }

  // Now use the iteration formula to compute the Legendre functions
  //   (n+1) P_{n+1}(mu) - (2n+1) mu P_n(mu) + n P_{n-1}(mu) = 0
  // See Courant and Hilbert, Methods of Mathematical Physics,
  // vol. 1, p. 86.
  double P_prev = 1.0;  // P_0(mu)
  double P_this = mu;   // P_1(mu)
  double P_next;
  for(int count = 1; count < coef.size() - 1; ++count)
  {
    // Note: the notation here is bad---P[count + 1] is the coefficient
    // multiplying the Legendre polynomial P_{count + 1}(mu).
    P_next = ((2*count + 1)*mu*P_this - count*P_prev)/(count + 1.0);
    sum += (count + 1.5)*P_next*coef.at(count + 1);

    // shift the values
    P_prev = P_this;
    P_this = P_next;
  }

  if(sum < 0.0) sum = 0.0;

  return sum;
}
// ----------- one_d_Legendre::expand -----------
void one_d_Legendre::expand()
// Expand a Legendre model into a linked list.
{
  dd_link link;  // for creating new links

  // initialize the list with mu = -1
  link = dd_link(-1.0, f(-1.0));
  insert(end(), link);

  // add a link with mu = 1
  link = dd_link(1.0, f(1.0));
  insert(end(), link);

  thicken();
  renorm();
}
// ----------- one_d_Legendre::copy ---------------------------
void one_d_Legendre::copy( one_d_Legendre& list_2 )
// Copies a list
{
  dd_list::copy( list_2 );
  weight = list_2.weight;
  interp_type = list_2.interp_type;
}
// ----------- one_d_Legendre::list_interp --------------------
void one_d_Legendre::list_interp(double e_in, one_d_Legendre& left_list,
  one_d_Legendre& right_list )
// interpolate this list between left_list and right_list.
{
  if(!empty())
  {
    SevereError("one_d_Legendre::list_interp",
        "Trying to fill a one_d_Legendre which has already been made");
  }
  // how many coefficients
  int num_coef = ( left_list.coef.size() > right_list.coef.size() ) ?
    left_list.coef.size() : right_list.coef.size();

  coef.get_space( num_coef );

  // interpolate to get the coefficients
  for ( int count = 0; count < num_coef; ++count )
  {
    double left_coef = ( count < left_list.coef.size() ) ?
      left_list.coef[ count ] : 0.0;
    double right_coef = ( count < right_list.coef.size() ) ?
      right_list.coef[ count ] : 0.0;
    double this_coef;

    if ( right_list.interp_type == 1 ) // histogram
    {
      coef[ count ] = left_coef;
    }
    else if ( right_list.interp_type == 2 ) // lin-lin
    {
      this_coef = Lin_Lin( e_in, left_list.E_in(), left_coef,
        right_list.E_in(), right_coef );
      coef[ count ] = this_coef;
    }
    else if ( right_list.interp_type == 3 ) // log-lin
    {
      this_coef = Log_Lin( e_in, left_list.E_in(), left_coef,
        right_list.E_in(), right_coef );
      coef[ count ] = this_coef;
    }
    else if ( right_list.interp_type == 4 ) // lin-log
    {
      this_coef = Lin_Log( e_in, left_list.E_in(), left_coef,
        right_list.E_in(), right_coef );
      coef[ count ] = this_coef;
    }
    else if ( right_list.interp_type == 5 ) // log-log
    {
      this_coef = Log_Log( e_in, left_list.E_in(), left_coef,
        right_list.E_in(), right_coef );
      coef[ count ] = this_coef;
    }
    else
    {
      Unimplemented("one_d_Legendre::list_interp",
        pastenum("Implement for type: ",right_list.interp_type));
    }
  }
  interp_type = right_list.interp_type;

  // expand the 1-d model
  E_in() = e_in;
  expand();

  // set up the equiprobable bins
  get_bins();
}
// ****************** class one_d_isotropic *****************
void one_d_isotropic::initiate(double e_in)
// Make a list for this incident energy
{
  E_in() = e_in;

  // put in the isotropic 1-d links
  dd_link link(-1.0, 0.5);
  insert(end(), link);
  link.mu() = 1.0;
  insert(end(), link);
}
// ----------- one_d_isotropic::copy ---------------------------
void one_d_isotropic::copy( one_d_isotropic& list_2 )
// Copies a list
{
  dd_list::copy( list_2 );
  weight = list_2.weight;
  interp_type = list_2.interp_type;
}
// ****************** class one_d_hat *****************
double one_d_hat::f(double E_out)
{
  if((E_out < left_scale) || (E_out > right_scale))
  {
    // we are outside the range
    return 0.0;
  }
  else if(E_out <= top_scale)
  {
    return (E_out - left_scale)/(top_scale - left_scale);
  }
  else
  {
    return (right_scale - E_out)/(right_scale - top_scale);
  }
}
// ------------ one_d_hat::initiate -----------------------
void one_d_hat::initiate(double left_end, double top,
  double right_end)
// set up a hat-shaped energy distribution
{
    // a1 = left end, a2 = top, and a3 = right end
  min_E_out = left_end;
  max_E_out = right_end;

  if(max_E_out <= min_E_out)
  {
    SevereError("one_d_hat::initiate","energies out of order");
  }

  left_scale = 0.0;
  top_scale = (top - right_end)/(left_end - right_end);
  left_scale = 1.0;

  // Expand a hat model into a linked list.
  dd_link link;  // for creating new links

  // the first link
  link = dd_link(0.0, 0.0);
  insert(end(), link);

  // a check
  if((top_scale <= 0) || (top_scale >= 1.0))
  {
    SevereError("one_d_hat::initiate","top of hat out of range");
  }

  link = dd_link(top_scale, 1.0);
  insert(end(), link);

  link = dd_link(1.0, 0.0);
  insert(end(), link);

  renorm();
}

