/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1843 $
 * $Date: 2006-04-20 11:38:46 -0700 (Thu, 20 Apr 2006) $
 * $Author: dbrown $
 * $Id: distrib_1d.hpp 1843 2006-04-20 18:38:46Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// header for the classes for energy and angle distributions
// These classes are derived from distrib_base

#ifndef ONE_D_DISTRIB_CLASS
#define ONE_D_DISTRIB_CLASS

#include "distrib_base.hpp"
#include "math_util.hpp"  // Param class for the Legendre coefficients
#include "messaging.hpp"
#include "ENDF_file.hpp"

// -------------------- class one_d_table ------------------
class one_d_table : public distrib_base
//! Used for tables of pairs at one incident neutron energy (i.e., energy of emitted particle and probability density).
{
private:
  // parameters used in lin-log or log-lin interpolation
  int interp_;
  double b_;
  double x_0_;
  double y_0_;

public:

  //!Function for lin-log or log-lin interpolation between links
  double f(double E);

  one_d_table(): interp_(0), b_(0.0), x_0_(0.0), y_0_(0.0) {}
  ~one_d_table() {}

  //!A version of expand_interp which allows lin-log and log-lin interpolation
  void expand_interp( double Max_E );

  //! Copies a list
  void copy( one_d_table& list_2 );

  //! Returns the total multiplicity for lists of (energy for a gamma line, its multiplicity).
  double line_multiple( );

  //! Interpolates between two lists
  void list_interp(double e_in, one_d_table& left_list,
    one_d_table& right_list);

  //! Reads pair-wise data into a 2-column list
  void read_data( int NP, ENDF_file& inFile, double max_E );
};

// -------------------- class one_d_evap ------------------
class one_d_evap : public distrib_base
//! Used for the evaporation model at one incident neutron energy
{
private:
  double U_;
  double Theta_;

public:

  double f(double E);

  one_d_evap();
  ~one_d_evap() {} 

  //! Sets up the parameters
  void initiate(double e_in, double Theta, double U);

  //! Copies a list
  void copy( one_d_evap& list_2 );

  //! Inserts the links for the extrema
  void extrema();

  //! Interpolates this list between left_list and right_list.
  void list_interp(double e_in, one_d_evap& left_list,
    one_d_evap& right_list);
};

// -------------------- class one_d_Maxwell ------------------
class one_d_Maxwell : public distrib_base
//! Used for the Maxwell model at one incident energy.
{
private:
  double U_;
  double Theta_;

public:

  double f(double E);

  one_d_Maxwell();
  ~one_d_Maxwell() {} 

  //! Sets up the parameters
  void initiate(double e_in, double Theta, double U);

  //! Inserts the links for the extrema
  void extrema();

  //! Copies a list
  void copy( one_d_Maxwell& list_2 );

  //! Interpolates this list between left_list and right_list.
  void list_interp(double e_in,
    one_d_Maxwell& left_list, one_d_Maxwell& right_list);
};

// -------------------- class one_d_Watt ------------------
class one_d_Watt : public distrib_base
//! Used for the Watt model at one incident neutron energy.
{
private:
  double a_;
  double b_;
  double U_;

public:
  // interp_type is used for interpolation of the parameter a
  // We also need to store the type of interpolation for b
  int interp_type_b;

  double f(double E);

  one_d_Watt();
  ~one_d_Watt() {} 

  //! Sets parameters E_in, U, and a
  void set_EUa( double e_in, double U, double a );

  //! Sets up a list for b between 2 incident energies
  void intermediate_b(double e_in, double U, double a, int a_interp,
    double b, int b_interp);

  //! Sets up a list for b at an incident energy
  void same_b( double b, int b_interp );

  //! Fills out the list
  void initiate();

  // Put in a link at the maximum.  It occurs
  // at the solution of beta*tanh(beta) = a*b/2 if
  // beta = sqrt(b*E').
  //! Puts in a link at the maximum
  void extrema();

  //! Copies a list
  void copy( one_d_Watt& list_2 );

 //! Interpolates this list between left_list and right_list
  void list_interp(double e_in,
    one_d_Watt& left_list, one_d_Watt& right_list);
};

//! Used to locate the maximum of the Watt function
double Watt_max(double beta, double dummy);

// -------------------- class one_d_Madland ------------------
class one_d_Madland : public distrib_base
//! Used for the Madland-Nix model at one incident neutron energy.
{
private:
  double EFL_;  // average energy of light fragment
  double EFH_;  // average energy of heavy fragment
  double TM_;   // Madland-Nix parameter
  double E_max;
  double g( double E_out, double E_frag );  // used by f
  double Gamma_a;  // Gamma(1.5)

public:

  double f(double E)
  {
    return 0.5*( g(E, EFL_) + g(E, EFH_));
  }

  one_d_Madland();
  ~one_d_Madland() {} 

  //! Sets up the parameters
  void initiate(double e_in, double EFL, double EFH, double TM);

  //! Copies a list
  void copy( one_d_Madland& list_2 );

 // Interpolates this list between left_list and right_list
  void list_interp(double e_in,
    one_d_Madland& left_list, one_d_Madland& right_list);
};

// -------------------- class one_d_Legendre ---------------
class one_d_Legendre : public distrib_base
//! Used for the Legendre angular distribution at one incident energy.
{
public:
  Param coef;  // the Legendre coefficients

  double f(double mu);

  one_d_Legendre():coef(1,0.0) {}
  ~one_d_Legendre() {} 

  //! makes a 1-d list for one incident energy
  void expand();

  //! Copies a list
  void copy( one_d_Legendre& list_2 );

  //! Interpolates this list between left_list and right_list
  // It is expected this is lin-log interpolation.
  void list_interp(double e_in,
    one_d_Legendre& left_list, one_d_Legendre& right_list );
};

// -------------------- class one_d_isotropic ---------------
class one_d_isotropic : public distrib_base
//! Used for the isotropic angular distribution at one incident energy.
{
public:

  double f(double E)
  {
    SevereError("one_d_isotropic::f","Write this function!");
    return 0.0;
  }

  one_d_isotropic() {}
  ~one_d_isotropic() {}

 //! sets up a list
  void initiate(double e_in);

  //! Copies a list
  void copy( one_d_isotropic& list_2 );
};

// -------------------- class one_d_hat ------------------
class one_d_hat : public distrib_base
//! This class is a hat function energy distribution, and is used for debugging.
{
private:
  // The model parameters are scaled to unit base: 0 <= E' <= 1
  double left_scale;  // location of the left end of the hat
  double top_scale;   // location of the top of the hat
  double right_scale; // location of the right end of the hat

public:

  double f(double E_out);

  one_d_hat() {}
  ~one_d_hat() {} 

  //! Sets up the parameters and makes the list
  void initiate(double left_end, double top, double right_end);

  //! Performs the interpolation
  void list_interp(double e_in, one_d_hat& left_list,
    one_d_hat& right_list)
  {
    SevereError("one_d_hats::list_interp",
        "Write code to interpolate between one_d_hats?");
  }
};

#endif
