/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1849 $
 * $Date: 2006-04-21 16:08:09 -0700 (Fri, 21 Apr 2006) $
 * $Author: hedstrom $
 * $Id: distrib_base.cpp 1849 2006-04-21 23:08:09Z hedstrom $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the distrib_base class

#include "distrib_base.hpp"
#include "messaging.hpp"
#include "endl_formats.hpp"
#include "endl_precision.hpp"

extern ENDLClass ENDL;

// *********** distrib_base class ***********************
// ----------- distrib_base::print -----------------
//! to print the list
void distrib_base::print()
{
  cout.setf(ios::scientific,ios::floatfield);

  cout << "E_in: " << ENDL.data( E_in() ) << endl;
  cout << "weight: " << weight << endl;
  dd_list::print();
}
// ----------- distrib_base::thicken -----------------
//! Expand the entire list to achieve the tolerance
void distrib_base::thicken( )
{
  // point to the the last true node
  dd_list::iterator last = end();
  --last;
  thicken( begin(), last );
}

// ----------- distrib_base::thicken -----------------
//! add links between first and last until linear interpolation
//! is good to within tol_1d
void distrib_base::thicken( dd_list::iterator first,
  dd_list::iterator last )
{
  int Max_List_Size = static_cast<int>( Global.Value("max_list_size") );
  double abs_tol;
  double biggest = 0.0;

  // set the noise level
  dd_list::iterator this_link;
  for(this_link = first; this_link != last; ++this_link)
  {
    if(abs(this_link->y) > biggest)
    {
      biggest = abs(this_link->y);
    }
  }
  // they could be all zeros
  if( biggest == 0.0 )
  {
    return;
  }
  abs_tol = cut_off*biggest;

  double E_mid;
  double p_mid;
  double true_p;

  dd_link new_link;  // for creating new links

  // *** In this version we test only the accuracy of the    ***
  // *** linear interpolation at the midpoint energy between ***
  // *** two consecutive links.                              ***

  // loop through the list
  for(this_link = first; this_link != last; ++this_link)
  {
    dd_list::iterator next_link = this_link;
    ++next_link;
    // keep checking until the interval passes the test
    for(;;)
    {
      // get the local energy epsilon (smallest dE can see in output file)
      double local_eps = ENDL_EPSILON( this_link->E_out() );
      if( next_link->E_out() - this_link->E_out() < local_eps )
      {
        break;
      }

      // the midpoint energy
      E_mid = 0.5*(this_link->E_out() + next_link->E_out());

      // the linear interpolation
      dd_link test_link(this_link->x, this_link->y);
      p_mid = linlin_interp(E_mid, test_link, *next_link);
      // the true probability
      true_p = f(E_mid);

      // test against relative and absolute errors
      if(abs(p_mid - true_p) < tol_1d*abs(true_p) ||
         abs(p_mid - true_p) < abs_tol  )
      {
        break;
      }
      else
      {
        // make a new link and insert it before next_link
        new_link = dd_link(E_mid, true_p);
        insert(next_link, new_link);
        --next_link;  // the new link is now the next link
      }
      // *** a safely check ***
      if( size() >= Max_List_Size )
      {
        print();
        SevereError("distrib_base::thicken",pastenum("got ",
             Max_List_Size)+" links in 1-d thicken routine");
      }
    }
  }
}

// ----------- distrib_base::get_norm -----------------
//! get the probability for this list.
//! This routine is used for double differential data
double distrib_base::get_norm()
{
  dd_list::iterator link = begin();  // the first link

  double norm = 0.0;
  double last_E = link->E_out();
  double last_P = link->y;
  double next_E;
  double next_P;

  // catch bad calls to routine
  if (empty()) 
  {
    Warning("distrib_base::get_norm",
      "Can't get norm with no elements in distrib_base!");
    return 0.0;
  }
  
  // if there is only one link, the norm is easy to do, in a sense
  if (size()==1) 
  {
    Warning("distrib_base::get_norm", "Only one element in distrib_base!");
    return link->y;
  }
  
  // loop through the list
  for(++link; link != end(); ++link)
  {
    // get the next energy and probability
    next_E = link->E_out();
    next_P = link->y;

    // accumulate the probability
    norm += 0.5*(next_P + last_P)*(next_E - last_E);

    // save the values for the next interval
    last_E = next_E;
    last_P = next_P;
  }

  return norm;
}

// ----------- distrib_base::renorm -----------------
//! ensure that the total probability is 1
void distrib_base::renorm()
{
  double norm = get_norm();
  dd_list::iterator link = begin();

  // error check
  if(norm == 0.0)
  {
    Warning("distrib_base::renorm",pastenum("Zero norm in renorm routine for E_in: ",E_in( )));
    // fudge something
    link->y = 1.0;
    max_P = 1.0;
  }
  else
  {
    // Now, renormalize and find the maximum y
    max_P = 0.0;
    for( ; link != end(); ++link)
    {
      link->y /= norm;
      if( link->y > max_P )
      {
        max_P = link->y;
      }
    }
  }
}

// ----------- distrib_base::renorm -----------------
//! ensure that the total probability is 1, with the norm given
void distrib_base::renorm(double Norm)
{
  // error check
  if(Norm == 0.0)
  {
    SevereError("distrib_base::renorm(double Norm)","Zero norm in renorm(double Norm) routine");
  }
  // Now, normalize
  *this *= 1.0/Norm;
}

// ----------- distrib_base::get_bins -----------------
//! make the equally probable energy bins
void distrib_base::get_bins()
{
  // for double-differential data the norm may not be 1
  double Norm = get_norm( );
  if( Norm <= 0.0 )
  {
    print( );
    SevereError("distrib_base::get_bins","zero norm");
  }
  double d_prob = Norm/num_bins;
  double cum_prob = 0.0;

  double prev_x = begin()->x;
  double prev_y = begin()->y;
  equi_prob.push_back( prev_x );  // [0]-th entry

  dd_list::iterator link = begin();
  ++link;
  double next_x = link->x;
  double next_y = link->y;

  // find the j-th bin
  for(int j = 1; j < num_bins; ++j)
  {
    double target = j*d_prob;
    double trapezoid = 0.5*(next_y + prev_y)*(next_x - prev_x);
    while(cum_prob + trapezoid < target)
    {
      cum_prob += trapezoid;
      prev_x = next_x;
      prev_y = next_y;
      ++link;
      if(link == end())
      {
        SevereError("distrib_base::get_bins","failure in get_bins");
      }
      next_x = link->x;
      next_y = link->y;
      trapezoid = 0.5*(next_y + prev_y)*(next_x - prev_x);
    }
    // solve the quadratic equation for x:
    //   target = cum_prob + 
    //      {integral from prev_x to x}{prev_y + slope*(u - prev_x)} du
    // with slope = (next_y - prev_y)/(next_x - prev_x)
    // {actually, it's better to solve for 1/(x - prev_x) and invert}
    double slope = (next_y - prev_y)/(next_x - prev_x);
    equi_prob.push_back( prev_x + 2*( target - cum_prob )/
      ( prev_y + sqrt( prev_y*prev_y + 2*slope*( target - cum_prob ) ) ) );
  }
  // the last bin boundary
  link = end();
  --link;
  equi_prob.push_back( link->x );
}

// ----------- distrib_base::mirror -----------------
//! Reverse the order of the links with x replaced by -x.
//! This routine is for angular distributions of the residual
//! in elastic scattering (mu, probability) -> (-mu, probability).
void distrib_base::mirror()
{
  // first, reverse the list
  reverse();

  // replace mu by -mu
  for(dd_list::iterator link = begin(); link != end(); ++link)
  {
    link->mu() *=-1.0;  // Mirror the "x" axis value about zero
  }
}

// ----------- distrib_base::widen_delta -----------------
//! This routine widens the delta functions
void distrib_base::widen_delta( )
{
  dd_list::iterator next_link;
  dd_list::iterator prev_link;
  dd_list::iterator link = begin();

  // for a single line with multiplicity 0 we need to keep the line
  if( ( size( ) == 1 ) && ( begin()->y == 0.0 ) ){
    begin()->y = 1.0;
  }

  // loop through the discrete lines
  for(; link != end( ); link = next_link)
  {
    next_link = link;
    ++next_link;
    prev_link = link;
    --prev_link;

    // do not widen null delta-functions
    if( link->y > 0.0)
    {
      // what is the separation between states?
      double dE_L = ( link == begin( ) ) ? link->E_out() :
        link->E_out() - prev_link->E_out();

      // set width of delta function (basically smallest dE we
      // can see in output file)
      double dE = 0.5*DELTA_WIDTH( link->E_out() );

      if( dE_L <= dE )
      {
        SevereError("distrib_base::widen_delta", 
            pastenum("Spikes out of order left at ",link->E_out()));
      }

      // insert the left base of the triangle
      dd_link new_link = dd_link(link->E_out() - dE, 0.0);
      insert(link, new_link);
      ++prev_link;

      double dE_R = ( next_link == end( ) ) ? link->E_out( ) :
        next_link->E_out( ) - link->E_out( );

      if( dE_R <= 0.0 )
      {
        SevereError("distrib_base::widen_delta", 
            pastenum("Spikes out of order right at ",link->E_out()));
      }

      // check for indistinguishability of lines
      while( ( dE_R <= 3*dE ) && ( next_link != end( ) ) )
      {
        // combine the lines
        next_link->y += link->y;
        erase( link );
        link = next_link;
        ++next_link;
        if( next_link == end( ) )
        {
          break;
        }
        else
        {
          dE_R = next_link->E_out( ) - link->E_out( );
        }
      }

      // insert the base of the triangle
      new_link = dd_link(link->E_out() + dE, 0.0);
      insert(next_link, new_link);

      // get the width of the triangle
      dE = link->E_out( ) + dE - prev_link->E_out( );

      // scale it
      link->y /= dE/2;

    }
  }
}
// ----------- distrib_base::list_interp -----------------
//! Interpolate this list between left_list and right_list
//! using the equiprobable bins.
void distrib_base::list_interp(double e_in, distrib_base& left_list,
  distrib_base& right_list)
{
  Warning("","You are using distrib_base::list_interp, not a good idea");
  if(!empty())
  {
    SevereError("distrib_base::list_interp",
        "Trying to fill a distrib_base which has already been made");
  }
  double alpha = (e_in - left_list.E_in())/
                 (right_list.E_in() - left_list.E_in());
  if((alpha <= 0.0) || (alpha >= 1.0))
  {
    SevereError("distrib_base::list_interp","Attempt to extrapolate");
  }
  E_in() = e_in;

  // fill the equiprob able bins
  for(int j = 0; j <= num_bins; ++j)
  {
    equi_prob.push_back( (1.0 - alpha)*left_list.equi_prob[j] +
      alpha*right_list.equi_prob[j] );
  }
  // now make the intermediate list (piecewise constant)
  double dp = 1.0/num_bins;
  for(int j = 0; j < num_bins; ++j)
  {
    double dE = equi_prob[j+1] - equi_prob[j];
    if(dE == 0.0) SevereError("distrib_base::list_interp"," dE = 0");
    dd_link new_link = dd_link(equi_prob[j], dp/dE);
    insert(end(), new_link);
    new_link.x = equi_prob[j+1];
    insert(end(), new_link);
  }
  widen_jumps();
}
// ----------- distrib_base::check_equiprob -----------------
//! check the accuracy of 2d linear interpolation based on the
//! equally probable bins
bool distrib_base::check_equiprob(distrib_base& left_list,
  distrib_base& right_list, double tol_2d)
{
  bool is_OK = true;
  // Use the maximum energy as a scale for the relative error.
  double E_max = equi_prob[num_bins];

  // location of this list relative to the others
  double alpha = (E_in() - left_list.E_in())/
    (right_list.E_in() - left_list.E_in());
  if((alpha <= 0.0) || (alpha >= 1.0))
  {
    SevereError("distrib_base::check_equiprob","bad interpolation encountered");
  }
  for(int j = 0; j <= num_bins; ++j)
  {
    // the average
    double average = (1.0 - alpha)*left_list.equi_prob[j] +
      alpha*right_list.equi_prob[j];
    if(abs(equi_prob[j] - average) > tol_2d*E_max)
    {
      is_OK = false;
    }
  }
  return is_OK;
}
// ----------- distrib_base::check_interp_2d -----------------
//! This routine was once used to check unit-based interpolation.
//! check the accuracy of 2d linear interpolation
//! We check the values at the links because they were interpolated
//! We also check half way between the links.
//! These checks are to within tol_2d; if they fail, then thicken and
//! thin this list to within tol_1d
//! But values below cut_off_2d times the maximum are ignored
bool distrib_base::check_interp_2d(double tol_2d, double cut_off_2d)
{
  SevereError("distrib_base::check_interp_2d",
    "Write code in distrib_base for checking unit-based interpolation");
  return false;
  /*  ***  What follows is some old code ***
   *  bool is_OK = true;
   *  double noise = 0.0;
   *  double E_out;
   *  dd_link good_link;
   *  dd_list::iterator left_link;
   *  dd_list::iterator next_link;

   *  //< set up a one_d_distrib of good values
   *  dd_list good_list();

   *  if(debug_on)
   *  {
   *    cout<< "Interpolated:\n";
   *    print();
   *  }

   *  //< make the good intermediate list
   *  left_link = begin();
   *  next_link = left_link;
   *  ++next_link;
   *  for(; next_link != end(); ++left_link, ++next_link)
   *  {
   *    // for the link at the node
   *    E_out = left_link->E_out();
   *    good_link.x = E_out;
   *    good_link.y = f(E_out);
   *    if(abs(good_link.y) > noise)
   *    {
   *      noise = good_link.y;
   *    }
   *    good_list.insert(good_list.end(), good_link);

   *    // for the link in the middle
   *    E_out = 0.5*(E_out + next_link->E_out());
   *    good_link.x = E_out;
   *    good_link.y = f(E_out);
   *    if(abs(good_link.y) > noise)
   *    {
   *      noise = good_link.y;
   *    }
   *    good_list.insert(good_list.end(), good_link);
   *  }

   *  // the last good link
   *  E_out = left_link->E_out();
   *  good_link.x = E_out;
   *  good_link.y = f(E_out);
   *  if(abs(good_link.y) > noise)
   *  {
   *    noise = good_link.y;
   *  }
   *  good_list.insert(good_list.end(), good_link);

   *  if(debug_on)
   *  {
   *    cout << "Good intermediate values\n";
   *    good_list.print();
   *  }

   *  // make the total probability be 1
   *  good_list.renorm();

   *  if(debug_on)
   *  {
   *    cout << "renormed intermediate values\n";
   *    good_list.print();
   *  }

   *  // set the level below which accuracy is ignored
   *  noise *= cut_off_2d;

   *  // check the midpoint values
   *  dd_list::iterator good_ptr = good_list.begin();
   *  left_link = begin();
   *  next_link = left_link;
   *  ++next_link;
   *  for( ; next_link != end();
   *    ++left_link, ++next_link, ++good_ptr)
   *  {
   *    // test the left link
   *    if((abs(good_ptr->y - left_link->y) > tol_2d*good_ptr->y) &&
   *       (abs(good_ptr->y) >= noise))
   *    {
   *      is_OK = false;
   *    }

   *    // go on to the middle values
   *    ++good_ptr;

   *    // are we still OK?
   *    is_OK &= check_interp(left_link, good_ptr,
   *      next_link, tol_2d, noise);
   *  }
   *  // test the last link
   *  is_OK &= (abs(good_ptr->y - left_link->y) <=
   *               tol_2d*abs(good_ptr->y)) ||
   *           (abs(good_ptr->y) <= noise);

   *  if(!is_OK)
   *  {
   *    // get the unnormalized values again
   *    for(left_link = begin(); left_link != end();
   *      ++left_link)
   *    {
   *      left_link->y = f(left_link->E_out());
   *    }
   *    thicken();
   *    //    cout << "mid_list:" << endl;
   *    //    print();
   *    thinit();
   *    renorm();
   *  }
   *  return is_OK;
   */
}
