/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1887 $
 * $Date: 2006-08-23 11:34:44 -0700 (Wed, 23 Aug 2006) $
 * $Author: dbrown $
 * $Id: endl_formats.cpp 1887 2006-08-23 18:34:44Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// Implementation of writing routines
#include "endl_formats.hpp"
#include "global_params.hpp"
#include "messaging.hpp"

extern bdflsClass bdfls;
extern GlobalParameterClass Global;
extern map< MTCPair, YoList > yoMap;

#define ENDL_FIELD_WIDTH     ENDL_DATA_PRECISION+7

// --------------- constructor -------------------
ENDLClass::ENDLClass()
{
  //Set some defaults
  incident_particle = 0;
  twelve = ENDL_FIELD_WIDTH;
  excitation_level = 0;
  atomic_weight = 0.0;
  lifetime = 0.0;
  temp = 0.0;
  zero = 0.0;
  C=0;I=0;S=0;T=0;LR=0;
  X[0]=0.0;  //Typically the Q value
  X[1]=0.0;
  X[2]=0.0;
  X[3]=0.0;
  X[4]=0.0;
  intermediate_C = 0;

  //Create the ENDL end_of_file string - 71 spaces then a 1
  string s_eof(71,' ');
  string s_one="1";
  eof_line=s_eof+s_one;

  // set up the yoList <-> MT,C map
  fill_yoMap();
}

// --------------- destructor -------------------
ENDLClass::~ENDLClass()
{
}

// --------------- global -------------------
// The global function is called to convert the global ENDF data
// used by ENDL into acceptable units.
// This routine should be called only once per evaluation.
void ENDLClass::global(int eZA, double eAWR, double eTEMP, double eELIS, int yi, int Date )
{
  //First we set up some conversion factors
  eV2MeV = 1.0e-6;      //eV to MeV conversion
  K2MeV = bdfls.nuclear_constants[10];
  N2AMU = 1.00866492;   //unit neutron to AMU

  ZA=eZA;
  date=Date;
  atomic_weight=eAWR*N2AMU;      // Convert the atomic weight to AMU
  temp=eTEMP*K2MeV;              // Convert the temperature to MeV
  excitation_level=eELIS*eV2MeV; // Convert the excitation level to MeV

  // Set the maximum incident energy
  incident_particle = yi;
  // fix for elastic reactions
  for (int i=0;i<8;++i) {
    yoMap[MTCPair( 2, 10 )].yos[i]=0;
    if (i==incident_particle) yoMap[MTCPair( 2, 10 )].yos[i]=1;
  }

  switch ( incident_particle )
  {
    case 1 : //incident neutron
      Max_E_in = Global.Value( "Max_n_energy" );
      break;
    case 2 :
      Max_E_in = Global.Value( "Max_cp_energy" );
      break;
    case 3 :
      Max_E_in = Global.Value( "Max_cp_energy" );
      break;
    case 4 :
      Max_E_in = Global.Value( "Max_cp_energy" );
      break;
    case 5 :
      Max_E_in = Global.Value( "Max_cp_energy" );
      break;
    case 6 :
      Max_E_in = Global.Value( "Max_cp_energy" );
      break;
    case 7 : //incident gamma, but might also be decay data 
      Max_E_in = Global.Value( "Max_g_energy" );
      break;
    default:
      SevereError("ENDLClass::global","Invalid incoming particle!");
  }      

  //Now we get the target mass and life
  mass_life_list::iterator MassLifePtr;  
  bool found = bdfls.Mass_Life.at(ZA, MassLifePtr);
  if(!found || (MassLifePtr->Mass() < 0.0) || (MassLifePtr->LifeTime() < 0.0))
  {
    FatalError("ENDLClass::global",
        pastenum("missing mass or lifetime for ZA: ",ZA));
  }
  lifetime = MassLifePtr->LifeTime();
  target_mass = MassLifePtr->Mass();

  //Now we get the mass of the incident particle
  found = bdfls.Mass_Life.at( yi, MassLifePtr );
  if( !found )
  {
    FatalError( "ENDLClass::global",
        pastenum( "missing mass or lifetime for yi: ", yi ) );
  }
  projectile_mass = MassLifePtr->Mass( );
  if( ( yi > 1 ) && ( yi < 7 ) )
  {
    Warning( "ENDLClass::global",
	     "do you want to subtract off the mass of the electrons?" );
  }
}

// ---------------------- reaction -------------------------
//! The reaction function is called to convert the ENDF data
//! used by ENDL for each reaction into acceptable units.
//! This routine should be called only once per reaction product, 
//! during the processing of the MF=3 file.
void ENDLClass::reaction(int lr, double eQM, double eQI)
{
  // The cross section file gives the Q values
  if ( F != 3 ) Warning("ENDLClass::reaction","Should only be called when reading MF=3 files");
  
  //Convert the T number to ENDL C and S numbers
  set_c_number( );
  set_LR( lr );
  set_s_number( );
  if ( LR > 0 ) adjust_c_number_for_breakup();

  //Determine whether to write an ENDL file or not...
  write_file = true ;
  if ( C == -1 ) //Flag to turn off ENDL file writing
  {
    Info("ENDLClass::reaction","Reaction has no ENDL equivalent, will not write an ENDL file");
    write_file = false ;
    return;
  }
  else if ( C < -1 ) //More serious flag to turn off ENDL file writing
  {
    write_file = false ;
    Warning("ENDLClass::reaction",
      pastenum("Illegal C number :",C)+", will not write an ENDL file");
    return;
  }

  //Convert the Q value to MeV
  QM=eQM*eV2MeV;
  QI=eQI*eV2MeV;
  //For cross sections there is no outgoing particle
  outgoing_particle = 0;

  set_QValue();

  if( threshold > Max_E_in ) {
    Info( "ENDLClass::reaction", 
        pastenum( "Threshold above Max_energy, will not write an ENDL file: threshold=", threshold )+
        pastenum( ", Max_E_in=", Max_E_in ) );
    write_file = false;
  }
}

// -------------------- set_outgoing_ZA ---------------------
//! set the outgoing particle using its ZA
void ENDLClass::set_outgoing_ZA( int eZAP )
{
  set_yo( za_to_yo( eZAP ) );
}

// -------------------- za_to_yo ---------------------
//! Get the particle number corresponding to given ZA
int ENDLClass::za_to_yo( int za )
{
  switch ( za )
  {
    case 1:
      return 1;
    case 1001:
      return 2;
    case 1002:
      return 3;
    case 1003:
      return 4;
    case 2003:
      return 5;
    case 2004:
      return 6;
    case 0:
      return 7;
    default:
      if (za<0) Warning("ENDLClass::za_to_yo",pastenum("Illegal residual ZA: ",za));
      Info("ENDLClass::za_to_yo",pastenum("Data for the residual ",za));
      write_file = false;
  }
  return 99;
}

// -------------------- yo_to_za ---------------------
//! Get the ZA corresponding to particle number yo
int ENDLClass::yo_to_za( int yo )
{
  switch( yo )
  {
    case 1:
      return 1;
    case 2:
      return 1001;
    case 3:
      return 1002;
    case 4:
      return 1003;
    case 5:
      return 2003;
    case 6:
      return 2004;
    case 7:
      return 0;
    default:
      SevereError("ENDLClass::yo_to_za",pastenum("no ZA for yo: ", yo));
  }
  return -1;// should never get here
}

// -------------------- get_ZAP ---------------------
//! Identify the outgoing particle from the mt number
int ENDLClass::get_ZAP( int mt )
{
  //The outgoing particles is almost always a neutron, except in a few
  //cases
  if      (( mt>=600 && mt<=649 ) || mt == 103)
  {
    return 1001;  //proton
  }
  else if (( mt>=650 && mt<=699 ) || mt == 104)
  {
    return 1002;  //deuteron
  }
  else if (( mt>=700 && mt<=749 ) || mt == 105)
  {
    return 1003;  //triton
  }
  else if (( mt>=750 && mt<=799 ) || mt == 106)
  {
    return 2003;  //helium-3
  } 
  else if (( mt>=800 && mt<=849 ) || mt == 107)
  {
    return 2004;  //alpha
  }
  else
  {
    return 1;     //neutron
  }
}

// -------------------- set_yo ---------------------
void ENDLClass::set_yo( int yo )
{
  outgoing_particle = yo;

  // the maximum outgoing particle energy
  switch ( yo ) {
    case 0 :
      Max_E_out = Max_E_in;
      break;
    case 1 :
      Max_E_out = Global.Value( "Max_n_energy" );
      break;
    case 2 :
      Max_E_out = Global.Value( "Max_cp_energy" );
      break;
    case 3 :
      Max_E_out = Global.Value( "Max_cp_energy" );
      break;
    case 4 :
      Max_E_out = Global.Value( "Max_cp_energy" );
      break;
    case 5 :
      Max_E_out = Global.Value( "Max_cp_energy" );
      break;
    case 6 :
      Max_E_out = Global.Value( "Max_cp_energy" );
      break;
    case 7 :
      Max_E_out = Global.Value( "Max_g_energy" );
      break;
    case 99 :
      // this is the phony yo for the residual
      break;
    default:
      SevereError("ENDLClass::set_yo","Invalid outgoing particle!");
  }      
}

// -------------------- get_resid ---------------------
//! Calculates the residual ZA and the compound nucleus
int ENDLClass::get_resid( )
{
  int ZA_in = yo_to_za( incident_particle );
  current_za = ZA + ZA_in;

  // special for charged-particle scattering
  if( C == 9 )
  {
    res_za = ZA;
  }
  else
  {
    // initialize
    res_za = current_za;
    // subtract off the outgoing particle ZAs
    for( int yout = 1; yout < 7; ++yout )
    {
      int ZA_out = yo_to_za( yout );
      res_za -= ZA_out * yoMap[ MTCPair(T,C) ][ yout ];
    }
  }
  return res_za;
}
  
// -------------------- set_I_number ---------------------
void ENDLClass::set_I_number( int eI )
{
  I=eI;

  //We can generate the ENDL file name
  file_name=endl_file_name(outgoing_particle,C,I,S);

  //Now we can build the header lines
  ostringstream os_1;

  os_1.setf(ios::right);
  os_1.setf(ios::scientific);
  os_1<<
    setw(6)<<ZA<<
    setw(3)<<incident_particle<<
    setw(3)<<outgoing_particle<<
    setw(13)<<setprecision(5)<<atomic_weight<<
    setw(7)<<date<<
    setw(3)<<ENDL_FIELD_WIDTH<<
    setw(13)<<setprecision(5)<<excitation_level<<
    setw(13)<<setprecision(5)<<lifetime<<
    setw(13)<<setprecision(5)<<temp;  

  header_line_1=os_1.str(); // Copy output string stream to normal string

  //remove the "e" of the exponent
  for( int i=0; i<header_line_1.length(); i++ )
    {
      if ( header_line_1.at(i)=='e' )
	{				     
	  header_line_1.erase(i,1) ;
	}
    }
  
  ostringstream os_2;

  os_2.setf(ios::right);
  os_2.setf(ios::scientific);
  os_2<<
    setw(2)<<C<<
    setw(3)<<I<<
    setw(3)<<S ;
  //The format is dependant on the sign of the number.
  //We only consider the issue for X[0] since all other
  //header values SHOULD be positive.
  if ( X[0] < 0.0 ) os_2<<setw(13)<<setprecision(4)<<X[0];
  if ( X[0] >=0.0 ) os_2<<setw(13)<<setprecision(5)<<X[0];
  if ( X[1] < 0.0 ) os_2<<setw(13)<<setprecision(4)<<X[1];
  if ( X[1] >=0.0 ) os_2<<setw(13)<<setprecision(5)<<X[1];
  if ( X[2] < 0.0 ) os_2<<setw(13)<<setprecision(4)<<X[2];
  if ( X[2] >=0.0 ) os_2<<setw(13)<<setprecision(5)<<X[2];
  if ( X[3] < 0.0 ) os_2<<setw(13)<<setprecision(4)<<X[3];
  if ( X[3] >=0.0 ) os_2<<setw(13)<<setprecision(5)<<X[3];
  if ( X[4] < 0.0 ) os_2<<setw(13)<<setprecision(4)<<X[4];
  if ( X[4] >=0.0 ) os_2<<setw(13)<<setprecision(5)<<X[4];

  header_line_2=os_2.str(); // Copy output string stream to normal string

  //remove the "e" of the exponent
  for( int i=0; i<header_line_2.length(); i++ )
    {
      if ( header_line_2.at(i)=='e' )
	{
	  header_line_2.erase(i,1) ;
	}
    }

}
// -------------------- new_file -------------------------------
bool ENDLClass::new_file( )
{
  return !( append || S==9 || S==1 );
}

// ------------------------- set_s_number ---------------------------
void ENDLClass::set_s_number( )
{
  if ( ( LR > 0 ) && ( LR != 31 ) ) S = 9;
  else
  {
    if ( T>=50  && T<91  || //yi,nprime cross sections to specific states
         T>=600 && T<649 || //yi,p cross sections to specific states
         T>=650 && T<699 || //yi,d cross sections to specific states
         T>=700 && T<749 || //yi,t cross sections to specific states
         T>=750 && T<799 || //yi,3He cross sections to specific states
         T>=800 && T<849  ) //yi,alpha cross sections to specific states
    {
      S = 1;
    }
    else         
    {
      S = 0;
    }
  }
}

// ------------------------- set_QValue ---------------------------
void ENDLClass::set_QValue( )
{
  // ENDL uses mass difference Q value of target's ground state, 
  // ENDF uses mass difference Q value of target's actual state
  double effQM = QM-excitation_level;
  X[2]=0; 
  X[3]=0;
  switch ( S )
  {
    case 0:
      X[0]=effQM;
      X[1]=0;
      break;
    case 1:
      X[0]=effQM;
      X[1]=QM - QI;
      break;
    case 9:
      X[0]=effQM;
      X[1]=QM - QI;
      X[2]=intermediate_C; 
      break;
    default:
      X[0]=effQM;
      X[1]=0;
      break;
  }
  // some reactions should never have thresholds, no matter what the evaluator
  // may have put down
  if ( C == 1 || C == 10 ) 
  {
    X[0] = 0.0;
    X[1] = 0.0;
  }
  
  // Now set the threshold.  This might be overridden in later processing because the 
  // data may start at a higher energy.  This is our first, best, guess.
  threshold = max( 0.0, - effQM * ( 1.0 + projectile_mass/target_mass ) );
  
  return;
}
// ----------------------- set_c_number ------------------------
void ENDLClass::set_c_number( )
{
  for ( map< MTCPair, YoList >::iterator it = yoMap.begin(); it != yoMap.end(); ++it )
  {
    if ( it->first.MT == T )
    {
      C = it->first.C;
      if ( C < 0 ) Info("ENDLClass::set_c_number",
        pastenum("MT = ",T)+" is for "+it->second.reactionName+", so not translating it");
      if ( ( C == 11 ) && ( Global.Value("sum_inelastic") != 0.0 ) ) {
        Warning("ENDLClass::set_c_number","Using only total inelastic");
        C = -1; //Do not write this out - inelastic cross section
      }
      return;
    }   
  }

  // Oops, if we get here, we've got a problem
  C = -42; // make sure we don't process any thing else with this MT!
  Warning("ENDLClass::set_c_number",
        pastenum("MT is not valid ENDF value, MT =",T) );
}

// ----------------------- adjust_c_number_for_breakup ------------------------
void ENDLClass::adjust_c_number_for_breakup()
{
  // save the originally assigned C number of the reaction that resulted in the nucleus that breaks up
  intermediate_C = C;

  YoList correction;
  // Figure out the adjustment to the outgoing particle list
  switch ( LR )
  {
    case 1: // Hafta parse the MF6 file completely in this case
      SevereError( "ENDLClass::adjust_c_number_for_breakup", 
        "Cannot determine C number when LR = 1 for complex breakup reactions");
    case 22: // a
        correction[6]=1;
        break;
    case 23: // 3a
        correction[6]=3;
        break;
    case 24: // na
        correction[1]=1;
        correction[6]=1;
        break;
    case 25: // 2na
        correction[1]=2;
        correction[6]=1;
        break;
    case 28: // p
        correction[2]=1;
        break;
    case 29: // 2a
        correction[6]=2;
        break;
    case 30: // n2a
        correction[1]=1;
        correction[6]=2;
        break;
    case 31:   // g, this is a special case, hidden in the footnote 
        break; // on p. 3.6 of the endf manual
    case 32: // d
        correction[3]=1;
        break;
    case 33: // t
        correction[4]=1;
        break;
    case 34: // 3He
        correction[5]=1;
        break;
    case 35: // d2a
        correction[3]=1;
        correction[6]=2;
        break;
    case 36: // t2a
        correction[4]=1;
        correction[6]=2;
        break;
    case 39: // ic
      Unimplemented("ENDLClass::adjust_c_number_for_breakup",
        pastenum("Cannot process special breakup reactions, LR=",LR));
      return;
    case 40: // ep pair formation
      Unimplemented("ENDLClass::adjust_c_number_for_breakup",
        pastenum("Cannot process special breakup reactions, LR=",LR));
      return;
    case 0:  // no changes
      return;
    default:
      SevereError( "ENDLClass::adjust_c_number_for_breakup",
        pastenum("Cannot determine C number, illegal LR, LR = ",LR) );
      break;
  }
  // our correction becomes the new outgoing partile list.
  correction += yoMap[ MTCPair(T,C) ];
  // figure out it's MT & C number, if any
  for ( map< MTCPair, YoList >::iterator it = yoMap.begin(); it != yoMap.end(); ++it )
  {
    if ( it->second == correction )
    {
      C = it->first.C;
      T = it->first.MT;
      if ( C < 0 ) Info("ENDLClass::adjust_c_number_for_breakup",
        pastenum("MT = ",T)+" is for "+it->second.reactionName+", so not translating it");
      return;
    }
  }
  // oops, we couldn't find it.  We can't translate this data.
  Warning("ENDLClass::adjust_c_number_for_breakup", "Reaction products have no corresponding C or MT");
  C=-1;
}

// ----------------------- 1 data ------------------------
string ENDLClass::data( double d )
{
  string s_data;
  ostringstream os_data;
  os_data.setf(ios::right);
  os_data.setf(ios::scientific);
  os_data<<
    setw(ENDL_FIELD_WIDTH)<<setprecision(ENDL_DATA_PRECISION)<<d;    

  s_data=os_data.str();
  return( s_data );
}
// ----------------------- 2 data ------------------------
string ENDLClass::data( double d1, double d2 )
{
  string s_data;
  ostringstream os_data;
  os_data.setf(ios::right);
  os_data.setf(ios::scientific);
  os_data<<
    setw(ENDL_FIELD_WIDTH)<<setprecision(ENDL_DATA_PRECISION)<<d1<<" "<<
    setw(ENDL_FIELD_WIDTH)<<setprecision(ENDL_DATA_PRECISION)<<d2;    

  s_data=os_data.str();
  return( s_data );
}

// ----------------------- 3 data ------------------------
string ENDLClass::data( double d1, double d2, double d3 )
{
  string s_data = data( d1, d2 );
  ostringstream os_data;
  os_data.setf(ios::right);
  os_data.setf(ios::scientific);
  os_data<<
    s_data<<" "<<
    setw(ENDL_FIELD_WIDTH)<<setprecision(ENDL_DATA_PRECISION)<<d3;

  s_data=os_data.str();
  return( s_data );
}

// ----------------------- 4 data ------------------------
string ENDLClass::data( double d1, double d2, double d3, double d4 )
{
  string s_data = data( d1, d2, d3 );
  ostringstream os_data;
  os_data.setf(ios::right);
  os_data.setf(ios::scientific);
  os_data<<
    s_data<<" "<<
    setw(ENDL_FIELD_WIDTH)<<setprecision(ENDL_DATA_PRECISION)<<d4;    

  s_data=os_data.str();
  return( s_data );
}

// ----------------------- yo_mult ------------------------
//! Function that returns the multiplicity of this outgoing particle.
int ENDLClass::yo_mult( )
{
  return yoMap[ MTCPair(T,C) ][ outgoing_particle ];
}

// ----------------------- get_Z ------------------------
//! get the Z of the incident particle
int ENDLClass::get_Z( int yi )
{
  int Z;
  switch ( incident_particle )
  {
  case 1:
    Z = 0;
    break;
  case 2:
    Z = 1;
    break;
  case 3:
    Z = 1;
    break;
  case 4:
    Z = 1;
    break;
  case 5:
    Z = 2;
    break;
  case 6:
    Z = 2;
    break;
  case 7:
    Z = 0;
    break;
  default:
    FatalError( "ENDLClass::get_Z",
               pastenum( "bad incident particle", yi ) );
  }
  return Z;
}

// ----------------------- fill_yoMap ------------------------
void ENDLClass::fill_yoMap()
{
  if ( !yoMap.empty() ) return; // already filled
  yoMap[MTCPair( 1, 1 )] = YoList( "(n,total)" );
  yoMap[MTCPair( 2, 10 )] = YoList( "(z,elastic)" );
  yoMap[MTCPair( 3, 55 )] = YoList( "(z,non-elastic)" );
  yoMap[MTCPair( 4, -1 )] = YoList( "(z,n), redundant", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 5, -1 )] = YoList( "(z,anything)" );
  yoMap[MTCPair( 10, -1 )] = YoList( "(z,continuum)" );
  yoMap[MTCPair( 11, 32 )] = YoList( "(n,2nd)", 0, 2, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 16, 12 )] = YoList( "(z,2n)", 0, 2, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 17, 13 )] = YoList( "(z,3n)", 0, 3, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 18, 15 )] = YoList( "(z,f)", -1, 0, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 19, -1 )] = YoList( "(n,f) -- $1^{st}$ chance fission.", -1, 0, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 20, -1 )] = YoList( "(n,nf) -- $2^{nd}$ chance fission.", -1, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 21, -1 )] = YoList( "(n,2nf) -- $3^{rd}$ chance fission.", -1, 2, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 22, 26 )] = YoList( "(z,na)", 0, 1, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 23, 36 )] = YoList( "(z,n3a)", 0, 1, 0, 0, 0, 0, 3, 0 );
  yoMap[MTCPair( 24, 33 )] = YoList( "(z,2na)", 0, 2, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 25, 16 )] = YoList( "(z,3na)", 0, 3, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 27, -1 )] = YoList( "(n,abs)" );
  yoMap[MTCPair( 28, 20 )] = YoList( "(z,np)", 0, 1, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 29, 27 )] = YoList( "(z,n2a)", 0, 1, 0, 0, 0, 0, 2, 0 );
  yoMap[MTCPair( 30, -1 )] = YoList( "(z,2n2a)", 0, 2, 0, 0, 0, 0, 2, 0 );
  yoMap[MTCPair( 32, 22 )] = YoList( "(z,nd)", 0, 1, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 33, 24 )] = YoList( "(z,nt)", 0, 1, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 34, 25 )] = YoList( "(z,nH)", 0, 1, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 35, -1 )] = YoList( "(z,nd2a)", 0, 1, 0, 1, 0, 0, 2, 0 );
  yoMap[MTCPair( 36, -1 )] = YoList( "(z,nt2a)", 0, 1, 0, 0, 1, 0, 2, 0 );
  yoMap[MTCPair( 37, 14 )] = YoList( "(z,4n)", 0, 4, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 38, -1 )] = YoList( "(n,3nf) -- $4^{th}$ chance fission.", -1, 3, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 41, 29 )] = YoList( "(z,2np)", 0, 2, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 42, 16 )] = YoList( "(z,3np)", 0, 3, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 44, 17 )] = YoList( "(n,n2p)", 0, 1, 2, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 45, 34 )] = YoList( "(n,npa)", 0, 1, 1, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 50, 11 )] = YoList( "(z,n[0]) -- excitation of g.s.", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 51, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 52, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 53, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 54, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 55, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 56, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 57, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 58, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 59, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 60, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 61, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 62, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 63, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 64, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 65, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 66, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 67, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 68, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 69, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 70, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 71, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 72, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 73, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 74, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 75, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 76, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 77, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 78, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 79, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 80, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 81, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 82, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 83, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 84, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 85, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 86, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 87, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 88, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 89, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 90, 11 )] = YoList( "(z,n[n]) -- excitation of n$^{th}$ excited state", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 91, 11 )] = YoList( "(z,n[c]) -- excitation to continuum", 0, 1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 101, -1 )] = YoList( "(n,disappearance)" );
  yoMap[MTCPair( 102, 46 )] = YoList( "(z,g)", 0, 0, 0, 0, 0, 0, 0, 1 );
  yoMap[MTCPair( 103, 40 )] = YoList( "(z,p)", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 104, 41 )] = YoList( "(z,d)", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 105, 42 )] = YoList( "(z,t)", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 106, 44 )] = YoList( "(z,H)", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 107, 45 )] = YoList( "(z,a)", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 108, 37 )] = YoList( "(z,2a)", 0, 0, 0, 0, 0, 0, 2, 0 );
  yoMap[MTCPair( 109, -1 )] = YoList( "(z,3a)", 0, 0, 0, 0, 0, 0, 3, 0 );
  yoMap[MTCPair( 111, 18 )] = YoList( "(z,2p)", 0, 0, 2, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 112, 48 )] = YoList( "(z,pa)", 0, 0, 1, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 113, 42 )] = YoList( "(z,t2a)", 0, 0, 0, 0, 1, 0, 2, 0 );
  yoMap[MTCPair( 114, -1 )] = YoList( "(z,d2a)", 0, 0, 0, 1, 0, 0, 2, 0 );
  yoMap[MTCPair( 115, 19 )] = YoList( "(z,pd)", 0, 0, 1, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 116, 39 )] = YoList( "(z,pt)", 0, 0, 1, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 117, 47 )] = YoList( "(z,da)", 0, 0, 0, 1, 0, 0, 1, 0 );
  yoMap[MTCPair( 151, -1 )] = YoList( "(n,resonance)" );
  yoMap[MTCPair( 201, -1 )] = YoList( "(z,Xn)", 0, -1, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 202, -1 )] = YoList( "(z,Xg)", 0, 0, 0, 0, 0, 0, 0, -1 );
  yoMap[MTCPair( 203, 50 )] = YoList( "(z,Xp)", 0, 0, -1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 204, 51 )] = YoList( "(z,Xd)", 0, 0, 0, -1, 0, 0, 0, 0 );
  yoMap[MTCPair( 205, 52 )] = YoList( "(z,Xt)", 0, 0, 0, 0, -1, 0, 0, 0 );
  yoMap[MTCPair( 206, 53 )] = YoList( "(z,XH)", 0, 0, 0, 0, 0, -1, 0, 0 );
  yoMap[MTCPair( 207, 54 )] = YoList( "(z,Xa)", 0, 0, 0, 0, 0, 0, -1, 0 );
  yoMap[MTCPair( 208, -1 )] = YoList( "Various meson and antiparticle production $sigma$'s" );
  yoMap[MTCPair( 209, -1 )] = YoList( "Various meson and antiparticle production $sigma$'s" );
  yoMap[MTCPair( 210, -1 )] = YoList( "Various meson and antiparticle production $sigma$'s" );
  yoMap[MTCPair( 211, -1 )] = YoList( "Various meson and antiparticle production $sigma$'s" );
  yoMap[MTCPair( 212, -1 )] = YoList( "Various meson and antiparticle production $sigma$'s" );
  yoMap[MTCPair( 213, -1 )] = YoList( "Various meson and antiparticle production $sigma$'s" );
  yoMap[MTCPair( 214, -1 )] = YoList( "Various meson and antiparticle production $sigma$'s" );
  yoMap[MTCPair( 215, -1 )] = YoList( "Various meson and antiparticle production $sigma$'s" );
  yoMap[MTCPair( 216, -1 )] = YoList( "Various meson and antiparticle production $sigma$'s" );
  yoMap[MTCPair( 217, -1 )] = YoList( "Various meson and antiparticle production $sigma$'s" );
  yoMap[MTCPair( 251, -1 )] = YoList( "Various elastic neutrons scattering parameters." );
  yoMap[MTCPair( 252, -1 )] = YoList( "Various elastic neutrons scattering parameters." );
  yoMap[MTCPair( 253, -1 )] = YoList( "Various elastic neutrons scattering parameters." );
  yoMap[MTCPair( 301, -1 )] = YoList( "Energy release for total and partial $sigma$'s." );
  yoMap[MTCPair( 451, -1 )] = YoList( "Heading or title information, MF=1 only." );
  yoMap[MTCPair( 452, 15 )] = YoList( "(z,f) $bar{nu}$ total, i.e. prompt plus delayed, fission.", -1, 0, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 454, 15 )] = YoList( "(z,f) Independent fission product yields.", -1, 0, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 455, 15 )] = YoList( "(z,f) $bar{nu}$ for delayed fission.", -1, 0, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 456, 15 )] = YoList( "(z,f) $bar{nu}$ for prompt fission.", -1, 0, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 457, -1 )] = YoList( "(z,f) Radioactive decay data.", -1, 0, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 458, 15 )] = YoList( "(z,f) Energy release in fission for incident $n$'s.", -1, 0, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 459, 15 )] = YoList( "(z,f) Cumulative fission product yields.", -1, 0, 0, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 500, -1 )] = YoList( "Total charged particle stopping power." );
  yoMap[MTCPair( 501, 70 )] = YoList( "Total photon interaction $sigma$." );
  yoMap[MTCPair( 502, 71 )] = YoList( "Photon coherent scattering." );
  yoMap[MTCPair( 504, 72 )] = YoList( "Photon incoherent scattering." );
  yoMap[MTCPair( 505, -1 )] = YoList( "Imaginary scattering factor." );
  yoMap[MTCPair( 506, -1 )] = YoList( "Real scattering factor." );
  yoMap[MTCPair( 515, -1 )] = YoList( "Pair production, electron field." );
  yoMap[MTCPair( 516, 74 )] = YoList( "Pair production." );
  yoMap[MTCPair( 517, -1 )] = YoList( "Pair production, nuclear field." );
  yoMap[MTCPair( 522, 73 )] = YoList( "Photoelectric absorption." );
  yoMap[MTCPair( 534, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 535, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 536, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 537, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 538, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 539, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 540, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 541, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 542, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 543, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 544, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 545, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 546, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 547, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 548, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 549, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 550, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 551, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 552, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 553, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 554, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 555, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 556, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 557, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 558, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 559, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 560, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 561, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 562, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 563, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 564, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 565, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 566, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 567, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 568, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 569, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 570, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 571, -1 )] = YoList( "Various subshell photoelectric $sigma$'s." );
  yoMap[MTCPair( 600, 40 )] = YoList( "(z,p[0])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 601, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 602, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 603, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 604, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 605, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 606, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 607, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 608, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 609, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 610, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 611, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 612, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 613, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 614, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 615, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 616, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 617, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 618, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 619, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 620, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 621, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 622, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 623, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 624, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 625, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 626, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 627, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 628, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 629, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 630, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 631, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 632, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 633, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 634, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 635, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 636, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 637, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 638, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 639, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 640, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 641, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 642, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 643, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 644, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 645, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 646, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 647, 40 )] = YoList( "(z,p[n])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 649, 40 )] = YoList( "(z,p[c])", 0, 0, 1, 0, 0, 0, 0, 0 );
  yoMap[MTCPair( 650, 41 )] = YoList( "(z,d[0])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 651, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 652, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 653, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 654, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 655, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 656, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 657, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 658, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 659, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 660, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 661, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 662, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 663, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 664, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 665, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 666, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 667, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 668, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 669, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 670, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 671, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 672, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 673, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 674, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 675, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 676, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 677, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 678, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 679, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 680, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 681, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 682, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 683, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 684, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 685, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 686, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 687, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 688, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 689, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 690, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 691, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 692, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 693, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 694, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 695, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 696, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 697, 41 )] = YoList( "(z,d[n])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 699, 41 )] = YoList( "(z,d[c])", 0, 0, 0, 1, 0, 0, 0, 0 );
  yoMap[MTCPair( 700, 42 )] = YoList( "(z,t[0])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 701, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 702, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 703, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 704, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 705, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 706, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 707, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 708, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 709, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 710, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 711, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 712, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 713, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 714, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 715, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 716, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 717, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 718, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 719, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 720, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 721, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 722, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 723, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 724, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 725, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 726, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 727, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 728, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 729, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 730, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 731, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 732, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 733, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 734, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 735, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 736, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 737, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 738, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 739, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 740, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 741, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 742, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 743, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 744, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 745, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 746, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 747, 42 )] = YoList( "(z,t[n])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 749, 42 )] = YoList( "(z,t[c])", 0, 0, 0, 0, 1, 0, 0, 0 );
  yoMap[MTCPair( 750, 44 )] = YoList( "(z,H[0])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 751, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 752, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 753, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 754, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 755, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 756, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 757, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 758, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 759, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 760, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 761, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 762, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 763, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 764, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 765, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 766, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 767, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 768, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 769, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 770, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 771, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 772, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 773, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 774, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 775, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 776, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 777, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 778, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 779, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 780, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 781, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 782, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 783, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 784, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 785, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 786, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 787, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 788, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 789, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 790, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 791, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 792, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 793, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 794, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 795, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 796, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 797, 44 )] = YoList( "(z,H[n])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 799, 44 )] = YoList( "(z,H[c])", 0, 0, 0, 0, 0, 1, 0, 0 );
  yoMap[MTCPair( 800, 45 )] = YoList( "(z,a[0])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 801, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 802, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 803, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 804, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 805, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 806, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 807, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 808, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 809, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 810, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 811, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 812, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 813, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 814, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 815, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 816, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 817, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 818, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 819, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 820, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 821, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 822, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 823, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 824, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 825, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 826, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 827, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 828, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 829, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 830, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 831, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 832, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 833, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 834, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 835, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 836, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 837, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 838, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 839, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 840, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 841, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 842, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 843, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 844, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 845, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 846, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 847, 45 )] = YoList( "(z,a[n])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 849, 45 )] = YoList( "(z,a[c])", 0, 0, 0, 0, 0, 0, 1, 0 );
  yoMap[MTCPair( 851, -1 )] = YoList( "Lumped reaction covariances." );
}

