/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1833 $
 * $Date: 2006-04-11 16:15:50 -0700 (Tue, 11 Apr 2006) $
 * $Author: hedstrom $
 * $Id: endl_formats.hpp 1833 2006-04-11 23:15:50Z hedstrom $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// routines for writing endl files

#ifndef ENDL_FORMATS
#define ENDL_FORMATS

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <string>
#include <cmath>
#include <vector>
#include <map>
#include <sstream>
#include "bdfls_tools.hpp"
#include "file_names.hpp"

using namespace std;

// --------------------- MTCPair ---------------------
class MTCPair
//! Class for storing MT <-> C number associations.  
//! Note: each MT gets a unique C number, but not necessarily the other way around 
{
public:
    int MT;
    int C;
    MTCPair( int mt, int c ): MT(mt), C(c) {}
    MTCPair( const MTCPair& other ): MT(other.MT), C(other.C) {}
    bool operator==( const MTCPair& other ) const { return MT == other.MT; }
    bool operator<( const MTCPair& other ) const { return MT < other.MT; }
};


// --------------------- YoList ---------------------
class YoList
//! Class for storing the outgoing particle list (not including the residual)
{
public:
    vector<int> yos;
    string reactionName;
    YoList( string rxnName="", bool fission=false, 
        int nN=0, int nP=0, int nD=0, int nT=0, int nH=0, int nA=0, int nG=0):
        reactionName(rxnName), yos(8,0)
        {
            yos[0]=int(fission);
            yos[1]=nN;
            yos[2]=nP;
            yos[3]=nD;
            yos[4]=nT;
            yos[5]=nH;
            yos[6]=nA;
            yos[7]=nG;
        }
    YoList( const YoList& other ): reactionName(other.reactionName), yos(other.yos) {}
    bool operator==( const YoList& other ) const { return yos == other.yos; }
    bool operator<( const YoList& other ) const { return yos < other.yos; }
    int operator[]( int i ) const { return yos[i]; }
    int& operator[]( int i ) { return yos[i]; }
    YoList& operator+=(const YoList& other)
    {
        for(unsigned int i=0;i<yos.size();++i) {yos[i]=yos[i]+other.yos[i];}
        return *this;
    }
    bool fission() { return yos[0]!=0; }
    string str() 
    { 
        stringstream ss; 
        for(unsigned int i=0;i<yos.size();++i) {ss<<yos[i];}
        return ss.str()+" "+reactionName;
    }
};



// --------------------- ENDLClass ---------------------
class ENDLClass
//! Class for writing ENDL files
{
private:

  // to set up the yoList <-> MT,C map
  void fill_yoMap();

  // This map controls the association of yos to Mt & C numbers
  map< MTCPair, YoList > yoMap;

public:
  //Some of the variables are valid for all reactions
  int ZA, incident_particle,date,twelve;
  double atomic_weight, lifetime, temp, zero;
  double projectile_mass, target_mass;
  double Max_E_in;  // maximum incident energy for this particle
  //Some variables are reaction dependent
  int outgoing_particle;
  double Max_E_out;  // maximum energy for outgoing particle
  int res_za;  // to keep track of the residual ZA
  int intermediate_C;  // to keep track of the intermediate C number, for breakup reactions
  int current_za;  // the current compund nucleus
  double excitation_level;
  double threshold;
  int C, S, I, T, F, LR, yi;
  double QM,QI,X[5];
  // Some ENDL angular data is in center-of mass frame, some lab
  bool angles_CM;
  //Conversion constants
  double K2MeV,N2AMU,eV2MeV;
  //ENDL header lines
  string header_line_1, header_line_2, eof_line;
  string file_name;
  bool write_file;
  bool append;
  
  ENDLClass();
  ~ENDLClass();

  //! Converts and stores global data for ENDL.
  void global( int eZA, double eAWR, double eTEMP, double eELIS, int yi, int Date );

  //! Converts reaction data to proper units for ENDL.
  void reaction( int lr, double eQM, double eQI );

  //! Returns the particle number corresponding to given ZA
  int za_to_yo( int za );

  //! Returns the ZA corresponding to the particle number
  int yo_to_za( int yo );

  //! Function to set the ENDL yo identifier.
  void set_yo( int yo );

  //! sets the break-up flag for special reactions
  //! Identifies the outgoing particle from the MT number
  int get_ZAP( int mt );

  //! set the outgoing particle using its ZA
  void set_outgoing_ZA( int eZAP );

  //! Calculates the residual ZA
  int get_resid( );

  //! Determine ENDL s number.
  void set_s_number( );

  //! Hard-wires the ENDL s identifier.
  inline void set_s_number( int es )
  {
    S = es;
  }

  //! Set the Q-value and X1, depending on the S number
  void set_QValue( );

  //! Converts an ENDF reaction identifier (MT), to an ENDL reaction c number.
  void set_c_number( );
  
  //! Function to hard-wire the ENDL c identifier.
  inline void set_c_number( int ec)
  {
    C = ec;
  }

  inline void set_LR( int lr )
  {
    LR=lr;
  }

  //! Reassigns the c number to include the products of residual decay if the residual breaks up
  void adjust_c_number_for_breakup();

  //! Function to set the ENDL X_0 value
  inline void set_x0( double x0)
  {
    X[0] = x0;
  }

  //! Function to set the ENDL X_1 value
  inline void set_x1( double x1)
  {
    X[1] = x1;
  }

  //! Sets the ENDL I number and builds the header lines.
  void set_I_number( int eI );

  //! Tests whether a new ENDL file is needed.
  bool new_file( );

  // Function that writes out the one double
  string data( double d );
  
  //! Function that writes out the two column data in ENDL format.
  string data( double d1, double d2 );
  
  //! Function that writes out the three column data in ENDL format.
  string data( double d1, double d2 , double d3 );
  
  //! Function that writes out the four column data in ENDL format.
  string data( double d1, double d2 , double d3, double d4);

  //! Function that returns the multiplicity of this outgoing particle.
  int yo_mult( );
 
  //! Function that returns the charge of this incident particle.
  int get_Z( int yi );
};

#endif
