/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1890 $
 * $Date: 2006-08-29 11:45:16 -0700 (Tue, 29 Aug 2006) $
 * $Author: dbrown $
 * $Id: fete.cpp 1890 2006-08-29 18:45:16Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

#include <sstream>
#include <string>
#include <vector>
#include <map>
#include <stdio.h>
#include <glob.h>
#include "convert.hpp"
#include "version.hpp"
#include "b6stream.hpp"
#include "bdfls_tools.hpp"
#include "endl_formats.hpp"
#include "load_data.hpp"
#include "global_params.hpp"
#include "messaging.hpp"
#include "mf12classes.hpp"
#include "mf13classes.hpp"
#include "logger.hpp"

#define CLEAN_TEMP_FILES

using namespace std;

#ifdef CLEAN_TEMP_FILES

// --------------- glob ---------------
//! Gizmo to glob for files patching a pattern
vector< string > glob(string pattern){
    glob_t globbuf;
    glob(pattern.c_str(), GLOB_ERR, NULL, &globbuf);
    vector< string > names;
    for (int i=0; i<globbuf.gl_pathc; ++i)
        names.push_back(static_cast<string>( globbuf.gl_pathv[i]) );
    return names;
}

// --------------- clean_files ---------------
//! Cleans files
bool clean_files( string pattern )
{
  bool result=true;
  vector<string> filelist = glob(pattern);
  for( unsigned int i=0; i<filelist.size(); ++i )
  {
    result = result && remove(filelist[i].c_str())!=-1;
  }
  return result;
}

#endif



// --------------- global data ---------------
// We need the global parameters for the translation available globally
GlobalParameterClass Global;

// We need the bdfls data to available globally
bdflsClass bdfls;

// This object holds the reaction details for each MT
// We'll pass members of this map around globally
ENDLClass ENDL;

// We need an (n,z[i]) storage array for cascades
map< int, cascade > cascade_gammas;

// We need to hold the C55 gammas to the end
C55_gammas C55;

// At the end, we write out the evaluation documentation, version info and
// any processing errors.  This holds that data.
MessageLogger messageLog;



// --------------- main ---------------
int main( int argc, char* argv[] )
{
  //Pretty banner to set off start of program & print the version information
  cout << "+" << 40*string("-") << "+"<<endl;
  cout << "|" << center("fete: from ENDF to endl...",40)<< "|"<<endl;
  cout << "|" << center("Version "+string(FETE_VERSION),40) << "|" <<endl;
  cout << "|" << center("Revision "+string(FETE_REVISION),40) << "|" <<endl;
  cout << "|" << center("Last SVN commit "+string(FETE_VERSION_DATE),40) << "|" <<endl;
  cout << "+" << 40*string("-") << "+"<<endl;

#ifdef CLEAN_TEMP_FILES

  if (!clean_files("yo*")) cout<<"\n\nRemoving old yo* files failed"<<endl;

#endif

  //Read in any user requested settings
  string inFile = Global.read_command_line(argc, argv);

  //Print the parameter values used in the translation
  Global.print();

  bdfls.read();

  b6stream Eval(inFile);

  // print out the data for each reaction
  if ( MESSAGELEVEL_INFO )
  {
    Eval.print_MT_MF( );
  }

  // Here is where we loop over the all the reaction files
  for( iMF_array_list::iterator this_mt = Eval.MT_MF_list.begin();
      this_mt != Eval.MT_MF_list.end(); ++this_mt)
  {
    if ( ( this_mt->MT() != 151 ) && ( ( this_mt->MT() < 450 ) || ( this_mt->MT() > 461 ) ) )
        cout << endl << " ------ Processing MT = " << this_mt->MT() << " ------ " << endl;
    
    int mt = this_mt->MT();
    bool skip_mt = false;
    bool skip_xs = false;

    // check for more detailed particle data
    if( 
        ( mt ==   4  && ( Eval.MT_MF_list.found(  50 ) || Eval.MT_MF_list.found(  91 ) ) )  ||
        ( mt == 103  && ( Eval.MT_MF_list.found( 600 ) || Eval.MT_MF_list.found( 649 ) ) )  ||
        ( mt == 104  && ( Eval.MT_MF_list.found( 650 ) || Eval.MT_MF_list.found( 699 ) ) )  ||
        ( mt == 105  && ( Eval.MT_MF_list.found( 700 ) || Eval.MT_MF_list.found( 749 ) ) )  ||
        ( mt == 106  && ( Eval.MT_MF_list.found( 750 ) || Eval.MT_MF_list.found( 799 ) ) )  ||
        ( mt == 107  && ( Eval.MT_MF_list.found( 800 ) || Eval.MT_MF_list.found( 849 ) ) ) 
      )
    {
      if ( Global.Value( "translate_all" ) == 0 )
      {
        Info( "main", pastenum("Skipping  mt: ", mt) + ".  Use detailed reaction data");
        skip_xs = true;
        skip_mt = true;
      }
    }
    if( !skip_mt )
    {
      ENDL.global( Eval.ZA, Eval.AWR, Eval.TEMP, Eval.ELIS, Eval.yi, Eval.date );
      ENDL.T  = mt;
      ENDL.C  = 0;
      ENDL.S  = 0;
      ENDL.LR = 0;
      ENDL.write_file = true;
      ENDL.append = false;   // the default is to start a new file
      bool xs_exist = false; // no cross section data yet
      
      for( iMF_array_link::iterator this_mf = this_mt->begin( ); 
           this_mf != this_mt->end( ); ++this_mf )
      {
        int mf = *this_mf;
        // We can accept mf13 files for mt = 3 or mt = 4.
        // For all other cases we also need the cross sections (mf3).
        if( ( mf == 3 ) || ( ( mf <= 13 ) && ( ( mt >= 3 ) && ( mt <= 4 ) ) ) ) xs_exist = true;
        // In ENDL some angular distributions are in CM and some in lab frame
        if( mf == 4 ) ENDL.angles_CM = this_mt->angle_cm( );
        if( xs_exist && Global.Value( "translate_all" ) > 0 )
        { 
          ENDL.F = mf;
          if( !( mf==3 && skip_xs ) ) LoadData( mf, mt );
        }
        else if ( mt < 151 ) Info("main",pastenum("Skipping mt: ",mt)+
            pastenum(" mf: ",mf)+".  No cross section data");
      }
    }
  }
  if( ( C55.has_MT3 ) || ( C55.has_MT4 ) )
  {
    C55.write_endl( );
  }
  messageLog.write();

#ifdef CLEAN_TEMP_FILES

  if (!clean_files("mf??mt???")) cout<<"\n\nRemoving temporary mf*mt* files failed"<<endl;

#endif

}
