/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1888 $
 * $Date: 2006-08-28 10:37:08 -0700 (Mon, 28 Aug 2006) $
 * $Author: dbrown $
 * $Id: global_params.cpp 1888 2006-08-28 17:37:08Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

//Implementation of the GlobalParameterClass

#include "global_params.hpp"
#include "convert.hpp"         // converting strings to numbers

// *********** ss_link class ***********************
// ----------- ss_link::print ---------------
// print the data in a link
void ss_link::print()
{
  cout << "   " << x << "   " << y << endl;
}

// *********** ss_list class ***********************
// ----------- ss_list::at ---------------
// Function to get the link at "x"
// Returns "false" if the entry is not found.
bool ss_list::at(string x, ss_list::iterator& link)
{
  for( link = begin(); 
       link != end();
       ++link )
  {
    if ( link->name( ) == x ) 
    {
      return true;
    }
  }
  return false;
}

// ----------- ss_list::print ---------------
void ss_list::print()
{
  for( ss_list::iterator link = begin(); 
       link != end();
       ++link)
  {
    link->print();
  }
  cout<<endl;
}

// *********** GlobalParameterClass class ***********************
// ----------- GlobalParameterClass constructor ---------------
//Default constructor loads the default Parameter list and then reads in the 
//default parameter file and loads them into the Parameter list.
GlobalParameterClass::GlobalParameterClass()   
{
  // Here is the "standard" list of translation switches
  // Interpolation options
  set("tol_1d", 0.005);         //! accuracy of interpolation in 1-d lists
  set("cut_off_1d", 0.001);     //! don't impose tolerance below this value
  set("tol_2d", 0.015);         //! accuracy of interpolation in 2-d lists
  set("cut_off_2d", 0.003);     //! don't impose 2-d tolerance below this value
  set("dE_tol", 0.01);          //! minimum incident energy difference for 2-d lists
  // 3d list options
  set("num_mu_3d", 2);          //! number of mu values in 3-d lists is 2*num_mu_3d+1
  set("num_E_in_3d", 12);       //! number of incident energies for 3d expansion
  set("log_log_tol", 0.005);    //! relative error in converting log-log to lin-lin
  // MF=5 options
  set("num_mf5_bins", 32);      //! number of equiprobable energy bins
  set("mf5_shift", 0.00002);    //! amount by which the first distribution is above threshold
  set("mf5_tol", 0.00001);      //! energy tolerance in ENDF/B-VI files
  // Other options specific to one data type
  set("sum_inelastic", 0);      //! 0: use the mt51--mt91 individual inelastic data; 1: use the mt04 total inelastic data
  set("Kalbach_i10", 0);        //! if > 0, calculate the i=10 average energy
  set("split_gammas", 0);       //! if > 0, split discrete gammas from continuum
  set("skip_gammas", 0);        //! if > 0, skip the gamma data
  set("fission_Q", 1);          //! 0: use ENDF fission Q value; otherwise: subtract Q for delayed fission decays
  // General processing options
  set("skip_date", 1);          //! if == 1, don't bother processing eval date field
  set("translate_all", 1);      //! if == 1, attempt to translate all data, even if we normally would ignore it
  set("skip_logging", 0);       //! if == 1, don't log errors in documentation.txt file
  set("message_level",0);       //!three levels of messages - info, warn and severe
                                // info = 0 (all messages written)
                                // warn = 1 (only warnings and errors)
                                // severe >=2 (errors only, these cause exits anyway!)
  set("endl_datafield_precision",8); //! 8 sig-figs in all endl data fields
  set("max_list_size", 500);    //! maximum length of a 1-d list
  set("Max_n_energy", 20.0);    //! maximum neutron energy for ENDL files
  set("Max_cp_energy", 30.0);   //! maximum charged particle energy for ENDL
  set("Max_g_energy", 30.0);    //! maximum gamma for ENDL files
  set("mf5_weight_tol", 0.001); //! tolerance on the total weight for mf5 models
  set("max_coulomb_mu", 0.94);  //! maximum mu for charged-particle scattering
  set("smallest_dE", 1.0e-16);  //! dE to use in widening a jump at E = 0
  set("jump_factor", 2.0);      //! factor of ENDL_EPSILON to use for jump width
  set("Min_n_energy", 1e-11); //! minimum incident neutron energy for ENDL files
}
// ----------- GlobalParameterClassss_link destructor --------
//Default destructor.
GlobalParameterClass::~GlobalParameterClass()  
{ 
}
// ----------- GlobalParameterClass::set ---------------
//Sets the parameter values and creates a new parameter if one doesn't exist.
void GlobalParameterClass::set(string gp, double gp_num)
{
  //First thing we need to do is to strip all extraneous blanks from the string gp
  gp = tolower( remove_extra_blanks( gp ) );
  ostringstream gp_ostr; 
  gp_ostr <<  gp_num;
  string gp_val = gp_ostr.str();
  //We test for the existence of the parameter name in the list.
  //If it is new, we add a link.  If not, we reset the existing
  //value in the link to gp_val.
  ss_list::iterator SSlist;

  if ( Parameters.at(gp,SSlist) )
    {
      SSlist->value()=gp_val;  //reset the parameter value in this link
    }
  else
    {
      ss_link SSlink;
      SSlink.name() = gp;  //Load into link
      SSlink.value() = gp_val;
      Parameters.insert(Parameters.end(), SSlink); //Insert link into list
    }
}
// ----------- GlobalParameterClass::set ---------------
//Sets the parameter values and creates a new parameter if one doesn't exist.
void GlobalParameterClass::set(string gp, string gp_val)
{
  //First thing we need to do is to strip all extraneous blanks from the strings
  gp     = remove_extra_blanks( gp );
  gp_val = remove_extra_blanks( gp_val );
  
  //We test for the existence of the parameter name in the list.
  //If it is new, we add a link.  If not, we reset the existing
  //value in the link to gp_val.
  ss_list::iterator SSlist;

  if ( Parameters.at(gp,SSlist) )
  {
      SSlist->value()=gp_val;  //reset the parameter value in this link
  }  
  else
  {
      ss_link SSlink;
      SSlink.name() = gp;  //Load into link
      SSlink.value() = gp_val;
      Parameters.insert(Parameters.end(), SSlink); //Insert link into list
  }
}
// ----------- GlobalParameterClass::print ---------------
// Print utility
void GlobalParameterClass::print( )
{
  cout<<"Info : List of global parameters used in this translation:"<<endl;
  Parameters.print();
}
// ----------- GlobalParameterClass::read_command_line ---------------
// Function used to parse the command line for parameters and settings
string GlobalParameterClass::read_command_line(int argc, char* argv[] )
{
  //First see if the user just wants help.  This is the only way to have a non-"-option value" pair
  string help1("-help"), help2("--help"), help3("-h");
  for( int j=1; j<argc; ++j)
  {
    string sarg(argv[j]);
    if( sarg == help1 || sarg == help2 || sarg == help3 )
    {
      cout << center(40*string("="),42) <<endl;
	  cout << center("Help for fete.",40)<<endl;
      cout << center(40*string("="),42) <<endl;
      cout << "Usage: fete [-option1 value]"<<endl<<endl;
      cout << "To set the file to translate, use the -f option (e.g. fete -f fred.endf)  "<<endl;
      cout << "Other options are set by prepending '-' to the options listed below:"<<endl;
      Parameters.print();
      exit(1);
    }
  }
  
  //Next, we test the command line inputs for the "-" indicator on every other input
  for( int j=1; j<argc; j+=2 )
  {
    if ( argv[j][0] != '-' ) 
	 { 
       cerr<<"fete Fatal Error GlobalParameterClass::read_command_line:"<<endl;
	   cerr<<"    Unable to process command line."<<endl;
	   cerr<<"    "<<argv[j]<<" is not a valid switch!"<<endl;
	   exit(-10);
	 }
    else
	 {
	   //Move the character array into string arrays
	   spar[j] = argv[j]; 
	   if (j+1<argc) spar[j+1]=argv[j+1];
      else spar[j+1]="";
	 }
  };

  //Then we will loop through argv to see if the user wants to use a private input file
  //or an assortment of input files, each possibly overwriting earlier settings...
  string privateParameterFile("");
  for( int j=1; j<argc;j+=2 ) if( spar[j] == "-p" || spar[j] == "-param" ) privateParameterFile=spar[j+1];
  if (privateParameterFile!="")
  {
    cout<<"Info : Private parameter input file: "<<privateParameterFile<<endl;
	read_file( privateParameterFile );
  }  
  else 
  {
    //Now, we read in the standard input file, if it is there
    read_file("fete_options.inp");
  }

  //Now we loop through the argument list and set parameters listed on the command line
  //We perform this loop AFTER the user file so that the command line settings are
  //always used.
  for( int j=1; j<argc;j+=2 )
  {
    if( spar[j] != "-f" && spar[j] != "-file" && spar[j] != "-p" && spar[j] != "-param" ) //not a param file request
    {
      string pval = spar[j+1];
      string key = spar[j];
      key.erase(0,1);
      set(key,pval);
    }
  }
  
  //Then we will loop through argv to see if the user wants to use a different default ENDF file name
  for( int j=1; j<argc;j+=2 )
  {
    if( spar[j] == "-f" || spar[j] == "-file" )
	{
      cout<<"Info : ENDF input file: "<<spar[j+1]<<endl;
      string file = spar[j+1];
      return file;
    }
  }
  return "";
}
// ----------- GlobalParameterClass::read_file ---------------
// Reads in a file and looks for parameter names and values...
void GlobalParameterClass::read_file(string file)
{
  string linebuff, origbuff;
  
  ifstream paramfile( file.c_str() );  //open the file
  
  if ( !paramfile )
    {
      cout<<"Info: GlobalParameterClass::read_file: There is no parameter file named: "<<file<<", using internal defaults"<<endl;
    }
  else
    {
      cout<<"Info : Reading the parameter file: "<<file<<" ..."<<endl;
      while( !paramfile.eof() )  //Read until end of file
	{ 
	  getline( paramfile, linebuff );  //Get a line
	  origbuff=linebuff;     //Make a copy in case there's trouble
	  linebuff = remove_extra_blanks( linebuff );  //remove extraneous blanks
	  if( linebuff.find('#') <= linebuff.size() )  //loook for comment marker
	    {
	      linebuff = linebuff.erase(linebuff.find('#'),linebuff.size()); //erase
	      linebuff=remove_extra_blanks( linebuff ); //new trailing blanks can show up
	    }

	  if( linebuff.size() > 0 ) //We are done with this line if it is blank
	    {
	      int nblanks=0; //counter for number of blanks left in string
	      for( int i=0;i<linebuff.size();i++ )
		{
		  if ( linebuff.at(i) == ' ' ) ++nblanks; //count the blanks
		}
	      if( nblanks != 1 ) //There should be exactly one left (two words)
		{
		  cerr
		    <<"fete Fatal Error GlobalParameterClass::read_file: Could not parse the following line from "
		    <<"the parameter file, "<<file<<endl;
		  cerr<<origbuff<<endl;  //Show original line in file
		  exit(-10);
		}
	      else  //Looks like we have a winner!
		{
		  string sparam 
		    = linebuff.substr(0,linebuff.find(' ')); //load name
		  string svalue 
		    = linebuff.substr(linebuff.find(' ')+1,linebuff.size()); //load value
		  set(sparam, svalue);
		}
	    }
	}
    }
}
// ----------- GlobalParameterClass::Value ---------------
// Returns the value of the requested parameter.
double GlobalParameterClass::Value( string ParamName )
{
  ss_list::iterator XYptr;
  string parm = tolower( ParamName );
  if( !Parameters.at(parm,XYptr) )
    {
      cerr<<"fete Fatal Error GlobalParameterClass::Value:"<<
        " no parameter called "<<ParamName<<" found!"<<endl;
      exit(-50);
    }
  return( static_cast<double>( stof(XYptr->value()) ) );
}
// ----------- GlobalParameterClass::Flag ---------------
// Returns the string value of the parameter.
string GlobalParameterClass::Flag( string ParamName )
{
  ss_list::iterator XYptr;
  string parm = tolower( ParamName );
  if( !Parameters.at(parm,XYptr) )
    {
      cerr<<"fete Fatal Error GlobalParameterClass::Flag:"<<
        " no parameter called "<<ParamName<<" found!"<<endl;
      exit(-50);
    }
  return( XYptr->value() );
}

