/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1873 $
 * $Date: 2006-05-17 08:45:49 -0700 (Wed, 17 May 2006) $
 * $Author: dbrown $
 * $Id: kalbach.cpp 1873 2006-05-17 15:45:49Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the class three_d_kalbach

#include <cmath>

#include "kalbach.hpp"
#include "convert.hpp"
#include "messaging.hpp"

extern ENDLClass ENDL;

// ********* for class Kalbach_data *************
// ------------ Kalbach_data::read_data ----------
void Kalbach_data::read_data( mf6_file& inFile, int num_E_out )
// read in the data for one incident energy
{
  string sbuff;
  Kalbach_triple triple;
 
  for ( int count = 0; count < num_E_out; ++count )
  {
    inFile >> sbuff;
    // for the particle energy in MeV
    triple.E_out = ENDL.eV2MeV*stod( sbuff );
    inFile >> sbuff;
    triple.Prob_E_out = stod( sbuff );
    inFile >> sbuff;
    triple.compound_frac = stod( sbuff );
    insert(end(), triple);
  }
  inFile.ignore(66,'\n');  // Must clear rest of line after using >>
}
// ----------- Kalbach_data::find_data -----------------
Kalbach_data::iterator Kalbach_data::find_data(double E_out)
// Find E_out in the data list and interpolate the data.
// *** This routine is writen for histograms; for linear ***
// *** interpolation, write new code. ***
{
  const double eps = 1.0e-6;  // a tolerance on real numbers
  Kalbach_data::iterator data_ptr = end();
  for(--data_ptr; ; --data_ptr)
  {
    if(( E_out >= data_ptr->E_out ) ||
      ( (data_ptr == begin()) && (E_out >= (1 - eps)*data_ptr->E_out)) )
    {
      return data_ptr;
    }

    // Is E_out below the first one?
    if(data_ptr == begin())
    {
      Warning("Kalbach_data::find_data",pastenum("E_out too small in Kalbach: ",E_out));
      return data_ptr;
    }
  }
}

// ********* for class one_d_Kalbach ************
// ----------- one_d_Kalbach::one_E_row ----------------
void one_d_Kalbach::one_E_row( double min_E, double max_E )
// for a given mu make new links for secondary energies
// min_E <= E <= max_E
{
  // a data check
  if( min_E > max_E )
  {
    SevereError( "one_d_Kalbach::one_E_row", "energies out of order" );
  }

  // check for forward in lab frame, backward in c-m frame
  double E_lab_trans = mu( ) * mu( ) * map_->E_transl;  // min c-m energy

  if( ( mu( ) > 0.0 ) && ( max_E < E_lab_trans ) )
  {
    // E_cm is only decreasing
    cm_E_decr( min_E, max_E );
  }
  else if( ( mu( ) > 0.0 ) && ( min_E < E_lab_trans ) &&
	   ( max_E > E_lab_trans ) )
  {
    // E_cm switches direction
    cm_E_decr( min_E, E_lab_trans );
    cm_E_incr( E_lab_trans, max_E );
  }
  else
  {
    // E_cm is increasing
    cm_E_incr( min_E, max_E );
  }
}
// ----------- one_d_Kalbach::cm_E_incr -----------------
// For this mu and this range of lab energies the cm energies increase.
// Spefically, we are in the region E_lab > map_->E_transl - E_cm
void one_d_Kalbach::cm_E_incr( double min_E, double max_E )
{
  Pair lab_Emu; // (energy, cosine) in lab coordinates
  Pair cm_Emu;  // (energy, cosine) in center-of-mass coordinates
  // for new links
  double E_out;
  double Prob;
  dd_link e_out_link;
  double dE;    // to widen the histogram

  lab_Emu.mu = mu( );

  // For given E_cm and mu_lab the algebra gives 2 values of E_lab.
  // We want the larger one if mu_lab > 0:
  int sigma = ( mu( ) > 0.0 ) ? 1 : -1;

  // insert the first point
  dE = ENDL_EPSILON( min_E );
  lab_Emu.E =  min_E + dE;
  // in center-of-mass coordinates
  cm_Emu = map_->lab_to_cm( lab_Emu );
  // E_f0_r_ptr points to the Kalbach data
  E_f0_r_ptr = data_ptr__->find_data( max(cm_Emu.E,0.0) );
  // insert this data
  Prob = f( lab_Emu.E );
  e_out_link = dd_link( lab_Emu.E, Prob );
  insert( end( ), e_out_link );

  // identify the largest cm outgoing energy
  dE = ENDL_EPSILON( max_E );
  lab_Emu.E = max_E - dE;
  cm_Emu = map_->lab_to_cm( lab_Emu );
  double E_cm_max = cm_Emu.E;

  // The Kalbach data is given as a histogram in cm coordinates.
  // Loop until we get to E_cm_max
  Kalbach_data::iterator next_data;
  for( next_data = E_f0_r_ptr; ; )
  {
    ++next_data;  // the next Kalbach triplet
    one_d_Kalbach::iterator this_link;
    one_d_Kalbach::iterator next_link;

    // Are we finished?
    if( ( next_data == data_ptr__->end( ) ) ||
        ( next_data->E_out > E_cm_max ) )
    {
      // insert a last link to finish the list
      this_link = end( );
      --this_link;

      // insert the new last link
      Prob = f( max_E );
      e_out_link = dd_link( max_E, Prob );
      insert( end( ), e_out_link );

      next_link = end( );
      --next_link;
      thicken( this_link, next_link );
      break;
    }

    // point to the current last link
    this_link = end( );
    --this_link;

    // get the end of the histogram
    E_out = map_->get_lab_E( next_data->E_out, mu(), sigma );
    dE = ENDL_EPSILON( E_out );
    // insert the new last link
    Prob = f( E_out - dE );
    e_out_link = dd_link( E_out - dE, Prob );
    insert( end( ), e_out_link );

    next_link = end( );
    --next_link;
    thicken( this_link, next_link );

    // start the next step of the histogram
    E_f0_r_ptr = next_data;
    Prob = f( E_out + dE );
    e_out_link = dd_link( E_out + dE, Prob );
    insert( end( ), e_out_link );
  }

  // We currently do not do thickening with respect to mu,
  // but this is the place to set up the equiprabable bins
  // if and when this thickening gets implemented.
  //    if( bin_interp )
  //    {
  //      get_bins( );
  //    }

}
// ----------- one_d_Kalbach::cm_E_decr -----------------
// For this mu and this range of lab energies the cm energies decrease.
// Specifically, we are in the region E_lab < map_->E_transl - E_cm
void one_d_Kalbach::cm_E_decr( double min_E, double max_E )
{
  Pair lab_Emu; // (energy, cosine) in lab coordinates
  Pair cm_Emu;  // (energy, cosine) in center-of-mass coordinates
  // for new links
  double E_out;
  double Prob;
  dd_link e_out_link;
  double dE;    // to widen the histogram

  lab_Emu.mu = mu( );

  // For given E_cm and mu_lab the algebra gives 2 values of E_lab.
  // We want the smaller one
  int sigma = -1;

  // A first cut at a shift away from the singularities
  dE = ENDL_EPSILON( min_E );
  // We may need something smaller
  double E_diff = max_E - min_E;
  if( 4*dE > E_diff ){
    dE = E_diff/4.0;
  }

  // insert the first point
  lab_Emu.E =  min_E + dE;
  // in center-of-mass coordinates
  cm_Emu = map_->lab_to_cm( lab_Emu );
  // E_f0_r_ptr points to the Kalbach data
  E_f0_r_ptr = data_ptr__->find_data( max(0.0,cm_Emu.E) );
  // insert this data
  Prob = f( lab_Emu.E );
  e_out_link = dd_link( lab_Emu.E, Prob );
  insert( end( ), e_out_link );

  // identify the smallest cm outgoing energy
  dE = ENDL_EPSILON( max_E );
  lab_Emu.E = max_E - dE;
  cm_Emu = map_->lab_to_cm( lab_Emu );
  double E_cm_min = cm_Emu.E;

  // The Kalbach data is given as a histogram in cm coordinates.
  // Loop until we get to E_cm_min
  Kalbach_data::iterator this_data;
  for( this_data = E_f0_r_ptr; ; )
  {
    one_d_Kalbach::iterator this_link;
    one_d_Kalbach::iterator next_link;

    // Are we finished?
    if( ( this_data == data_ptr__->begin( ) ) ||
        ( this_data->E_out < E_cm_min ) )
    {
      // insert a last link to finish the list
      this_link = end( );
      --this_link;

      // insert the new last link
      Prob = f( max_E );
      e_out_link = dd_link( max_E, Prob );
      insert( end( ), e_out_link );

      next_link = end( );
      --next_link;
      thicken( this_link, next_link );
      break;
    }

    // point to the current last link
    this_link = end( );
    --this_link;

    // get the start of the histogram
    E_out = map_->get_lab_E( this_data->E_out, mu(), sigma );
    dE = ENDL_EPSILON( E_out );
    // insert the new last link
    Prob = f( E_out - dE );
    e_out_link = dd_link( E_out - dE, Prob );
    insert( end( ), e_out_link );

    next_link = end( );
    --next_link;
    thicken( this_link, next_link );

    // start the next step of the histogram (the previous cm energy)
    --this_data;  // the next Kalbach triplet
    E_f0_r_ptr = this_data;
    Prob = f( E_out + dE );
    e_out_link = dd_link( E_out + dE, Prob );
    insert( end( ), e_out_link );
  }
}
// ----------- one_d_Kalbach::copy_params -----------------
// Copy the parameters
void one_d_Kalbach::copy_params( Param *params )
{
  params__ = params;
}

// ----------- one_d_Kalbach::copy_data -----------------
// Copy the parameters
void one_d_Kalbach::copy_data( list< Kalbach_data >::iterator data_ptr )
{
  data_ptr__ = data_ptr;
}

// ----------- one_d_Kalbach::get_a ---------------
// Evaluate the Kalbach a function for the slope value
double one_d_Kalbach::get_a( double E_out )
{
  return Get_a( E_out, *params__ );
}

// ----------- one_d_Kalbach::f -----------------
double one_d_Kalbach::f( double E )
// Evaluate the Kalbach probability function for
// cosine mu and energy E_out (laboratory coordinates).
{
  Pair lab_Emu; // (energy, cosine) in lab coordinates
  Pair cm_Emu;  // (energy, cosine) in center-of-mass coordinates
  double Prob;  // return value

  lab_Emu.mu = mu( );
  lab_Emu.E = E;

  // in center-of-mass coordinates
  cm_Emu = map_->lab_to_cm( lab_Emu );

  if( E_f0_r_ptr == NULL )
  {
    E_f0_r_ptr = data_ptr__->find_data( max(0.0,cm_Emu.E) );
  }

  double slope_a = get_a( cm_Emu.E );
  // DAB 2/3/05 -- factor of 0.5 missing from older version of ENDF manual
  Prob = 0.5*(E_f0_r_ptr->Prob_E_out*( slope_a/sinh( slope_a ) )*
       ( cosh( slope_a*cm_Emu.mu ) +
     E_f0_r_ptr->compound_frac*sinh( slope_a*cm_Emu.mu ) ));  

  // in laboratory coordinates
  Prob *= map_->J_cm_to_lab( cm_Emu.E, lab_Emu.E );
  return Prob;
}

// ********* for class two_d_Kalbach *************
// ----------- two_d_Kalbach::new_one_d ------
two_d_Kalbach::iterator
  two_d_Kalbach::new_one_d( two_d_Kalbach::iterator where )
{
  one_d_Kalbach new_link;
  insert( where, new_link );

  // point to the new link
  two_d_Kalbach::iterator link_ptr = where;
  --link_ptr;

  // set up the parameters
  link_ptr->copy_params( params_ );
  link_ptr->copy_map( &map );
  link_ptr->copy_data( data_ptr_ );

  return link_ptr;
}
// ----------- two_d_Kalbach::set_params ------
void two_d_Kalbach::set_params(
  list< Kalbach_data >::iterator data_ptr, Param *params )
{
  data_ptr_ = data_ptr;
  params_ = params;
}

// ********* for class three_d_Kalbach *************
// ----------- three_d_Kalbach::three_d_Kalbach ------
three_d_Kalbach::three_d_Kalbach( int targ_ZA, int proj_ZA,
  int eject_ZA, int this_mult )
// constructor with given target, projectile, and ejectum.
{
  // this model makes no sense for natural targets
  if( targ_ZA % 1000 == 0 )
  {
    SevereError( "three_d_Kalbach::three_d_Kalbach",
		 " the Kalbach model makes no sense for natural targets" );
  }

  this_mult_ = this_mult;

  A.set(targ_ZA);
  a.set(proj_ZA);
  b.set(eject_ZA);
  C.set(targ_ZA + proj_ZA);
  B.set(C.ZA - eject_ZA);

  // get the separation energies
  S_a = get_S(A, a);

  // We need 5 parameters for the Kalbach function
  Params.get_space( 5 );
  Params[0] = 0.0;  // e_a depends on the incident energy
  Params[1] = ( B.AWR + b.AWR ) / B.AWR;  // needed to calculate epsilon_b
  Params[2] = get_S( B, b );  //  S_b;

  // Params[3] is the Kalbach M_a factor
  switch(a.ZA)
  {
  case 1:  // neutron
    Params[3] = 1.0;
    break;
  case 1001:  // proton
    Params[3] = 1.0;
    break;
  case 1002:  // deuteron
    Params[3] = 1.0;
    break;
  case 1003:  // tritron
    Params[3] = 0.0;
    break;
  case 2003:  // helium 3
    Params[3] = 0.0;
    break;
  case 2004:  // alpha
    Params[3] = 0.0;
    break;
  default:
    Unimplemented("three_d_Kalbach::three_d_Kalbach",
        pastenum("M for particle ",a.ZA)+" not implemented");
  }

  // Params[4] is the Kalbach m_b factor
  switch(b.ZA)
  {
  case 1:  // neutron
    Params[4] = 0.5;
    break;
  case 1001:  // proton
    Params[4] = 1.0;
    break;
  case 1002:  // deuteron
    Params[4] = 1.0;
    break;
  case 1003:  // tritron
    Params[4] = 1.0;
    break;
  case 2003:  // helium 3
    Params[4] = 1.0;
    break;
  case 2004:  // alpha
    Params[4] = 2.0;
    break;
  default:
    Unimplemented("three_d_Kalbach::three_d_Kalbach",
        pastenum("M for particle ",b.ZA)+" not implemented");
  }

}
// ----------- three_d_Kalbach::expand_data -----------------
void three_d_Kalbach::expand_data(int num_E_in, mf6_file& inFile)
// read in all of the ENDF/B-VI data and expand it to a list
// num_E_in is the number of data sets
{
  // read in the data
  read_data( inFile, num_E_in );
  if(!ENDL.write_file )
  {
    return;
  }

  if( Global.Value( "Kalbach_i10" ) > 0 )
  {
    // Check the endep calculation of <E'>
    check_i10( );
  }

  // make a link for each incident energy
  for( Data_ptr = triples.begin( ); Data_ptr != triples.end( );
    ++Data_ptr )
  {
    one_e_in();
  }
}
// ----------- three_d_Kalbach::one_e_in -----------------
void three_d_Kalbach::one_e_in( )
// handle the data for one incident energy
{
  // this E_1 corresponds to a new link in the 3d list
  two_d_Kalbach e_in_link;
  e_in_link.E_in( ) = Data_ptr->E_in;
  insert( end( ), e_in_link );

  // make a pointer to the new 3-d link
  three_d_Kalbach::iterator this_link = end();
  --this_link;

  // copy the parameters
  this_link->set_params( Data_ptr, &Params );

  // Expand this data into a 3-d link
  expand_E_in( this_link );

  // add a link to the cosine list
  one_d_table new_cos;
  cosines.insert(cosines.end(), new_cos);

  // expand this cosine link
  two_d_list<one_d_table>::iterator cos_ptr = cosines.end();
  --cos_ptr;
  this_link->make_cos_link( Data_ptr->E_in, cos_ptr );

}

// ----------- three_d_Kalbach::expand_E_in -----------------
void three_d_Kalbach::expand_E_in( three_d_Kalbach::iterator e_in_link )
// expand the double differential data for 1 incident energy
{
  // center-of-mass incident energy plus separation energy of particle.
  // This e_a is used later by the get_a routine
  Params[0] = e_in_link->E_in()*A.AWR/(A.AWR + a.AWR) + S_a;

  // For ENDL we need to go through cosines first and then E_out.
  // But the ENDF/B6 data has E_out first.

  // for the lowest particle energy in MeV
  Kalbach_data::iterator this_data = Data_ptr->begin( );
  double first_E_out = this_data->E_out;

  // for the highest particle energy in MeV
  this_data = Data_ptr->end( );
  --this_data;
  double last_E_out = this_data->E_out;

  // to map between lab and center-of-mass coordinates
  e_in_link->map.set_map( &A, &a, &b, e_in_link->E_in() );

  // identify the geometry of the (E', mu) region in lab coordinates.
  e_in_link->get_geom( first_E_out, last_E_out );
  if( e_in_link->geom_.fastest == Both_ways )
  {
    // The fastest outgoing particles can have mu < 0
    if( e_in_link->geom_.slowest == Straight )
    {
      // there are outgoing particles with CM speed 0
      e_in_link->full_range( last_E_out );
    }
    else if( e_in_link->geom_.slowest == Forward )
    {
      // some outgoing particles are only forward in the lab frame
      e_in_link->both_ways_hole( first_E_out, last_E_out );
    }
    else if( e_in_link->geom_.slowest == Transition )
    {
      // the slowest outgoing particles move at the CM translational speed
      e_in_link->both_ways_trans( first_E_out, last_E_out );
    }
    else
    {
      // all outgoing particles move faster than the CM translational speed
      e_in_link->all_both_ways( first_E_out, last_E_out );
    }
  }
  else if( e_in_link->geom_.fastest == Transition )
  {
    // the fastest outgoing particles move at the CM translational speed
    if( e_in_link->geom_.slowest == Straight )
    {
      // there are outgoing particles with CM speed 0
      e_in_link->full_trans( last_E_out );
    }
    else
    {
      // the slowest outgoing particles have postive CM speed
      e_in_link->trans_hole( first_E_out, last_E_out );
    }
  }
  else
  {
    // the fastest outgoing particles all move forward in the lab frame
    if( e_in_link->geom_.slowest == Straight )
    {
      // there are outgoing particles with CM speed 0
      e_in_link->full_forward( 0.0, last_E_out );
    }
    else
    {
      // the slowest outgoing particles have postive CM speed
      e_in_link->forward_hole( first_E_out, last_E_out );
    }
  }
}

// ----------- three_d_Kalbach::read_data -----------------
void three_d_Kalbach::read_data( mf6_file& inFile, int num_E_in )
// read in all of the ENDF/B-VI data
// num_E_in is the number of data sets
{
  Kalbach_data one_set;  // Kalbach data for 1 incident energy
  list<Kalbach_data>::iterator data_ptr;  // point to the data

  double ZERO;  // zero
  double E_1;   // incident neutron energy
  int ND;   // number of discrete energies
  int NA;   // number of triples  			//FIXME!!!
  int NW;   // number of words in the record (3*NA)  	//FIXME!!!
  int NEP;  // number of secondary energies

  string linebuff;
  for(int count = 0; count < num_E_in; ++count)
  {
    getline( inFile, linebuff ); //read next line
    read_ddiiii(&linebuff, 
	      &ZERO, &E_1,   // incident energy
	      &ND,           // number of discrete energies
	      &NA,           // number of angular parameters (1)
	      &NW,           // number of table entries
	      &NEP);         // number of E_out values
    if(ND > 0)
    {
      SevereError("three_d_Kalbach::read_data","Implement discrete energies for Kalbach");
    }

    // build the list for this incident energy (in MeV)
    one_set.E_in = E_1*ENDL.eV2MeV;
    triples.insert(triples.end(), one_set);

    // point to the new link
    data_ptr = triples.end();
    --data_ptr;
    // fill this list
    data_ptr->read_data( inFile, NEP );
  }
}
/*
// --------- three_d_Kalbach::fill_in_data -----------------
void three_d_Kalbach::fill_in_data();
// If the emission jumps from all forward to all directions,
// insert a smoother transition.
{
  // The translational energy from the initial collisiton is lambda*E_in
  double lambda = (A.AWR)*(b.AWR)/((A.AWR + a.AWR)*(A.AWR + a.AWR));

  // look for a jump from all forward
  for(list<Kalbach_data>::iterator data_ptr = triples.begin();
    data_ptr != triples.end(); ++data_ptr)
  {
    Kalbach_data::iterator last_set = data_ptr.end();
    --last_set;

  }

}
*/
// ----------- three_d_Kalbach::get_S -----------------
double three_d_Kalbach::get_S(Nuclei& targ_A, Nuclei& proj_a)
// evaluate the Kalbach S function for the separation energy
{
  double S;

  double I_a;
  switch(proj_a.ZA)
  {
  case 1:  // neutron
    I_a = 0.0;
    break;
  case 1001:  // proton
    I_a = 0.0;
    break;
  case 1002:  // deuteron
    I_a = 2.22;
    break;
  case 1003:  // tritron
    I_a = 8.48;
    break;
  case 2003:  // helium 3
    I_a = 7.72;
    break;
  case 2004:  // alpha
    I_a = 28.3;
    break;
  default:
    Unimplemented("three_d_Kalbach::get_S",
        pastenum("particle ",proj_a.ZA)+" not implemented in get_S");
  }

  const double coef_1 = 15.68;
  const double coef_2 = 28.07;
  const double coef_3 = 18.56;
  const double coef_4 = 33.22;
  const double coef_5 = 0.717;
  const double coef_6 = 1.211;

  S = coef_1*(C.A - targ_A.A) -
      coef_2*((C.N - C.Z)*(C.N - C.Z)/C.A -
              (targ_A.N - targ_A.Z)*(targ_A.N - targ_A.Z)/
                 targ_A.A) -
      coef_3*(exp(2.0/3.0*log(1.0*C.A)) -
              exp(2.0/3.0*log(1.0*targ_A.A))) +
      coef_4*((C.N - C.Z)*(C.N - C.Z)/
                 exp(4.0/3.0*log(1.0*C.A)) -
              (targ_A.N - targ_A.Z)*(targ_A.N - targ_A.Z)/
                 exp(4.0/3.0*log(1.0*targ_A.A))) -
      coef_5*(C.Z*C.Z/exp(1.0/3.0*log(1.0*C.A)) -
              targ_A.Z*targ_A.Z/exp(1.0/3.0*log(1.0*targ_A.A))) +
      coef_6*(C.Z*C.Z/C.A -
              targ_A.Z*A.Z/targ_A.A) - I_a;

  return S;
}
// ----------- three_d_Kalbach::renorm -----------------
void three_d_Kalbach::renorm( ){
  for (three_d_list< two_d_Kalbach >::iterator it=begin();it!=end();++it) it->renorm();
}

// ----------- three_d_Kalbach::check_i10 -----------------
void three_d_Kalbach::check_i10( )
// Write an i=10 file of average energy for outgoing particle.
  // What we want is
  // (1/2) * \int_\Omega E_b P( E_b, \mu ) \, dE_b \, d\mu
  // where everything is in laboratory coordinates.  But we are
  // given P is center-of-mass coordinates.  It is better to do
  // the integration in center-of-mass coordinates, for 2 reasons.
  // First, this avoids having to compute the Jacobian of the mapping
  // to laboratory coordinates.  Second, in center-of-mass coordinates
  // the domain of integration is a rectangle, -1 <= \mu <= 1, and
  // 0 <= E_b <= E_{b \max}.  From equations (6.4) and (6.5) of the
  // ENDF/B-VI manual, the resulting integral is
  // (1/2) * \int_0^{E_b\max} f_0 \int_{-1}^1 [E_b + \gamma^2 E_a +
  //      2 \gamma \mu \sqrt{E_a E_b}] [a / sinh(a) ]
  //      [cosh(a \mu) + r sinh(a \mu)] \, d\mu \, dE_b.
  // Here, \gamma = sqrt{a.AWR * b.AWR}/(A.AWR + a.AWR), the values of
  // f_0 and r are given in the mf6 file, and a is from the formula on
  // page 6.5 of the  ENDF/B-VI manual.
  //
  // We can do the integration over \mu by hand, and the result is
  // \int_0^{E_b\max} f_0 { E_b + \gamma^2 E_a +
  //   [ 2 r \gamma \sqrt{E_a E_b} / sinh(a) ][ cosh(a) - (1/a)sinh(a) ] dE_b.
  // Our routine performs this last integration.
  // (Note that the integrand looks as if it has a 0/0 singularity
  // at a = 0, but this is removable since cosh(a) - (1/a)sinh(a)
  // has Taylor series a^2/3 + a^4/30 + ....)
{
  one_d_table i10_data;
  double gamma = sqrt( a.AWR * b.AWR )/( A.AWR + a.AWR );

  // for adaptive quadrature
  quad_list adapt_quad( av_energy );

  // make a link for each incident energy
  for( Data_ptr = triples.begin( ); Data_ptr != triples.end();
    ++Data_ptr )
  {
    dd_link new_link;
    new_link.x = Data_ptr->E_in;
    double E_a_coef = gamma*gamma*new_link.x; // \gamma^2 E_a
    double Er_sum = 0;    // used to integrate sqrt(E)*(coth(a) - 1/a)
    double E0_sum = 0;    // used to integrate f_0 * { E_b + \gamma^2 E_a }
    double norm_sum = 0;  // used to integrate f_0
    Kalbach_data::iterator this_data = Data_ptr->begin();
    double prev_E_out = this_data->E_out;
    double prev_f0 = this_data->Prob_E_out;
    double prev_r = this_data->compound_frac;
    double integral_coef = 2 * prev_f0 * gamma * sqrt( new_link.x );

    // loop over the (E_out, f0, r) triples
    for( ++this_data ; this_data != Data_ptr->end( ); ++this_data )
    {
      double E_out = this_data->E_out;
      // the data is histograms
      norm_sum += prev_f0 * ( E_out - prev_E_out );

      E0_sum += prev_f0 * ( E_out - prev_E_out ) *
                ( 0.5*( E_out + prev_E_out ) + E_a_coef );

      Er_sum += integral_coef * prev_r *
          adapt_quad.Simp_quad( prev_E_out, E_out, Params, 1.0e-8 );

      prev_f0 = this_data->Prob_E_out;
      prev_r = this_data->compound_frac;
      prev_E_out = E_out;
    }
    new_link.y = this_mult_ * ( E0_sum + Er_sum ) / norm_sum;
    i10_data.insert( i10_data.end( ), new_link );
  }
  i10_data.write_endl( 10 );
}

// *************** Get_a routine *********************
double Get_a( double E_out, Param& params )
// evaluate the Kalbach a function for the slope value
// Note that:
  // params[0] = e_a
  // params[1] = ( B.AWR + b.AWR ) / B.AWR
  // params[2] = S_b
  // params[3] = M_a
  // params[4] = m_b
{
  double slope_a;
  double e_a = params[0];
  if( e_a == 0.0 )
  {
    slope_a = 0.0;
  }
  else
  {
    // center-of-mass energy plus separation energy for exit particle
    double e_b = E_out*params[1] + params[2];

    const double E_t1 = 130.0;  // MeV
    double R_1 = ( e_a < E_t1 ) ? e_a : E_t1;
    double X_1 = R_1*e_b/e_a;

    const double E_t3 = 41.0;  // MeV
    double R_3 = ( e_a < E_t3 ) ? e_a : E_t3;
    double X_3 = R_3*e_b/e_a;

    const double C_1 = 0.04;   // per MeV
    const double C_2 = 1.8e-6;  // per MeV^3
    const double C_3 = 6.7e-7;  // per MeV^4

    slope_a = X_1*(C_1 + C_2*X_1*X_1) +
       C_3*params[3]*params[4]*X_3*X_3*X_3;
  }
  return slope_a;
}

// **************** average energy integrand *****************
// integrand for the average energy integral
double av_energy( double E_b, Param& params )
{
  double slope_a = Get_a( E_b, params );

  double ans;
  ans = ( slope_a == 0.0 ) ? 0.0 :
    sqrt( slope_a ) * ( cosh( slope_a ) / sinh( slope_a ) - 
                       ( 1.0 / slope_a ) );

  return ans;
}
