/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1735 $
 * $Date: 2006-02-09 13:47:26 -0800 (Thu, 09 Feb 2006) $
 * $Author: dbrown $
 * $Id: kalbach.hpp 1735 2006-02-09 21:47:26Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// header for the class three_d_Kalbach

#ifndef THREE_D_KALBACH
#define THREE_D_KALBACH

#include "mf6classes.hpp"
#include "math_util.hpp"

// ----------- class Kalbach_triple -----------------
//! a single set of Kalbach data
class Kalbach_triple
{
public:
  double E_out;
  double Prob_E_out;
  double compound_frac;

  //! Default constructor
  Kalbach_triple() {}

  //! Default destructor
  ~Kalbach_triple() {}
};

// ----------- class Kalbach_data -----------------
//! Class to read and hold the Kalbach data
class Kalbach_data : public list< Kalbach_triple >
{
public:
  double E_in;

  //! Default constructor
  Kalbach_data() {}

  //! Default destructor
  ~Kalbach_data() {}

  //! Reads in the data for one incident energy
  void read_data( mf6_file& inFile, int num_E_out );

  //! Finds E_out in the data list
  Kalbach_data::iterator find_data(double E_out);
};

// ------------------- Get_a routine ------------------
//! Evaluate the Kalbach a function for the slope value
double Get_a( double E_out, Param& params );

// ----------- class one_d_Kalbach -----------------
//! Class for pairs (E', probability) for one E and mu
class one_d_Kalbach : public mf6_one_d
{
private:
  // point to the Kalbach data for this incident energy
  list< Kalbach_data >::iterator data_ptr__;

  // Kalbach parameters (E', f0, r)
  Kalbach_data::iterator E_f0_r_ptr;

  Param *params__;   // 5 parameters for the quadrature routine

public:
  //! Default constructor
  one_d_Kalbach( )
  {
  }

  //! We need a special one_E_row to handle the histograms
  void one_E_row( double min_E, double max_E );

  //! for E_cm increasing with E_lab
  void cm_E_incr( double min_E, double max_E );

  //! for E_cm decreeasing with increasing E_lab
  void cm_E_decr( double min_E, double max_E );

  //! Copy the parameters
  void copy_params( Param *params );

  //! Copy the pointer to the data
  void copy_data(  list< Kalbach_data >::iterator data_ptr );

  //! Evaluate the Kalbach a function for the slope value
  double get_a( double E_out );

  //! The Kalbach distribution function
  double f( double E );
};

// ----------- class two_d_Kalbach -----------------
//! Class for one incident energy for the ENDF three dimension Kalbach data
class two_d_Kalbach : public three_d_link< one_d_Kalbach >
{
private:
  list< Kalbach_data >::iterator data_ptr_; // point to the current data

  Param *params_;   // 5 parameters for the quadrature routine

public:

  // ********* implement virtual functions *******************
  //! Append a new one_d_phase and set its parameters
  two_d_Kalbach::iterator new_one_d( two_d_Kalbach::iterator where );
  // ********** end of virtual functions *********************

  //! Set the parameters
  void set_params( list< Kalbach_data >::iterator data_ptr,
    Param *params );

};

// ----------- class three_d_Kalbach -----------------
//! Class to handle the ENDF three dimension Kalbach data
class three_d_Kalbach : public three_d_list< two_d_Kalbach >
{
private:
  double S_a;  // Kalbach S_a value for this reaction
  double M_a;  // Kalbach M value for the incident particle
  double m_b;  // Kalbach m value for the ejected particle
  int this_mult_; // multiplicity of particle (if constant)
  Param Params;   // 5 parameters for the quadrature routine

  list<Kalbach_data> triples;  // the lists of ENDF/B-VI data

  list<Kalbach_data>::iterator Data_ptr; // point to the current data

  //! Evaluates the Kalbach S function for the separation energy
  double get_S(Nuclei& targ_A, Nuclei& proj_a);

  //! Evaluate the Kalbach a function for the slope value
  double get_a(double E_out);

  //! Expands the double differential data for single incident energy
  void expand_E_in( three_d_Kalbach::iterator e_in_link );

  //! Reads in all of the ENDF/B-VI data from an intermediate file
  void read_data( mf6_file& inFile, int num_E_in );

  //! Expands the data for one incident energy
  void one_e_in( );

public:
  three_d_Kalbach( int targ_ZA, int proj_ZA, int eject_ZA, int this_mult );
  ~three_d_Kalbach() {}

  //! Reads the data from an intermediate file and expands it to a list
  void expand_data(int num_E_in, mf6_file& inFile);

  void renorm(void);

  //! Calculates the average energy of the outgoing particle
  void check_i10( );
};

// ------------------- av_energy routine ------------------
//! Integrand for the average energy integral
double av_energy( double E_b, Param& params );

#endif
