/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1868 $
 * $Date: 2006-05-15 12:21:47 -0700 (Mon, 15 May 2006) $
 * $Author: dbrown $
 * $Id: list_1d.cpp 1868 2006-05-15 19:21:47Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the one_d_list class

#include "list_1d.hpp"
#include "math_util.hpp"
#include "global_params.hpp"
#include "record_types.hpp"   // used in read_data
#include "messaging.hpp"
#include "endl_precision.hpp"
#include "endl_formats.hpp"

extern GlobalParameterClass Global;
extern ENDLClass ENDL;

#define IS_ODD(n)  ((n) & 1)
#define IS_EVEN(n) (!(IS_ODD(n)))
#define ENDL_FIELD_WIDTH     ENDL_DATA_PRECISION+7

using namespace std;

// *********** dd_list::interpolation functions ******
// -------------------- dd_list::interp ----------------
double dd_list::interp(double E, dd_link& left_link,
  dd_link& right_link, int Interp_Type)
{

  if( Interp_Type == 1 ) // histogram
  {
    return hist_interp( E, left_link, right_link );
  }
  else if( ( Interp_Type == 2 ) ||     // lin-lin
           ( Interp_Type == -7 ) )     // expanded to lin-lin
  {
    return linlin_interp( E, left_link, right_link );
  }
  else if( Interp_Type == 3 ) // log-lin
  {
    return loglin_interp( E, left_link, right_link );
  }
  else if( Interp_Type == 4 ) // lin-log
  {
    return linlog_interp( E, left_link, right_link );
  }
  else if( Interp_Type == 5 ) // log-log
  {
    return loglog_interp( E, left_link, right_link );
  }
  else
  {
    Unimplemented("dd_list::interp",
        pastenum("Implement dd_list::interp 1 for Interp_Type: ",Interp_Type));
  }
  return -1; // should not get here
}
// -------------------- dd_list::hist_interp ----------------
double dd_list::hist_interp( double E, dd_link& left_link,
    dd_link& right_link )
// perform histogram interpolation
{
  // check for improper input
  if( (E < left_link.x) || (E > right_link.x) ||
       (right_link.x < left_link.x) )
  {
    SevereError("dd_list::hist_interp","improper input in hist_interp 1");
  }

  return left_link.y;
}
// -------------------- dd_list::linlin_interp --------------
double dd_list::linlin_interp( double E, dd_link& left_link,
    dd_link& right_link )
{
  double E_diff = right_link.E_out() - left_link.E_out();

  // check for division by 0
  if(E_diff == 0.0)
  {
    Warning("dd_list::linlin_interp","attempted division by 0 in linlin_interp 1");
    return (left_link.y+right_link.y)/2.0;
  }

  double alpha = (E - left_link.E_out())/E_diff;

  // check for extrapolation
  if((alpha < 0.0) || (alpha > 1.0))
  {
    Warning("dd_list::linlin_interp",
        pastenum("extrapolation in linlin_interp: alpha=",alpha)
        +pastenum(" at E=",E));
  }

  return (alpha*right_link.y +
    (1.0 - alpha)*left_link.y);
}
// -------------------- dd_list::loglin_interp --------------
double dd_list::loglin_interp( double E, dd_link& left_link,
    dd_link& right_link )
// interpolation for y = a + b*log(x/x_0)
{
  double x_ratio = right_link.x / left_link.x;

  // check for division by 0 and bad logarithm
  if( (x_ratio <= 0.0) || (x_ratio == 1.0) )
  {
    SevereError("dd_list::loglin_interp",
        pastenum("bad x_ratio in loglin_interp 1: ",x_ratio));
  }

  double b = ( right_link.y - left_link.y ) / log( x_ratio );

  // check for extrapolation
  if( ( (E > left_link.x) && (E > right_link.x) ) ||
      ( (E < left_link.x) && (E < right_link.x) ) )
  {
    SevereError("dd_list::loglin_interp",
        "extrapolation in loglin_interp 1");
  }

  return ( left_link.y + b*log( E / left_link.x ) );
}
// -------------------- dd_list::linlog_interp ------------------
double dd_list::linlog_interp( double E, dd_link& left_link,
    dd_link& right_link )
// interpolation for log(y) = log(y_0) + b*(x - x_0)
{
  // check for proper inputs
  if( (left_link.y <= 0.0) || (right_link.y <= 0.0) ||
      (left_link.x == right_link.x) )
  {
    SevereError("dd_list::linlog_interp",
        "Bad inputs to linlog_interp 1 ");
  }

  double b = log( right_link.y / left_link.y ) /
    ( right_link.x - left_link.x );

  // check for extrapolation
  if( ( (E > left_link.x) && (E > right_link.x) ) ||
      ( (E < left_link.x) && (E < right_link.x) ) )
  {
    SevereError("dd_list::linlog_interp",
        "extrapolation in linlog_interp 1");
  }

  return ( left_link.y * exp( b * ( E - right_link.x ) ) );
}
// -------------------- dd_list::loglog_interp -------------------
double dd_list::loglog_interp( double E, dd_link& left_link,
    dd_link& right_link )
// interpolation for log(y) = log(y_0) + b*log(x/x_0)
{
  if( (right_link.y <= 0.0) || (left_link.y <= 0.0) ||
      (right_link.x <= 0.0) || (left_link.x <= 0.0) ||
      (right_link.x == left_link.x) )
  {
    SevereError("dd_list::loglog_interp",
        "Bad inputs to loglog_interp 1 ");
  }

  double b = log( right_link.y / left_link.y ) /
    log( right_link.x / left_link.x );

  // check for extrapolation
  if( ( (E > left_link.x) && (E > right_link.x) ) ||
      ( (E < left_link.x) && (E < right_link.x) ) )
  {
    SevereError("dd_list::loglog_interp",
        "extrapolation in loglog_interp 1");
  }

  return ( left_link.y * exp( b*log( E / left_link.x ) ) );
}

// ***************************************************
//              dd_list::zeroin function
// ***************************************************
double dd_list::zeroin(double (*func)(double, double), 
	      double target,
	      dd_link& BB, 
	      dd_link& CC,
              double alpha, 
	      double tol)
// find a root of func(x, alpha) = target
// between BB and CC.
// Return the root with accuracy of tol + 4*EPS*|root|

/* This routine is a translation from the Brent zeroin
 * from netlib.  The original is available by e-mail:
 *   e-mail: netlib@ornl.gov
 *   subject: send zeroin from go
*/
{
  //Stash the original 1d_links
  dd_link B = BB;
  dd_link C = CC;

  const int MAX_ITER = 100;  // the maximum number of tries

  double rel_tol;     // bound on relative error
  double diff_cb;   // (c - b)/2
  double new_width; // |diff_cb|

  double num;    // fraction for the secant rule
  double denom;

  int num_poor = 0;  // how many poor improvements

  // use a reasonable tolerance
  double use_tol = (tol < EPS) ? EPS : tol;

  // the current interval
  double old_width = abs(C.x - B.x);

  // shift by the target
  B.y -= target;
  C.y -= target;

  // worse error
  double worse = (abs(B.y) < abs(C.y)) ? abs(C.y) : abs(B.y);

  // the oldest estimate
  dd_link A = C;

  for(int count = 0; count < MAX_ITER; ++count)
  {
    // make B be the best estimate so far
    if(abs(C.y) < abs(B.y))
    {
      A = B;
      B = C;
      C = A;
    }

    // how close are we?
    diff_cb = 0.5*(C.x - B.x);
    new_width = abs(diff_cb);
    rel_tol = use_tol*abs(B.x) + EPS;
    if(new_width <= rel_tol)
    {
      if(B.y*C.y > 0.0)
      {
	SevereError("dd_list::zeroin","No root found" );
      }
      else if(abs(B.y) > worse)
      {
	SevereError("dd_list::zeroin","This looks like a pole.");
      }
      else
      {
	return B.x;
      }
    }

    // set up the secant iteration
    num = (B.x - A.x)*B.y;
    denom = A.y - B.y;

    // arrange so that num >= 0
    if(num < 0.0)
    {
      num *= -1;
      denom *= -1;
    }

    // save the best so far
    A = B;

    // have we had too many poor ones?
    ++num_poor;
    if((num_poor >= 4) && (8.0*new_width < old_width))
    {
      num_poor = 0;
      old_width = new_width;
    }

    // which type of iteration?
    if(num_poor >= 4)
    {
      B.x = 0.5*(C.x + B.x);
    }
    else if(num < abs(denom)*rel_tol)
    {
      // too small a change
      B.x += (diff_cb > 0.0) ? rel_tol : -rel_tol;
    }
    else if(num < denom*diff_cb)
    {
      // secant rule if x between B and (C + B)/2
      B.x += num/denom;
    }
    else
    {
      // bisection
      B.x = 0.5*(C.x + B.x);
    }

    // the new function value
    B.y = func(B.x, alpha) - target;

    // did we hit it?
    if(B.y == 0.0)
    {
      return B.x;
    }

    // which old point do we keep?
    if(B.y*C.y > 0.0)
    {
      C = A;
    }
    //    cout << B.x << "  " << B.y << endl;
  }

  // if we got here, there were too many iterations
  Warning("dd_list::zeroin","Too many iterations in zeroin");
  return B.x;  //However, we MUST return something or exit()?
}

// *********** dd_list class ***********************
// ----------- dd_list::unit_base -----------------
void dd_list::unit_base( )
// map the energies to 0 <= E' <= 1
{
  dd_list::iterator link = begin( );  // the first link
  dd_list::iterator last_link = end( );
  --last_link;
    
  // save the minimum and maximum outgoing energies
  min_E_out = link->E_out( );  // lowest energy
  max_E_out = last_link->E_out( );  // highest
    
  double scale = max_E_out - min_E_out;  // energy range
    
  // loop through the list
  for( ; link != end( ); ++link )
  {
    // scale this link
    link->E_out( ) = ( link->E_out( ) - min_E_out )/scale;
  }
}
// ----------- dd_list::rescale -----------------
void dd_list::rescale( )
// map the energies from 0 <= E' <= 1 to  min_E_out <= E' <= max_E_out
{
  double scale = max_E_out - min_E_out;  // energy range
    
  // loop through the list
  for(dd_list::iterator link = begin( ); link != end( ); ++link)
  {
    // scale this link
    link->E_out( ) = min_E_out + scale*link->E_out( );
  }
}
// ----------- dd_list::read_data -----------------
void dd_list::read_data( ifstream& inFile, int NP, double x_factor,
  double y_factor )
// read in the weight for this model and scale by (x_factor, y_factor)
{
string linebuff;
string strbuff;
dd_link XYdata;  // for links in our lists

for( int iNP=0; iNP < NP; iNP++ ) //Loop over NP data pairs
  {
    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 )
    {
      getline( inFile, linebuff); //Read in new line
    }
    strbuff = linebuff.substr(iP*22, 22); //Grab the right pair
    read_dd(&strbuff, &XYdata.x, &XYdata.y );
    // convert to MeV
    XYdata.x *= x_factor;
    XYdata.y *= y_factor;
    insert(end(), XYdata);
  }
}
// ----------- dd_list::print -----------------
void dd_list::print()
// to print the list
{
  cout.setf(ios::scientific,ios::floatfield);

  // loop through the list
  for(dd_list::iterator link = begin(); link != end();
    ++link)
  {
    cout << ENDL.data( link->x, link->y ) << endl;
  }
  cout << endl;
}

// ----------- dd_list::needs_widening ------------
//! Tests if data actually needs jump widening
bool dd_list::needs_widening( )
{
  dd_list::iterator itr = begin();
  double last_x=itr->x;
  ++itr;
  for(; itr!=end(); ++itr)
  {
    if (itr->x >= last_x+ENDL_EPSILON(last_x)) return true;
  }
  return false;
}

// ----------- dd_list::widen_jumps ------------
//! shift duplicate x-values in the list
void dd_list::widen_jumps( double cluster_min )
{
  // scale dE by the smoother used for histograms
  double local_eps, jump_width;

  // container to hold links to elements we need to widen
  dd_list::iterator cluster_start = begin();
  dd_list::iterator cluster_end = begin();
  dd_list::iterator this_link = begin();
  int cluster_size=1;

  for( ++this_link; this_link != end( ); ++this_link )
  {
    // get the local energy epsilon (smallest dE can see in output file)
    local_eps = ENDL_EPSILON( cluster_start->x );
    jump_width = ENDL_JUMP_WIDTH( cluster_start->x );

    // this_link is too close to the cluster_end, so add it to the cluster
    if( abs( this_link->x - cluster_end->x ) <= local_eps )
    {
      cluster_end = this_link;
      ++cluster_size;
    }
    // this_link is far from the cluster_end, so let's process the cluster
    else 
    {
      if ( cluster_size > 1 ) 
        widen_cluster( cluster_start, cluster_end, cluster_size, jump_width, cluster_min );
      // clear the clusters
      cluster_start = ++cluster_end;
      cluster_size  = 1;
    }
  }
  // process the last cluster since doesn't get processed otherwise
  if ( cluster_size > 1 ) 
    widen_cluster( cluster_start, cluster_end, cluster_size, jump_width, cluster_min );
}

// ----------- dd_list::widen_cluster ------------
//! Utility code used by widen_jumps
void dd_list::widen_cluster( dd_list::iterator cluster_start, 
  dd_list::iterator cluster_end, int cluster_size, double jump_width, double cluster_min )
{

  if( ( cluster_size == 1 ) || ( cluster_start==cluster_end ) ) return;
  if ( cluster_size>4 ) 
  {
    Warning("dd_list::widen_cluster",pastenum("Found ",cluster_size)+
      pastenum(" links in a cluster, check ENDL_EPSILON and ENDL_JUMP_WIDTH values or the data is broken at x=",
      cluster_start->x));
    return;
  }

  if ( cluster_start->x <= cluster_min+jump_width )
  {
    dd_list::iterator cluster_middle = cluster_start; 
    for (int ijump = 1; ijump<=cluster_size; ++ijump)
    {
      ++cluster_middle; 
      cluster_middle->x += static_cast<double>(ijump)*jump_width;
    }
  }
  else
  {
    switch (cluster_size)
    {
      case 2:
        cluster_start->x -= 0.5*jump_width;
        cluster_end->x   += 0.5*jump_width;
        break;
      case 3:
        cluster_start->x -= jump_width;
        cluster_end->x   += jump_width;
        break;
      case 4:
        {
        dd_list::iterator cluster_left = cluster_start;
        dd_list::iterator cluster_right = cluster_end;
        ++cluster_left;--cluster_right;
        cluster_start->x -= 1.5*jump_width;
        cluster_left->x  -= 0.5*jump_width;
        cluster_right->x += 0.5*jump_width;
        cluster_end->x   += 1.5*jump_width;
        }
        break;
      default:
        Warning("dd_list::widen_cluster",pastenum("Found ",cluster_size)+
        pastenum(" links in a cluster, check ENDL_EPSILON and ENDL_JUMP_WIDTH values or the data is broken at x=",
        cluster_start->x)); 
        return;
    }
  }
}

// ----------- dd_list::thinit -----------------
void dd_list::thinit()
// thin the list to within tolerance: tol
{
  // declare pointers to the test links
  dd_list::iterator left_link;
  dd_list::iterator mid_link;
  dd_list::iterator right_link;
  dd_list::iterator all_done;
  bool thin_ok;  // for test on 1 link
  bool do_thin;  // thin OK at this level
  bool prev_thin;  // thin OK at the previous level
  const double cut_off = Global.Value("cut_off_1d"); // don't impose accuracy below this value
  const double tol_1d = Global.Value("tol_1d");  // accuracy of linear interpolation

  //  quit at the last valid link
  all_done = end();
  --all_done;

  // set the noise level
  double noise = 0.0;
  // don't use it for delta functions
  if( ( ENDL.F != 12 ) || ( ENDL.F != 13 ) )
  {
    for(left_link = begin(); left_link != all_done; ++left_link)
    {
      if(abs(left_link->y) > noise)
      {
        noise = abs(left_link->y);
      }
    }
    noise *= cut_off;
  }

  // loop through the list of left-hand anchors
  for(left_link = begin(); left_link != all_done; ++left_link)
  {
    // make sure that we have data to check
    mid_link = left_link;
    ++mid_link;
    right_link = mid_link;
    ++right_link;
    if((mid_link == end()) || (right_link == end()))
    {
      return;  // no more data
    }

    // the default is no thinning
    prev_thin = false;

    // how long a span may we delete?
    for( ; right_link != end(); ++right_link)
    {
      // the default is don't delete links at this level
      do_thin = false;

      // check all of the intermediate links
      mid_link = left_link;
      for(++mid_link; mid_link != right_link; ++mid_link)
      {
        thin_ok = check_interp(left_link, mid_link, right_link,
          tol_1d, noise);
        if(thin_ok)
        {
          // at least one deletion is possible
          do_thin = true;
        }
        else
        {
          // this section fails
          do_thin = false;
          break;
        }
      }

      // did they all pass?
      if(do_thin)
      {
        prev_thin = true;
      }
      else
      {
        break;
      }
    }

    dd_list::iterator left_next = left_link;
    ++left_next;
    dd_list::iterator right_prev = right_link;
    --right_prev;

    if((prev_thin) && (!do_thin))
    {
      // deletion is not allowed here but it is OK at the previous level
      erase(left_next, right_prev);
    }
    else if((do_thin) && (right_link == end()) &&
      (mid_link == all_done))
    {
      // we are at the end and all links passed
      erase(left_next, mid_link);
    }
  }
}
// -------------------- dd_list::copy ----------------
void dd_list::copy( dd_list& list_2 )
// Copy one list into another
{
  // loop through list_2
  for( dd_list::iterator link_ptr = list_2.begin();
    link_ptr != list_2.end(); ++link_ptr )
  {
    dd_link new_link = dd_link( link_ptr->x, link_ptr->y );
    insert( end(), new_link );
  }
  tag = list_2.tag;
}
// ----------- dd_list::chop ------------
void dd_list::chop( double max_E )
// Cuts off links with x-value exceeding max_E
{
  dd_list::iterator list_ptr = begin( );
  double dE = ENDL_EPSILON( max_E );

  // Make sure list starts below max_E
  if( begin( )->x  > max_E )
  {
    SevereError( "dd_list::chop", "all energies too big" );
  }

  // Find the first link with incident energy too big
  for( ; list_ptr != end( ); ++list_ptr )
  {
    if( list_ptr->x >= max_E + dE )
    {
      break;
    }
  }
  
  // Move the last link to max_E, then erase all the links above it
  if( list_ptr != end( ) )
  {
    list_ptr->y = evaluate(max_E);
    list_ptr->x = max_E; 
    ++list_ptr;
    if( list_ptr != end( ) ) erase( list_ptr, end( ) );
  }
  
}
// ------------------------- write_endl -----------------
//! ENDL file writing function
void dd_list::write_endl( int I )
{
  if( !ENDL.write_file ) return;

  // chop off the energies that we don't use
  // for 2-column data the first entry is the incident energy
  chop( ENDL.Max_E_in );

  ENDL.set_I_number(I);

  fstream endl_file;
  string file_name = ENDL.file_name;

  if(empty())
  {
    Warning("dd_list::write_endl","The file "+file_name+" is empty.");
    return;
  }

  if ( ENDL.new_file() )
  {
    endl_file.open(file_name.c_str(),ios::out);
    Info("dd_list::write_endl","Opening ENDL file "+file_name);
  }
  else
  {
    endl_file.open(file_name.c_str(),ios::out|ios::app);
    Info("dd_list::write_endl","Appending ENDL file "+file_name);
  }

  //Write out the header information
  endl_file<<ENDL.header_line_1<<"\n";
  endl_file<<ENDL.header_line_2<<"\n";
  endl_file.setf(ios::scientific,ios::floatfield);
  

  // check for & fix double value of x, except for cross sections
  double old_x = -2.0;
  bool list_needs_widening=false;
  for(dd_list::iterator link = begin(); link != end(); ++link)
  {
    if( (link->x == old_x) && (ENDL.F != 3) )
    {
      Warning("dd_list::write_endl",
        pastenum("Widening the jump in dd_list at: ",old_x)+
        pastenum(" x: " ,old_x));
      list_needs_widening=true;
    }
    old_x = link->x;
  }
  if (list_needs_widening) widen_jumps();

  //Loop through the data and write to file
  for(dd_list::iterator link = begin(); link != end(); ++link)
  {
    endl_file<<ENDL.data( link->x, link->y )<<endl;
  }

  //Put on end of file/section line into output file
  endl_file<<ENDL.eof_line<<endl;
  
  //Finally, we close the output file
  Info("dd_list::write_endl","Closing ENDL file "+file_name);
  endl_file.close();
}

// ----------- dd_list::operator+= ------------
dd_list& dd_list::operator+=(dd_list& list_2)
{
  // ensure that the x-values are identical
  fill_in_lists( *this, list_2 );

  // now go through the lists
  dd_list::iterator L1 = begin();
  dd_list::iterator L2 = list_2.begin();
  for( ; L1 != end(); ++L1, ++L2 )
  {
    L1->y += L2->y;
  }

  return *this;
}

// ----------- dd_list::line_sum ------------
dd_list& dd_list::line_sum( dd_list& list_2, double tol )
{
  dd_link new_link;  // for creating new links
  dd_list::iterator L1 = begin();
  dd_list::iterator L2 = list_2.begin();
  double E1 = L1->x;
  double E2 = L2->x;
  bool L1_done = empty( );  // have we finished the first list?

  // go through the lists
  for( ; L2 != list_2.end( ); )
  {
    if( L1_done || ( E2 < ( 1.0 - tol )*E1 ) )
    {
      new_link.x = E2;
      new_link.y = L2->y;
      insert( L1, new_link);
      ++L2;
      E2 = L2->x;
    }
    else if( !L1_done )
    {
      if( E1 < ( 1.0 - tol )*E2 )
      {
        ++L1;
        E1 = L1->x;
      }
      else   //  E1 == E2
      {
        L1->y += L2->y;
        ++L1;
        ++L2;
        E1 = L1->x;
        E2 = L2->x;
      }
      if( L1 == end( ) )
      {
        L1_done = true;
      }
    }
  }

  return *this;
}

// ----------- dd_list::join_lines ------------
//! If two adjacent lines have the same frequency, this routine adds
//! their intensities.
void dd_list::join_lines( )
{
  if( size( ) < 2 ){
    return;
  }
  dd_list::iterator this_ptr = begin();
  dd_list::iterator next_ptr = this_ptr;
  ++next_ptr;
  for(  ; this_ptr != end( ); this_ptr = next_ptr, ++next_ptr )
  {
    if( this_ptr->x == next_ptr->x )
    {
      next_ptr->y += this_ptr->y;
      erase( this_ptr );
    }
  }
}

// ----------- dd_list::operator*= ------------
dd_list& dd_list::operator*=(double X)
{
  for(dd_list::iterator L1 = begin(); L1 != end(); ++L1 )
  {
    L1->y *= X;
  }
  return *this;
}

// ----------- dd_list::operator*= ------------
dd_list& dd_list::operator*=(dd_list& list_2)
{
  for( dd_list::iterator itr = begin() ; itr != end(); ++itr )
  {
    itr->y *= list_2.evaluate( itr->x );
  }

  return *this;
}

// ----------- dd_list::operator/= ------------
dd_list& dd_list::operator/=(dd_list& list_2)
{
  double y;
  for( dd_list::iterator itr = begin() ; itr != end(); ++itr )
  {
    y = list_2.evaluate( itr->x ); 
    if ( y == 0.0 ) 
    {
      if ( itr->y != 0.0 )
        Warning("dd_list::operator/=","division by 0 in operator /=");
      itr->y = 0;
    }
    else
    {
      itr->y /= y;
    }
  }

  return *this;
}

// ----------- dd_list::find_next -----------------
//! Find the link with the next larger x-value
dd_list::iterator dd_list::find_next( double X )
{
  if( empty( ) )
  {
    return end( );
  }

  dd_list::iterator next_link;
  for( next_link = begin( ); next_link != end( ); ++next_link )
  {
    if( next_link->x >= X )
    {
      return next_link;
    }
  }

  // we didn't find it
  return end( );
}

// ----------- dd_list::evaluate -----------------
//! Find the y-value corresponding to this x using lin-lin interpolation
double dd_list::evaluate( double X )
{
  if( empty( ) )
  {
    SevereError("dd_list::evaluate",
        "Attempt to use evaluate on an empty list");
  }

  // find the first bigger gamma energy
  dd_list::iterator next_link = find_next( X );

  if( ( next_link == begin( ) ) && ( next_link->x != X ) )
  {
    Warning("dd_list::evaluate",
        pastenum("Attempt to evaluate below the first link, x=",X));
    return next_link->y;
  }
  else if( next_link == end( ) )
  {
    Warning("dd_list::evaluate",
        pastenum("Attempt to evaluate past the last link, x=",X));
    return (--next_link)->y;
  }

  if (next_link->x == X)
  {
    return next_link->y;
  }
  else
  {
    dd_list::iterator prev_link = next_link;
    --prev_link;
    return linlin_interp( X, *prev_link, *next_link );
  }
}

// ----------- dd_list::check_interp -----------------
//! check the accuracy of linear interpolation
//! Ignore values below the noise.
bool dd_list::check_interp(dd_list::iterator left_link,
    dd_list::iterator mid_link,
    dd_list::iterator right_link, double tol, double noise)
{
  // the interpolated probability
  double P_interp = linlin_interp( mid_link->E_out(), *left_link, 
    *right_link );

  return ( ( abs(P_interp - mid_link->y) < tol*mid_link->y ) ||
           ( ( abs(left_link->y) <= noise ) &&
             ( abs(mid_link->y) <= noise ) &&
             ( abs(right_link->y) <= noise ) ) );
}

// ----------- dd_list::expand_interp -----------------
//! Expand the list to permit linear-linear interpolation
void dd_list::expand_interp( double Max_E )
{
  if(INT[0] == -7)
  {
    //    SevereError("dd_list::expand_interp","expand_interp called twice");
    return;
  }

  dd_list::iterator left_link = begin();

  for(int count = 1; count <= NBT.size(); ++count)
  {
    // pointer to the end of this segment
    dd_list::iterator right_link;
    if(count == NBT.size())
    {
      right_link = end();
      --right_link;
    }
    else
    {
      int j = NBT[count] - NBT[count-1];

      // get the corresponding link
      right_link = left_link;
      for(int k = 0; k < j; ++k)
      {
        ++right_link;
      }
    }
    // convert to lin-lin
    switch(INT[count-1])
    {
    case 1: // histogram
      Hist_2_LinLin(left_link, right_link, Max_E);
      break;
    case 2: // lin-lin
      break;
    case 5: // log-log
      LogLog_2_LinLin(left_link, right_link);
      break;
    default:
      Unimplemented("dd_list::expand_interp",
        pastenum("Implement interpolation model ",INT[count-1]));
    }
    // save for the next set
    left_link = right_link;
  }
  // flag to ensure that we don't do this twice on the same data
  INT[0] = -7;
} 

// ----------- dd_list::expand_interp -----------------
//! This version is used for gammas, in which LEP is the interpolation type
//! for use with the gamma energy
void dd_list::expand_interp( int LEP, double Max_E )
{
  if ( LEP == 2 ) // lin-lin
  {
  }
  else if ( LEP == 1 ) // histogram
  {
    dd_list::iterator data_ptr = begin( );

    // We could be finished already
    if( data_ptr == end( ) )
    {
      return;
    }

    // loop through the list
    dd_list::iterator prev_ptr = data_ptr;
    ++data_ptr;
    dd_list::iterator next_ptr = data_ptr;
    ++next_ptr;
    for(  ; data_ptr != end( );
        prev_ptr = data_ptr, data_ptr = next_ptr, ++next_ptr )
    {
      if( prev_ptr->y != data_ptr->y )
      {
        // how wide the jump
        double dE_L = data_ptr->x - prev_ptr->x;
        double dE_R = ( next_ptr == end( ) ) ? data_ptr->x :
          next_ptr->x - data_ptr->x;
        double dE = ( dE_L < dE_R ) ? dE_L : dE_R;

        if(dE <= 0.0)
        {
          SevereError("dd_list::expand_interp","bad histogram data");
        }

        // set smoothing dE to be smallest dE we can see in output file
        dE = ENDL_EPSILON( data_ptr->x );

        if( data_ptr->x < Max_E - dE )
        {
          dd_link new_link;
          insert( data_ptr, new_link );
          dd_list::iterator new_ptr = data_ptr;
          --new_ptr;  //  points to the new link

          new_ptr->x = data_ptr->x - dE;
          data_ptr->x += dE;
          new_ptr->y = prev_ptr->y;
        }
        else
        {
          data_ptr->x = Max_E;
          data_ptr->y = prev_ptr->y;
          erase( next_ptr, end( ) );
          break;
        }
      }
    }
  }
  else
  {
    Unimplemented("dd_list::expand_interp( int LEP, double Max_E )",
        pastenum("Implement interpolation type: ",LEP));
  }
} 

// ----------- dd_list::Hist_2_LinLin -----------------
//! Expand histogram data to linear-linear between the two consecutive links.
//! If we need a jump, insert a new link
void dd_list::Hist_2_LinLin( dd_list::iterator left_link, double Max_E )
{
  if ( left_link == end() ) SevereError( "dd_list::Hist_2_LinLin",
    "Trying to expand links, but starting link == end()!");

  dd_link new_link;  // for making new links

  dd_list::iterator jump_link = left_link;
  ++jump_link;

  // special treatment of the end of the list
  if( jump_link == end( ) )
  {
    if(left_link->x < Max_E)
    {
      // add a final link
      new_link.x = Max_E;
      new_link.y = left_link->y;
      insert( jump_link, new_link );
    }
  }
  else
  {
    // is there a jump?
    if( jump_link->y != left_link->y )
    {
      // make a new link
      dd_list::iterator right_link;
      right_link = jump_link;
      ++right_link;
      double dE_L = jump_link->x - left_link->x;
      double dE_R = ( right_link == end( ) ) ? jump_link->x :
        right_link->x - jump_link->x;
      double dE = ( dE_L < dE_R ) ? dE_L : dE_R;

      // set width of jump to be smallest dE we can see in output file
      double dEmin = ENDL_EPSILON( jump_link->x );

      if( dE < 0.0 )
      {
        SevereError("dd_list::Hist_2_LinLin","energies out of order");
      }

      // the original x-values may be too close to see in the output file
      if( dE_R <= dEmin )
      {
        jump_link->x = right_link->x - dEmin;
        jump_link->y = left_link->y;
      }
      else if( dE > dEmin )
      {
        // add a link
        new_link.x = jump_link->x - dEmin;
        new_link.y = left_link->y;
        insert( jump_link, new_link );
        jump_link->x += dEmin;
      }
    }
  }
}

// ----------- dd_list::Hist_2_LinLin -----------------
//! Expand histogram data to linear-linear between the two links
//! first_link and last_link.
void dd_list::Hist_2_LinLin( dd_list::iterator first_link,
  dd_list::iterator last_link, double Max_E )
{
  dd_list::iterator next_link = first_link;
  ++next_link;
  for ( dd_list::iterator this_link = first_link; this_link != last_link;
    this_link = next_link, ++next_link )
  {
    if( this_link == end( ) )
    {
      SevereError("dd_list::Hist_2_LinLin","bad link");
    }
    Hist_2_LinLin( this_link, Max_E );
  }
}

// ----------- dd_list::LogLog_2_LinLin -----------------
//! Expand the log-log data to linear-linear between the two 
//! consecutive links.
void dd_list::LogLog_2_LinLin(dd_list::iterator left_link)
{
  double tol = Global.Value("log_log_tol");

  dd_list::iterator right_link = left_link;
  ++right_link;
  int num_add; // the number of intermediate links to add

  // Make sure that x values are positive
  if ( left_link->x <= 0.0 || right_link->x <= 0.0 )
  {
    SevereError("dd_list::LogLog_2_LinLin","x must be larger than zero!");
  }
  // Make sure that y values are positive
  if ( left_link->y <= 0.0 || right_link->y <= 0.0 )
  {
    SevereError("dd_list::LogLog_2_LinLin","y must be larger than zero!");
  }
  // Make sure these links are ordered
  if ( left_link->x > right_link->x )
  {
    SevereError("dd_list::LogLog_2_LinLin","x values not ascending!");
  }

  // Calculate slope of loglog
  double alpha = log(right_link->y/left_link->y)/
    log(right_link->x/left_link->x); 

  // we may not need to add intermediate links
  double xi = right_link->x/left_link->x;
  double maxE = maxE_loglog( xi, alpha );
  if(maxE <= tol)
  {
    num_add = 0;
  }
  else
  {
    // dummy links for zeroin
    dd_link low_link = dd_link(1.0, 0.0);  // low estimate for xi
    dd_link high_link = dd_link(xi, maxE); // high estimate for xi

    // A Taylor series often gives a better estimate
    double delta = sqrt(8.0*tol/abs(alpha*(alpha - 1.0)));
    if(1.0 + delta < xi)
    {
      maxE = maxE_loglog( 1.0 + delta, alpha );
      if(maxE < tol)
      {
      // this is a lower bound
        low_link.x = 1.0 + delta;
        low_link.y = maxE;
      }
      else
      {
        // this is an upper bound
        high_link.x = 1.0 + delta;
        high_link.y = maxE;
      }
    }
    // get the ideal interval
    xi = zeroin(&maxE_loglog, tol, low_link, high_link, alpha, 0.00001);

    num_add = static_cast<int>
      (floor(log(right_link->x/left_link->x)/log(xi)));

    // add the intermediate links
    double eta = pow(right_link->x/left_link->x, 1.0/(num_add + 1.0));
    double eta_alpha = pow(eta, alpha);
    double x = left_link->x;
    double y = left_link->y;
    for(int count = 0; count < num_add; ++count)
    {
      x *= eta;
      y *= eta_alpha;
      dd_link link;
      link.x = x;
      link.y = y;
      insert(right_link, link);
    }
  }
}

// ----------- dd_list::LogLog_2_LinLin -----------------
//! Expand the log-log data to linear-linear between the two links
//! first_link and last_link.
void dd_list::LogLog_2_LinLin(dd_list::iterator first_link,
  dd_list::iterator last_link)
{
  dd_list::iterator next_link = first_link;
  ++next_link;
  for (dd_list::iterator this_link = first_link; this_link != last_link;
    this_link = next_link, ++next_link)
  {
    if((this_link == 0) || (this_link == end()))
    {
      SevereError("dd_list::LogLog_2_LinLin"," bad link");
    }
    LogLog_2_LinLin(this_link);
  }
}

// ---------------- fill_in_lists -------------------------
//! This routine fills in to ensure that the x-values in list_1 and
//! list_2 are the same.
void fill_in_lists( dd_list& list_1, dd_list& list_2 )
{
  if( list_1.empty( ) || list_2.empty( ) )
  {
    cout << "list_1\n";
    list_1.print( );
    cout << "list_2\n";
    list_2.print( );
    SevereError("fill_in_lists","attempt to fill in an empty 1d list");
  }

  // widen any jumps
  list_1.widen_jumps( );
  list_2.widen_jumps( );

  // pad the start with zeros if necessary
  dd_list::iterator L1 = list_1.begin( );
  dd_list::iterator L2 = list_2.begin( );

  double E1 = L1->x;
  double E2 = L2->x;

  if( E1 < E2 )
  {
    list_2.pad_head( E1 );
  }
  else if( E2 < E1 )
  {
    list_1.pad_head( E2 );
  }

  // pad the tail with zeros if necessary
  L1 = list_1.end( );
  L2 = list_2.end( );
  --L1;
  --L2;

  E1 = L1->x;
  E2 = L2->x;

  if( E1 > E2 )
  {
    list_2.pad_tail( E1 );
  }
  else if( E2 > E1 )
  {
    list_1.pad_tail( E2 );
  }

  // make a list of the common E_in values
  list< double > E_in_list;
  list_1.collect_E_in( E_in_list );
  list_2.collect_E_in( E_in_list );

  // sort the list
  E_in_list.sort( );

  // remove duplicates
  E_in_list.unique( );
  list_1.unique( );
  list_2.unique( );

  // fill each list with values from E_in_list
  list_1.fill_with( E_in_list );
  list_2.fill_with( E_in_list );
}

// ----------- dd_list::collect_E_in ------------
//! Append to E_in_list the incident energies in this list
void dd_list::collect_E_in( list< double >& E_in_list )
{
  for( dd_list::iterator this_E = begin( ); this_E != end( );
       ++this_E )
  {
    double E_in = this_E->E_in( );
    E_in_list.push_back( E_in );
  }
}

// ----------- dd_list::pad_head ------------
//! insert zero links from E0 to the head
void dd_list::pad_head( double E0 )
{
  dd_link new_link;  // for creating new links

  double first_E = begin( )->x;
  // smallest dE can see in output file
  double dE = ENDL_EPSILON( first_E );

  // insert a jump if needed
  new_link.y = 0.0;
  if( ( begin( )->y != 0.0 ) && ( first_E - E0 > dE ) )
  {
    new_link.x = first_E - dE;
    push_front( new_link );
  }

  // insert a new head
  new_link.x = E0;
  push_front( new_link );
}

// ----------- dd_list::pad_tail ------------
//! insert zero links from the tail to E0
void dd_list::pad_tail( double E0 )
{
  dd_link new_link;  // for creating new links
  dd_list::iterator L = end( );
  --L;
  double last_E = L->x;
  // smallest dE can see in output file
  double dE = ENDL_EPSILON( last_E );

  // insert a jump if needed
  new_link.y = 0.0;
  if( ( L->y != 0.0 ) && ( E0 - last_E > dE ) )
  {
    new_link.x = last_E + dE;
    push_back( new_link );
  }

  // insert a new tail
  new_link.x = E0;
  push_back( new_link );
}

// ----------- dd_list::unique ------------
//! remove duplicate energies, overrides list<>::unique()
void dd_list::unique( )
{
  dd_list::iterator this_link = begin( );
  dd_list::iterator prev_link = this_link;
  double local_epsilon;

  for( ++this_link; this_link != end( );
       prev_link = this_link, ++this_link )
  {
    if( abs(this_link->x - prev_link->x) == local_epsilon )
    {
      this_link->y = (this_link->y + prev_link->y)/2.0;
      erase( prev_link );
    }
  }
}

// ----------- dd_list::fill_with ------------
//! make sure that there are links for each energy in E_in_list
void dd_list::fill_with( list< double >& E_in_list )
{
  dd_list::iterator this_link = begin( );
  dd_list::iterator prev_link = this_link;

  // go through the E_in_list
  for( list< double >::iterator E_in_ptr = E_in_list.begin( ); 
       E_in_ptr != E_in_list.end( ); ++E_in_ptr )
  {
    double E_in = *E_in_ptr;

    // if there is a node here, just go on to the next
    if( this_link->x == E_in )
    {
      prev_link = this_link;
      ++this_link;
    }
    else
    {
      // interpolate and make a new link
      dd_link new_link;  // for creating new links
      new_link.x = E_in;
      new_link.y = linlin_interp( E_in, *prev_link, *this_link );
      insert( this_link, new_link );
    }
  }
}
