/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1867 $
 * $Date: 2006-05-15 10:19:04 -0700 (Mon, 15 May 2006) $
 * $Author: dbrown $
 * $Id: list_1d.hpp 1867 2006-05-15 17:19:04Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// header for the classes used in making lists
#ifndef ONE_D_LIST_CLASS
#define ONE_D_LIST_CLASS

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cmath>
#include <list>
#include <utility>
#include <vector>
#include <string>

using namespace std;

// ------------------------ class dd_link ---------------
//!The basic link holds a pair of doubles
class dd_link
{
public:
  double x;
  double y;

  //!Default inline constructor is zero.
  inline dd_link( )
  {
    x = 0.0;
    y = 0.0;
  }  
  //!Inline copy constructor
  inline dd_link(double xx, double yy)
  {
    x = xx;
    y = yy;
  }

  //!Inline default desctructor is blank.
  inline ~dd_link() {}

  //!Common name used to access the X value of this link.
  inline double& E_in()
  {
    return x;
  }
  //!Common name used to access the X value of this link.
  inline double& E_out()
  {
    return x;
  }
  //!Common name used to access the X value of this link.
  inline double& mu()
  {
    return x;
  }

  //!Common name used to access the Y value of this link.
  inline double& Prob()
  {
    return y;
  }

  //!Common name used to access the Y value of this link.
  inline double& x_sec()
  {
    return y;
  }
  
  //!Equality testing.
  inline bool operator==(const dd_link& other) const
  {
    return ( x == other.x && y == other.y );
  }

};

//               **************************************
//                  Here is the principal linked list
//               **************************************

// ----------------------- class dd_list -------------------
//!List of dd_links.
class dd_list : public list< dd_link >
{
private:

public:
  //! A label for the list.
  double tag;

  double min_E_out; // needed to recover from scaling to unit base
  double max_E_out;

  //! Energy at which the interpolation type changes
  vector<int> NBT;

  // the interpolation types
  //  1: histogram
  //  2: linear-linear
  //  3: y linear in log(x)
  //  4: log(y) linear in x
  //  5: log-log
  //  6: special for charged particle cross sections (see 0.19)
  //  11-15: corresponding points (see 0.20)
  //  21-25: unit base (see 0.20)
  //! Interpolation type used.
  vector<int> INT;

  //!Default constructor
  dd_list(): tag(0.0), min_E_out(0.0), max_E_out(20.0), INT(0), NBT(0) {}

  //!Common name used to access the label for this list.
  inline double& E_in()
  {
    return tag;
  }

  //! To scale the energies in the list to make 0 <= E' <= 1
  void unit_base();

  //! To expand the energies from unit base to min_E_out <= E' <= max_E_out
  void rescale();

  //! Method for printing this list.
  void print();

  //! Copy one list into another
  void copy( dd_list& list_2 );

  //! Adds one list to another, with interpolation.
  dd_list& operator+=(dd_list& list_2);

  //! Add list_2 as discrete lines (for gammas).  Here, tol measures the sameness of the energies
  dd_list& line_sum( dd_list& list_2, double tol );

  //! If two adjacent lines have the same frequency, this routine adds their intensities.
  void join_lines( );

  //! Scales the Y value of this list.
  dd_list& operator*=(double X);

  //! Multiplies the Y values by the Y values in a second list.
  dd_list& operator*=(dd_list& list_2);

  //! Divides the Y values in one list by the values in another.
  dd_list& operator/=(dd_list& list_2);

  //! Tests if data actually needs jump widening
  bool needs_widening( );

  //! Shift duplicate x-values in the list.
  void widen_jumps( double cluster_min=1e-11 );

  //! Utility code used by widen_jumps
  void widen_cluster( dd_list::iterator cluster_start, 
    dd_list::iterator cluster_end, int cluster_size, 
    double jump_width, double cluster_min );

  //! Thin the list to within global tolerance (tol_1d).
  void thinit();

  //! Insert into the list NP data pairs from inFile.
  void read_data( ifstream& inFile, int NP, double x_factor,
    double y_factor);

  //! Find the link with the next larger x-value
  dd_list::iterator find_next( double X );

  //! Find the y-value corresponding to this x using lin-lin interpolation.
  double evaluate( double X );

  //! Chops off values above the maximum ENDL energy
  void chop( double max_E );

  //! Writes the 2-column data to an ENDL file, e.g., cross sections
  void write_endl( int I );

  //! This function performs Interp_Type interpolation between two dd_links.
  double interp( double E, dd_link& left_link,
    dd_link& right_link, int Interp_Type );

  //! This function performs histogram interpolation between two dd_links.
  double hist_interp( double E, dd_link& left_link,
    dd_link& right_link );

  //! This function performs lin-lin interpolation between two dd_links.
  double linlin_interp( double E, dd_link& left_link,
    dd_link& right_link );

  //! This function performs lin-log interpolation between two dd_links.
  double linlog_interp( double E, dd_link& left_link,
    dd_link& right_link );

  //! This function performs log-lin interpolation between two dd_links.
  double loglin_interp( double E, dd_link& left_link,
    dd_link& right_link );

  //! This function performs log-log interpolation between two dd_links.
  double loglog_interp( double E, dd_link& left_link,
    dd_link& right_link );

  //find a root of func(x, alpha) = target
  //between BB and CC.
  //! Return the root with accuracy of tol + 4*tol*|root|
  double zeroin(double (*func)(double, double), 
		double target,
		dd_link& BB, 
		dd_link& CC,
		double alpha, 
		double tol);

  //! Checks the accuracy of linear interpolation.
  // Ignore values below the noise.
  bool check_interp(dd_list::iterator left_link, dd_list::iterator mid_link,
		    dd_list::iterator right_link, double tol, double noise);

  //! Expand the list to permit linear-linear interpolation.
  //! Max_E is the maximum energy for histogram data.
  void expand_interp( double Max_E);

  //! This version is used for gammas, in which LEP is the interpolation type
  //! Max_E is the maximum energy for histogram data.
  void expand_interp( int LEP, double Max_E );

  //! Expand histogram data to linear-linear between the two cosecutive links.
  void Hist_2_LinLin(dd_list::iterator left_link, double Max_E);

  //! Expand histogram data to linear-linear between the two links.
  void Hist_2_LinLin(dd_list::iterator first_link,
    dd_list::iterator last_link, double Max_E );

  //! Expand the log-log data to linear-linear between the two consecutive links.
  void LogLog_2_LinLin(dd_list::iterator left_link);

 //! Expand the log-log data to linear-linear between the two links.
  void LogLog_2_LinLin(dd_list::iterator first_link,iterator last_link);

  //! This function expands log-log data to linear-linear for an entire link list.
  void LogLog_2_LinLin( );

  //! Appends to E_in_list the incident energies in this list
  void collect_E_in( list< double >& E_in_list );

  // Used by the fill_in_lists routine
  void pad_head( double E0 );

  // Used by the fill_in_lists routine
  void pad_tail( double E0 );

  // Deletes links with duplicate x values
  void unique( );

  // make sure that there are links for each energy in E_in_list
  void fill_with( list< double >& E_in_list );
};

// -------------------- fill_in_lists ---------------
//! This routine fills in to ensure that the x-values in both lists are the same.
void fill_in_lists( dd_list& list_1, dd_list& list_2 );
#endif
