/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1882 $
 * $Date: 2006-05-24 15:20:29 -0700 (Wed, 24 May 2006) $
 * $Author: dbrown $
 * $Id: list_2d.hpp 1882 2006-05-24 22:20:29Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// header for the class two_d_list

#ifndef TWO_D_LIST_CLASS
#define TWO_D_LIST_CLASS

#include "distrib_1d.hpp"
#include "endl_formats.hpp"
#include "record_types.hpp"
#include "ENDF_file.hpp"
#include "global_params.hpp"
#include "endl_precision.hpp"

using namespace std;

extern ENDLClass ENDL;
extern GlobalParameterClass Global;

// ----------- class two_d_list -----------------
//! 2d list template composed of 2d links
template <class one_d> class two_d_list : public list<one_d>
{

private:
  bool debug_on;        //!< for checking the 2-d interpolation
  double tol_2d;        //!< tolerance for 2-d interpolation
  double cut_off_2d;    //!< below this value ignore tolerance
  double dE_tol;        //!< minimum relative difference in incident energy
  double _tag;          //!< used with double differential data
                        //! switch for type of interpolation to be used
  bool _unit_base;      // if true, thicken using unit-based interpolation;
                        // if false, thicken based on equiprobable bins

protected:
  dd_list _weight;      //!< weight for this model

public:
  typedef typename list<one_d>::iterator two_d_iterator;

  //! Default constructor
  two_d_list()
  {
    // debug_on = true;  // for checking the 2-d interpolation
    debug_on = false;

    // set the parameters
    tol_2d = Global.Value("tol_2d");
    cut_off_2d = Global.Value("cut_off_2d");
    dE_tol = Global.Value("dE_tol");
    _tag=0.0;
    _unit_base=false;
  }

  //! The usual tag for this list
  inline double& E_in( )
  {
    return _tag;
  }

  //! The tag for mf14 gamma data
  inline double& E_gamma( )
  {
    return _tag;
  }

  //! Contains the interpolation regions
  vector<int> NBT;

  //! Contains the interpolation type for each region
  vector<int> INT;

  //! Simple printing utility
  void print()
  {
  // loop through the list
    for(two_d_iterator list_ptr = two_d_list::begin();
      list_ptr != two_d_list::end(); ++list_ptr)
    {
      cout << "Energy " << ENDL.data( list_ptr->E_in( ) ) << endl;
      list_ptr->print();
    }
    cout << endl;
  }

  // ----------- rescale -----------------
  //! Scale the energies of the emitted particle from [0, 1]
  void rescale()
  {
  // loop through the list
    for(two_d_iterator list_ptr = two_d_list::begin();
      list_ptr != two_d_list::end(); ++list_ptr)
    {
      list_ptr->rescale();
    }
  }

  // ----------- renorm -----------------
  //! Normalize probabilities such that integrated probability is unity
  void renorm()
  {
  // loop through the list
    for(two_d_iterator list_ptr = two_d_list::begin();
      list_ptr != two_d_list::end(); ++list_ptr)
    {
      list_ptr->renorm();
    }
  }

  // ----------- set_up_endl -----------------
  //! set up and open an ENDL file
  void set_up_endl( int i_number, fstream& endl_file )
  {
    ENDL.set_I_number(i_number);
    string file_name = ENDL.file_name;

    if(two_d_list::empty())
    {
      Warning("two_d_list::set_up_endl","The file "+file_name+" is empty.");
      return;
    }

    if ( ENDL.new_file() )
    {
      endl_file.open(file_name.c_str(),ios::out);
      Info("two_d_list::write_endl","Opening ENDL file "+file_name);
    }
    else
    {
      endl_file.open(file_name.c_str(),ios::out|ios::app);
      Info("two_d_list::write_endl","Appending ENDL file "+file_name);
    }

    //Shove the header info in
    endl_file << ENDL.header_line_1 << endl;
    endl_file << ENDL.header_line_2 << endl;
    endl_file.setf(ios::scientific,ios::floatfield);
  }

  // ----------- close_file -----------------
  //! close an ENDL file
  void close_file( fstream& endl_file )
  {
    //Put on end of file/section line into output file
    endl_file << ENDL.eof_line << endl;

    //Finally, we close the output file
    Info("two_d_list::close_file","Closing ENDL file "+ENDL.file_name);
    endl_file.close();
  }

  // ----------- write_endl -----------------
  //! Open an ENDL file and write the data.
  void write_endl( int i_number )
  {
    // set up the file
    fstream endl_file;
    set_up_endl( i_number, endl_file );

    // print the data
    out_data( i_number, endl_file, 0.0 );

    // finish writing and close the file
    close_file( endl_file );
  }

  // ----------- out_data -----------------
  // The list holds angular distributions if i_number = 1
  // and isotropic energy distributions if i_number = 4.
  // L_order is the Legendre order for i4 data
  //! Writes out the ENDL data
  void out_data( int i_number, fstream& endl_file, double L_order )
  {
  
    // loop over the data and write it out
    for(two_d_iterator link = two_d_list::begin(); 
        link != two_d_list::end(); ++link)
    {
      // write out
      for(dd_list::iterator prob_link = link->begin();
          prob_link != link->end(); ++prob_link)
      {
        switch(i_number)
        {
          case 1:  // angular distributions
            endl_file << ENDL.data( link->E_in(), 
              prob_link->mu(), prob_link->y )
	       	  << endl;
            break;
          case 4:  // energy distributions
            endl_file << ENDL.data( link->E_in(), 
              prob_link->E_out(), L_order, prob_link->y )
		     <<endl;
            break;
          default:
            SevereError("two_d_list::out_data",
              pastenum("Improper i_number: ",i_number));
        }
      }
    }
  }

  // ----------- mirror -----------------
  //! Used to reflect all of the lists mu -> -mu
  void mirror()
  {
  // loop through the list
    for(two_d_iterator list = two_d_list::begin();
      list != two_d_list::end(); ++list)
    {
      list->mirror();
    }   
  }

  // ----------- widen_jumps -----------------
  //! Widens the jumps for each member 1-d list and for 
  //! the main 2d list
  void widen_jumps( double cluster_min=0.0 )
  {
    // loop through the 1d lists that make up the two_d_list first
    for( two_d_iterator list1d = two_d_list::begin( );
      list1d != two_d_list::end( ); ++list1d )
    {
      list1d->widen_jumps( 0.0 );
    }   

    // Now for the main 2d list

    // scale dE by the smoother used for histograms
    double local_eps, jump_width;

    // container to hold links to elements we need to widen
    two_d_iterator cluster_start = two_d_list::begin();
    two_d_iterator cluster_end = two_d_list::begin();
    int cluster_size=1;

    for( two_d_iterator this_link = two_d_list::begin(); 
      this_link != two_d_list::end( ); ++this_link )
    {
      // get the local energy epsilon (smallest dE can see in output file)
      local_eps = ENDL_EPSILON( this_link->tag );
      jump_width = ENDL_JUMP_WIDTH( this_link->tag );

      // this_link is too close to the cluster_end, so add it to the cluster
      if( abs( this_link->tag - cluster_end->tag ) <= local_eps )
      {
        cluster_end = this_link;
        ++cluster_size;
      }
      // this_link is far from the cluster_end, so let's process the cluster
      else 
      {
        if ( cluster_size > 1 ) 
          widen_cluster( cluster_start, cluster_end, cluster_size, jump_width, cluster_min );
        // clear the clusters
        cluster_start = ++cluster_end;
        cluster_size  = 1;
      }
    }
    // process the last cluster since doesn't get processed otherwise
    if ( cluster_size > 1 ) 
      widen_cluster( cluster_start, cluster_end, cluster_size, jump_width, cluster_min );
  }

// ----------- widen_cluster -----------------
//! Utility code used by widen_jumps
void widen_cluster( two_d_iterator cluster_start, two_d_iterator cluster_end, 
  int cluster_size, double jump_width, double cluster_min )
{

  if( ( cluster_size == 1 ) || ( cluster_start==cluster_end ) )  return;
  if ( cluster_size>4 ) 
  {
    Warning("two_d_list::widen_cluster",pastenum("Found ",cluster_size)+
      pastenum(" links in a cluster, check ENDL_EPSILON and ENDL_JUMP_WIDTH values or the data is broken at tag=",
      cluster_start->tag)); 
  }

  if ( cluster_start->tag <= cluster_min+jump_width )
  {
    two_d_iterator cluster_middle = cluster_start; 
    for (int ijump = 1; ijump<=cluster_size; ++ijump)
    {
      ++cluster_middle; 
      cluster_middle->tag += static_cast<double>(ijump)*jump_width;
    }
  }
  else
  {
    switch (cluster_size)
    {
      case 2:
        cluster_start->tag -= 0.5*jump_width;
        cluster_end->tag   += 0.5*jump_width;
        break;
      case 3:
        cluster_start->tag -= jump_width;
        cluster_end->tag   += jump_width;
        break;
      case 4:
        {
        two_d_iterator cluster_left = cluster_start;
        two_d_iterator cluster_right = cluster_end;
        ++cluster_left;--cluster_right;
        cluster_start->tag -= 1.5*jump_width;
        cluster_left->tag  -= 0.5*jump_width;
        cluster_right->tag += 0.5*jump_width;
        cluster_end->tag   += 1.5*jump_width;
        }
        break;
      default:
        Warning("two_d_list::widen_cluster",pastenum("Found ",cluster_size)+
        pastenum(" links in a cluster, check ENDL_EPSILON and ENDL_JUMP_WIDTH values or the data is broken at tag=",
        cluster_start->tag)); 
        return;
    }
  }
}

  // ----------- weight_jumps -----------------
  //! Insert 1-d lists where the weight has jumps
  void weight_jumps()
  {
    one_d new_dist;  // for inserting new lists

    // widen the jumps in the weights
    _weight.widen_jumps( );

    // look for jumps in _weight
    two_d_iterator next_list = two_d_list::begin( );
    dd_list::iterator wt_ptr = _weight.begin( );
    for( ++wt_ptr ;
	 ( wt_ptr != _weight.end( ) ) && ( next_list != two_d_list::end( ) );
        ++wt_ptr )
    {
      double wt_E = wt_ptr->x;
      // find the list with next larger incident energy
      for( ; next_list->E_in( ) < wt_E; ++next_list )
      {
        if( next_list == two_d_list::end( ) )
        {
          // We went past the maximum energy
          break;
        }
      }

      // we might already have a list at essentinally this energy
      double local_eps = ENDL_EPSILON( wt_E );
      if( next_list->E_in( ) < wt_E + 1.1*local_eps )
      {
        next_list->E_in( ) = wt_E;
        next_list->weight = wt_ptr->y;
      }
      else
      {
        // insert a new list
        two_d_iterator prev_list = next_list;
        --prev_list;
        insert( next_list, new_dist );
        two_d_iterator mid_list = prev_list;
        ++mid_list;
        mid_list->list_interp( wt_E, *prev_list, *next_list );
        mid_list->weight = wt_ptr->y;
      }
    }   
  }

  // ----------- thinit -----------------
  //! Thins the member lists to within some tolerance
  void thinit()
  {
    for( two_d_iterator list_ptr = two_d_list::begin(); 
      list_ptr != two_d_list::end(); ++list_ptr )
    {
      list_ptr->thinit( );
    }
  }

  // ----------- thicken -----------------
  //! Expands a list to some tolerance
  void thicken()
  {
    bool interp_OK;
    two_d_iterator list = two_d_list::begin();
    two_d_iterator next_list = list;
    ++next_list;

    // loop through the 1-d lists
    for( ; next_list != two_d_list::end(); ++list, next_list = list, ++next_list)
    {
      // keep checking until the interval passes the test
      for(;;)
      {
        interp_OK = check_interp(list);

        if(interp_OK)
        {
	  break;
        }
        // *** a safely check ***
        const int Max_Size = 2000;
        if(two_d_list::size() >= Max_Size)
        {
          SevereError("two_d_list::thicken",
            pastenum("got ", Max_Size)+" links in 2-d thicken routine");
        }
      }
    }
  }

  // ----------- set_interp -----------------
  //! Sets the interpolation type for each list
  void set_interp()
  {
    int count = 0;
    int INT_index = 0;
    // loop through the list
    for(two_d_iterator this_link = two_d_list::begin();
      this_link != two_d_list::end(); ++this_link, ++count )
    {
      // do we change interpolation types?
      if ( count >= NBT[ INT_index ] )
      {
        ++INT_index;
        if ( INT_index >= INT.size() )
        {
          SevereError("two_d_list::set_interp"," INT_index overflow");
        }
      }
      this_link->interp_type = INT[ INT_index ];
    }   
  }


  // ----------- chop_highE -----------------
  //! Truncates list to the maximum incident energy, E_max
  void chop_highE( double Max_energy )
  {
    one_d new_list;
    two_d_iterator last_ptr = two_d_list::end();
    --last_ptr;
     two_d_iterator prev_ptr = last_ptr;
    --prev_ptr;
    insert( last_ptr, new_list );
    two_d_iterator new_ptr = last_ptr;
    --new_ptr;  // point to the new list
    new_ptr->list_interp( Max_energy, *prev_ptr, *last_ptr );
    erase( last_ptr, two_d_list::end() );  // delete the list above Max_energy
  }

  // ----------- check_interp -----------------
  //! Checks the accuracy of linear interpolation
  // between left_link and ++left_link
  // return: true if (worst relative error) <= tol_2d
  // if not, insert a new 1-d list good to within tol_1d
  bool check_interp(two_d_iterator left_list)
  {
    // is there a next list?
    two_d_iterator next_list = left_list;
    ++next_list;
    if( next_list == two_d_list::end( ) )
    {
      SevereError("two_d_list::check_interp",
        "attempt to check past the end of the list");
    }

    // lin-lin interpolation is always good
    if(( next_list->interp_type == 2 )||( next_list->interp_type == 22 ))
    {
      return( true );
    }

    // set width of jump as smallest dE we can see in output file
    double min_tol = ENDL_EPSILON( next_list->E_in( ) );

    // accept the link if dE is very small, but not too small
    if((next_list->E_in() - left_list->E_in() < dE_tol*next_list->E_in()) &&
        (next_list->E_in() - left_list->E_in() < min_tol) )
    {
      return(true);
    }

    // set the energy for an intermediate list
    double E_in;
    double dE;  // for histogram data
    if( next_list->interp_type == 1 ) // histogram data
    {
      two_d_iterator next_next_list = next_list;
      ++next_next_list;
      double dE_L = next_list->E_in( ) - left_list->E_in( );
      double dE_R = ( next_next_list == two_d_list::end( ) ) ?
        next_list->E_in( ) :
        next_next_list->E_in( ) - next_list->E_in( );
      dE = ( dE_L < dE_R ) ? dE_L : dE_R;

      if(dE <= 0.0)
      {
        left_list->print( );
        next_list->print( );
        next_next_list->print( );
        SevereError("two_d_list::check_interp","energies out of order");
      }

      E_in = next_list->E_in( );
      // if next_list->E_in is too close to next_next_list->E_in,
      // erase next_list
      if( dE_R <= min_tol )
      {
        erase( next_list );
        next_list = next_next_list;
      }
    }
    else
    {
      E_in = 0.5*( left_list->E_in( ) + next_list->E_in( ) );
    }
    // insert an intermediate list with incident energy E_in
    one_d mid_dist;
    insert(next_list, mid_dist);

    // expand the intermediate list and check it
    two_d_iterator mid_list = next_list;
    --mid_list;
    mid_list->E_in( ) = E_in;

    // get the weight for the model if this makes sense
    if ( ENDL.F ==5 )
    {
      mid_list->weight = _weight.evaluate( E_in );
    }

    if( next_list->interp_type == 1 ) // histogram data
    {
      // copy the next list
      mid_list->copy( *left_list );
      if( E_in > ENDL.Max_E_in - min_tol )
      {
        mid_list->E_in( ) = ENDL.Max_E_in;  // reset the incident energy tag
        erase( next_list );
      }
      else
      {
        mid_list->E_in( ) = E_in - min_tol;  // reset the incident energy tag
        next_list->interp_type = 2;  // now use lin-lin
        next_list->E_in( ) += min_tol;
      }
      mid_list->interp_type = 2;
      return( true );
    }

    mid_list->list_interp(E_in, *left_list, *next_list);

    bool is_OK;
    if(mid_list->bin_interp)
    {
      is_OK = mid_list->check_equiprob(*left_list, *next_list, tol_2d);
    }
    else
    {
      is_OK = mid_list->check_interp_2d(tol_2d, cut_off_2d);
    }

    if(is_OK)
    {
      // we don't need this list
      erase(mid_list);
    }
    else
    {
      if(debug_on)
      {
        cout << endl << " insert list " << endl;
        mid_list->print();
      }
    }
    return is_OK;
  }

  // ----------- collect_Ein -----------------
  //! Ensures that E_incident contains every incident energy
  void collect_Ein( list< double >& E_incident )
  {
  	// DAB 2/16/05 -- previous algorithm did not always pick up 
    // all energies in the list. 
    for( two_d_iterator list_ptr = two_d_list::begin( ); 
         list_ptr != two_d_list::end( ); ++list_ptr ){
			E_incident.push_back(list_ptr->E_in());
	  }
    E_incident.sort();   //STL gizmo to sort the list
    E_incident.unique(); //STL gizmo to remove duplicates
  }

  // ----------- copy_data -----------------
  //! copies the weights and the tabular data
  template <class One_d> void copy_data( two_d_list<One_d>& copy_from )
  {
    one_d new_dist;

    for( typename two_d_list<One_d>::iterator list_ptr = copy_from.begin( );
       list_ptr != copy_from.end( ); ++list_ptr )
    {
      insert( two_d_list::end( ), new_dist );

      // insert the data
      two_d_iterator target_ptr = two_d_list::end( );
      --target_ptr;
      target_ptr->E_in( ) = list_ptr->E_in( );
      target_ptr->weight = list_ptr->weight;
      target_ptr->interp_type = target_ptr->interp_type;
      for( dd_list::iterator link_ptr = list_ptr->begin();
        link_ptr != list_ptr->end(); ++link_ptr )
      {
        dd_link new_link = dd_link( link_ptr->x, link_ptr->y );
        target_ptr->insert( target_ptr->end(), new_link );
      }
    }
  }

  // ----------- fill_in_list -----------------
  //! Insures that 2d list has a link for each "E_in" in fill_from
  void fill_in_list( list< double >& fill_from )
  {
    const double fill_tol = Global.Value( "mf5_tol" );  // tolerance for mf5 data

    one_d mid_dist;  // for adding new distributions
    two_d_iterator mid_list;
    dd_link XYdata;  // for links in our lists

    two_d_iterator list_ptr = two_d_list::begin( );
    list< double >::iterator fill_ptr = fill_from.begin( );
    double list_E = list_ptr->E_in( );
    double fill_E = *fill_ptr;
    if( fill_E < ( 1 - fill_tol )*list_E )
    {
      insert( list_ptr, mid_dist );
      // make an initial list with 2 zero links
      mid_list = list_ptr;
      --mid_list;
      mid_list->weight = 0.0;
      mid_list->E_in( ) = fill_E;
      XYdata.x = 0.0;
      XYdata.y = 0.0;
      mid_list->insert( mid_list->end( ), XYdata );

      XYdata.x = 1.0e-8;
      mid_list->insert( mid_list->end( ), XYdata );
      list_ptr = mid_list; // set list_ptr to the start
      list_E = fill_E;
      ++fill_ptr;
    }
    else if( fill_E < ( 1 + fill_tol )*list_E )
    {
      // the lists start at the same incident energies, go to the next
      ++fill_ptr;
    }
    // Go through the fill_from list
    two_d_iterator prev_ptr = list_ptr;
    ++list_ptr;
    for( ; fill_ptr != fill_from.end();  )
    {
      // test for failure
      if( list_ptr == two_d_list::end() )
      {
        SevereError("two_d_list::fill_in_list",
            "Unexpected end of 2d list in fill_in_list");
      }
      list_E = list_ptr->E_in();
      fill_E = *fill_ptr;
      if ( fill_E < list_E )
      {
				// insert a link at energy fill_E
        insert(list_ptr, mid_dist);
        // expand the intermediate list
        mid_list = list_ptr;
        --mid_list;
        mid_list->list_interp(fill_E, *prev_ptr, *list_ptr);
        if( ENDL.F == 5 )
				{
          mid_list->weight = _weight.evaluate( fill_E );
				}
        ++fill_ptr;
      }
      else if ( fill_E > list_E )
      {
        // go on to the next link
        prev_ptr = list_ptr;
        ++list_ptr;
      }
      else
      {
        // go on to the next link in both lists
        prev_ptr = list_ptr;
        ++list_ptr;
        ++fill_ptr;
      }
    }
  }
  // ----------- use_weight -----------------
  //! Scales the 1-d lists by their weights
  void use_weight( )
  {
    for( two_d_iterator list_ptr = two_d_list::begin(); 
      list_ptr != two_d_list::end(); ++list_ptr )
    {
      // do not zero out energy distributions
      if( list_ptr->weight > 0.0 )
      {
        *list_ptr *= list_ptr->weight;
      }
    }
  }

};

// ************* sum_lists function *****************
// ----------- sum_lists -----------------
//! Insures that 2d list has a link for each E_in in to_add
template <class one_d, class One_d> void sum_lists(
  two_d_list<one_d>& sum, two_d_list<One_d>& to_add )
{
  const double fill_tol = Global.Value( "mf5_tol" );  // tolerance for mf5 data
  typename two_d_list<one_d>::iterator sum_ptr = sum.begin();
  typename two_d_list<One_d>::iterator add_ptr = to_add.begin();
  for( ;
    sum_ptr != sum.end(), add_ptr != to_add.end();
    ++sum_ptr, ++add_ptr )
  {
    if ( abs(sum_ptr->E_in() - add_ptr->E_in()) >
      2*fill_tol*add_ptr->E_in() )
    {
      SevereError("two_d_list::sum_lists",
        pastenum("energy mismatch in sum_lists; E_in: ",sum_ptr->E_in())+
        pastenum(" ",add_ptr->E_in()));
    }
    // check for zero weight
    if( add_ptr->weight == 0.0 )
    {
      // usually do nothing
      if( sum_ptr->weight == 0.0 )
      {
        // take the average
        *sum_ptr += *add_ptr;
        // remove duplicates but not for gammas
        if( ( ENDL.F < 12 ) || ( ENDL.F > 15 ) )
        {
          sum_ptr->thinit();
        }
      }
    }
    else if( sum_ptr->weight == 0.0 )
    {
      // replace *sum_ptr data by *add_ptr data
      // save the next pointer and erase this one
      typename two_d_list<one_d>::iterator next_ptr = sum_ptr;
      ++next_ptr;
      sum.erase( sum_ptr );
      // insert a new one_d list
      one_d new_one_d;
      sum.insert( next_ptr, new_one_d );
      // reset sum_ptr
      sum_ptr = next_ptr;
      --sum_ptr;
      // copy the *add_ptr data
      sum_ptr->dd_list::copy( *add_ptr );
      sum_ptr->E_in( ) = add_ptr->E_in( );
      sum_ptr->weight = add_ptr->weight;
    }
    else
    {
      // add the lists with weight
      *add_ptr *= add_ptr->weight;
      *sum_ptr += *add_ptr;
      // remove duplicates but not for gammas
      if( ( ENDL.F < 12 ) || ( ENDL.F > 15 ) )
      {
        sum_ptr->thinit();
      }
      sum_ptr->weight += add_ptr->weight;
    }
  }
}
// ************* sum_lines function *****************
// ----------- sum_lines -----------------
//! Add 2 lists as discrete lines, no interpolation
template <class one_d, class One_d> void sum_lines(
  two_d_list<one_d>& sum, two_d_list<One_d>& to_add )
{
  const double fill_tol = Global.Value( "mf5_tol" );  // tolerance for mf5 data
  typename two_d_list<one_d>::iterator sum_ptr = sum.begin();
  typename two_d_list<One_d>::iterator add_ptr = to_add.begin();
  for( ;
    sum_ptr != sum.end(), add_ptr != to_add.end();
    ++sum_ptr, ++add_ptr )
  {
    if ( abs(sum_ptr->E_in() - add_ptr->E_in()) >
      2*fill_tol*add_ptr->E_in() )
    {
      SevereError("two_d_list::sum_lines",
        pastenum("energy mismatch in sum_lines; E_in: ",sum_ptr->E_in())+
        pastenum(" ",add_ptr->E_in()));
    }
    // add the lists
    sum_ptr->line_sum( *add_ptr, fill_tol );
  }
}

// *********** class two_d_table ****************
//! Class that holds 2d tabular data as a function of incident energy
// ----------- class two_d_table -----------------
// derive this class from two_d_list
class two_d_table : public two_d_list< one_d_table >
{
};


#endif
