/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1735 $
 * $Date: 2006-02-09 13:47:26 -0800 (Thu, 09 Feb 2006) $
 * $Author: dbrown $
 * $Id: list_3d.cpp 1735 2006-02-09 21:47:26Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the classes for 3-d linked lists

#include "list_3d.hpp"
#include "endl_formats.hpp"
#include "global_params.hpp"

extern ENDLClass ENDL;
extern GlobalParameterClass Global;

// ********* for class three_d_link *************
// ----------- three_d_link::make_cos_link -----------------
void three_d_link::make_cos_link(double e_in,
  two_d_arbitrary::iterator new_cos)
// construct the (cosine, probability) list and normalize
{
  new_cos->E_in() = e_in;

  // loop through the double differential list
  for(two_d_arbitrary::iterator this_cos = begin(); this_cos != end();
    ++this_cos)
  {
    // make a new (cosine, probability) link
    double prob = this_cos->get_norm();
    dd_link new_link(this_cos->mu(), prob);
    new_cos->insert(new_cos->end(), new_link);

    // renormalize the double differential data
    this_cos->renorm(prob);
  }

  // normalize the (cosine, probability) list
  new_cos->renorm();
}

// ********* for class three_d_list *************

// ----------- three_d_list constructor ------------
// the default constructor
three_d_list::three_d_list()
{
  // debug_on = true;  // for checking the 3-d interpolation
  debug_on = false;

  // how many angles to use = 2*Num_mu+1
  Num_mu = static_cast<int>( Global.Value("num_mu_3d")); //Load from Global parameters
}
// ----------- three_d_list::print -----------------
void three_d_list::print()
// to print the list
{
// loop through the list
  for(three_d_list::iterator this_e_in = begin();
      this_e_in != end(); ++this_e_in)
    {
      cout<<"----------- E_in = " << this_e_in->E_in() << endl;
      this_e_in->print();
    }   
  cout << endl;
} 

// ----------- three_d_list::write_endl -----------------
void three_d_list::write_endl()
  // The three_d_list is used to hold the double differential data
  // These data are stored in two different files - i=1 and 1=3.
{

  ENDL.set_I_number( 3 );

  fstream endl_file;
  string file_name = ENDL.file_name;

  if(empty())
  {
    Warning("three_d_list::write_endl","The file "+file_name+" is empty.");
    return;
  }

  if ( ENDL.new_file() )
  {
    endl_file.open(file_name.c_str(),ios::out);
    Info("three_d_list::write_endl","Opening ENDL file "+file_name);
  }
  else
  {
    endl_file.open(file_name.c_str(),ios::out|ios::app);
    Info("three_d_list::write_endl","Appending ENDL file "+file_name);
  }

  //Shove the header info in
  endl_file<<ENDL.header_line_1<<endl;
  endl_file<<ENDL.header_line_2<<endl;
  endl_file.setf(ios::scientific,ios::floatfield);
  
  //Now loop over the data and write it out
  double old_E_in = -1.0;
  for(three_d_list::iterator E_in_link = begin();
      E_in_link != end(); 
      ++E_in_link)
    {
      // check for double value of E_in
      if(E_in_link->E_in() == old_E_in)
      {
        Warning("three_d_list::write_endl",
            pastenum("Widen the jump in three_d_list at E_in: ",old_E_in));
      }
      old_E_in = E_in_link->E_in();
      double old_mu = -2.0;
      for(two_d_list<one_d_table>::iterator cos_link = E_in_link->begin();
	  cos_link != E_in_link->end(); 
	  ++cos_link)
	{
          // check for double value of x
          if(cos_link->mu() == old_mu)
          {
            Warning("three_d_list::write_endl",
                pastenum("Widen the jump in cos_list at E_in: ",old_E_in)+
                pastenum(" mu: ",old_mu));
          }
          old_mu = cos_link->mu();
          double old_x = -2.0;
	  for(dd_list::iterator E_out_link = cos_link->begin();
	      E_out_link != cos_link->end(); 
	      ++E_out_link)
	    {
              // check for double value of x
              if(E_out_link->x == old_x)
              {
                Warning("three_d_list::write_endl",
                    pastenum("Widen the jump in one_d_list at E_in: ",old_E_in)+
                    pastenum(" mu: ",old_mu)+
                    pastenum(" x: ",old_x));
              }
              old_x = E_out_link->x;
	      endl_file << ENDL.data( E_in_link->E_in(), 
				    cos_link->mu(), 
				    E_out_link->E_out(), 
				    E_out_link->y )
		       <<endl;
	    }
	}
    }
  
  //Put on end of file/section line into output file
  endl_file<<ENDL.eof_line<<endl;
  
  //Finally, we close the output file
  Info("three_d_list::write_endl","Closing ENDL file "+file_name);
  endl_file.close();
  
  //************************************************
  //Now the cosine data
  cosines.write_endl(1);
}

