/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1853 $
 * $Date: 2006-04-21 16:09:59 -0700 (Fri, 21 Apr 2006) $
 * $Author: hedstrom $
 * $Id: list_3d.hpp 1853 2006-04-21 23:09:59Z hedstrom $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// header for the class three_d_list

#ifndef THREE_D_LIST_CLASS
#define THREE_D_LIST_CLASS

#include "list_2d.hpp"
#include "endl_formats.hpp"
#include "global_params.hpp"
#include "mappings.hpp"
#include "nuclei.hpp"

extern ENDLClass ENDL;
extern GlobalParameterClass Global;

// ********** enum Direction *************************
//! Type to identify the direction in laboratory coordinates
//! of an outgoing particle with given energy in center-of-mass
//! coordinates.  The options are:
//!   Straight: only $\mu = 1$;
//!   Forward: $0 < \mu_{\rm min} \le \mu \le 1$;
//!   Transition: $0 \le \mu \le 1$;
//!   Both_ways: $-1 \le \mu \le 1$.

enum  Direction{ Straight, Forward, Transition, Both_ways };

// ********* class Geometry *************
//! The double differential data is typically given for a range
//! of energies in center-of-mass coordinates.  The coding in this
//! file is written in terms of what direction outgoing particle
//! can go in the laboratory frame for the minimum and maximum
//! center-of-mass energies.  The Straight option is to handle
//! the case of the minimum center-of-mass energy being zero.
class Geometry
{
public:
  Direction fastest;  // mu range for the most energetic particles
  Direction slowest;  // mu range for the least energetic particles
};

// ********* for class three_d_link *************
//! This is the link from which the 3d linked lists are composed.
template< class one_d > class three_d_link : public two_d_list< one_d >
{
public:
  int Num_mu;    // The number of mu values is 2*Num_mu + 1.

  mappings map;  // data needed for map to laboratory frame

  Geometry geom_;  // geometry of the (E', mu) domain in lab frame

  // ----------- three_d_link constructor ------------
  three_d_link()
  {
    // how many angles to use = 2*Num_mu+1
    Num_mu = static_cast<int>( Global.Value("num_mu_3d")); //Load from Global parameters
  }

  // ********* start of virtual functions ********************
  //! Virtual function to insert a new 1-d list depends on the model
  virtual typename three_d_link< one_d >::iterator 
    new_one_d( typename three_d_link< one_d >::iterator where ) = 0;
  // ********** end of virtual functions *********************

  // ----------- three_d_link::make_cos_link -----------------
  //! Constructs the (cosine, probability) list and normalize
  void make_cos_link(double e_in, two_d_table::iterator new_cos)
  {
    new_cos->E_in() = e_in;

    // loop through the double differential list
    for( typename three_d_link::iterator this_cos = three_d_link::begin();
      this_cos != three_d_link::end(); ++this_cos )
    {
      // make a new (cosine, probability) link
      double prob = this_cos->get_norm();
      dd_link new_link(this_cos->mu(), prob);
      new_cos->insert(new_cos->end(), new_link);

      // if the probability is zero, the distribution is arbitrary
      if( prob == 0.0 )
      {
        this_cos->begin( )->y = 1.0;
	this_cos->widen_jumps( );
	this_cos->renorm( );
      }
      else
      {
        this_cos->renorm(prob);
      }
    }

    // normalize the (cosine, probability) list
    new_cos->renorm();
  }

  // ----------- three_d_link::get_geom -----------------
  void get_geom( double first_E_out, double last_E_out )
  // determine how an (E', mu) rectangle in the center-of-mass frame
  // maps to the lab frame.
  {
    if(last_E_out <= first_E_out)
    {
      SevereError( "three_d_link::get_geom", "energies out of order" );
    }

    if(last_E_out > map.E_transl)
    {  // The fastest particles can go backwards in the lab frame.
      geom_.fastest = Both_ways;
      if(first_E_out == 0.0)
      {
        geom_.slowest = Straight;
      }
      else if(first_E_out < map.E_transl)
      {
        // the slowest particles only go forward
        geom_.slowest = Forward;
      }
      else if(first_E_out == map.E_transl)
      {
        geom_.slowest = Transition;
      }
      else
      {
        // evetn the slowest particles can go backward
        geom_.slowest = Both_ways;
      }
    }
    else if(last_E_out == map.E_transl)
    {  // special case
      geom_.fastest = Transition;
      if(first_E_out == 0.0)
      {
        geom_.slowest = Straight;
      }
      else
      {
        geom_.slowest = Forward;
      }
    }
    else
    {  // only forward emission
      geom_.fastest = Forward;
      if(first_E_out == 0.0)
      {
        geom_.slowest = Straight;
      }
      else
      {
        geom_.slowest = Forward;
      }
    }
  }
  // ----------- three_d_link::full_range -----------------
  void full_range( double last_E_out )
  // expand the double differential data for 1 incident energy
  // in the case when the speed of outgoing particles in
  // center-of-mass coordinates ranges from zero to greater than
  // the speed of the frame.
  // last_E_out is the maximum CM emission energy in MeV
  {
    // For ENDL we need to go through cosines first and then E_out.
    // But the ENDF/B6 data has E_out first.
    const double dmu = 1.0/Num_mu;
    double this_mu;

    for (int count = -Num_mu; count <= Num_mu; ++count)
    {
      this_mu = count*dmu;
      // mu corresponds to a new 2d link in the list for this E_in
      typename three_d_link< one_d >::iterator mu_link =
        new_one_d( three_d_link::end( ) );
      mu_link->mu() = this_mu;
 
      // make E_out links in the 1d list for this mu (in MeV)
      double E_dummy;  // for the call to get_E_range
      double max_E_out; // maximum lab emission energy for this mu
      map.get_E_range(last_E_out, this_mu, &E_dummy, &max_E_out);

      // add energy links for this mu
      mu_link->one_E_row( 0.0, max_E_out );
    }
  }

  // ----------- three_d_link::all_both_ways -----------------
  void all_both_ways( double first_E_out, double last_E_out )
  // expand the double differential data for 1 incident energy
  // in the case when even the slowest outgoing particles are 
  // fast enough to go backward in the lab frame.
  // first_E_out is the minimum CM emission energy, and
  // last_E_out is the maximum
  {
    // For ENDL we need to go through cosines first and then E_out.
    // But the ENDF/B6 data has E_out first.
    const double dmu = 1.0/Num_mu;
    double mu;

    for (int count = -Num_mu; count <= Num_mu; ++count)
    {
      mu = count*dmu;
      // mu corresponds to a new 2d link in the list for this E_in
      typename three_d_link< one_d >::iterator mu_link =
        new_one_d(  three_d_link::end( ) );
      mu_link->mu() = mu;

      // make E_out links in the 1d list for this mu (in MeV)
      double E_dummy;  // for the call to get_E_range
      double max_E_out; // maximum lab emission energy for this mu
      map.get_E_range(last_E_out, mu, &E_dummy, &max_E_out);

      // get the minimum CM emission energy in MeV
      double min_E_out; // minimum lab emission energy for this mu
      map.get_E_range(first_E_out, mu, &E_dummy, &min_E_out);

      // add energy links for this mu
      mu_link->one_E_row( min_E_out, max_E_out );
    }
  }

  // ----------- three_d_link::both_ways_trans -----------------
  void both_ways_trans( double first_E_out, double last_E_out )
  // expand the double differential data for 1 incident energy
  // in the case when the slowest outgoing particles are right
  // at the translational speed of the center of mass.
  // first_E_out is the minimum CM emission energy, and
  // last_E_out is the maximum
  {
    // For ENDL we need to go through cosines first and then E_out.
    // But the ENDF/B6 data has E_out first.
    const double dmu = 1.0/Num_mu;
    double mu;

    for (int count = -Num_mu; count <= Num_mu; ++count)
    {
      mu = count*dmu;
      // mu corresponds to a new 2d link in the list for this E_in
      typename three_d_link< one_d >::iterator mu_link =
        new_one_d(  three_d_link::end( ) );
      mu_link->mu() = mu;

      // make E_out links in the 1d list for this mu (in MeV)
      double E_dummy;  // for the call to get_E_range
      double max_E_out; // maximum lab emission energy for this mu
      map.get_E_range(last_E_out, mu, &E_dummy, &max_E_out);

      double min_E_out; // minimum lab emission energy for this mu
      if(mu <= 0)
      {
        min_E_out = 0.0;
      }
      else
      {
        // get the minimum Lab emission energy in MeV
        map.get_E_range(first_E_out, mu, &E_dummy, &min_E_out);
      }

      // add energy links for this mu
      mu_link->one_E_row( min_E_out, max_E_out );
    }
  }

  // ----------- three_d_link::full_trans -----------------
  void full_trans( double last_E_out )
  // expand the double differential data for 1 incident energy
  // in the case when the slowest outgoing particles have speed
  // zero in the center-of-mass coordinates and the fastest ones
  // move at the translational speed of the center of mass.
  // last_E_out is the maximum CM emission energy
  {
    // For ENDL we need to go through cosines first and then E_out.
    // But the ENDF/B6 data has E_out first.

    // pick a positive minimum mu
    double mu_floor = 0.001;
    const double dmu = (1.0 - mu_floor)/(2*Num_mu);
    double mu;

    for (int count = 0; count <= 2*Num_mu; ++count)
    {
      mu = mu_floor + count*dmu;
      // mu corresponds to a new 2d link in the list for this E_in
      typename three_d_link< one_d >::iterator mu_link =
        new_one_d(  three_d_link::end( ) );
      mu_link->mu() = mu;

      double E_dummy;  // for the call to get_E_range
      double max_E_out; // maximum lab emission energy for this mu
      map.get_E_range(last_E_out, mu, &E_dummy, &max_E_out);

      // add energy links for this mu
      mu_link->one_E_row( 0.0, max_E_out );
    }
  }

  // ----------- three_d_link::full_forward -----------------
  void full_forward( double first_E_out, double last_E_out )
  // expand the double differential data for 1 incident energy
  // in the case when the slowest outgoing particles have speed
  // zero in the center-of-mass frame and the fastest ones
  // are slower than the translational speed of the center of mass.
  // first_E_out is the minimum CM emission energy, and
  // last_E_out is the maximum
  {
    // For ENDL we need to go through cosines first and then E_out.
    // But the ENDF/B6 data has E_out first.

    // pick a minimum mu
    double mu_floor = map.get_min_mu(last_E_out);
    double dmu = (1.0 - mu_floor);
    // add a little to get a non-null list
    mu_floor += dmu * 0.001;
    dmu = (1.0 - mu_floor)/(2*Num_mu);
    double mu;

    for (int count = 0; count <= 2*Num_mu; ++count)
    {
      mu = mu_floor + count*dmu;
      if ( mu > 1.0 )
      {
        mu = 1.0;
      }
      // mu corresponds to a new 2d link in the list for this E_in
      typename three_d_link< one_d >::iterator mu_link =
        new_one_d(  three_d_link::end( ) );
      mu_link->mu() = mu;

      // make E_out links in the 1d list for this mu (in MeV)
      // get the minimum and maximum lab emission energy in MeV
      double min_E_out;  // minimum lab emission energy for this mu
      double max_E_out;  // maximum lab emission energy for this mu
      map.get_E_range(last_E_out, mu, &min_E_out, &max_E_out);

      // add energy links for this mu
      mu_link->one_E_row( min_E_out, max_E_out );
    }
  }

// ----------- three_d_link::forward_hole -----------------
  void forward_hole( double first_E_out, double last_E_out )
  // expand the double differential data for 1 incident energy
  // in the case when the outgoing particles all have speed
  // larger than zero in center-of-mass coordinates and less
  // than the translational speed of the center of mass.
  // first_E_out is the minimum CM emission energy, and
  // last_E_out is the maximum
  {
    // For ENDL we need to go through cosines first and then E_out.
    // But the ENDF/B6 data has E_out first.

    // pick a minimum mu
    double mu_floor = map.get_min_mu(last_E_out);
    double dmu = (1.0 - mu_floor);
    // add a little to get a non-null list
    mu_floor += dmu * 0.001;
    dmu = (1.0 - mu_floor)/(2*Num_mu);
    double mu;

    // bifurcation mu
    // get the minimum lab emission energy in MeV
    double mu_split = map.get_min_mu(first_E_out);

    for (int count = 0; count <= 2*Num_mu; ++count)
    {
      mu = mu_floor + count*dmu;
      // mu corresponds to a new 2d link in the list for this E_in
      typename three_d_link< one_d >::iterator mu_link =
        new_one_d( three_d_link::end( ) );
      mu_link->mu() = mu;

      // make E_out links in the 1d list for this mu (in MeV)
      // get the minimum and maximum lab emission energy in MeV
      double min_E_out;  // minimum lab emission energy for this mu
      double max_E_out; // maximum lab emission energy for this mu
      map.get_E_range(last_E_out, mu, &min_E_out,
        &max_E_out);

      // add energy links for this mu
      if(mu < mu_split)
      {
        mu_link->one_E_row( min_E_out, max_E_out );
      }
      else
      {
        // get the maximum lab emission energy in MeV
        double min_E_split;  // minimum of the hole
        double max_E_split; // maximum of the hole
        map.get_E_range(first_E_out, mu, &min_E_split, &max_E_split);

        // for E below the hole
        mu_link->one_E_row( min_E_out, min_E_split );

        // 2 zeros for the hole
        double de = 0.0001*(max_E_split - min_E_split);
        dd_link e_out_link(min_E_split + de, 0.0);
        mu_link->insert(mu_link->end(), e_out_link);
        e_out_link.x = max_E_split - de;
        mu_link->insert(mu_link->end(), e_out_link);

        // for E above the hole
        mu_link->one_E_row( max_E_split, max_E_out );
      }
    }
  }

  // ----------- three_d_link::trans_hole -----------------
  void trans_hole( double first_E_out, double last_E_out )
  // expand the double differential data for 1 incident energy
  // in the case when the fastest outgoing particles in
  // center-of-mass coordinates move at the translational speed 
  // of the center of mass and the slowest ones have positive speed.
  // first_E_out is the minimum CM emission energy, and
  // last_E_out is the maximum
  {
    // For ENDL we need to go through cosines first and then E_out.
    // But the ENDF/B6 data has E_out first.

    // pick a minimum mu
    double mu_floor = 0.001;
    const double dmu = (1.0 - mu_floor)/(2*Num_mu);
    double mu;

    // get the minimum CM emission energy in MeV
    // bifurcation mu
    double mu_split = map.get_min_mu(first_E_out);

    for (int count = 0; count <= 2*Num_mu; ++count)
    {
      mu = mu_floor + count*dmu;
      // mu corresponds to a new 2d link in the list for this E_in
      typename three_d_link< one_d >::iterator mu_link =
        new_one_d(  three_d_link::end( ) );
      mu_link->mu() = mu;

      // make E_out links in the 1d list for this mu (in MeV)
      // get the minimum and maximum lab emission energy in MeV
      double E_dummy;
      double max_E_out; // maximum lab emission energy for this mu
      map.get_E_range(last_E_out, mu, &E_dummy, &max_E_out);

      // add energy links for this mu
      if(mu < mu_split)
      {
        mu_link->one_E_row( 0.0, max_E_out );
      }
      else
      {
        double min_E_split;  // minimum of the hole
        double max_E_split; // maximum of the hole
        // get the maximum lab emission energy in MeV
        map.get_E_range(first_E_out, mu, &min_E_split, &max_E_split);

        // for E below the hole
        mu_link->one_E_row( 0.0, min_E_split );

        // 2 zeros for the hole
        double de = 0.0001*(max_E_split - min_E_split);
        dd_link e_out_link(min_E_split + de, 0.0);
        mu_link->insert(mu_link->end(), e_out_link);
        e_out_link.x = max_E_split - de;
        mu_link->insert(mu_link->end(), e_out_link);

        // for E above the hole
        mu_link->one_E_row( max_E_split, max_E_out );
      }
    }
  }

  // ----------- three_d_link::both_ways_hole -----------------
  void both_ways_hole( double first_E_out, double last_E_out )
  // expand the double differential data for 1 incident energy
  // in the case when the fastest outgoing particles in
  // center-of-mass coordinates move faster than the translational
  // speed of the center of mass and the slowest ones have
  // positive speed less than the translational speed.
  // first_E_out is the minimum CM emission energy, and
  // last_E_out is the maximum
  {
    // For ENDL we need to go through cosines first and then E_out.
    // But the ENDF/B6 data has E_out first.

    const double dmu = 1.0/Num_mu;
    double mu;

    // bifurcation mu
    double mu_split = map.get_min_mu(first_E_out);

    for (int count = -Num_mu; count <= Num_mu; ++count)
    {
      mu = count*dmu;
      // mu corresponds to a new 2d link in the list for this E_in
      typename three_d_link< one_d >::iterator mu_link =
        new_one_d(  three_d_link::end( ) );
      mu_link->mu() = mu;

      // make E_out links in the 1d list for this mu (in MeV)
      // get the minimum and maximum lab emission energy in MeV
      double E_dummy;
      double max_E_out; // maximum lab emission energy for this mu
      map.get_E_range(last_E_out, mu, &E_dummy, &max_E_out);

      // add energy links for this mu
      if(mu < mu_split)
      {
        mu_link->one_E_row( 0.0, max_E_out );
      }
      else
      {
        // for the lowest CM particle energy in MeV
        // get the maximum lab emission energy in MeV
        double min_E_split;  // minimum of the hole
        double max_E_split; // maximum of the hole
        map.get_E_range(first_E_out, mu, &min_E_split,
          &max_E_split);

        // for E below the hole
        mu_link->one_E_row( 0.0, min_E_split );

        // 2 zeros for the hole
        double de = 0.0001*(max_E_split - min_E_split);
        dd_link e_out_link(min_E_split + de, 0.0);
        mu_link->insert(mu_link->end(), e_out_link);
        e_out_link.x = max_E_split;
        mu_link->insert(mu_link->end(), e_out_link);

        // for E above the hole
        mu_link->one_E_row( max_E_split, max_E_out );
      }
    }
  }
  // ----------- three_d_link::write_endl_link ---------------
  void write_endl_link( fstream& endl_file )
  // print this link in ENDL format
  {
    double old_mu = -2.0;
    for( typename three_d_link< one_d >::iterator cos_link = three_d_link::begin();
	  cos_link != three_d_link::end(); ++cos_link )
    {
      // check for double value of x
      if(cos_link->mu() == old_mu)
      {
        Warning("three_d_list::write_endl_link",
            pastenum("Widen the jump in cos_list at E_in: ", three_d_link::E_in( ) )+
            pastenum(" mu: ",old_mu));
      }
      old_mu = cos_link->mu();
      double old_x = -2.0;
      for(dd_list::iterator E_out_link = cos_link->begin();
	  E_out_link != cos_link->end(); 
	  ++E_out_link)
      {
        // check for double value of x
        if(E_out_link->x == old_x)
        {
          Warning("three_d_link::write_endl_link",
              pastenum("Widen the jump in one_d_list at E_in: ", three_d_link::E_in( ) )+
              pastenum(" mu: ",old_mu)+
              pastenum(" x: ",old_x));
        }
        old_x = E_out_link->x;
        endl_file << ENDL.data( three_d_link::E_in(), 
	       	    cos_link->mu(), 
		    E_out_link->E_out(), 
	       	    E_out_link->y )
	          <<endl;
      }
    }
  }
};

// ********* for class three_d_list *************
// ----------- class three_d_list -----------------
//! Class to handle the double differential data.
//! It is assumed in applications that two_d is derived from three_d_link< one_d >.
template < class two_d > class three_d_list :
  public list< two_d >
{
private:
  bool debug_on;   // for checking the 2-d interpolation

public:
  vector<int> NBT;  //  locations of changes in interpolation rules
  vector<int> INT;  //  the interpolation rules

  Target A;
  Projectile a;
  Residual B;
  Product b;
  Compound C;

  two_d_table cosines;  // the cosines used for each E_in

  // ----------- three_d_list constructor ------------
  three_d_list()
  {
    // debug_on = true;  // for checking the 3-d interpolation
    debug_on = false;
  }

  // ----------- three_d_list::print -----------------
  //! Prints the 3d list
  void print()
  {
    // loop through the list
    for( typename three_d_list< two_d >::iterator this_e_in = three_d_list::begin();
      this_e_in != three_d_list::end(); ++this_e_in )
    {
      cout<<"----------- E_in = " << ENDL.data( this_e_in->E_in()) << endl;
      this_e_in->print();
    }   
    cout << endl;
  }
  
/*
  // ----------- three_d_list::prune_zeronorm_data -------------
  //! clean the list(s) of entries that have zero probability,
  //! they only confuse downstream codes and add nothing to the data 
  void prune_zeronorm_data() {
    if ( cosines.size() != three_d_list::size() ) 
        SevereError("three_d_list::prune_zeronorm_data",
          "three_d_list and three_d_list::cosines dimensions don't match");
    typename three_d_list< two_d >::iterator it1 = three_d_list::begin();
    two_d_table::iterator                    it2 = cosines.begin();    
    for ( int iE = 0; iE < three_d_list::size(); ++iE )
    {
      if ( ( it1 == three_d_list::end() ) || (it2 == cosines.end() ) ) 
          SevereError( "three_d_list::prune_zeronorm_data","fell off edge of E list!");
      // Loop through list of mu's in lock-step, 
      // removing zero norm sets as we go
      typename two_d::iterator it1mu = it1->begin();
      dd_list::iterator        it2mu = it2->begin();    
      for ( int imu = 0; imu < it1->size(); ++imu )
      {
        if ( ( it1mu == it1->end() ) || (it2mu == it2->end() ) ) 
            SevereError( "three_d_list::prune_zeronorm_data","fell off edge of mu list!");
        if ( it2mu->y <= 0. )
        {
          cout << "Norm 0.0? "<<it2mu->y<<endl;
          it1mu = it1->erase(it1mu);
          it2mu = it2->erase(it2mu);       
        }
        else
        {
          ++it1mu;
          ++it2mu;       
        }
      }
      it1->renorm();
      ++it1;
      ++it2;       
    }
    cosines.renorm();
  }
*/
  
  // ----------- three_d_list::write_endl -----------------
  //! Writes the ENDL i=1 and i=3 files
  void write_endl()
  {

    ENDL.set_I_number( 3 );

    fstream endl_file;
    string file_name = ENDL.file_name;

    // clean the list(s) of entries that have zero probability, 
    // they only confuse downstream codes and add nothing to the data 
//    three_d_list::prune_zeronorm_data();

    if(three_d_list::empty())
    {
      Warning("three_d_list::write_endl","The file "+file_name+" is empty.");
      return;
    }

    if ( ENDL.new_file() )
    {
      endl_file.open(file_name.c_str(),ios::out);
      Info("three_d_list::write_endl","Opening ENDL file "+file_name);
    }
    else
    {
      endl_file.open(file_name.c_str(),ios::out|ios::app);
      Info("three_d_list::write_endl","Appending ENDL file "+file_name);
    }

    //Shove the header info in
    endl_file<<ENDL.header_line_1<<endl;
    endl_file<<ENDL.header_line_2<<endl;
    endl_file.setf(ios::scientific,ios::floatfield);
  
    //Now loop over the data and write it out
    double old_E_in = -1.0;
    for( typename three_d_list< two_d >::iterator E_in_link = three_d_list::begin( );
      E_in_link != three_d_list::end( ); 
      ++E_in_link )
    {
      // check for double value of E_in
      if( E_in_link->E_in( ) == old_E_in )
      {
        Warning("three_d_list::write_endl",
            pastenum("Widen the jump in three_d_list at E_in: ",old_E_in));
      }
      old_E_in = E_in_link->E_in( );

      // print the data for this E_in in ENDL format
      E_in_link->write_endl_link( endl_file );
    }
  
    //Put on end of file/section line into output file
    endl_file<<ENDL.eof_line<<endl;
  
    //Finally, we close the output file
    Info("three_d_list::write_endl","Closing ENDL file "+file_name);
    endl_file.close();
  
    //Now the cosine data
    cosines.write_endl(1);
  }

};

#endif
