/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1735 $
 * $Date: 2006-02-09 13:47:26 -0800 (Thu, 09 Feb 2006) $
 * $Author: dbrown $
 * $Id: mappings.cpp 1735 2006-02-09 21:47:26Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the mappings class

#include <fstream>  // standard file stream package
#include <iostream>

#include "mappings.hpp"
#include "messaging.hpp"

using namespace std;

// Set Use_cm true if you want to keep Kalbach data 
// in the center-of-mass frame, false for lab frame.
const bool Use_cm = false;

// ********* Pair class **********************
// ------------- constructor --------------------
Pair::Pair()
{
  E = 0.0;
  mu = 0.0;
}
// ------------- copy constructor -----------------
Pair::Pair(double energy, double cosine)
{
  E = energy;
  mu = cosine;
}
// ------------- destructor --------------------
Pair::~Pair()
{
}
// ------------- subtraction -------------------
Pair operator-(const Pair& A, const Pair& B)
// return A - B
{
  Pair diff = Pair(A.E - B.E, A.mu - B.mu);
  return diff;
}
// ------------- Pair_area -------------------
double Pair_area(const Pair& A, const Pair& B)
// area of the triangle generated by A, B, and O
{
  // use the vector cross product
  double area = 0.5*(A.E*B.mu - B.E*A.mu);
  return area;
}

// ********* mappings class **********************
// ------------- default constructor --------------------
mappings::mappings( )
{
  _E_in = 0.0;

  E_transl = 0.0;
}
// ------------- constructor --------------------
mappings::mappings(Target *target, Projectile *projectile,
    Product *ejectum, double E_in)
{
  _target = target;
  _projectile = projectile;
  _ejectum = ejectum;
  _E_in = E_in;

  // translational energy from initial collision
  // AWR is the atomic weight.
  // It is zero if we stay in center-of-mass frame
  E_transl = ( Use_cm ) ? 0.0 :
    (projectile->AWR)*(ejectum->AWR)*E_in/
    ((projectile->AWR + target->AWR)*
      (projectile->AWR + target->AWR));
}
// ------------- destructor --------------------
mappings::~mappings()
{
}
// ------------- mappings::set_map --------------------
void mappings::set_map(Target *target, Projectile *projectile,
    Product *ejectum, double E_in)
{
  _target = target;
  _projectile = projectile;
  _ejectum = ejectum;
  _E_in = E_in;

  // translational energy from initial collision
  // AWR is the atomic weight.
  // It is zero if we stay in center-of-mass frame
  E_transl = ( Use_cm ) ? 0.0 :
    (projectile->AWR)*(ejectum->AWR)*E_in/
    ((projectile->AWR + target->AWR)*
      (projectile->AWR + target->AWR));
}
// ------------- lab_to_cm --------------------
Pair mappings::lab_to_cm(Pair& lab)
{
  // this mapping is ill posed if cm.E is near 0
  const double Floor = 1.0e-8 * E_transl;
  if( Use_cm )
  {
    return lab;  // no change of coordinates
  }
  else
  {
    Pair cm;

    cm.E = lab.E + E_transl - 2*(lab.mu)*sqrt(E_transl*lab.E);
    if( cm.E < Floor )
    {
      cm.mu = 0.0;
    }
    else
    {
      cm.mu = (lab.mu)*sqrt(lab.E/cm.E) - sqrt(E_transl/cm.E);
    }
    return cm;
  }
}
// ------------- cm_to_lab --------------------
Pair mappings::cm_to_lab(Pair& cm)
// map from center-of-mass to Lab
{
  if( Use_cm )
  {
    return cm;  // no change of coordinates
  }
  else
  {
    Pair lab;

    lab.E = cm.E + E_transl + 2*(cm.mu)*sqrt(E_transl*cm.E);

    if(lab.E == 0.0)
    {
      lab.mu = 0.0;  //  pick an angle
    }
    else
    {
      lab.mu = (cm.mu)*sqrt(cm.E/lab.E) + sqrt(E_transl/lab.E);
    }
    return lab;
  }
}
// ------------- J_cm_to_lab --------------------
double mappings::J_cm_to_lab( double E_cm, double E_lab )
// Jacobian of the mapping
{
  if( Use_cm )
  {
    return 1.0;  // no change of coordinates
  }
  else
  {
    const double J_max = 1.0e7;

    // don't divide by zero
    double J = ( E_cm*J_max*J_max < E_lab )? J = J_max :
      sqrt( E_lab / E_cm );

    return J;
  }
}
// ------------- get_min_mu --------------------
double mappings::get_min_mu(double E_cm)
// return the minimal cosine (lab frame) for this exit energy
{
  double smallest;

  if(E_cm > E_transl)
  {
    smallest = -1.0;
  }
  else if(E_cm == E_transl)
  {
    smallest = 0.0;
  }
  else
  {
    smallest = sqrt(1.0 - E_cm/E_transl);
  }
  return smallest;
}

// ------------- get_E_range --------------------
void mappings::get_E_range(double E_cm, double mu, double *E_back,
  double *E_ahead)
// Return the lab frame exit energies for given center-of-mass exit
// energy and lab angle mu.
// If there are 2 such energies, the smaller one is E_back.
// If there is only 1, it is E_ahead and E_back = -1.
// If there are none, then both E_ahead and E_back are set to -1.
{
  double V_1 = E_cm - (1 - mu*mu)*E_transl;

  if(V_1 < 0.0)
  {
    Warning("mappings::get_E_range",pastenum("no particles with mu: ",mu));
    *E_back = -1.0;
    *E_ahead = -1.0;
  }
  else
  {
    V_1 = sqrt(V_1);
    double V_2 = mu*sqrt(E_transl);
    if(E_cm < E_transl)
    {
      // there are two forward velocities
      *E_back = (V_2 - V_1)*(V_2 - V_1);
      *E_ahead = (V_2 + V_1)*(V_2 + V_1);
    }
    else
    {
      // there is just one velocity
      *E_back = -1.0;
      *E_ahead = (V_2 + V_1)*(V_2 + V_1);
    }
  }
}
// ------------- mappings::get_lab_E --------------------
// Calculates the larger lab energy for given c-m energy and lab mu
double mappings::get_lab_E( double E_cm, double mu_lab, int sigma )
{
  // With V_0 = sqrt(E_transl) and alpha = tan(theta_lab),
  // find the intersection of the line
  //   y = alpha * (x + V_0)
  // with the circle x^2 + y^2 = E_cm,
  // sigma specifies the sign of the square root

  double E_lab;

  // special for mu = 0
  if( mu_lab == 0.0 )
  {
    E_lab = E_cm - E_transl;
    if( E_lab < 0.0 )
    {
      SevereError("mappings::lab_E_big", "negative E_lab");
    }
  }
  else
  {
    double alpha_sq = ( 1 - mu_lab*mu_lab ) / ( mu_lab*mu_lab );
    double alpha = sqrt( alpha_sq );
    if( mu_lab < 0.0 )
    {
      alpha *= -1.0;
    }
    double V_0 = sqrt( E_transl );
    double radical = alpha_sq*alpha_sq*E_transl +
      (1 + alpha_sq)*(E_cm - alpha_sq*E_transl);
    if( radical < 0.0 )
    {
      SevereError("mappings::get_lab_E", "no E_lab found");
    }
    double x = ( -alpha_sq*V_0 + sigma*sqrt( radical ) ) / ( 1 + alpha_sq );
    double y = alpha * ( x + V_0 );
    E_lab = ( x + V_0 )*( x + V_0 ) + y*y;
  }
  return E_lab;
}

