/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1872 $
 * $Date: 2006-05-17 08:34:30 -0700 (Wed, 17 May 2006) $
 * $Author: dbrown $
 * $Id: math_util.cpp 1872 2006-05-17 15:34:30Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of mymatrix class

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>

#include "math_util.hpp"
#include "messaging.hpp"
#include "ENDF_file.hpp"

using namespace std;
extern ENDLClass ENDL;

// *************** Param class *******************

// ------------- Param::set_zero -------------
void Param::set_zero( )
// Set parameters to zero
{
  for(Param::iterator itr=begin(); itr!=end(); ++itr)
  {
    *itr = 0.0;
  }
}

// ------------- Param::print() -------------
void Param::print()
{
  cout.setf(ios::scientific,ios::floatfield);
  for(Param::iterator itr=begin(); itr!=end(); ++itr)
  {
    cout << ENDL.data( *itr ) << "  ";
  }
  cout << endl;
}

// ------------- Param::operator*= -------------
Param& Param::operator*=( double alpha )
{
  for(Param::iterator itr=begin(); itr!=end(); ++itr)
  {
    *itr *= alpha;
  }
  return *this;
}

// ------------- Param::operator+= -------------
Param& Param::operator+=( Param& to_add )
{
  if( size() != to_add.size() )
  {
    SevereError( "Param::operator+=", "incompatible sizes" );
  }
  for( unsigned int count = 0; count < size(); ++count )
  {
    at( count ) += to_add[ count ];
  }
  return *this;
}

// *************** basic interpolation routines *****************
// ------------------- Lin_Lin ------------------------
double Lin_Lin( double x_mid, double x_0, double y_0,
  double x_1, double y_1 )
{
  double alpha = ( x_mid - x_0 ) / ( x_1 - x_0 );
  if((alpha <= 0.0) || (alpha >= 1.0))
  {
    SevereError("Lin_Lin","Attempt to extrapolate");
  }
  return (1.0 - alpha)*y_0 + alpha*y_1;
}
// ------------------- Log_Lin ------------------------
double Log_Lin( double x_mid, double x_0, double y_0,
  double x_1, double y_1 )
{
  double x_ratio = x_1 / x_0;

  // check for division by 0 and bad logarithm
  if( (x_ratio <= 0.0) || (x_ratio == 1.0) )
  {
    SevereError("Log_Lin",pastenum("bad x_ratio:",x_ratio));
  }

  double b = ( y_1 - y_0 ) / log( x_ratio );

  // check for extrapolation
  if( ( (x_mid > x_0) && (x_mid > x_1) ) ||
      ( (x_mid < x_0) && (x_mid < x_1) ) )
  {
    SevereError("Log_Lin","Attempt to extrapolate");
  }
  return y_0 + b*log( x_mid / x_0 );
}
// ------------------- Lin_Log ------------------------
double Lin_Log( double x_mid, double x_0, double y_0,
  double x_1, double y_1 )
// interpolation for log(y) = log(y_0) + b*(x - x_0)
{
  // check for proper inputs
  if( (y_0 <= 0.0) || (y_1 <= 0.0) ||
      (x_1 == x_0) )
  {
    if ( ( x_mid == x_0 ) && ( y_0 == y_1 ) ) return y_0;
    SevereError("Lin_Log", 
        "Bad inputs, (x_0,y_0), (x_1,y_1), x_mid: "+
        pastenum("(",x_0)+
        pastenum(",",y_0)+
        pastenum("), (",x_1)+
        pastenum(",",y_1)+
        pastenum("), ",x_mid)
    );
  }
  double b = log( y_1 / y_0 ) / ( x_1 - x_0 );

  // check for extrapolation
  if( ( (x_mid > x_0) && (x_mid > x_1) ) ||
      ( (x_mid < x_0) && (x_mid < x_1) ) )
  {
    SevereError("Lin_Log","Attempt to extrapolate");
  }

  return ( y_0 * exp( b * ( x_mid - x_0 ) ) );
}
// ------------------- Log_Log ------------------------
double Log_Log( double x_mid, double x_0, double y_0,
  double x_1, double y_1 )
// interpolation for log(y) = log(y_0) + b*log(x/x_0)
{
  if( (y_0 <= 0.0) || (y_1 <= 0.0) ||
      (x_0 <= 0.0) || (x_1 <= 0.0) ||
      (x_1 == x_0) )
  {
    if ( ( x_mid == x_0 ) && ( y_0 == y_1 ) && ( x_0 == x_1 ) ) return y_0;
    SevereError("Log_Log","Bad inputs");
  }

  double b = log( y_1 / y_0 ) /
    log( x_1 / x_0 );

  // check for extrapolation
  if( ( (x_mid > x_0) && (x_mid > x_1) ) ||
      ( (x_mid < x_0) && (x_mid < x_1) ) )
  {
    SevereError("Log_Log","Attempt to extrapolate");
  }

  return y_0 * exp( b * log(x_mid/x_0) );
}

// ---------------- maxE_loglog ----------------------------
double maxE_loglog( double xi, double alpha )
// get the maximum error for linear interpolation of x^alpha
// on the interval 1 < x < xi.
// This routine is used in linear interpolation of loglog data.
{
  if ( xi <= 1.0) //Make sure xi is not below interval
  {
    SevereError("maxE_loglog","xi must be larger than 1!");
  }

  //Calculate x_tilde - place where error is largest in interval
  double x_tilde;
  double u_tilde;  // y value at x_tilde
  double log_xi = log(xi);

  if ( abs(alpha*log_xi) <= 0.00001 )  //Take care of pole at 0
  {
    x_tilde = (1.0 - xi + alpha*log_xi)/((alpha - 1.0)*log_xi);
  }
  else if ( abs((alpha-1.0)*log_xi) <= 0.00001 )  //Take care of pole at +1
  {
    x_tilde = (alpha*xi*log_xi)/(xi - 1.0 + (alpha-1.0)*xi*log_xi);
  }
  else
  {
    x_tilde =
      alpha*(pow(xi,alpha) - xi)/((alpha - 1.0)*(pow(xi,alpha) - 1.0));
  }
  u_tilde = pow(x_tilde, alpha);
  //  cerr << "x_tilde = " << x_tilde << "  u_tilde = " << u_tilde << endl;

  //Comput v_tilde (and beta) --- y value of linlin extrap at x_tilde
  double beta = (pow(xi,alpha)-1.0)/(xi-1.0);
  //  cerr<<"beta = "<<beta<<endl;
  double v_tilde = 1.0 + beta*(x_tilde - 1.0);

  //Relative error at x_tilde
  double rel_err = abs((v_tilde - u_tilde)/u_tilde);
  //  cerr << "v_tilde = " << v_tilde << "  rel_err = " << rel_err << endl;

  return rel_err;
}

// ******* translation of Slatec incomplete gamma function *****
// local variables:
static int KONTRL = 2;
static char StrLevel[4][32] = { { "Informative message" },
  { "Informative message" },
  { "recoverable error" },
  { "fatal error" } };

// local routines
double c_d1mach( int I );
double c_d9gmit( double a, double x, double algap1, double sgngam, double alx );
double c_d9lgic( double a, double x, double alx );
double c_d9lgit( double a, double x, double algap1 );
double c_d9lgmc( double x );
double c_dcsevl( double x, double *cs, int n );
void c_dgamlm( double *xmin, double *xmax );
double c_dgamr( double x );
void c_dlgams( double x, double *dlgam, double *sgngam );
double c_dlngam( double x );
int c_initds( double *os, int nos, double eta );
void c_xermsg( char *Lib, char *Routine, char *Msg, int nError, int level );
void c_xsetf( int k ); 
int c_xgetf( void );
static void c_xerrPrint( char *Lib, char *Routine, char *Msg, int nError,
  int level );

/*
********************** c_d1mach *****************************
*/
double c_d1mach( int I )
// machine-dependent parameters
{

        static double D1MACH[5] = { DBL_MIN, DBL_MAX, DBL_EPSILON / 2.,
                           DBL_EPSILON, 3.01029995663981198e-01 };

        if( ( I < 1 ) || ( I > 5 ) ) c_xermsg( "cSLATEC", "d1mach",
           "I out of bounds", 1, 2 );

        return( D1MACH[I-1] );
}

/*
*********************** c_d9gmit ********************************
*/
double c_d9gmit( double a, double x, double algap1, double sgngam,
  double alx ) 
// Compute Tricomi's incomplete Gamma function for small arguments.
{

        static int FirstCall = 1;
        static double eps, bot;
        int k, m, ma;
        double ae, aeps, algs, alg2, fk, s, sgng2, t, te, d9gmit;

        if( FirstCall ) {
                FirstCall = 0;
                eps = 0.5 * c_d1mach( 3 );
                bot = log( c_d1mach( 1 ) );
        }

        if( x <= 0. ) c_xermsg( "cSLATEC", "c_d9gmit", "X should be > 0", 1, 2 );

        ma = static_cast<int>( a + 0.5 );
        if( a < 0. ) ma--;
        aeps = a - ma;

        ae = a;
        if( a < -0.5 ) ae = aeps;

        t = 1.;
        te = ae;
        s = t;
        for( k = 1; k <= 200; k++ ) {
                fk = k;
                te = -x * te / fk;
                t = te / ( ae + fk );
                s = s + t;
                if( fabs( t ) < eps * fabs( s ) ) break;
        }
        if( k > 200 ) c_xermsg( "cSLATEC", "c_d9gmit",
         "No convergence in 200 terms of taylor-s series", 2, 2 );

        if( a >= -0.5 ) {
                algs = -algap1 + log(s);
                d9gmit = exp( algs ); }
        else {
                algs = -c_dlngam( 1. + aeps ) + log( s );
                s = 1.;
                m = -ma - 1;
                if( m != 0 ) {
                        t = 1.;
                        for( k = 0; k < m; k++ ) {
                                t = x * t / ( aeps - ( m - k ) );
                                s = s + t;
                                if( fabs( t ) < eps * fabs( s ) ) break;
                        }
                }
                d9gmit = 0.;
                algs = -ma * log( x ) + algs;
                if( ( s == 0. ) && ( aeps == 0. ) ) return( exp( algs ) );
                sgng2 = 1.;
                if( s < 0. ) sgng2 = -1.;
                alg2 = -x - algap1 + log( fabs ( s ) );
                if( alg2 > bot ) d9gmit = sgng2 * exp( alg2 );
                if( algs > bot ) d9gmit = d9gmit + exp( algs );
        }
        return( d9gmit );
}

/*
****************************** c_d9lgic ****************************
*/
double c_d9lgic( double a, double x, double alx ) 
// Compute the log complementary incomplete Gamma function
//            for large X and for A .LE. X.
{

        static double eps = DBL_EPSILON / 2.;
        int k;
        double fk, p, r, s, t, xma, xpa;

        xpa = x + 1. - a;
        xma = x - 1. - a;

        r = 0.;
        p = 1.;
        s = p;
        for( k = 1; k <= 300; k++ ) {
                fk = k;
                t = fk * ( a - fk ) * ( 1. + r );
                r = -t / ( ( xma + 2. * fk ) * ( xpa + 2. * fk ) + t );
                p = r * p;
                s = s + p;
                if( fabs( p ) < eps * s ) break;
        }
        if( k > 300 ) c_xermsg ( "cSLATEC", "c_d9lgic", 
            "no convergence in 300 terms of continued fraction", 1, 2 );
        return( a * alx - x + log( s / xpa ) );
}

/*
****************************** c_d9lgit ******************************
*/
double c_d9lgit( double a, double x, double algap1 ) 
// Compute the logarithm of Tricomi's incomplete Gamma
//            function with Perron's continued fraction for large X and
//            A .GE. X.
{

        static int FirstCall = 1;
        static double eps, sqeps;
        int k;
        double ax, a1x, fk, hstar, p, r, s, t;

        if( FirstCall ) {
                FirstCall = 0;
                eps = 0.5 * c_d1mach( 3 );
                sqeps = sqrt( c_d1mach( 4 ) );
        }

        if( ( x <= 0. ) || ( a < x ) ) c_xermsg( "cSLATEC", "c_d9lgit",
           "x should be > 0.0 and <= a", 2, 2 );

        ax = a + x;
        a1x = ax + 1.0;
        r = 0.;
        p = 1.;
        s = p;
        for( k = 1; k <= 200; k++ ) {
                fk = k;
                t = ( a + fk ) * x * ( 1. + r );
                r = t / ( ( ax + fk ) * ( a1x + fk ) - t );
                p = r * p;
                s = s + p;
                if( fabs( p ) < eps * s) break;
        }
        if( k > 200 ) c_xermsg ( "cSLATEC", "c_d9lgit",
             "no convergence in 200 terms of continued fraction", 3, 2 );
        hstar = 1.0 - x * s / a1x;
        if( hstar < sqeps) c_xermsg ( "cSLATEC", "c_d9lgit",
               "result less than half precision", 1, 1 );
        return( -x - algap1 - log( hstar ) );
}

/*
*************************** c_d9lgmc *********************************
*/
double c_d9lgmc( double x ) 
// Compute the log Gamma correction factor so that
//            LOG(DGAMMA(X)) = LOG(SQRT(2*PI)) + (X-5.)*LOG(X) - X
//            + D9LGMC(X).
{


        static int FirstCall = 1;
        static int nalgm;
        static double xbig, xmax, algmcs[15] = {
                +0.1666389480451863247205729650822e+00,
                -0.1384948176067563840732986059135e-04,
                +0.9810825646924729426157171547487e-08,
                -0.1809129475572494194263306266719e-10,
                +0.6221098041892605227126015543416e-13,
                -0.3399615005417721944303330599666e-15,
                +0.2683181998482698748957538846666e-17,
                -0.2868042435334643284144622399999e-19,
                +0.3962837061046434803679306666666e-21,
                -0.6831888753985766870111999999999e-23,
                +0.1429227355942498147573333333333e-24,
                -0.3547598158101070547199999999999e-26,
                +0.1025680058010470912000000000000e-27,
                -0.3401102254316748799999999999999e-29,
                +0.1276642195630062933333333333333e-30 };
        double d9lgmc;

        if( FirstCall ) {
                FirstCall = 0;
                nalgm = c_initds( algmcs, 15, c_d1mach( 3 ) );
                xbig = 1.0 / sqrt( c_d1mach( 3 ) );
                xmax = c_d1mach( 2 ) / 12.;
                if( xmax > 1. / ( 12. * c_d1mach( 1 ) ) )
                   xmax = 1. / ( 12. * c_d1mach( 1 ) );
                xmax = exp( xmax );
        }

        if( x < 10. ) c_xermsg( "cSLATEC", "c_d9lgmc", "x must be >= 10", 1, 2 );
        if( x < xmax ) {
                d9lgmc = 1. / ( 12. * x );
                if( x < xbig ) d9lgmc =
                  c_dcsevl( 2.0 * ( 10. / x ) * ( 10. / x ) - 1., algmcs, nalgm ) / x;
        }
        else {
                d9lgmc = 0.;
                c_xermsg( "cSLATEC", "c_d9lgmc",
                    "x so big d9lgmc underflows", 2, 1 );
        }
        return( d9lgmc );
}

/*
************************* c_dcsevl ***********************************
*/
double c_dcsevl( double x, double *cs, int n ) 
// Evaluate the N-term Chebyshev series CS at X.
{

        static int FirstCall = 1;
        static double onepl;
        int i, ni;
        double b0, b1, b2, twox;

        if( FirstCall ) {
                FirstCall = 0;
                onepl = 1.0 + c_d1mach( 4 );
        }

        if( n < 1 ) c_xermsg( "cSLATEC", "c_dcsevl",
             "number of terms <= 0", 2, 2 );
        if( n > 1000 ) c_xermsg( "cSLATEC", "c_dcsevl",
             "number of terms > 1000", 3, 2 );
        if( fabs( x ) > onepl ) c_xermsg( "cSLATEC", "c_dcsevl",
             "x outside the interval (-1, +1)", 1, 1 );

        b1 = 0.;
        b0 = 0.;
        twox = 2.0 * x;
        for( i = 0; i <= n; i++ ) {
                b2 = b1;
                b1 = b0;
                ni = n - i;
                b0 = twox * b1 - b2 + cs[ ni ];
        }
        return( 0.5 * ( b0 - b2 ) );
}

/*
************************** c_dgamit **********************************
*/
double c_dgamit( double a, double x ) 
// Calculate Tricomi's form of the incomplete Gamma function,
//   x^{-a} / Gamma(a) * int_0^x t^(a-1) * exp{-t} dt.
{

        static int FirstCall = 1;
        static double alneps, sqeps;//, bot;
        double aeps, ainta, algap1, alng, alx, h, sga, sgngam, t, dgamit;

        if( FirstCall ) {
                FirstCall = 0;
                alneps = -log( c_d1mach( 3 ) );
                sqeps = sqrt( c_d1mach( 4 ) );
                //bot = log( c_d1mach( 1 ) );
        }

        if( x < 0. ) c_xermsg( "cSLATEC", "c_dgamit", "x is negative", 2, 2 );
        if( x != 0. ) alx = log( x );
        sga = 1.;
        if( a < 0. ) sga = -1.;
        ainta = (double) ( (int) ( a + 0.5 * sga ) );
        aeps = a - ainta;

        if( x <= 0. ) {
                dgamit = 0.;
                if( ( ainta > 0. ) || ( aeps != 0. ) )
                    dgamit = c_dgamr( a + 1.0 ); }
        else if( x <= 1. ) {
                        if( ( a >= -0.5 ) || ( aeps != 0. ) )
                            c_dlgams( a + 1., &algap1, &sgngam );
                        dgamit = c_d9gmit( a, x, algap1, sgngam, alx ); }
        else if( a >= x ) {
                t = c_d9lgit( a, x, c_dlngam( a + 1. ) );
                dgamit = exp( t ); }
        else {
                alng = c_d9lgic( a, x, alx );
                h = 1.;
                if( ( aeps != 0. ) || ( ainta > 0. ) ) {
                        c_dlgams( a + 1., &algap1, &sgngam );
                        t = log( fabs( a ) ) + alng - algap1;
                        if( t > alneps ) goto l60;
                        if( t > -alneps) h = 1. - sga * sgngam * exp( t );
                        if( fabs( h ) <= sqeps )
                           c_xermsg( "cSLATEC", "c_dgamit",
                                    "result lt half precision", 1, 1 );
                }
                t = -a * alx + log( fabs( h ) );
                dgamit = exp( t );
                if( h < 0. ) dgamit = -dgamit;
                return( dgamit );

l60:    t = t - a * alx;
                dgamit = -sga * sgngam * exp( t );
        }
        return( dgamit );
}

/*
*********************** c_dgamlm *************************************
*/
void c_dgamlm( double *xmin, double *xmax ) 
// Compute the minimum and maximum bounds for the argument in
//            the Gamma function.
{

        int i;
        double alnbig, alnsml, xln, xold;

        alnsml = log( c_d1mach( 1 ) );
        *xmin = -alnsml;
        for( i = 0; i < 10; i++ ) {
                xold = *xmin;
                xln = log( *xmin );
                *xmin = *xmin - *xmin * ( ( *xmin + 0.5 ) * xln - *xmin - 0.2258
 + alnsml ) / ( *xmin * xln + 0.5 );
                if( fabs( *xmin - xold ) < 0.005 ) break;
        }
        if( i == 10 )
           c_xermsg( "cSLATEC", "c_dgamlm", "unable to find xmin", 1, 2 );
        *xmin = -*xmin + 0.01;
        alnbig = log( c_d1mach( 2 ) );
        *xmax = alnbig;
        for( i = 0; i < 10; i++ ) {
                xold = *xmax;
                xln = log( *xmax );
                *xmax = *xmax - *xmax * ( ( *xmax - 0.5 ) * xln - *xmax + 0.9189
 - alnbig ) / ( *xmax * xln - 0.5 );
                if( fabs( *xmax - xold ) < 0.005 ) break;
        }
        if( i == 10 )
           c_xermsg( "cSLATEC", "c_dgamlm", "unable to find xmax", 2, 2 );
        *xmax = *xmax - 0.01;
        if( ( -*xmax + 1. ) > *xmin ) *xmin = -*xmax + 1.;
}
/*
**************************** c_dgamma ********************************
*/
double c_dgamma( double x ) 
// Compute the complete Gamma function.
{

        static int FirstCall = 1, ngam;
        static double pi = 3.14159265358979323846264338327950,
                sq2pil = 0.91893853320467274178032973640562,
                 xmin, xmax, dxrel;
        static double gamcs[42] = {
                +0.8571195590989331421920062399942e-02,
                +0.4415381324841006757191315771652e-02,
                +0.5685043681599363378632664588789e-01,
                -0.4219835396418560501012500186624e-02,
                +0.1326808181212460220584006796352e-02,
                -0.1893024529798880432523947023886e-03,
                +0.3606925327441245256578082217225e-04,
                -0.6056761904460864218485548290365e-05,
                +0.1055829546302283344731823509093e-05,
                -0.1811967365542384048291855891166e-06,
                +0.3117724964715322277790254593169e-07,
                -0.5354219639019687140874081024347e-08,
                +0.9193275519859588946887786825940e-09,
                -0.1577941280288339761767423273953e-09,
                +0.2707980622934954543266540433089e-10,
                -0.4646818653825730144081661058933e-11,
                +0.7973350192007419656460767175359e-12,
                -0.1368078209830916025799499172309e-12,
                +0.2347319486563800657233471771688e-13,
                -0.4027432614949066932766570534699e-14,
                +0.6910051747372100912138336975257e-15,
                -0.1185584500221992907052387126192e-15,
                +0.2034148542496373955201026051932e-16,
                -0.3490054341717405849274012949108e-17,
                +0.5987993856485305567135051066026e-18,
                -0.1027378057872228074490069778431e-18,
                +0.1762702816060529824942759660748e-19,
                -0.3024320653735306260958772112042e-20,
                +0.5188914660218397839717833550506e-21,
                -0.8902770842456576692449251601066e-22,
                +0.1527474068493342602274596891306e-22,
                -0.2620731256187362900257328332799e-23,
                +0.4496464047830538670331046570666e-24,
                -0.7714712731336877911703901525333e-25,
                +0.1323635453126044036486572714666e-25,
                -0.2270999412942928816702313813333e-26,
                +0.3896418998003991449320816639999e-27,
                -0.6685198115125953327792127999999e-28,
                +0.1146998663140024384347613866666e-28,
                -0.1967938586345134677295103999999e-29,
                +0.3376448816585338090334890666666e-30,
                -0.5793070335782135784625493333333e-31 };
        int i, n;
        double sinpiy, y, dgamma;

        if( FirstCall ) {
                FirstCall = 0;
                ngam = c_initds( gamcs, 42, 0.1 * c_d1mach( 3 ) );
                c_dgamlm( &xmin, &xmax );
                dxrel = sqrt( c_d1mach( 4 ) );
        }

        y = fabs( x );
        if( y <= 10 ) {
                n = static_cast<int>( x );
                if( x < 0. ) n--;
                y = x - n;
                n--;
                dgamma = 0.9375 + c_dcsevl( 2. * y - 1., gamcs, ngam );
                if( n != 0 ) {
                        if( n < 0 ) {
                                n = -n;
                                if( x == 0. )
                                   c_xermsg( "cSLATEC", "c_dgamma",
                                       "x is 0", 4, 2 );
                                if( ( x <  0.0 ) && ( x + n - 2 == 0. ) )
                                    c_xermsg( "cSLATEC", "c_dgamma", 
                                      "x is a negative integer", 4, 2 );
                                if( ( x < -0.5 ) && ( fabs( ( x - ( (int) ( x - 
0.5 ) ) ) / x ) < dxrel ) )
                                     c_xermsg( "cSLATEC", "c_dgamma",
                "answer < half precision because x too near negative integer", 1, 1 );
                                for( i = 0; i < n; i++ ) dgamma /= ( x + i ); }
                        else {
                                for( i = 1; i <= n; i++ ) dgamma *= ( y + i );
                        }
                } }
        else {
                if( x > xmax )
                    c_xermsg( "cSLATEC", "c_dgamma",
                         "x so big gamma overflows", 3, 2 );
                dgamma = 0.;
                if(x  < xmin )
                    c_xermsg( "cSLATEC", "c_dgamma", "u so small gamma underflows",
                       2, 1 );
                if(x >= xmin ) {
                        dgamma = exp( ( y - 0.5 ) * log( y ) - y + sq2pil + c_d9lgmc( y ) );
                        if( x <= 0. ) {
                                if( fabs( ( x - (int) (x - 0.5 ) ) / x ) < dxrel
 )
                                 c_xermsg( "cSLATEC", "c_dgammae",
                    "answer lt half precision, x too near negative integer", 1, 1 );
                                sinpiy = sin( pi * y );
                                if( sinpiy == 0. )
                      c_xermsg( "c_SLATEC", "c_dgamma", "x is a negative integer",
                            4, 2 );
                                dgamma = -pi / ( y * sinpiy * dgamma );
                        }
                }
        }
        return( dgamma );
}

/*
************************* c_dgamr ***********************************
*/
double c_dgamr( double x ) 
// Compute the reciprocal of the Gamma function.
{

        int irold;
        double dgamr, alngx, sgngx;

        dgamr = 0.;
        if( ( x > 0. ) || ( (double) ( (int) x ) != x ) ) {
                irold = c_xgetf( );
                c_xsetf( 1 );
                if( fabs( x ) <= 10. ) {
                        dgamr = 1. / c_dgamma( x );
                        c_xsetf( irold ); }
                else {
                        c_dlgams( x, &alngx, &sgngx );
                        c_xsetf( irold );
                        dgamr = sgngx * exp( -alngx );
                }
        }
        return( dgamr );
}

/*
*************************** c_dlgams *********************************
*/
void c_dlgams( double x, double *dlgam, double *sgngam ) 
// Compute the logarithm of the absolute value of the Gamma
//            function.
{

        *dlgam = c_dlngam( x );
        *sgngam = 1.;
        if( x <= 0. ) {
                if( ( ( -(int) x ) % 2 ) == 0 ) *sgngam = -1.;
        }
}

/*
************************ c_dlngam ************************************
*/
double c_dlngam( double x ) 
// Compute the logarithm of the absolute value of the Gamma
//            function.
{

        static int FirstCall = 1;
        static double sq2pil = 0.91893853320467274178032973640562,
                  sqpi2l = 0.225791352644727432363097614947441,
                  pi     = 3.14159265358979323846264338327950, xmax, dxrel;
        double sinpiy, y, temp;

        if( FirstCall ) {
                FirstCall = 0;
                temp = 1. / log( c_d1mach( 2 ) );
                xmax = temp * c_d1mach( 2 );
                dxrel = sqrt( c_d1mach( 4 ) );
        }

        y = fabs( x );
        if( y <= 10. ) return( log( fabs( c_dgamma( x ) ) ) );

        if( y > xmax )
             c_xermsg( "cSLATEC", "c_dlngam", "abs(x) so big c_dlngam overflows",
                   2, 2 );
        if( x > 0 ) return( sq2pil + ( x - 0.5 ) * log( x ) - x + c_d9lgmc( y ) 
);

        sinpiy = fabs( sin( pi * y ) );
        if( sinpiy == 0. )
              c_xermsg( "cSLATEC", "c_dlngam", "x is a negative integer", 3, 2 );
        if( fabs( ( x - (int) ( x -0.5 ) ) / x ) < dxrel )
              c_xermsg( "c_SLATEC", "c_dlngam",
                "answer < half precision because x too near negative integer",
                     1, 1 );
        return( sqpi2l + ( x - 0.5 ) * log( y ) - x - log( sinpiy ) - c_d9lgmc( y ) );
}

/*
**************************** c_initds ********************************
*/
int c_initds( double *os, int nos, double eta ) 
// Determine the number of terms needed in an orthogonal
//            polynomial series so that it meets a specified accuracy.
{

        int i, ii;
        double err;

        if( nos < 1 )
            c_xermsg( "cSLATEC", "c_initds", "number of coefficients is less than 1",
                2, 1 );
        err = 0.;
        for( ii = 1; ii <= nos; ii++ ) {
                i = nos - ii;
                err += fabs( os[i] );
                if( err > eta ) break;
        }
        if( i >= nos )
            c_xermsg( "cSLATEC", "c_initds",
                "Chebyshev series too short for specified accuracy", 1, 1 );
        return( i );
} 

/*
************************ c_xermsg **********************************
*/
void c_xermsg( char *Lib, char *Routine, char *Msg, int nError, int level ) 
// Process an error message depending on the severity level
{

        char Str[1024];

        if( ( level < -1 ) || ( level > 2 ) ) {
                sprintf( Str, "invalid level = %d", level );
                c_xerrPrint( Lib, "c_xermsg", Str, 1, 2 );
                exit( EXIT_FAILURE );
        }

        c_xerrPrint( Lib, Routine, Msg, nError, level );
        if( level == 2 ) {
                exit( EXIT_FAILURE );
        }
}
/*
************************** c_xerrPrint ********************************
*/
void c_xerrPrint( char *Lib, char *Routine, char *Msg, int nError, int level ) 
// Print an error message
{

        fprintf( stderr, "\n" );
        fprintf( stderr, "*** Message from routine %s in library %s\n", Routine,
 Lib );
        fprintf( stderr, "*** %s\n", StrLevel[level+1] );
        fprintf( stderr, "*   %s\n", Msg );
        fprintf( stderr, "*   Error number = %d\n", nError );
        fprintf( stderr, "*\n" );
        fprintf( stderr, "*** End of message\n" );
        fprintf( stderr, "\n" );
}

/*
*********************** c_xsetf ***********************************
*/
void c_xsetf( int k ) 
// set the error severity level to k
{

        char Str[1024];

        if( abs( k ) > 2 ) {
                sprintf( Str, "invalid k = %d value", k );
                c_xerrPrint( "cSlatec", "c_xsetf", Str, 1, 2 );
        }
        KONTRL = abs( k );
        return;
}

/*
****************************** c_xgetf **************************
*/
int c_xgetf( void ) { return( KONTRL ); }

// ************* expint function *******************************
/*********************************************************************
   Returns the exponential integral function
   E_n(x) = int_1^infinity e^{-x*t}/t^n dt,     for x > 0.
   C.A. Bertulani        May/15/2000
*********************************************************************/
#define EULER 0.5772156649      /* Euler's constant gamma */
#define MAXIT 100               /* Maximum allowed number of iterations. */
#define FPMIN 1.0e-30  // close to the smallest representable floting-point number.
#define EPS 6.0e-8    // Desired relative error, not smaller than the machine precision.

double expint(int n, double x)
{
  int i,ii,nm1;
  double a,b,c,d,del,fact,h,psi,ans;

  nm1=n-1;
  if (n < 0 || x < 0.0 || (x==0.0 && (n==0 || n==1)))
  {
    SevereError("expint", "Bad arguments");
  }
  else
  {
    if (n == 0)
    {
      ans=exp(-x)/x;   /* Special case */
    }
    else if (x == 0.0)
    {
      ans=1.0/nm1;  /* Another special case */
    }
    else if (x > 1.0)
    {   /* Lentz's algorithm */
      b=x+n;
      c=1.0/FPMIN;
      d=1.0/b;
      h=d;
      for (i=1;i<=MAXIT;i++)
      {
        a = -i*(nm1+i);
        b += 2.0;
        d=1.0/(a*d+b);  /* Denominators cannot be zero */
        c=b+a/c;
        del=c*d;
        h *= del;
        if (fabs(del-1.0) < EPS)
        {
          ans=h*exp(-x);
          return ans;
        }
      }
      SevereError("expint", "Continued fraction failed");
    }
    else
    {
      ans = (nm1!=0) ? 1.0/nm1 : -log(x)-EULER; /* Set first term */
      fact=1.0;
      for (i=1;i<=MAXIT;i++)
      {
        fact *= -x/i;
        if (i != nm1)
        {
          del = -fact/(i-nm1);
        }
        else
        {
          psi = -EULER;  /* Compute psi(n) */
          for (ii=1;ii<=nm1;ii++)
          {
            psi += 1.0/ii;
          }
          del=fact*(-log(x)+psi);
        }
        ans += del;
        if (fabs(del) < fabs(ans)*EPS)
        {
          return ans;
        }
      }
      SevereError("expint", "series failed");
    }
  }
  return ans;
}
#undef EPS
#undef EULER
#undef MAXIT
#undef FPMIN

// ************ for adaptive quadrature ******************
// ------------------- constructor ------------------------
quad_list::quad_list( double ( *F )( double x, Param& params ) )
{
  F_ = F;
}
// ------------------- quad_list::initialize ------------------
// Set up the first link and get a rough approximate integral
void quad_list::initialize( double A, double B, Param& params,
  double tol )
{
  // start with a clean slate
  if( !empty( ) )
  {
    SevereError("quad_list::initialize", "Start with an empty list");
  }

  A_ = A;
  B_ = B;
  EPS_ = 2.0e-14;
  tol_ = ( tol < EPS_ ) ? EPS_ : tol;
  params_ = params;

  quad_link first_link;

  first_link.a = A_;
  first_link.b = B_;
  first_link.fa = F_( A_, params_ );
  first_link.fb = F_( B_, params_ );
  double x = 0.5*( A_ + B_ );
  first_link.fm = F_( x, params_ );
  insert( end( ), first_link );

  // for the approximate integral
  is_ = first_link.fa + first_link.fb + first_link.fm;

  // some "random" other points
  x = A_ + 0.9501 * ( B_ - A_ );
  is_ += F_( x, params_ );

  x = A_ + 0.2311 * ( B_ - A_ );
  is_ += F_( x, params_ );

  x = A_ + 0.6068 * ( B_ - A_ );
  is_ += F_( x, params_ );

  x = A_ + 0.4860 * ( B_ - A_ );
  is_ += F_( x, params_ );

  x = A_ + 0.8913 * ( B_ - A_ );
  is_ += F_( x, params_ );

  is_ *= ( B_ - A_ ) / 8;

  // ensure a nonzero approximation
  if( is_ == 0 )
  {
    is_ = B_ - A_;
  }

  // scale to account for the tolerance
  is_ *= tol_ / EPS_;

  // initialize the integral
  sum_ = 0.0;

  warning_set = false;
}
// ------------------- quad_list::test_int ------------------
// Evaluate the integral over one interval and test its accuracy.
void quad_list::test_int( quad_list::iterator link_ptr )
{
  double m = ( link_ptr->a + link_ptr->b ) / 2;
  double h = ( link_ptr->b - link_ptr->a ) / 4;
  double fml = F_( link_ptr->a + h, params_ );
  double fmr = F_( link_ptr->b - h, params_ );

  // Simpson rule on original interval
  double i1 = ( link_ptr->fa + 4*link_ptr->fm + link_ptr->fb ) * h / 1.5;

  // Simpson rule on split interval
  double i2 = ( link_ptr->fa + 2*link_ptr->fm + link_ptr->fb +
                4*( fml + fmr ) ) * h / 3;

  // Do a Richardson extrapolation
  i1 = ( 16*i2 - i1 ) / 15;

  // test the accuracy
  if( ( is_ + (i1 - i2 ) == is_ ) ||
      ( m <= link_ptr->a ) ||
      ( link_ptr->b <= m ) )
  {
    if( ( m <= link_ptr->a ) || ( link_ptr->b <= m ) )
    {
      warning_set = true;
    }
    sum_ += i1;  // update the running integral
    erase( link_ptr );  // we are done with this interval
  }
  else
  {
    // split this interval
    quad_link left_half;
    left_half.a = link_ptr->a;
    left_half.b = m;
    left_half.fa = link_ptr->fa;
    left_half.fb = link_ptr->fm;
    left_half.fm = fml;
    insert( link_ptr, left_half );

    // update the right half-interval
    link_ptr->a = m;
    link_ptr->fa = link_ptr->fm;
    link_ptr->fm = fmr;
  }
}

// ------------------- quad_list::Simp_quad ------------------
// Evaluate the integral
double quad_list::Simp_quad( double A, double B, Param& params,
  double tol )
{
  int count = 0;
  initialize( A, B, params, tol );

  // process all of the subintervals
  while( !empty( ) )
  {
    quad_list::iterator link_ptr = begin( );
    // cout << "a: " << link_ptr->a << " b: " << link_ptr->b << endl;
    ++count;
    test_int( link_ptr );
  }
  if( warning_set )
  {
    Warning("quad_list::Simp_quad", "The integral may not be accurate");
  }
  // cout << "intervals: " << count << endl;
  return sum_;
}
