/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1806 $
 * $Date: 2006-03-27 15:37:19 -0800 (Mon, 27 Mar 2006) $
 * $Author: dbrown $
 * $Id: math_util.hpp 1806 2006-03-27 23:37:19Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// header for my math routines
#ifndef MATH_UTIL
#define MATH_UTIL

#include <iostream>
#include <cmath>
#include <cstdlib>
#include <list>
#include <vector>

const double EPS = 1.0e-14; // machine precision

using namespace std;

// ----------- class Param -----------------
//! The parameters of a probability function
class Param:public vector< double >
{
public:

  Param( ):vector< double >(1,0.0){}
  Param( int dim, double val):vector< double >(dim,val){}

  //! Allocate space
  void get_space( int N ){resize(N);}

  //! Set parameters to zero
  void set_zero( );

  void print( );

  // operator *=, used in interpolation
  Param& operator*=( double alpha );

  // operator +=, used in interpolation
  Param& operator+=( Param& to_add );
};

// *************** basic interpolation routines *****************
//! Interpolates between two points as lin-lin
double Lin_Lin( double x_mid, double x_0, double y_0,
  double x_1, double y_1 );

//! Interpolates between two points as lin-log
double Lin_Log( double x_mid, double x_0, double y_0,
  double x_1, double y_1 );

//! Interpolates between two points as log-lin
double Log_Lin( double x_mid, double x_0, double y_0,
  double x_1, double y_1 );

//! Interpolates between two points as log-log
double Log_Log( double x_mid, double x_0, double y_0,
  double x_1, double y_1 );

//!Determine maximum error in lin-lin interpolation of log-log data.
double maxE_loglog( double xi, double alpha );

// ******* translation of Slatec incomplete gamma function *****
// int_0^x t^{a-1} exp{-t} dt / ( Gamma(a) * x^a )
//! Incomplete gamma function
double c_dgamit( double a, double x );

// Gamma(a) = int_0^infinity t^{a-1} exp{-t} dt
//! Gamma function
double c_dgamma( double x );

// ************ Bertulani exponential integral function ***********
//! Exponential integral function
double expint(int n, double x);

// ************ for adaptive quadrature ******************
//! Classes for Walter Gander's adaptive Simpson quadrature routine.
//! See W. Gander and W. Gautschi, "Adaptive quadrature--revisited", BIT 40 (2000), 84-101.
// --------------- quad_link ---------------------------
//! Link for the list used by quadrature class quad_list
struct quad_link
{
public:
  double a;   // left end point
  double b;   // right end point
  double fa;  // f(a)
  double fb;  // f(b)
  double fm;  // f( 0.5*(a + b) )
};

// --------------- quad_list ---------------------------
//! The linked list used to hold the quadrature routine Simp_quad
class quad_list : public list< quad_link >
{
private:
  double (*F_)( double x, Param& params );  // the function to integrate
  double A_;   // left end point
  double B_;   // right end point
  Param params_;  // the function paramters
  double is_;  // a rough approximation to the integral
  double tol_; // the tolerance
  double EPS_; // machine epsilon
  double sum_; // the value of the integral
  bool warning_set;

  //! Set up the first link and get a rough approximate integral
  void initialize( double A, double B, Param& params,
    double tol );

  //! To test one interval
  void test_int( quad_list::iterator link_ptr );

public:
  quad_list( double ( *F )( double x, Param& params ) );

  ~quad_list( )
    {}

  //! Evaluate the integral
  double Simp_quad( double A, double B, Param& params,
    double tol );

};

#endif
