/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1897 $
 * $Date: 2006-09-08 11:04:02 -0700 (Fri, 08 Sep 2006) $
 * $Author: dbrown $
 * $Id: mf12classes.cpp 1897 2006-09-08 18:04:02Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the mf12 class for 2-d linked lists

#include <cmath>
#include <map>

#include "mf12classes.hpp"
#include "endl_formats.hpp"
#include "mf15classes.hpp"

extern ENDLClass ENDL;
extern GlobalParameterClass Global;
extern map< int, cascade > cascade_gammas;

// ********* for class MF12_base *************
// ----------- MF12_base::read_mult -----------------
//! handle the total gamma multiplicity
void MF12_base::read_mult( multiplicity& multiple )
{
  int NR, NP;

  inFile->read_line2(
    &NR, // number of interpolation regions
   	&NP  // number of data points
  );          

  // read the interpolation information
  inFile->get_regions( NR, multiple.NBT, multiple.INT );

  //The first data is the total multiplicity
  multiple.read_data( NP, *inFile );
}

// ----------- MF12_base::find_link -----------------
//! Find the 1-d list with incident energy E_in.
MF12_base::iterator MF12_base::find_link(double E_in,
  double EG)
{
  MF12_base::iterator e_in_link = begin();

  while( e_in_link->E_in() < E_in )
  {
    ++e_in_link;

    if(e_in_link == end())
    {
      SevereError("MF12_base::find_link",
        pastenum("mf12 incident energy ",E_in)+" not found");
    }
  }
  // check for data omission in the total multiplicity
  if( e_in_link->E_in() > E_in )
  {
    Warning("MF12_base::find_link",pastenum("for discrete gamma: ",EG)+
        pastenum(" there is an extra mf12 incident energy: " ,E_in));
    e_in_link = end();
  }
  return e_in_link;
}

// ----------- MF12_base::shift_gammas -----------------
//! Shift the gamma energies according to the atomic weight
void MF12_base::shift_gammas( )
{
  for( MF12_base::iterator e_in_link = begin();
    e_in_link != end( ); ++e_in_link )
  {
    // add this to each gamma energy
    double increment = ( AWR / ( AWR + 1 ) ) * e_in_link->E_in();
    dd_list::iterator next_ptr = e_in_link->begin( );
    ++next_ptr;
    for( dd_list::iterator link_ptr = e_in_link->begin( );
      link_ptr != e_in_link->end( ); link_ptr = next_ptr, ++next_ptr )
    {
      link_ptr->x += increment;
      if( link_ptr->x > ENDL.Max_E_out )
      {
        e_in_link->erase( link_ptr );
      }
    }
  }
}

// ----------- MF12_base::set_weights -----------------
//! set the weights for the discrete spectra
void MF12_base::set_weights( multiplicity& multiple )
{
  double e_in;
  double num;
  double denom;
  double fract;

  for( MF12_base::iterator list_ptr = begin( );
       list_ptr != end( ); ++list_ptr )
  {
    e_in = list_ptr->E_in( );
    // get the multiplicity of the lines
    num = list_ptr->line_multiple( );
    // get the total gamma multiplicity
    denom = multiple.evaluate( e_in );
    if( denom == 0.0 )
    {
      fract = 0.0;
    }
    else
    {
      fract = num/denom;
    }
    if( fract < 0.0 )
    {
      SevereError("MF12_base::set_weights",pastenum("bad weight: ",fract));
    }
    else if( fract > 1.0 )
    {
      Info("MF12_base::set_weights",
        pastenum("line multiplicity too big: ",fract)+", setting to 1.0");
      fract = 1.0;
    }
    list_ptr->weight = fract;
  }
}

// ----------- MF12_base::weight_mult -----------------
//! set the weights for the discrete spectra and save their multiplicities
void MF12_base::weight_mult( multiplicity& multiple, multiplicity& line_mult )
{
  double e_in;
  double num;
  double denom;
  double fract;
  dd_link XYdata;  // for adding links to line_mult

  for( MF12_base::iterator list_ptr = begin( );
       list_ptr != end( ); ++list_ptr )
  {
    e_in = list_ptr->E_in( );
    // get the multiplicity of the lines
    num = list_ptr->line_multiple( );

    // save this multiplicity
    XYdata.x = e_in;
    XYdata.y = num;
    line_mult.insert( line_mult.end( ), XYdata );

    // get the total gamma multiplicity
    denom = multiple.evaluate( e_in );
    if( denom == 0.0 )
    {
      fract = 0.0;
    }
    else
    {
      fract = num/denom;
    }
    if( ( fract < 0.0 ) || ( fract > 1.0 ) )
    {
      SevereError("MF12_base::weight_mult",pastenum("bad weight: ",fract));
    }
    list_ptr->weight = fract;
  }
}

// ----------- MF12_base::widen_deltas -----------------
//! Widen all of the delta-functions.
//! if clean_up is true, delete 1-d lists with zero norm
void MF12_base::widen_deltas( bool clean_up )
{
  MF12_base::iterator e_in_link = begin();
  MF12_base::iterator next_link = e_in_link;
  ++next_link;

  for( ; e_in_link != end( ); e_in_link = next_link, ++next_link )
  {
    e_in_link->widen_delta( );
    double Norm = e_in_link->get_norm( );

    // do we separate continuum from discrete gammas?
    if( Global.Value( "split_gammas" ) > 0 )
    {
      // for the multiplicity list
      dd_link XYdata;
      XYdata.x = e_in_link->E_in();
      XYdata.y = Norm;
      multiple.insert( multiple.end(), XYdata );
    }

    if( Norm > 0.0 )
    {
      e_in_link->renorm( Norm );
    }
    else if( clean_up )
    {
      erase( e_in_link );
    }
  }
}

// ----------- MF12_base::init_list -----------------
//! initialize the ENDL list from raw_list
void MF12_base::init_list( MF12_raw& raw_list )
{
  // initialize the 2-d lists of energy distributions using
  // the incident neutron energies from raw_list
  MF12_raw::iterator first_raw = raw_list.begin( );
  for( dd_list::iterator raw_ptr = first_raw->begin( );
    raw_ptr != first_raw->end( ); ++raw_ptr )
  {
    one_d_table new_link;
    new_link.E_in( ) = raw_ptr->E_in( );
    insert( end( ), new_link );
  }
}

// ----------- MF12_base::join_lines -----------------
//! If two adjacent lines have the same frequency, this routine adds their intensities.
void MF12_base::join_lines( )
{
  // loop over the gamma energies
  for( MF12_base::iterator E_gam_ptr = begin( );
       E_gam_ptr != end( ); ++E_gam_ptr )
  {
    E_gam_ptr->join_lines( );
  }
}

// ********* for class MF12_raw *************
// ----------- MF12_raw::one_line -----------------
void MF12_raw::one_line( double EG, int NR, int NP )
// Read the data for one gamma with energy EG
{
  dd_link XYdata;
  string linebuff;
  string strbuff;

  MF12_raw::iterator e_in_link;
  double E_in;
  double Prob;

  // keep the line if EG <= Max_g_energy (max outgoing energy for gammas)
  bool keepit = ( EG <= Global.Value( "Max_g_energy" ) );

  // compare with the maximum incident energy
  bool big_E_in = false;
  double dE = ENDL_EPSILON( ENDL.Max_E_in );

  // The gamma energies are in descending order
  multiplicity new_EG;
  MF12_raw::iterator new_EG_ptr;

  new_EG.E_gamma( ) = EG;
  if( keepit )
  {
    insert( begin( ), new_EG );
    // point to the new list
    new_EG_ptr = begin( );
    // read the interpolation regions
    inFile->get_regions( NR, new_EG_ptr->NBT, new_EG_ptr->INT );
  }
  else
  {
    Warning("MF12_raw::one_line",pastenum("Line above Max_g_energy, check that line not cut off.  EG = ",EG));
    // read and discard the interpolation regions
    inFile->get_regions( NR, new_EG.NBT, new_EG.INT );
  }

  // read the data for this line
  for( int iNP=0; iNP < NP; iNP++ ) //Loop over NP data pairs
  { 
    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 )
    {
      getline( *inFile, linebuff ); //Read in new line
    }
    strbuff = linebuff.substr(iP*22, 22);
    read_dd(&strbuff, &E_in, &Prob );
    if ( E_in==-999.0 || Prob==-999.0 ) //read_PAIR sends -999.0 if problem
    {
      SevereError("MF12_raw::one_line","read_PAIR fucked up!");
    }
    if( keepit && !big_E_in )
    {
      E_in *= ENDL.eV2MeV;
      XYdata.x = E_in;
      XYdata.y = Prob;
      new_EG_ptr->insert( new_EG_ptr->end( ), XYdata );
      // is this incident energy too big?
      if( E_in > ENDL.Max_E_in - dE )
      {
        new_EG_ptr->expand_interp( ENDL.Max_E_in );
        big_E_in = true;
        new_EG_ptr->chop( ENDL.Max_E_in );
      }
    }
  }
  if( keepit )
  {
    // expand histogram data
    new_EG_ptr->expand_interp( ENDL.Max_E_in );
    // the data may have jumps
    new_EG_ptr->widen_jumps( );
  }
}

// ----------- MF12_raw::fill_raw_lists -----------------
void MF12_raw::fill_raw_lists( )
// make sure that if any gamma energy has a
// ( E_in, multiplicity ) pair for some E_in, then all of
// the lists contain this E_in.
{
  // first widen the jumps
  widen_jumps( );

  // Build a list of incident energies by combining the incident energies of all the gammas
  list< double > E_incident;
  for (MF12_raw::iterator gamma_ptr = begin( ); gamma_ptr != end( ); ++gamma_ptr )
  {
    for (multiplicity::iterator itr=gamma_ptr->begin();itr!=gamma_ptr->end(); ++itr) 
      E_incident.push_back(itr->x);
  }
  E_incident.unique();
  E_incident.sort();

  // now ensure common E_in values in all of the gamma lists
  for( MF12_raw::iterator gamma_ptr = begin( );
      gamma_ptr != end( ); ++gamma_ptr )
  {
    for (list<double>::iterator Eitr=E_incident.begin(); Eitr!=E_incident.end();
    ++Eitr)
    {
      dd_list::iterator next_link = gamma_ptr->find_next(*Eitr);
      if ((next_link==gamma_ptr->end()) || (next_link->x != *Eitr))
        if ( gamma_ptr->begin()->x < *Eitr )  
        {
            // Eitr points to an energy that fits in the gamma's energy grid
            gamma_ptr->insert( next_link, dd_link( *Eitr, gamma_ptr->evaluate( *Eitr ) ) );
        }
        else             
        {
            // Eitr points to an energy that is below the start point of the gamma's energy grid, so multiplicity is zero
            gamma_ptr->insert( next_link, dd_link( *Eitr, 0.0 ) );
        }
    }
  }
}

// ----------- MF12_raw::fill_raw_lists -----------------
void MF12_raw::fill_raw_lists( MF15_list& cont_data )
// make sure that the ( E_in, multiplicity ) pairs in the
// ENDF raw list have E_in values matching the E_in tags
// in cont_data.
{
  list<double>::iterator Eitr;
  
  // Build a list of incident energies by combining the incident energies of all the gammas and continuum
  list< double > E_incident;
  cont_data.collect_Ein(E_incident);
  for (MF12_raw::iterator gamma_ptr = begin( ); gamma_ptr != end( ); ++gamma_ptr )
  {
    for (multiplicity::iterator itr=gamma_ptr->begin();itr!=gamma_ptr->end(); ++itr) 
      E_incident.push_back(itr->x);
  }
  E_incident.unique();
  E_incident.sort();

  // now ensure common E_in values in all of the gamma lists
  for( MF12_raw::iterator gamma_ptr = begin( );
      gamma_ptr != end( ); ++gamma_ptr )
  {
    for (Eitr=E_incident.begin(); Eitr!=E_incident.end(); ++Eitr)
    {
      dd_list::iterator next_link = gamma_ptr->find_next(*Eitr);
      if ((next_link==gamma_ptr->end()) || (next_link->x != *Eitr))
        gamma_ptr->insert(next_link,dd_link(*Eitr,gamma_ptr->evaluate(*Eitr)));
    }
  }

  E_incident.unique(); // somehow need to redo it??

  // now ensure common E_in values in are also in the continuum data, contin_data
  Eitr = E_incident.begin( );
  for(MF15_list::iterator cont_ptr = cont_data.begin( );
       ( cont_ptr != cont_data.end( ) ) &&
       ( Eitr != E_incident.end( ) ); )
  {
    if( cont_ptr->E_in( ) < *Eitr )
    {
      // go to the next
      ++cont_ptr;
    }
    else if( cont_ptr->E_in( ) == *Eitr )
    {
      // just go to the next links
      ++cont_ptr;
      ++Eitr;
    }
    else //  cont_ptr->E_in( ) > *Eitr
    {
      // insert an intermediate list
      one_d_table new_continuum;
      cont_data.insert( cont_ptr, new_continuum );
      MF15_list::iterator insert_ptr = cont_ptr;
      --insert_ptr;  // now points to the new link
      insert_ptr->E_in( ) = *Eitr;
      MF15_list::iterator prev_ptr = insert_ptr;
      --prev_ptr;
      insert_ptr->list_interp( *Eitr, *prev_ptr, *cont_ptr );
      ++Eitr;
    }
  }
}

// ----------- MF12_raw::insert_cross -----------------
void MF12_raw::insert_cross( list< double >& crossings )
// insert ( E_in, multiplicity ) pairs for each of the crossings
// neutron energies
{
  // loop over the gamma energies
  for( MF12_raw::iterator E_gam_ptr = begin( );
    E_gam_ptr != end( ); ++E_gam_ptr )
  {
    list< double >::iterator cross_ptr = crossings.begin( );
    dd_list::iterator prev_ptr = E_gam_ptr->begin( );
    dd_list::iterator next_ptr = prev_ptr;
    ++next_ptr;

    double E_prev = prev_ptr->x;
    double E_next = next_ptr->x;
    double E_cross = *cross_ptr;
    // scan the lists
    for( ; ; )
    {
      if( E_cross < (1 - EPS)*E_prev )
      {
        ++cross_ptr;
        E_cross = *cross_ptr;
      }
      else if( E_cross < (1 + EPS)*E_prev )
      {
	// we essentially have equality
        ++cross_ptr;
        E_cross = *cross_ptr;
      }
      else if( E_cross < (1 - EPS)*E_next )
      {
	// are between 2 neutron energies
	// insert a new link
        dd_link XYdata;
        XYdata.x = E_cross;
        XYdata.y = E_gam_ptr->interp( E_cross, *prev_ptr, *next_ptr,
          E_gam_ptr->INT[ 0 ] );
        E_gam_ptr->insert( next_ptr, XYdata );
        E_prev = E_cross;
        ++prev_ptr;
        ++cross_ptr;
        E_cross = *cross_ptr;
      }
      else if( E_cross < (1 + EPS)*E_next )
      {
	// we essentially have equality
        ++cross_ptr;
        E_cross = *cross_ptr;
        prev_ptr = next_ptr;
        ++next_ptr;
        E_prev = E_next;
        E_next = next_ptr->x;
      }
      else
      {
	// E_cross > E_next
        prev_ptr = next_ptr;
        ++next_ptr;
        E_prev = E_next;
        E_next = next_ptr->x;
      }
      // are we done?
      if( ( cross_ptr == crossings.end( ) ) ||
          ( next_ptr == E_gam_ptr->end( ) ) )
      {
        break;
      }
    }
  }
}
// ----------- MF12_raw::duplicate_E -----------------
void MF12_raw::duplicate_E( )
// ENDF/B-VI sometimes has duplicate gamma lines from
// different cascades.  We combine them
{
  MF12_raw::iterator this_link = begin();
  MF12_raw::iterator next_link = this_link;
  ++next_link;

  // go through the list
  for( ; next_link != end( ); )
  {
    if( next_link->E_gamma( ) == this_link->E_gamma( ) )
    {
      // combine these lists
      (*this_link) += (*next_link);
      erase( next_link );
    }
    else
    {
      // test the next one
      ++this_link;
    }
    // update the next link
    next_link = this_link;
    ++next_link;
  }
}

// ********* for class MF12_list *************
// ----------- MF12_list::master -----------------
void MF12_list::master( double awr, int NK, mf12_file *InFile )
// handle an mf12 multiplicity file (LO = 1) in the isotropic case
{
  // save the input data
  movers.AWR = awr;
  mf12_NK = NK;

  // Use InFile for all of the mf12 input
  inFile = InFile;
  statics.inFile = InFile;
  raw_move.inFile = InFile;
  raw_static.inFile = InFile;

  // initialize the Booleans
  with_move = false;
  with_static = false;
  with_mf15 = false;

  // get the energy distributions
  read_data( );

  // proceed according to the type of data
  if( NK == 1 )
  {
    //Here, there is a single gamma from the reaction (except continuum)
    //The i=9 file is created in the write_xxxx routines with a 
    //multiplicity of one (or a ratio of cross sections).
    if( with_mf15 )
    {
      // there are only continuum distributions
      Info("MF12_list::master","mf12 continuum only");
      write_continuum( );
    }
    else if( with_static )
    {
      // there is just 1 line (static)
      Info("MF12_list::master","mf12 1 static line");
      write_static( );
    }
    else
    {
      // there is just 1 line (moving)
      Info("MF12_list::master","mf12 1 moving line");
      write_mover( );
    }
  }
  else
  {
    if( with_mf15 )
    {
      if( with_move )
      {
        // moving lines plus continuum
        Info("MF12_list::master","mf12 moving lines plus continuum");
        move_w_contin( );
      }
      else
      {
        // static lines plus continuum
        Info("MF12_list::master","mf12 static lines plus continuum");
        plus_contin( );
      }
    }
    else if( with_move )
    {
      // continuum represented by moving lines
      Info("MF12_list::master","mf12 static lines plus movers");
      do_movers( );
    }
    else
    {
      // only static lines
      Info("MF12_list::master","mf12 static lines only");
      just_statics( );
    }
  }
}
// ----------- MF12_list::read_data -----------------
void MF12_list::read_data( )
// This routine is used to read in ENDF/B-VI probabilities
// of discrete gammas,
// mf12_NK is the number of data sets, discrete gammas + continuum
{
  double EG; // gamma energy
  double ES; // nucleus excitation level
  int LP;    // 2: put into the raw_move list
             // 0, 1: stationary gamma
  int LF;    // 2 for discrete; 1 for continuum (in MF15 file)
  int NR;    // number of interpolation regions
  int NP;    // number of incident energies

  if( mf12_NK > 1 )
  {
    // the first block is the total multiplicity
    read_mult( multiple );

    if( ENDL.F == 13 )  //Need to convert cross section to multiplicity
    {
      multiple.divide_by_xs();
    }
  }
  
  for(int count = 0; count < mf12_NK; ++count)
  {
    inFile->gamma_line1(
      &EG,     // gamma energy
      &ES,     // excitation level
      &LP,     // energy-dependent gamma?
      &LF,     // discrete-continuum flag
      &NR,     // number of interpolation regions
      &NP      // number of (E, multiplicity) pairs
    );

    // change to MeV
    EG *= ENDL.eV2MeV;
    if(LF == 1)
    {
      // this data is for the continuum, read the mf15 file
      with_mf15 = true;
      continuum( NR, NP );
      break; // we should be done
    }
    else if( LP == 2 )
    {
      // a gamma to shift
      with_move = true;
      raw_move.one_line( EG, NR, NP );
    }
    else
    {
      // a fixed gamma
      with_static = true;
      raw_static.one_line( EG, NR, NP );
    }
  }
}
// ----------- MF12_list::clean_multiple -----------------
void MF12_list::clean_multiple( MF12_base& statics )
// make sure that the multiplicity is 0 if there is no distribution
{
  multiplicity::iterator this_ptr = multiple.begin( );
  double first_E = statics.begin( )->E_in( );

  while ( (this_ptr != multiple.end( )) && this_ptr->x < first_E )
  {
    this_ptr = multiple.erase( this_ptr );
  }
/*
  if( this_ptr->x < first_E )
  {
    for( ; this_ptr->x < first_E; ++this_ptr )
    {
      if( this_ptr == multiple.end( ) )
      {
        SevereError("MF12_list::clean_multiple","Unexplained failure");
      }
      this_ptr->x = first_E;
      this_ptr->y = 0.0;
    }
  }
*/
}
// ----------- MF12_list::write_continuum -----------------
void MF12_list::write_continuum( )
// write only the continuum data
{
  if( ENDL.F == 13 )
  {
    // convert from photon production cross section to multiplicity
    cont_data.multiple.divide_by_xs();
  }
  cont_data.multiple.widen_jumps();
  ENDL.set_yo( 7 );
  cont_data.multiple.write_endl( 9 );
  cont_data.widen_jumps();
  cont_data.renorm();
  cont_data.write_endl( 4 );
}
// ----------- MF12_list::write_statics -----------------
void MF12_list::write_statics( )
// write the static lines
{
  ENDL.set_yo( 7 );
  ENDL.set_s_number( 3 );
  ENDL.set_I_number( 9 );

  fstream endl_file;
  string file_name = ENDL.file_name;

  // combine duplicate lines
  raw_static.duplicate_E( );

  //Now loop over the raw static lines
  for( MF12_raw::iterator link = raw_static.begin();
     link != raw_static.end(); ++link)
  {
    // reset the header lines
    ENDL.set_x1( link->E_gamma() );

    if( ENDL.F == 13 )  //Need to convert cross section to multiplicity
    {
      link->divide_by_xs();
    }

    // print the multiplicity
    link->write_endl( 9 );

    // append the rest of the data
    ENDL.append = true;
  }
  //Reset "append"
  ENDL.append = false;

}
// ----------- MF12_list::write_static -----------------
void MF12_list::write_static( )
// write the one static line
{
  statics.init_list( raw_static ); // initialize the ENDL list
  // copy the mf12 data into ENDL order
  ENDL_order( raw_static, statics );
  statics.join_lines( );    // join any repeated lines
  statics.widen_deltas( true );  // widen the delta-functions

  ENDL.set_yo( 7 );
  // clean up and write out
  statics.widen_jumps();
  statics.renorm();
  statics.write_endl( 4 );

  //Since this routine only gets called when there is a single gamma...
  if( !multiple.empty( ) )
  {
    SevereError("MF12_list::write_static",
		" there should be no total multiplicity" );
  }

  // raw_static has just 1 multiplicity list
  MF12_raw::iterator mult_list = raw_static.begin( );
  mult_list->widen_jumps( );
  if( ENDL.F == 13 )  //Need to convert cross section to multiplicity
  {
    mult_list->divide_by_xs( );
  }

  mult_list->check_count( );   // Is multiplicity <= 1?
  mult_list->write_endl( 9 );  //Write the ENDL file
}
// ----------- MF12_list::write_mover -----------------
void MF12_list::write_mover( )
// write the one moving line
{
  movers.init_list( raw_move ); // initialize the ENDL list
  // copy the mf12 data into ENDL order
  ENDL_order( raw_move, movers );
  movers.join_lines( );    // join any repeated lines
  movers.widen_deltas( true );  // widen the delta-functions

  // move the moving line
  movers.shift_gammas( );
  
  // clean up and write out
  ENDL.set_yo( 7 );
  movers.widen_jumps();
  movers.renorm();
  movers.write_endl( 4 );

  //Since this routine only gets called when there is a single gamma...
  MF12_raw::iterator mult_list = raw_move.begin( );
  mult_list->widen_jumps( );
  if( ENDL.F == 13 )  //Need to convert cross section to multiplicity
  {
    mult_list->divide_by_xs( );
  }

  mult_list->write_endl( 9 );  //Write the ENDL file
}
// ----------- MF12_list::plus_contin -----------------
void MF12_list::plus_contin( )
// there is mf15 continuum plus static lines
{
  if( Global.Value( "split_gammas" ) > 0 )
  {
    // s=3 is used when there are separate discrete gammas
    ENDL.set_s_number( 3 );

    // write the continuum
    write_continuum( );

    // write the static lines
    ENDL.append = true;
    write_statics( );
  }
  else
  {
    // ensure that we have 2-d links for all energies
    raw_static.fill_raw_lists( cont_data );

    statics.init_list( raw_static ); // initialize the ENDL lists

    // copy the mf12 data into ENDL order
    ENDL_order( raw_static, statics );
    statics.join_lines( );    // join any repeated lines

    // set the weights for the discrete lines
    statics.set_weights( multiple );

    // widen the delta-functions
    // keep the ones with zero norm---they may have continuum data
    statics.widen_deltas( false );

    // set the weights for continuum
    set_mf15_wt( );

    // weight the statics
    statics.use_weight( );

    // widen jumps in the continuum before adding them to the statics
    cont_data.widen_jumps( );

    // do the weighted sum
    sum_lists( statics, cont_data ); 

// for testing
//cont_data.widen_jumps( );
//cont_data.renorm( );
//ENDL.set_yo( 7 );
//cont_data.write_endl( 4 );


    // data clean up and final output
    statics.widen_jumps( );
    statics.renorm( );
    ENDL.set_yo( 7 );
    statics.write_endl( 4 );

    // make sure that the multiplicity is 0 if there is no distribution
    clean_multiple( statics );

    // print the multiplicities
    multiple.widen_jumps( );
    multiple.write_endl( 9 );
  }
}
// ----------- MF12_list::join_move_static -----------------
void MF12_list::join_move_static( )
// join the moving lines with the static ones
{
  // ensure that we have 2-d links for all energies
  raw_static.fill_raw_lists( );
  raw_move.fill_raw_lists( );

  find_crossings( );  // where moving and static energies coincide
  statics.init_list( raw_static ); // initialize the ENDL lists
  movers.init_list( raw_static ); // initialize the ENDL lists

  // copy the mf12 data into ENDL order
  ENDL_order( raw_static, statics );
  ENDL_order( raw_move, movers );

  // move the moving lines
  movers.shift_gammas( );

  // weight the statics
  statics.use_weight( );
  // form the line sum
  sum_lines( statics, movers );
  statics.join_lines( );    // join any repeated lines
}
// ----------- MF12_list::move_w_contin -----------------
void MF12_list::move_w_contin( )
// there is mf15 continuum plus moving lines plus maybe static lines
{
  if( ( Global.Value( "split_gammas" ) > 0 ) || !with_static )
  {
    // a list for the multiplicity of the movers
    multiplicity move_mult;

    // combine the moving discrete gammas with the continuum
    // ensure that we have 2-d links for all energies
    raw_move.fill_raw_lists( cont_data );

    movers.init_list( raw_move ); // initialize the ENDL lists

    // copy the mf12 data into ENDL order
    ENDL_order( raw_move, movers );

    // set the weights of the movers and save their multiplicities
    movers.weight_mult( multiple, move_mult );

    // move the moving lines
    movers.shift_gammas( );

    // widen the delta-functions
    // keep the ones with zero norm---they may have continuum data
    movers.widen_deltas( false );

    set_mf15_wt( );   // set the weights for the continuum

//    // make sure that the mf15 data has all of the movers energies
//    list< double > E_incident;
//    movers.collect_Ein( E_incident );
//    cont_data.collect_Ein( E_incident );
//    E_incident.sort( );
//    E_incident.unique( );
//    cont_data.fill_in_list( E_incident );

    // weight the movers
    movers.use_weight( );
    
    // do the weighted sum
    sum_lists( movers, cont_data ); 

    // write the continuum
    ENDL.set_yo( 7 );
    if( with_static )
    {
      // s=3 is used when there are separate discrete gammas
      ENDL.set_s_number( 3 );
    }

    // data clean up and final output
    movers.widen_jumps( );
    movers.renorm( );
    movers.write_endl( 4 );

    // get the multiplicity
    move_mult += cont_data.multiple;

    if( ENDL.F == 13 )
    {
      // convert from photon production cross section to multiplicity
      move_mult.divide_by_xs();
    }
    move_mult.widen_jumps();

    ENDL.set_yo( 7 );
    move_mult.write_endl( 9 );

    // write the static lines
    if( with_static )
    {
      ENDL.append = true;
      write_statics( );
    }
  }
  else
  {
    // make sure that one line has all of the energies from the continuum
    list< double > E_contin;
    cont_data.collect_Ein( E_contin );
    raw_static.begin( )->collect_E_in( E_contin );
    E_contin.sort( );
    E_contin.unique( );

    raw_static.begin( )->fill_with( E_contin );

    // combine the static lines with the moving ones
    join_move_static( );

    // set the weights for continuum and discrete lines
    set_mf15_wt( );

    statics.widen_deltas( false );  // widen the delta-functions

    // make sure that the mf15 data has all of the energies from the lines
    list< double > E_incident;
    statics.collect_Ein( E_incident );
    cont_data.collect_Ein( E_incident );
    E_incident.sort( );
    E_incident.unique( );
    cont_data.fill_in_list( E_incident );

    // weight the lines
    statics.use_weight( );
    sum_lists( statics, cont_data ); // do the weighted sum

    // data clean up and final output
    ENDL.set_yo( 7 );
    statics.renorm( );
    statics.write_endl( 4 );

    // make sure that the multiplicity is 0 if there is no distribution
    clean_multiple( statics );

    // print the multiplicities
    multiple.widen_jumps( );
    multiple.write_endl( 9 );
  }
}
// ----------- MF12_list::do_movers -----------------
void MF12_list::do_movers( )
// there are moving lines
{
  if( Global.Value( "split_gammas" ) > 0 )
  {
    // separate the discrete gammas from the continuum

    // ensure that we have 2-d links for all energies
    raw_move.fill_raw_lists( );

    movers.init_list( raw_move ); // initialize the ENDL lists

    // copy the mf12 data into ENDL order
    ENDL_order( raw_move, movers );

    // move the moving lines
    movers.shift_gammas( );
    
    // join any repeated lines
    movers.join_lines( );    

    // widen the delta-functions and sum the multiplicities
    movers.widen_deltas( true );  

    // clean up and output
    ENDL.set_yo( 7 );
    ENDL.set_s_number( 3 );
    movers.widen_jumps( );  
    movers.renorm( );
    movers.write_endl( 4 );
    // print the multiplicities of movers
    if( ENDL.F == 13 )  //Need to convert cross section to multiplicity
    {
      movers.multiple.divide_by_xs();
    }
    movers.multiple.widen_jumps( );
    movers.multiple.write_endl( 9 );

    // append the static lines
    ENDL.append = true;
    write_statics( );
  }
  else
  {
    // first combine the static lines with the moving ones
    join_move_static( );
    statics.widen_deltas( true );  // widen the delta-functions

    // clean up and output
    ENDL.set_yo( 7 );
    statics.widen_jumps( );
    statics.renorm( );
    statics.write_endl( 4 );

    // make sure that the multiplicity is 0 if there is no distribution
    clean_multiple( statics );

    // print the multiplicities
    multiple.widen_jumps( );
    multiple.write_endl( 9 );
  }
}
// ----------- MF12_list::just_statics -----------------
void MF12_list::just_statics( )
// write the static lines
{
  if( Global.Value( "split_gammas" ) > 0 )
  {
    // write the static lines
    write_statics( );
  }
  else
  {
    // ensure that we have 2-d links for all energies
    raw_static.fill_raw_lists( );
    statics.init_list( raw_static ); // initialize the ENDL lists

    // copy the mf12 data into ENDL order
    ENDL_order( raw_static, statics );
    statics.join_lines( );    // join any repeated lines
    statics.widen_deltas( true );  // widen the delta-functions

    // clean up and output
    ENDL.set_yo( 7 );
    statics.widen_jumps();
    statics.renorm();
    statics.write_endl( 4 );

    // make sure that the multiplicity is 0 if there is no distribution
    clean_multiple( statics );

    // print the multiplicities
    multiple.unique( );
    multiple.widen_jumps( );
    multiple.write_endl( 9 );
  }
}
// ----------- MF12_list::find_crossings -----------------
void MF12_list::find_crossings( )
// Find the neutron energies at which an energy-dependent
// gamma coincides with a stationary gamma
{
  const double EPS = Global.Value( "mf5_tol" );  // tolerance for mf5 data

  list< double > crossings;  // the coincident energies
  double E_n;  // neutron energy for a gamma coincidence

  // go through the static gammas
  for( MF12_raw::iterator stat_ptr = raw_static.begin( );
    stat_ptr != raw_static.end( ); ++stat_ptr )
  {
    // go through the moving gammas
    for( MF12_raw::iterator move_ptr = raw_move.begin( );
      move_ptr != raw_move.end( ); ++move_ptr )
    {
      if( move_ptr->E_gamma( ) > ( 1 - EPS )*stat_ptr->E_gamma( ) )
      {
        break;
      }
      else
      {
        E_n = ( ( movers.AWR + 1 ) / movers.AWR ) *
          ( stat_ptr->E_gamma( ) - move_ptr->E_gamma( ) );
	// insert this energy in the list
        crossings.insert( crossings.end( ), E_n );
      }
    }
  }

  // put the crossing energies in increasing order
  crossings.sort( );

  // remove duplicates
  list< double >::iterator cross_ptr = crossings.begin( );
  list< double >::iterator next_ptr = cross_ptr;
  ++next_ptr;
  double prev_E_n = -1.0;
  
  for( ; cross_ptr !=  crossings.end( ); cross_ptr = next_ptr, ++next_ptr )
  {
    E_n = *cross_ptr;
    if( abs( E_n - prev_E_n ) < EPS*E_n )
    {
      crossings.erase( cross_ptr );
    }
    else
    {
      prev_E_n = E_n;
    }
  }
  
  // insert ( neutron energy, multiplicity ) pairs at each crossing energy
  raw_move.insert_cross( crossings );
  raw_static.insert_cross( crossings );
}
// ----------- MF12_list::set_mf15_wt -----------------
void MF12_list::set_mf15_wt( )
// set the weights for the continuum
{
  double num;
  double denom;
  double fract;  // num/denom, the weight
  double e_in;

  // set the weights in cont_data
  for( MF15_list::iterator cont_ptr = cont_data.begin( );
       cont_ptr != cont_data.end( ); ++cont_ptr )
  {
    e_in = cont_ptr->E_in( );
    num = cont_data.multiple.evaluate( e_in );
    denom = multiple.evaluate( e_in );
    if( denom == 0.0 )
    {
      fract = 0.0;
    }
    else
    {
      fract = num/denom;
    }
    if( ( fract < 0.0 ) || ( fract > 1.0 ) )
    {
      SevereError("MF12_list::set_mf15_wt",pastenum("bad weight: ",fract));
    }
    cont_ptr->weight = fract;
  }
}
// ----------- MF12_list::ENDL_order -----------------
void MF12_list::ENDL_order( MF12_raw& mf12_data, MF12_base& ENDL_data )
// copy the mf12_data to ENDL_data
{
  const double EPS = Global.Value( "mf5_tol" );  // tolerance for mf5 data

  dd_link XYdata;
  MF12_base::iterator e_in_link;
  double E_in;
  double EG;  // gamma energy
  dd_list::iterator data_ptr;

  // loop through the gammas
  for( MF12_raw::iterator raw_ptr = mf12_data.begin( );
    raw_ptr != mf12_data.end( ); ++raw_ptr )
  {
    EG = raw_ptr->E_gamma( ); // the gamma energy

    // loop over the (E_in, multiplicity) pairs
    for( data_ptr = raw_ptr->begin( ); data_ptr != raw_ptr->end( );
      ++data_ptr )
    {
      // form an (E_in, multiplicity) pair
      XYdata.x = EG;
      XYdata.y = data_ptr->y;
      E_in = data_ptr->x;

      // Find the 1-d list with this incident energy
      e_in_link = ENDL_data.find_link( E_in, EG );

      if( e_in_link != ENDL_data.end( ) )
      {
        // The gamma energies are now in ascending order
        e_in_link->insert( e_in_link->end( ), XYdata );
      }
    }
  }
}
// ----------- MF12_list::continuum -----------------
//! handle multiplicity and energy distributions for the continuum
//! The multiplicity for the continuum is at the end of the mf12 file
//! and we use it only if mf12_NK = 1
void MF12_list::continuum( int NR, int NP )
{
  // read the interpolation regions
  inFile->get_regions( NR, cont_data.multiple.NBT, cont_data.multiple.INT );

  cont_data.multiple.read_data( NP, *inFile );

  contin_file.open( 15, ENDL.T );  // open the mf15 file
  cont_data.read_data( contin_file );
  Info("MF12_list::continuum","Closing file "+endf_file_name(15,ENDL.T)+"...");
  contin_file.close( );
}

// ********* for class MF12_deexcite *************
// ----------- MF12_deexcite::master -----------------
//! handle an mf12 transition probability file (LO = 2)
//! For LG = 1 the data is pairs; for LG = 2 it is triplets.
void MF12_deexcite::master( double AWR, int LG, mf12_file& inFile )
{
  //  int NR,NP;
  //  double EG,ES;
  //  int LP,LF;

  double Max_energy = ENDL.Max_E_in;

  // get the threshold energy
  double e_in = ENDL.threshold;
  if( e_in > Max_energy )
  {
    // skip this data
    return;
  }

  // make a new list at the threshold energy
  one_d_table new_link;
  new_link.E_in() = e_in;
  insert(end(), new_link);
  MF12_deexcite::iterator e_in_link = end();
  --e_in_link;

  // get the discrete gammas
  read_data( AWR, LG, inFile );

  // stick it in the MF12_deexcite object
  stick_it( e_in_link );
  
  // Copy this link to another at Max_energy
  new_link.E_in() = Max_energy;
  insert(end(), new_link);
  MF12_deexcite::iterator new_link_ptr = end();
  --new_link_ptr;
  for(dd_list::iterator list_ptr = e_in_link->begin();
    list_ptr != e_in_link->end(); ++list_ptr)
  {
    dd_link XYdata(list_ptr->x, list_ptr->y);
    new_link_ptr->insert(new_link_ptr->end(), XYdata);
  }

  ENDL.set_yo(7);
  renorm();
  write_endl( 4 );
}

// ----------- MF12_deexcite::read_data -----------------
//! This routine is used to read in transition probabilities
void MF12_deexcite::read_data( double AWR, int LG, mf12_file& inFile )
{
  //  double ES;  // excitation level of the original nucleus
  double zero;
  int LP;
  int NT;  // the number of transitions in the file
  int dummy;
  double ES_i; // excitation level after gamma emission
  double prev_ESi;  // previous level
  dd_link XYdata;
  dd_link EPdata;
  string linebuff;
  string strbuff;
  double Prob, GammaConditionalProb;
  double cum_prob = 0.0;  // is the total probability 1?
  bool found_level;

  if (LG==2)
    Warning("MF12_deexcite::read_data","Ignoring internal conversion.  All energy from a transition goes to gamma production.");

  inFile.gamma_line1(
    &ES,     //  level energy
    &zero,
    &LP,     // do a Doppler shift?
    &dummy,  // zero
    &dummy,  // count of excitation levels
    &NT      // number of (ES_i, probability) pairs
  );
  
  if( LP == 2 )
  {
    if( ( ENDL.T > 50 ) && ( ENDL.T < 91 ) )
    {
      Warning("MF12_deexcite::read_data",
              "Typo: shifted gamma cascade makes no sense");
    }
    else
    {
      SevereError("MF12_deexcite::read_data",
                  "Implement shift for gamma cascade");
    }
  }
  ES *= ENDL.eV2MeV;  // Convert level energy to MeV
  cascade new_cascade;
  cascade_gammas.insert(make_pair(ENDL.T,new_cascade));
  cascade_gammas[ENDL.T].Level_Energy = ES;

  for(int count = 0; count < NT; ++count)
  {
    int iP;
    if(LG == 1)
    {
      iP = count % 3;  //There are 3 pairs per line
      if ( iP == 0 )
      {
        getline( inFile, linebuff); //Read in new line
      }
      strbuff = linebuff.substr(iP*22, 22);
      read_dd(&strbuff, &ES_i, &Prob );
      GammaConditionalProb = 1.0;
      if ( Prob > 1.0 ) 
        Warning( "MF12_deexcite::read_data", pastenum( "Prob = ", Prob ) + "> 1.0" );
    }
    else if(LG == 2)
    {
      iP = count % 2;  //There are 2 triples per line
      if ( iP == 0 )
      {
        getline( inFile, linebuff); //Read in new line
      }
      strbuff = linebuff.substr(iP*33, 33);
      read_ddd(&strbuff, &ES_i, &Prob, &GammaConditionalProb );
      // Note: we're throwing away the conditional gamma emission probability 
      // (GammaConditionalProb) and assuming a transition goes completely to the gamma
      // channel.  Really if GammaConditionalProb < 1, we should send some strength to 
      // internal conversion electrons.
      if ( Prob > 1.0 ) 
        Warning( "MF12_deexcite::read_data", pastenum( "Prob = ", Prob ) + "> 1.0" );
      if ( GammaConditionalProb > 1.0 ) 
        Warning( "MF12_deexcite::read_data", pastenum( "GammaConditionalProb = ", GammaConditionalProb ) + "> 1.0" );
    }
    else
    {
      SevereError("MF12_deexcite::read_data",pastenum("Bad value of LG: ",LG));
    }
    if ( ES_i==-999.0 || Prob==-999.0 ) //read_PAIR sends -999.0 if problem
    {
      SevereError("MF12_deexcite::read_data","read_PAIR messed up!");
    }
    ES_i *= ENDL.eV2MeV;
    EPdata.x = ES - ES_i;
    EPdata.y = Prob;
    cum_prob += Prob;

    // check for levels out of order
    if( ( count > 0 ) && ( ES_i >= prev_ESi ) )
    {
      SevereError( "MF12_deexcite::read_data",
		   pastenum( "levels out of order: ", ES_i ) );
    }
    prev_ESi = ES_i;
    
    //Insert the data into the appropriate list
    cascade_gammas[ENDL.T].insert(cascade_gammas[ENDL.T].end(),EPdata);
    
    double levelMatchTol = 0.00001; // 10 eV
    //Need to propagate the cascade if it didn't decay to ground state
    if ( ES_i > 0.0 )  
    { 
     found_level = false;  //flag to insure all level searches are successful
	  for( int level = ENDL.T; level >= ENDL.T-40; --level ) //Loop down thru MT's
	  { 
	    if( abs(cascade_gammas[level].Level_Energy - ES_i ) < levelMatchTol ) //Match is found
	    {
          found_level = true;
          cascade these = cascade_gammas[level];  // local copy
          these *= Prob;                          // scale by initiating probability
          cascade_gammas[ENDL.T].line_sum( these , .00001 ); // add them to list
          break;
	    }
	  }
	  if( !found_level ) //Better let somebody know 
	  {
	    SevereError("MF12_deexcite::read_data",
          pastenum( "Could not find matching level ", ES_i) +
          " in cascade data!");
	  }
    }

  }


  // Check total probability.
  if( abs( cum_prob - 1.0 ) > ENDL_EPSILON( 1.0 ) )
  {
    Info( "MF12_deexcite::read_data",
           pastenum( "total probability: ", cum_prob )+", so rescaling" );
    for ( int level = ENDL.T; level >= ENDL.T-40; --level ) 
    {
      cascade_gammas[ENDL.T] *= 1.0/cum_prob;
    }
  }

  //Each time we complete a level, we write out to an i=9 file
  cascade_gammas[ENDL.T].write_endl( );

}



// ----------- MF12_deexcite::stick_it -----------------
void MF12_deexcite::stick_it( MF12_deexcite::iterator e_in_link )
// Routine to stick the cascade gammas into the deexcite object
{
  dd_link XYdata;

//Loop through the EP pairs and stick them into the deexcite object...
  for( cascade::iterator link = cascade_gammas[ENDL.T].begin(); 
     link != cascade_gammas[ENDL.T].end();
     ++link)
    {
      XYdata.x = link->x;
      XYdata.y = link->y;
  
      one_d_table::iterator next_xy;
      if(e_in_link->empty())
	{
	  next_xy = e_in_link->end();
	}
      else
	{
	  // find the first bigger gamma energy
	  for(next_xy = e_in_link->begin();
	      next_xy != e_in_link->end();
	      ++next_xy)
	    {
	      if (next_xy->x >= XYdata.x)
		{
		  break;
		}
	    }
	}
      // is there positive gamma emission?
      if( XYdata.x <= 0.0 )
      {
        Warning( "MF12_deexcite::stick_it", "line with no energy" );
      }
      else
      {
        //Insert the data into the table
        e_in_link->insert(next_xy, XYdata);
      }
    }    
  // now widen the delta functions and renormalize
  e_in_link->widen_delta();
  e_in_link->renorm();
}

// ----------- cascade::print -----------------
void cascade::print()
{
  cout<<"Energy of this level = "<<Level_Energy<<endl;
  for(dd_list::iterator link = begin(); link != end();
      ++link)
    {
      cout << ENDL.data( link->x, link->y ) << endl;
    }
  cout<< "........................." <<endl;
}
      
// ----------- cascade::write_endl -----------------
void cascade::write_endl()
{
  ENDL.set_I_number(9);
  fstream endl_file;
  string file_name = ENDL.file_name;
  
  if(empty())
  {
    Warning("cascade::write_endl","The file "+file_name+" is empty.");
    return;
  }
  
  if ( ENDL.new_file() )
  {
    endl_file.open(file_name.c_str(),ios::out);
    Info("cascade::write_endl","Opening ENDL file "+file_name);
  }
  else
  {
    endl_file.open(file_name.c_str(),ios::out|ios::app);
    Info("cascade::write_endl","Appending ENDL file "+file_name);
  }
      
  //Write out the header information
  endl_file<<ENDL.header_line_1<<endl;
  endl_file<<ENDL.header_line_2<<endl;
  endl_file.setf(ios::scientific,ios::floatfield);
  
  //Loop through the data and determine the average number of gammas
  double mean = 0.0;
  for(dd_list::iterator link = begin(); link != end(); ++link)
  {
    mean += link->y ;
  }
  
  endl_file<<ENDL.data( ENDL.threshold, mean )<<endl;
  endl_file<<ENDL.data( ENDL.Max_E_in, mean )<<endl;
  
  //Put on end of file/section line into output file
  endl_file<<ENDL.eof_line<<endl;
  
  //Finally, we close the output file
	  Info("cascade::write_endl","Closing ENDL file "+file_name);
  endl_file.close();
  
}
