/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1887 $
 * $Date: 2006-08-23 11:34:44 -0700 (Wed, 23 Aug 2006) $
 * $Author: dbrown $
 * $Id: mf12classes.hpp 1887 2006-08-23 18:34:44Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// header for the MF=12 translation code

#ifndef MF12CLASSES
#define MF12CLASSES

#include "list_1d.hpp"
#include "list_2d.hpp"
#include "ENDF_file.hpp"
#include "mf15classes.hpp"
#include "multiplicity.hpp"

using namespace std;

// ----------- class MF12_raw -----------------
//! This class just holds raw ENDF data for gamma multiplicity files 
// (LO = 1).  That is, the tag for each 1-d list is the gamma energy,
// and each 1-d list consists of pairs (neutron energy, multiplicity),
// The MF12_list class uses one of these lists for energy-independent
// gammas and one for energy-dependent gammas.
class MF12_raw : public two_d_list< multiplicity >
{
public:
  mf12_file *inFile;

  // object for the total multiplicity
  multiplicity multiple;

  //! default constructor
  MF12_raw(): inFile(NULL), multiple(){}

  //! Read the data for one gamma
  void one_line( double EG, int NR, int NP );

  //! Makes sure that if any gamma energy has a ( E_in, multiplicity ) pair for some E_in, then all of the lists contain this E_in.
  void fill_raw_lists( );

  //! Makes sure that the ( E_in, multiplicity ) pairs in the ENDF raw list have E_in values matching the E_in tags in cont_data.
  void fill_raw_lists( MF15_list& cont_data );

  //! Inserts neutron energies at which moving gammas coincide with stationary ones
  void insert_cross( list< double >& crossings );

  //! ENDF/B-VI sometimes has duplicate gamma lines from different cascades.  We combine them.
  void duplicate_E( );
};

// ----------- class MF12_base -----------------
//! This is the base class for the MF12_list class
class MF12_base : public two_d_table
{
public:
  mf12_file *inFile;
  double AWR; // the atomic weight

  //  multiplicity multiple;
  multiplicity multiple;

  //! default constructor
  MF12_base(): inFile(NULL), AWR(0.0), multiple(){}

  //! Handles the total gamma multiplicity
  void read_mult( multiplicity& multiple );

  //! Find the 1-d list with incident energy E_in
  MF12_base::iterator find_link( double E_in, double EG );

  //! Shift the gamma energies according to the atomic weight
  void shift_gammas( );

  //! Set the weights for the discrete spectra
  void set_weights( multiplicity& multiple );

  // set the weights of the lines and save their multiplicities
  void weight_mult( multiplicity& multiple, multiplicity& move_mult );

  //! initialize the ENDL list from raw_list
  void init_list( MF12_raw& raw_list );

  //! Widen all of the delta-functions; if clean_up is true, delete 1-d lists with zero norm
  void widen_deltas( bool clean_up );

  //! If two adjacent lines have the same frequency, this routine adds their intensities.
  void join_lines( );

};

// ----------- class MF12_list -----------------
//! This class is for gamma multiplicity files (LO = 1)
class MF12_list : public MF12_base
{
public:
  mf15_file contin_file;   // continuum energy distributions

  int mf12_NK;       // total number of mf12 gammas, including continuum
  bool with_mf15;    // is there an mf15 continuum data file?
  bool with_move;    // are there gammas dependent on the neutron energy?
  bool with_static;  // are there gammas independent of the neutron energy?

  // continuum represented as energy-dependent lines in mf12 format
  MF12_raw raw_move;

  // continuum represented as energy-dependent lines in ENDL format
  MF12_base movers;

  // the energy-independent mf12 gamma data
  MF12_raw raw_static;

  // the energy-independent gammas in ENDL format
  MF12_base statics;

  // the continuum data
  MF15_list cont_data;
  
  //! default constructor
  MF12_list(): contin_file(), mf12_NK(0), with_mf15(false), with_move(false), 
    with_static(false), raw_move(), movers(), raw_static(), statics(), cont_data(){}

  //! Handles an mf12 multiplicity file
  void master( double awr, int NK, mf12_file *InFile );

  //! This routine is used to read in probabilities of gammas
  void read_data( );

  //! Makes sure that the multiplicity is 0 if there is no distribution
  void clean_multiple( MF12_base& statics );

  //! Writes only the continuum data
  void write_continuum( );

  //! Writes discrete gamma multiplicities
  void write_statics( );

  //! Writes the one static line
  void write_static( );

  //! Writes the one moving line
  void write_mover( );

  //! Writes the sum of mf15 continuum plus static lines
  void plus_contin( );

  //! Joins the moving lines with static ones
  void join_move_static( );

  //! Writes the sum of mf15 continuum plus moving lines
  void move_w_contin( );

  //! Writes the sum of static plus moving lines
  void do_movers( );

  //! Write the static lines
  void just_statics( );

  //! Finds the neutron energies at which an energy-dependent gamma coincides with a stationary gamma
  void find_crossings( );

  //! Reorders the mf12_data as ENDL_data
  void ENDL_order( MF12_raw& mf12_data, MF12_base& ENDL_data );

  //! Sets the weights for the continuum
  void set_mf15_wt( );

  //! Handles multiplicity and energy distributions for the continuum
  void continuum( int NR, int NP );

};


// ----------- class MF12_deexcite -----------------
//! This class is for nuclear transition probabilities (LO = 2)
// and its pricipal use is for gammas from (n, n' gamma) reactions.
class MF12_deexcite : public two_d_table
{
public:
  // Holds the energy level for this level
  double ES;

  list< dd_list > EP_pairs;

  //!default constructor
  MF12_deexcite(): ES(0.0), EP_pairs(0){}

  void print()
  {
    cout<<"Energy of this level is "<<ES<<endl;
    for(two_d_iterator list_ptr = begin();
      list_ptr != end(); ++list_ptr)
    {
      list_ptr->print();
    }
    cout << endl;
  }


  // Handles an mf12 transition probability file
  void master( double AWR, int LG, mf12_file& inFile );

  // This routine is used to read in transition probabilities
  void read_data( double AWR, int LG,
    mf12_file& inFile );

  //Sticks the cascade data into the MF12-deexcite object
  void stick_it( MF12_deexcite::iterator e_in_link );

};

// ----------- class cascade -----------------
class cascade: public dd_list
{
public:
  double Level_Energy; // Energy of a level above ground state

  //!default constructor
  cascade(): Level_Energy(0.0){}

  // Print utility
  void print();

  // Writes the multiplicity file
  void write_endl();
  
};

#endif
