/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1881 $
 * $Date: 2006-05-24 13:27:07 -0700 (Wed, 24 May 2006) $
 * $Author: dbrown $
 * $Id: mf13classes.cpp 1881 2006-05-24 20:27:07Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the classes for translating the MF=13 data
#include "mf13classes.hpp"
#include "endl_formats.hpp"
#include "mf4classes.hpp"
#include "mf15classes.hpp"
#include "global_params.hpp"
#include "mf14classes.hpp"

extern ENDLClass ENDL;
extern GlobalParameterClass Global;

// ********* for class MF13_list *************
// ----------- MF13_list::master -----------------
void MF13_list::master( int NK, mf13_file& inFile )
// handle an mf13 file
{
  int NR,NP;

  if(NK > 1)
  {
    // the first block is the total production of discrete gammas.
    // Skip over it
    inFile.read_line2(&NR, // number of interpolation regions
   	  &NP);            // number of data points

    multiplicity total;
    // read the interpolation information
    inFile.get_regions(NR, total.NBT, total.INT);

    //The first data is the total multiplicity
    total.read_data( NP, inFile);
  }
  // get the energy distributions
  read_data( NK, inFile );
}
// ----------- MF13_list::read_data -----------------
void MF13_list::read_data( int NK, mf13_file& inFile )
// This routine is used to read in ENDF/B-VI probabilities
// for discrete gamma production.
// NK is the number of data blocks (discrete gammas + continuum)
{
  double EG;  // gamma energy
  double ES;  // nucleus excitation level
  int LP;   // 2 for primary photons; 1 for nonprimaries;
            // 0 for we don't know.  We do not use this number
  int LF;   // 2 for discrete; 1 for continuum (MF15)
  int NR;   // number of words in the record (3*NA)
  int NP;   // number of incident energies

  with_discrete = false;  // the default is no discrete gammas

  for(int count = 0; count < NK; ++count)
  {
    // go through all of the incident energies for each gamma
    inFile.gamma_line1(&EG,     //  gamma energy
              &ES,     //  excitation level
	      &LP,     // is this a primary gamma?
	      &LF,     // discrete-continuum flag
	      &NR,     // number of interpolation regions
	      &NP);    // number of (E, gamma production) pairs

    if( LF == 1 )
    {
      continuum(NR, NP, inFile);
      break; // we should be done
    }
    // change to MeV
    EG *= ENDL.eV2MeV;

    // start a new two_d_link for this incident energy
    multiplicity new_link;
    new_link.E_gamma( ) = EG;
    // put it at the beginning---the gamma energies are decreasing
    insert( begin( ), new_link );

    // read the data for this gamma
    two_d_iterator e_gam_link = begin( );
    // read in the interpolation regions
    inFile.get_regions( NR, e_gam_link->NBT, e_gam_link->INT );
    e_gam_link->read_data( NP, inFile);
  }
  if( (NK > 1) || (LF == 2))
  {
    with_discrete = true;
    if( ENDL.F == 12 )
    {
      for(two_d_iterator link = begin(); link != end(); 
      ++link)
      {
        link->scale_by_xs( );
      }
    }

    // test for the presence of an mf14 file
    mf14_file angle_file;
    angle_file.open( 14, ENDL.T );
    if( angle_file )
    {
      int ZA;
      double AWR;
      int LI, LTT;
      int NK,NI;
      MF14_c55_list angles;
      angle_file.first_line(&ZA,   // target Z & A 
	      &AWR,            // target mass
	      &LI,             // isotropy flag
	      &LTT,            // Legendre flag
	      &NK,             // number of discreet photons (inc continuum)
	      &NI);            // number of isotropic photon dists

      if ( LI == 1 )  //all photons isotropic
      {
        isotropic();   // make isotropic distributions
      }
      else if ( LI == 0 ) // anisotropic distribution
      {
	// handle the angular distributions
        angles.master( angle_file, LTT, NK, NI );
      }
      else
      {
        SevereError("MF13_list::read_data",
            pastenum("Horrible problem! LI = ",LI)+" not known MF=14!");
      }
    }
    else
    {
      isotropic();   // make isotropic distributions
    }
  }
}
// ----------- MF13_list::write_endl -----------------
void MF13_list::write_endl()
// write a c = 55, i = 0, s = 3 ENDL file
{
  ENDL.set_c_number(55);
  ENDL.set_s_number(3);
  ENDL.set_yo(0);
  ENDL.set_I_number(0);

  fstream endl_file;
  string file_name = ENDL.file_name;

  if ( ENDL.new_file() )
  {
    endl_file.open(file_name.c_str(),ios::out);
    Info("MF13_list::write_endl","Opening ENDL file "+file_name);
  }
  else
  {
    endl_file.open(file_name.c_str(),ios::out|ios::app);
    Info("MF13_list::write_endl","Appending ENDL file "+file_name);
  }

  //Now loop over the data and write it out
  for(two_d_iterator link = begin(); link != end(); 
      ++link)
  {
    if( ENDL.F == 12 )
    {
      link->scale_by_xs( );
    }
    ENDL.set_x1(link->E_gamma());
    // rebuild the header
    ENDL.set_I_number(0);
    // Print the header info
    endl_file << ENDL.header_line_1 << endl;
    endl_file << ENDL.header_line_2 << endl;
    endl_file.setf(ios::scientific,ios::floatfield);

    for(dd_list::iterator prob_link = link->begin();
        prob_link != link->end(); ++prob_link)
    {
      endl_file << ENDL.data(prob_link->x, prob_link->y) << endl;
    } 
    //Put on end of file/section line into output file
    endl_file << ENDL.eof_line << endl;
  }
  
  // Finally, we close the output file
    Info("MF13_list::write_endl","Closing ENDL file "+file_name);
  endl_file.close();
}

// ----------- MF13_list::isotropic -----------------
void MF13_list::isotropic()
// write a c = 55, i = 1, s = 3 ENDL file
{
  ENDL.set_c_number(55);
  ENDL.set_s_number(3);
  ENDL.set_yo(7);

  ENDL.set_I_number(1);

  fstream endl_file;
  string file_name = ENDL.file_name;
  if ( ENDL.new_file() )
  {
    endl_file.open(file_name.c_str(),ios::out);
    Info("MF13_list::isotropic","Opening ENDL file "+file_name);
  }
  else
  {
    endl_file.open(file_name.c_str(),ios::out|ios::app);
    Info("MF13_list::isotropic","Appending ENDL file "+file_name);
  }

  // make an isotropic distribution
  two_d_isotropic angles;
  angles.expand_data(0);
  two_d_isotropic::iterator first_E_in = angles.begin();

  //Now loop over the gamma production data and get the thresholds
  for(two_d_iterator link = begin(); link != end(); 
      ++link)
  {
    // reset the header lines
    ENDL.set_x1(link->E_gamma());
    ENDL.set_I_number(1);

    first_E_in->E_in() = link->begin()->x;

    //Shove the header info in
    endl_file << ENDL.header_line_1 << endl;
    endl_file << ENDL.header_line_2 << endl;
    endl_file.setf(ios::scientific,ios::floatfield);

    // print the distribution
    angles.out_data( 1, endl_file, 0.0 );

    //Put on end of file/section line into output file
    endl_file << ENDL.eof_line << endl;
  }
  //Finally, we close the output file
    Info("MF13_list::isotropic","Closing ENDL file "+file_name);
  endl_file.close();
}
// ----------- MF13_list::continuum -----------------
void MF13_list::continuum( int NR, int NP, mf13_file& inFile )
// handle gamma production and energy distributions for the continuum
{
  // read the interpolation regions
  inFile.get_regions(NR, cont_multiple.NBT, cont_multiple.INT);
  //The first data is the total multiplicity
  cont_multiple.read_data( NP, inFile);

  mf15_file mf15File;   //Instantiate the input file
  mf15File.open(15, ENDL.T);

  cont_data.read_data( mf15File );
  Info("MF13_list::continuum","Closing file "+endf_file_name(15,ENDL.T)+"...");
  mf15File.close();
}
// ----------- MF13_list::write_continuum -----------------
void MF13_list::write_continuum( )
// Writes C55 data for the continuum
{
  ENDL.set_c_number( 55 );
  ENDL.set_s_number( 0 );  // for continuum
  ENDL.set_yo( 0 );
  ENDL.set_x0( 0.0 );  // ENDL convention for Q of c55 gammas

  cont_multiple.widen_jumps( );
  cont_multiple.write_endl( 0 );
}

// ***************** class C55_gammas *******************
// --------------- C55_gammas::write_endl -------------
void C55_gammas::write_endl( )
// Writes the c55 files
{
  // Reset the header information, overwrite the previous reaction
  ENDL.write_file = true;
  ENDL.set_x0( 0.0 );  // the Q value
  ENDL.T = 3;  // This is mt=3 plus mt=4 data
  ENDL.set_c_number(55);

  //The addition operators for different lists do not know how to
  //handle empty lists.  So, we insure non-empty lists before we add

  // --- for the continuum energy distributions ---
  ENDL.set_s_number(0);  // for gamma continuum
  ENDL.set_yo(7);

  if( !MT3_gammas.cont_data.empty( ) && !MT4_gammas.cont_data.empty( ) )
  {
    Info("C55_gammas::write_endl","we have distributions from both MT3 & MT4");
    // record the types of interpolation between 1-d lists
    MT3_gammas.cont_data.set_interp( );
    MT4_gammas.cont_data.set_interp( );
    set_weights( );  // weight according to photon production

    // for sum_lists, make the incident neutron energies agree
    list< double > E_incident;
    MT3_gammas.cont_data.collect_Ein( E_incident );
    MT4_gammas.cont_data.collect_Ein( E_incident );
    E_incident.sort( );
    E_incident.unique( );
    MT3_gammas.cont_data.fill_in_list( E_incident );
    MT4_gammas.cont_data.fill_in_list( E_incident );
    MT3_gammas.cont_data.use_weight( );  // scale by the weight

    sum_lists( MT3_gammas.cont_data, MT4_gammas.cont_data );

    MT3_gammas.cont_data.widen_jumps( );
    MT3_gammas.cont_data.renorm( );
    MT3_gammas.cont_data.write_endl( 4 );
  }
  else if ( !MT3_gammas.cont_data.empty( ) && MT4_gammas.cont_data.empty( ) )
  {
    Info("C55_gammas::write_endl","We have continuum from MT3");
    MT3_gammas.cont_data.widen_jumps( );
    MT3_gammas.cont_data.write_endl( 4 );
  }
  else if ( MT3_gammas.cont_data.empty( ) && !MT4_gammas.cont_data.empty( ) )
  {
    Info("C55_gammas::write_endl","We have continuum from MT4");
    MT4_gammas.cont_data.widen_jumps( );
    MT4_gammas.cont_data.write_endl( 4 );
  }
  else
  {
    Info("C55_gammas::write_endl","No continuum energy distributions.");
  }

  // --- for the continuum photon production cross sections ---
  ENDL.set_yo( 0 );
  ENDL.set_I_number( 0 );  // recompose the header
  if( !MT3_gammas.cont_multiple.empty( ) && !MT4_gammas.cont_multiple.empty( ) )
  {
    Info("C55_gammas::write_endl","We have continuum from both MT3 & MT4");
    MT3_gammas.cont_multiple += MT4_gammas.cont_multiple;
    MT3_gammas.write_continuum( );
  }
  else if ( !MT3_gammas.cont_multiple.empty( ) && MT4_gammas.cont_multiple.empty( ) )
  {
    Info("C55_gammas::write_endl","We have continuum from MT3");
    MT3_gammas.write_continuum( );
  }
  else if ( MT3_gammas.cont_multiple.empty( ) && !MT4_gammas.cont_multiple.empty( ) )
  {
    Info("C55_gammas::write_endl","We have continuum from MT4");
    MT4_gammas.write_continuum( );
  }
  else
  {
    Info("C55_gammas::write_endl","Has no continuum.");
  }


  // ---for the discrete lines---
  if( MT3_gammas.with_discrete && MT4_gammas.with_discrete )
  {
    Info("C55_gammas::write_endl","we have discrete from both MT3 & MT4");
    list< double > E_incident;
    MT3_gammas.collect_Ein( E_incident );
    MT4_gammas.collect_Ein( E_incident );
    E_incident.sort( );
    E_incident.unique( );
    MT3_gammas.fill_in_list( E_incident );
    MT4_gammas.fill_in_list( E_incident );

    MT3_gammas.use_weight( ); // scale by the weight
    sum_lists( MT3_gammas, MT4_gammas );  // weight MT4 and add
    MT3_gammas.write_endl( );
  }
  else if( MT3_gammas.with_discrete && !MT4_gammas.with_discrete )
  {
    Info("C55_gammas::write_endl","we have discrete from MT3");
    MT3_gammas.write_endl( );
  }
  else if(  !MT3_gammas.with_discrete && MT4_gammas.with_discrete )
  {
    Info("C55_gammas::write_endl","we have discrete from MT4");
    MT4_gammas.write_endl( );
  }
  else
  {
    Info("C55_gammas::write_endl","no discrete lines");
  }
}
// --------------- C55_gammas::set_weights -------------
void C55_gammas::set_weights( )
// Sets the weights of the energy distributions according to
// the photon production cross section.
{
  MF15_list::iterator cont_ptr;
  double e_in;

  // set the weights for the MT3 data
  for( cont_ptr = MT3_gammas.cont_data.begin( );
       cont_ptr != MT3_gammas.cont_data.end( ); ++cont_ptr )
  {
    e_in = cont_ptr->E_in( );
    cont_ptr->weight = MT3_gammas.cont_multiple.evaluate( e_in );
  }

  // set the weights for the MT4 data
  for( cont_ptr = MT4_gammas.cont_data.begin( );
       cont_ptr != MT4_gammas.cont_data.end( ); ++cont_ptr )
  {
    e_in = cont_ptr->E_in( );
    cont_ptr->weight = MT4_gammas.cont_multiple.evaluate( e_in );
  }
}
