/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1895 $
 * $Date: 2006-09-07 13:47:49 -0700 (Thu, 07 Sep 2006) $
 * $Author: dbrown $
 * $Id: mf14classes.cpp 1895 2006-09-07 20:47:49Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the classes for translating the MF=14 data

#include "endl_formats.hpp"
#include "mf4classes.hpp"
#include "mf14classes.hpp"
#include "global_params.hpp"
#include "messaging.hpp"
#include "endl_precision.hpp"

extern ENDLClass ENDL;
extern GlobalParameterClass Global;

using namespace std;

// ********* for class MF14_list *************
// ----------- MF14_list constructor ----------------
MF14_list::MF14_list( int NI, int NK, int order, mf12_file *InFile,
    mf14_file *Angle_File )
{
  inFile = InFile;  // the mf12 file
  angle_file = Angle_File; // the mf14 file

  mf14_NI = NI; // number of isotropic distributions
  mf14_NK = NK; // total number of gamma distributions
  num_aniso = NK - NI;  // number of anisotropic gammas
  aniso_count = 0;  // numer of anisotropic gammas read

  // set the Legendre order
  order_ = order;
  coefs_.reserve( order + 1 );
  coefs_.resize( order + 1 );
  coefs_[0].inFile = inFile;
}
// ----------- MF14_list destructor ----------------
MF14_list::~MF14_list()
{
}
// ----------- MF14_list::master -----------------
void MF14_list::master( int mf12_nk )
// Read the data from the mf12 multiplicity file and the mf14
// anisotropy file and print the results if ENDL i=4 format.
// mf12_NK is the number of lines in the mf12 file---it may
// be 1 more than mf14_NK (for the continuum).
{
  int ZA;
  int mf14_LI, mf14_LTT;

  mf12_NK = mf12_nk;

  // reread the first line or the mf14 file
  angle_file->first_line(
       &ZA,             // target Z & A 
       &AWR,            // target mass
	    &mf14_LI,        // isotropy flag
	    &mf14_LTT,       // Legendre flag
	    &mf14_NK,        // number of discreet photons (inc continuum)
	    &mf14_NI         // number of isotropic photon dists
  );       
  // skip the isotropic lines
  angle_file->skip_iso( mf14_NI );

  // is there an mf15 file?
  contin_file.open( 15, ENDL.T );
  with_mf15 = ( contin_file )? true : false;

  if(mf14_NK > 1)
  {
    // The first block is the total multiplicity of discrete gammas.
    // Use the MF12_base read_mult routine.
    coefs_[0].read_mult( multiple );

    // set up the 2-d lists for the higher Legendre coefficients
    for( int L_count = 0; L_count <= order_; ++L_count )
    {
      dd_list::iterator mult_ptr = multiple.begin();
      for( int count = 0; count < multiple.size();
        ++count, ++mult_ptr )
      {
        one_d_table new_link;
        new_link.E_in() = mult_ptr->E_in();
        coefs_[ L_count ].insert( coefs_[ L_count ].end(), new_link );
      }
    }

    // get the discrete lines and continuum
    read_data( );

    // If there is continuum data, add it to the discrete lines
    if( with_mf15 )
    {
      add_lists( );
    }
    else
    {
      // just widen the delta functions
      widen_delta();
    }
    
    // strip any incident energy points that have 0 multiplicity from 
    // both the mulitplicity and outgoing angular distributions
    strip_zero_multiplicity( );

    // write the i = 4 file
    ENDL.set_yo(7);
    renorm();
    write_endl( );

    // print the multiplicities
    ENDL.set_yo(7);
    if ( ENDL.C == 55 )
    {
      if ( ENDL.T == 12 )
      {
        // read the cross section and multiply to get photon production
        multiple.scale_by_xs();
      }
      ENDL.set_yo(0);
      multiple.write_endl(0);
    }
    else
    {
      if ( ENDL.T == 13 )
      {
        // read the cross section and divide to get the gamma multiplicity
        multiple.divide_by_xs();
      }
      multiple.write_endl(9);
    }
  }
}
// ----------- MF14_list::read_data -----------------
void MF14_list::read_data( )
// This routine is used to read in ENDF/B-VI probabilities
// of discrete and continuum gammas and their Legendre expansions,
{
  // read the first anisotropic gamma energy
  angle_file->read_EG( &mf14_EG, &mf14_NR, &mf14_NE );
  mf14_EG *= ENDL.eV2MeV;  // in Mev
  if( mf14_EG == 0.0 )
  {
    SevereError("MF14_list::read_data",
        "Implement Legendre expansion of the continuum");
  }
  ++aniso_count; // one anisotropic gamma energy read

  double EG;  // gamma energy
  double ES;  // nucleus excitation level
  int LP;   // 2 Doppler shift the photon energy
            // 0, 1: do not Doppler shift
  int LF;   // 2 for discrete; 1 for continuum (in MF15 file)
  int NR;   // number of interpolation regions
  int NP;   // number of incident energies
  bool do_shift;  // true if we Doppler shift the gamma energy
  double prev_E_in;

  // read the mf12 file
  for(int count = 0; count < mf12_NK; ++count)
  {
    // initialize the test for duplicate incident energies
    prev_E_in = -1.0;

    inFile->gamma_line1(
        &EG,     //  gamma energy
        &ES,     //  excitation level
	     &LP,     // do a Doppler shift?
	     &LF,     // discrete-continuum flag
	     &NR,     // number of interpolation regions
	     &NP      // number of (E, multiplicity) pairs
    );

    // change to MeV
    EG *= ENDL.eV2MeV;
    if(LP == 2)
    {
      // do a Doppler shift
      do_shift = true;
    }
    else
    {
      do_shift = false;
    }

    if(LF == 1)
    {
      // this data is for the continuum, read the mf15 file
      continuum( NR, NP );
    }
    else
    {
      one_line( EG, &prev_E_in, NR, NP, do_shift );
    }
  }

  // we want probability density, not multiplicity
  if ( with_mf15 )
  {
    for( int L_count = 0; L_count <= order_; ++L_count )
    {
      for( MF12_base::iterator e_in_link = coefs_[ L_count ].begin( );
  	   e_in_link != coefs_[ L_count ].end( ); ++e_in_link )
      {
        double E_in = e_in_link->E_in( );
        double scale_by = multiple.evaluate( E_in );
        (*e_in_link) *= 1.0/scale_by;
      }
    }
  }
}
// ----------- MF14_list::one_line -----------------
void MF14_list::one_line( double EG, double *prev_E_in, int NR,
  int NP, bool do_shift )
// read in data for one discrete line
{
  MF12_base::iterator e_in_link;
  dd_link XYdata;
  dd_link prev_XY;  // save for interpolation
  one_d_table new_link;
  string linebuff;
  string strbuff;
  double E_in;
  double use_E_in;
  double how_many;
  int L_count;  // count the Legendre order
  const double EPS = Global.Value( "mf5_tol" );  // tolerance for mf5 data
  double *coef_buff; // buffer for each E_in and its Legendre coefficients
  double **Coef_ptrs; // where in coef_buff for each E_in
  bool alloc_coef = false;  // no buffer allocated yet
  bool done = false;  // E_in is still below ENDL.Max_E_in
  double dE = ENDL_EPSILON( ENDL.Max_E_in );

  // read the interpolation regions
  inFile->get_regions(NR, coefs_[0].NBT, coefs_[0].INT);
  if( (NR > 1) || ( (coefs_[0].INT[0] != 1) && (coefs_[0].INT[0] != 2) ) )
  {
    Unimplemented("MF14_list::one_line","Implement interpolation");
  }

  // is this one of the anisotropic lines?
  if( abs( EG - mf14_EG ) <= EPS * EG )
  {
    // allocate space for the Legendre coefficients
    coef_buff = new double[ mf14_NE * ( order_ + 1 ) ];
    Coef_ptrs = new double* [ mf14_NE ];
    alloc_coef = true;
    read_Legendre( coef_buff, Coef_ptrs );
  }

  // read the data for this line
  for( int iNP=0; iNP < NP; iNP++ ) //Loop over NP data pairs
  { 
    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 )
    {
      getline( *inFile, linebuff ); //Read in new line
    }
    strbuff = linebuff.substr( iP*22, 22 );
    read_dd( &strbuff, &E_in, &how_many );
    if ( E_in==-999.0 || how_many==-999.0 ) //read_dd sends -999.0 if problem
    {
      SevereError( "MF14_list::one_line", "read_dd failed!" );
    }
    E_in *= ENDL.eV2MeV;

    // did we exceed the maximum E_in?
    if( done )
    {
      e_in_link = coefs_[ 0 ].end();
    }
    else if( E_in <= ENDL.Max_E_in )
    {
      use_E_in = E_in;
      if(mf12_NK > 1)
      {
        // Find the 1-d list with this incident energy
        e_in_link = coefs_[ 0 ].find_link( E_in, EG );
      }
      else
      {
        // make a new list
        new_link.E_in() = E_in;
        coefs_[ 0 ].insert(coefs_[ 0 ].end(), new_link);
        e_in_link = coefs_[ 0 ].end();
        --e_in_link;
      }
    }
    else
    {
      // E_in > ENDL.Max_E_in and we need to interpolate
      use_E_in = ENDL.Max_E_in;
      if(mf12_NK > 1)
      {
        // use ENDL.Max_E_in as the incident energy
        e_in_link = coefs_[ 0 ].find_link( use_E_in, EG );
      }
    }

    //  do we store this data?
    if( e_in_link != coefs_[ 0 ].end() )
    {
      if( do_shift )
      {
        XYdata.x = EG + AWR*E_in/(AWR + 1);
      }
      else
      {
        XYdata.x = EG;
      }
      XYdata.y = how_many;
      // we may have to interpolate
      if( E_in > ENDL.Max_E_in ) 
      {
        double E_diff = E_in - prev_XY.x;
        if( ( E_diff <= 0.0 ) || ( iNP == 0 ) )
        {
          SevereError( "MF14_list::one_line", "energies out of order" );
        }
        XYdata.y = prev_XY.y +
          ( ENDL.Max_E_in - prev_XY.x ) *
          ( XYdata.y - prev_XY.y ) / dE;
      }

      // save it if this is less than the maximum gamma energy
      if( XYdata.x < ENDL.Max_E_out + dE )
      {
        insert_pair( XYdata, e_in_link );
        // is this one of the anisotropic lines?
        if( alloc_coef )
        {
          insert_Legendre( use_E_in, XYdata, *prev_E_in, Coef_ptrs );
        }
        else
        {
          XYdata.y = 0.0;
          for( L_count = 1; L_count <= order_; ++L_count )
          {
            if(mf14_NK > 1)
            {
              // Find the 1-d list with this incident energy
              e_in_link = coefs_[ L_count ].find_link( use_E_in, EG );
            }
            else
            {
              // make a new list
              new_link.E_in() = use_E_in;
              coefs_[ L_count ].insert(coefs_[ L_count ].end(), new_link);
              e_in_link = coefs_[ L_count ].end();
              --e_in_link;
            }
            insert_pair( XYdata, e_in_link );
          }
        }
      }
    }
    // Save for the next data set
    *prev_E_in = use_E_in;
    prev_XY.x = XYdata.x;
    prev_XY.y = XYdata.y;

    // is this the last data to store?
    if( E_in > ENDL.Max_E_in - dE )
    {
      done = true;
    }
  }
  
  // is there more mf14 data?
  if( aniso_count < num_aniso )
  {
    // read the next anisotropic gamma energy
    angle_file->read_EG( &mf14_EG, &mf14_NR, &mf14_NE );
    mf14_EG *= ENDL.eV2MeV;  // in Mev
    if( mf14_EG == 0.0 )
    {
      Unimplemented("MF14_list::read_Legendre",
        "Implement Legendre expansion of the continuum");
    }
    ++aniso_count; // one more anisotropic gamma energy read
  }
  else
  {
    // make sure that mf14_EG is nonphysical
    mf14_EG = -1.0;
  }


  // delete storage if necessary
  if( alloc_coef )
  {
    delete [] coef_buff;
    delete [] Coef_ptrs;
  }
}
// ----------- MF14_list::insert_pair -----------------
void MF14_list::insert_pair( dd_link& XYdata,
  MF12_base::iterator e_in_link )
// Insert an (EG, probability) pair into a list
{
  // The gamma energies are usually in descending order
  // but the Doppler shift may introduce inversions
  one_d_table::iterator next_xy;
  if( e_in_link->empty() )
  {
    next_xy = e_in_link->end();
  }
  else
  {
    // find the first bigger gamma energy
    for( next_xy = e_in_link->begin();
      next_xy != e_in_link->end();
        ++next_xy )
    {
      if ( next_xy->x >= XYdata.x )
      {
        break;
      }
    }
  }
  e_in_link->insert( next_xy, XYdata );
}
// ----------- MF14_list::read_Legendre -----------------
void MF14_list::read_Legendre( double *coef_buff, double **Coef_ptrs )
// Read Legendre data for one gamma from an mf14 file
{
  double e_in;
  double *this_ptr = coef_buff;
  string linebuff;
  string strbuff;
  int NL;

  // read the interpolation regions
  angle_file->get_regions( mf14_NR, coefs_[1].NBT, coefs_[1].INT );
  if( (mf14_NR > 1) || ( (coefs_[1].INT[0] != 1) && (coefs_[1].INT[0] != 2) ) )
  {
    SevereError("MF14_list::read_Legendre","Implement interpolation");
  }

  // loop over the mf14 incident energies
  for( int count = 0; count < mf14_NE; ++count )
  {
    // read the incident neutron energy and the number of coefficients
    angle_file->Leg_head( &e_in, &NL );
    e_in *= ENDL.eV2MeV;  // in Mev
    *this_ptr = e_in;     // save the incident energy
    Coef_ptrs[ count ] = this_ptr;  // where to find this incident energy
    ++this_ptr;

    // read the Legendre coefficients for this incident energy
    int iNL;
    for( iNL = 0; iNL < NL; ++iNL, ++this_ptr )
    { 
      int iL=iNL % 6;  //There are 6 entries per line
      if ( iL == 0 )
      {
        getline( *angle_file, linebuff); //Read in new line
      }
      strbuff = linebuff.substr( iL*11, 11 );
      read_d( &strbuff, this_ptr );
    }
    // fill with zeros
    for( ; iNL < order_; ++iNL, ++this_ptr )
    {
      *this_ptr = 0.0;
    }
  }
}
// ----------- MF14_list::insert_Legendre -----------------
void MF14_list::insert_Legendre( double e_in, dd_link& XYdata,
  double prev_E_in, double **Coef_ptrs )
// Insert the Legendre data for one gamma from an mf14 file into our lists
// Coef_ptrs[ j ] points to the j-th m14 incident energy, and this is
// followed by order_ Legendre coefficients
{
  MF12_base::iterator e_in_link;
  double scale = XYdata.y;  // The ENDF mf14 file assumes c_0 = 1
  int L_count;              // count the Legendre order
  dd_link new_data;         // for new 1-d links
  one_d_table new_link;     // for new 1-d lists
  double *prev_coef_ptr;
  double *next_coef_ptr;

  new_data.x = XYdata.x;  // this gamma energy

  // first find the mf14 data for this incident energy
  int j;
  for( j = 0; ( ( j < mf14_NE ) && ( *Coef_ptrs[ j ] < e_in ) ); ++j )
  {
  }
  if( j >= mf14_NE )
  {
    string msg( pastenum("Incident energy ",e_in)+" not found" + pastenum(" for gamma ", XYdata.x) );
    SevereError("MF14_list::insert_Legendre",msg);
  }

  // there is no interpolation at the first incident energy
  if( j == 0 )
  {
    next_coef_ptr = Coef_ptrs[ 0 ];
    ++next_coef_ptr;
    for( L_count = 1; L_count <= order_; ++L_count, ++next_coef_ptr )
    {
      if(mf14_NK > 1)
      {
        // Find the 1-d list with this incident energy
        e_in_link = coefs_[ L_count ].find_link( e_in, mf14_EG );
        if ( e_in_link == coefs_[ L_count ].end() ) 
            SevereError("MF14_list::insert_Legendre","Cannot find link");
      }
      else
      {
        // make a new list
        new_link.E_in() = e_in;
        coefs_[ L_count ].insert(coefs_[ L_count ].end(), new_link);
        e_in_link = coefs_[ L_count ].end();
        --e_in_link;
      }
      XYdata.y = scale * (*next_coef_ptr);
      insert_pair( XYdata, e_in_link );
    }
  }
  else
  {
    prev_coef_ptr = Coef_ptrs[ j - 1 ];  // previous incident energy
    next_coef_ptr = Coef_ptrs[ j ];      // next incident energy
    double alpha = ( e_in - *prev_coef_ptr ) / 
        ( *next_coef_ptr - *prev_coef_ptr );
    for( L_count = 1; L_count <= order_; ++L_count )
    {
      if(mf14_NK > 1)
      {
        // Find the 1-d list with this incident energy
        e_in_link = coefs_[ L_count ].find_link( e_in, mf14_EG );
        if ( e_in_link == coefs_[ L_count ].end() ) 
            SevereError("MF14_list::insert_Legendre","Cannot find link");
      }
      else
      {
        // make a new list
        new_link.E_in() = e_in;
        coefs_[ L_count ].insert(coefs_[ L_count ].end(), new_link);
        e_in_link = coefs_[ L_count ].end();
        --e_in_link;
      }
      // point to the next Legendre coefficients
      ++next_coef_ptr;
      ++prev_coef_ptr;
      XYdata.y = scale * ( alpha * (*next_coef_ptr) + 
        ( 1.0 - alpha) * (*prev_coef_ptr) );
      insert_pair( XYdata, e_in_link );
    }
  }
}
// ----------- MF14_list::widen_delta -----------------
void MF14_list::widen_delta( )
// widen the delta functions for the discrete spectrum
{
  MF12_base::iterator e_in_link[ order_ + 1 ];  // Legendre coefficients
  int L_count;        // to count the Legendre order

  // point to the first incident energy for each Legendre order
  for( L_count = 0; L_count <= order_; ++L_count )
  {
    e_in_link[ L_count ] = coefs_[ L_count ].begin();
  }

  // pointers to links in a (gamma energy, probability) list
  dd_list::iterator this_link[ order_ + 1 ];
  dd_list::iterator next_link[ order_ + 1 ];
  dd_list::iterator prev_link;

  // loop through the incident energies
  bool quit = false;
  for( ; !quit; )
  {
    // start at the lowest gamma energy for each Legendre order
    for( L_count = 0; L_count <= order_; ++L_count )
    {
      this_link[ L_count ] = e_in_link[ L_count ]->begin();
      next_link[ L_count ] = e_in_link[ L_count ]->begin();
      ++next_link[ L_count ];
    }
    // loop through the discrete lines
    for( ; ; )
    {
      prev_link = this_link[ 0 ];
      --prev_link;

      // what is the separation between states?
      double dE_L = (this_link[ 0 ] == e_in_link[ 0 ]->begin()) ?
        this_link[ 0 ]->E_out() :
        this_link[ 0 ]->E_out() - prev_link->E_out();
      double dE_R = (next_link[ 0 ] == e_in_link[ 0 ]->end()) ?
        this_link[ 0 ]->E_out() :
        next_link[ 0 ]->E_out() - this_link[ 0 ]->E_out();
      double dE = (dE_L < dE_R) ? dE_L : dE_R;

      // set width of delta function (basically 5* smallest dE we
      // can see in output file)
      dE = ENDL_JUMP_WIDTH( this_link[ 0 ]->E_out() );

      dd_link new_link = dd_link(this_link[ 0 ]->E_out() - dE, 0.0);
      for( L_count = 0; L_count <= order_; ++L_count )
      {
        // scale the line
        this_link[ L_count ]->y /= dE;

        // insert the base of the triangle at each Legendre order
        e_in_link[ L_count ]->insert( this_link[ L_count ], new_link );
      }
      new_link = dd_link(this_link[ 0 ]->E_out() + dE, 0.0);
      for( L_count = 0; L_count <= order_; ++L_count )
      {
        e_in_link[ L_count ]->insert( next_link[ L_count ], new_link );
        this_link[ L_count ] = next_link[ L_count ];
        ++next_link[ L_count ];
      }
      if( this_link[ 0 ] == e_in_link[ 0 ]->end() )
      {
        break;
      }
    }
    // go to the next incident energy for each Legendre order
    for( L_count = 0; L_count <= order_; ++L_count )
    {
      ++e_in_link[ L_count ];
      if( e_in_link[ L_count ] == coefs_[ L_count ].end() )
      {
        quit = true;
        break;
      }
    }
  }
}
// ----------- MF14_list::thinit -----------------
void MF14_list::thinit( )
// remove redundant data points
{
  // this routine is based on the list_1d version
  // declare pointers to the test links
  MF12_base::iterator e_in_list[ order_ + 1 ];
  dd_list::iterator left_link[ order_ + 1 ];
  dd_list::iterator mid_link[ order_ + 1 ];
  dd_list::iterator right_link[ order_ + 1 ];
  dd_list::iterator all_done;
  bool thin_ok;  // for test on 1 link
  bool do_thin;  // thin OK at this level
  bool prev_thin;  // thin OK at the previous level
  const double cut_off = Global.Value("cut_off_1d"); // don't impose accuracy below this value
  const double tol_1d = Global.Value("tol_1d");  // accuracy of linear interpolation
  int L_order;  // the Legendre order

  // initialize the pointers to the lists of Legendre coefficients
  for( L_order = 0; L_order <= order_; ++L_order )
  {
    e_in_list[ L_order ] = coefs_[ L_order ].begin( );
  }

  // loop over the incident energies
  for( ; ;  )
  {

    //  quit at the last valid link
    all_done = e_in_list[ 0 ]->end( );
    --all_done;

    // set the noise level
    double noise = 0.0;

    // initialize the pointers to the Legendre coefficients
    for( L_order = 0; L_order <= order_; ++L_order )
    {
      left_link[ L_order ] = e_in_list[ L_order ]->begin( );
    }
    // loop through the list of left-hand anchors
    for( ; left_link[ 0 ] != all_done; ++left_link[ 0 ])
    {
      // make sure that we have data to check
      for( L_order = 0; L_order <= order_; ++L_order )
      {
        mid_link[ L_order ] = left_link[ L_order ];
        ++mid_link[ L_order ];
        right_link[ L_order ] = mid_link[ L_order ];
        ++right_link[ L_order ];
        if( ( mid_link[ L_order ] == e_in_list[ L_order ]->end() ) ||
            ( right_link[ L_order ] == e_in_list[ L_order ]->end() ) )
        {
          break;  // no more data for this incident energy
        }
      }

      // the default is no thinning
      prev_thin = false;

      // how long a span may we delete?
      for( ; right_link[ 0 ] != e_in_list[ 0 ]->end(); ++right_link[ 0 ] )
      {
        // the default is don't delete links at this level
        do_thin = false;

        // check all of the intermediate links at all orders
        for( L_order = 0; L_order <= order_; ++L_order )
        {
          mid_link[ L_order ] = left_link[ L_order ];
          ++mid_link[ L_order ];
        }
        for( ; mid_link[ 0 ] != right_link[ 0 ]; ++mid_link[ 0 ] )
        {
          for( L_order = 0; L_order <= order_; ++L_order )
          {
            thin_ok = e_in_list[ L_order ]->check_interp(
              left_link[ L_order ], mid_link[ L_order ],
              right_link[ L_order ], tol_1d, noise );

            if(thin_ok)
            {
              // at least one deletion is possible
              do_thin = true;
            }
            else
            {
              // this section fails
              do_thin = false;
              break;
            }
          }
          if( !do_thin )
          {
            break;
          }
          for( L_order = 1; L_order <= order_; ++L_order )
          {
            ++mid_link[ L_order ];
          }
        }

        // did they all pass?
        if(do_thin)
        {
          prev_thin = true;
          // move the right links
          for( L_order = 1; L_order <= order_; ++L_order )
          {
            ++right_link[ L_order ];
          }
        }
        else
        {
          break;
        }
      }

      // do the deletions
      dd_list::iterator left_next;
      dd_list::iterator right_prev;

      if((prev_thin) && (!do_thin))
      {
        // deletion is not allowed here but it is OK at the previous level
        for( L_order = 0; L_order <= order_; ++L_order )
        {
          left_next = left_link[ L_order ];
          ++left_next;
          right_prev = right_link[ L_order ];
          --right_prev;
          e_in_list[ L_order ]->erase(left_next, right_prev);
        }
      }
      else if( (do_thin) && (right_link[ 0 ] == e_in_list[ 0 ]->end() ) &&
        ( mid_link[ 0 ] == all_done) )
      {
        // we are at the end and all links passed
        for( L_order = 0; L_order <= order_; ++L_order )
        {
          left_next = left_link[ L_order ];
          ++left_next;
          e_in_list[ L_order ]->erase( left_next, mid_link[ L_order ] );
        }
      }

      // increment the pointers to the higher Legendre coefficients
      for( L_order = 1; L_order <= order_; ++L_order )
      {
        ++left_link[ L_order ];
      }
    }
    // go to the list for the next incident energy
    for( L_order = 0; L_order <= order_; ++L_order )
    {
      ++e_in_list[ L_order ];
      if( e_in_list[ L_order ] == coefs_[ L_order ].end( ) )
      {
        return;  // there is no more data
      }
    }
  }
}
// ----------- MF14_list::set_norms -----------------
void MF14_list::set_norms( )
// set the norms of the discrete and continuous spectra according
// to the given multiplicities
{
  // first get the probability of a continuum gamma
  continuum_dist.multiple /= multiple;
  double scale_by;
  
  // pointers to the data
  MF12_base::iterator disc_ptr[ order_ + 1 ];
  MF15_list::iterator contin_ptr = continuum_dist.begin( );
  int L_order;
  for( L_order = 0; L_order <= order_; ++L_order )
  {
    disc_ptr[ L_order ] = coefs_[ L_order ].begin();
  }

  // go through the lists
  for( ; contin_ptr != continuum_dist.end( ); ++contin_ptr )
  {
    double cont_wgt =  continuum_dist.multiple.evaluate( contin_ptr->E_in( ) );
    if( ( cont_wgt < 0.0 ) || ( cont_wgt > 1.0 ) )
    {
      SevereError("MF14_list::set_norms",pastenum("Bad weight =",cont_wgt));
    }
    double norm_now = contin_ptr->get_norm( );
    if( norm_now <= 0.0 )
    {
      Warning("MF14_list::set_norms","Zero norm");
    }
    else 
    {
      scale_by = cont_wgt / norm_now;
      *(contin_ptr) *= scale_by;
    }
    
    norm_now = disc_ptr[ 0 ]->get_norm( );
    if( norm_now <= 0.0 )
    {
      disc_ptr[ 0 ]->print();
      Warning("MF14_list::set_norms","Zero norm");
    }
    else
    {
      scale_by = ( 1.0 - cont_wgt ) / norm_now;
      for( int L_order = 0; L_order <= order_; ++L_order )
      {
        *(disc_ptr[ L_order ]) *= scale_by;
        ++disc_ptr[ L_order ];
      }
    }
  }
}

// ----------- MF14_list::strip_zero_multiplicity -----------
void MF14_list::strip_zero_multiplicity( )
{
    multiplicity::iterator multPtr = multiple.begin();
    while ( multPtr != multiple.end() )
    {
        if ( multPtr->y == 0.0 ) {
            double bad_Ein = multPtr->x;
            Info( "MF14_list::strip_zero_multiplicity", pastenum( "Erasing outgoing energy ", multPtr->x ) + pastenum( " because M(E) = ", multPtr->y ) );
            multPtr = multiple.erase( multPtr );
            for( int L_count = 0; L_count <= order_; ++L_count )
            {
                MF12_base::iterator list_ptr = coefs_[ L_count ].begin();
                while ( list_ptr != coefs_[ L_count ].end() )
                {
                     if ( abs( bad_Ein - list_ptr->E_in( ) ) < 1e-10 ) list_ptr = coefs_[ L_count ].erase( list_ptr );
                     else ++list_ptr;
                }
            }
        }
        else 
        {
            ++multPtr;
        }
    }
}

// ----------- MF14_list::renorm -----------------
void MF14_list::renorm( )
// normalize the probabilities to 1
{
  MF12_base::iterator e_in_link[ order_ + 1 ];  // Legendre coefficients
  int L_count;        // to count the Legendre order

  // point to the first incident energy for each Legendre order
  for( L_count = 0; L_count <= order_; ++L_count )
  {
    e_in_link[ L_count ] = coefs_[ L_count ].begin();
  }

  // loop through the incident energies
  for( ; ; )
  {
    // get the norm of the 0-order Legendre term
    double norm = e_in_link[ 0 ]->get_norm( );
    if( norm == 0.0 )
    {
      Warning("MF14_list::renorm",
        pastenum("No spectrum for E_in= ",e_in_link[ 0 ]->E_in( )));
    }
    else
    {
      // scale all Legendre orders
      for( L_count = 0; L_count <= order_; ++L_count )
      {
        *e_in_link[ L_count ] *= 1.0/norm;
      }
    }

    // go to the next incident energy for each Legendre order
    for( L_count = 0; L_count <= order_; ++L_count )
    {
      ++e_in_link[ L_count ];
      if( e_in_link[ L_count ] == coefs_[ L_count ].end() )
      {
        return;  // we did them all
      }
    }
  }
}
// ----------- MF14_list::add_lists -----------------
void MF14_list::add_lists( )
// combine the continuous spectrum with the discrete lines
{
  // ensure that we have continuum data for all discrete incident energies
  list< double > E_incident;
  coefs_[ 0 ].collect_Ein( E_incident );
  continuum_dist.collect_Ein( E_incident );
  E_incident.sort( );
  E_incident.unique( );
  continuum_dist.fill_in_list( E_incident );

  // delete any extra continuum incident energies
  extra_contin( );

  // then widen the delta functions
  widen_delta( );

  // set the weights of the continuum and discrete spectra
  set_norms( );

  // now do the addition
  MF12_base::iterator disc_ptr[ order_ + 1 ];
  int L_order;
  for( L_order = 0; L_order <= order_; ++L_order )
  {
    disc_ptr[ L_order ] = coefs_[ L_order ].begin();
  }
  for( MF15_list::iterator contin_ptr = continuum_dist.begin( );
    contin_ptr != continuum_dist.end(); ++contin_ptr )
  {
    // make the gamma energies match
    fill_in_lists( *contin_ptr, *(disc_ptr[ 0 ]) );
    for( L_order = 1; L_order <= order_; ++L_order )
    {
      fill_in_lists( *(disc_ptr[ 0 ]), *(disc_ptr[ L_order ]) );
    }
    // add the zero-order terms
    *disc_ptr[ 0 ] += *contin_ptr;
    // next Legendre terms
    for( L_order = 0; L_order <= order_; ++L_order )
    {
      ++disc_ptr[ L_order ];
    }
  }
  //check_Legendre( false );  // ensure same E_gamma for all Legendre orders
  thinit( );
  renorm( );
}
// ----------- MF14_list::extra_contin -----------------
void MF14_list::extra_contin( )
// Delete any extra continuum incident energies.
// We do this because there is no safe way to interpolate between
// lists of discrete lines.  (Data are not always given for all lines.)
{
  // pointer to the discrete data
  MF12_base::iterator disc_ptr = coefs_[ 0 ].begin();

  // pointers to the continuum data
  MF15_list::iterator contin_ptr = continuum_dist.begin( );
  MF15_list::iterator next_contin = contin_ptr;
  ++next_contin;

  const double EPS = Global.Value( "mf5_tol" );  // tolerance for mf5 data
  if( abs( disc_ptr->E_in( ) - contin_ptr->E_in( ) ) >
    EPS*contin_ptr->E_in( ) )
  {
    SevereError("MF14_list::extra_contin",
        "mf14 discrete and continuum lists should start together");
  }

  // go through the incident energies for the discrete data
  for(  ; disc_ptr != coefs_[ 0 ].end( ); )
  {
    if( disc_ptr->E_in( ) < ( 1.0 - EPS )*contin_ptr->E_in( ) )
    {
      // We have discrete data but no continuum data for this E_in
      SevereError("MF14_list::extra_contin",
        "For mf14 first expand the continuum data");
    }
    else if( disc_ptr->E_in( ) > ( 1.0 + EPS )*contin_ptr->E_in( ) )
    {
      // We have continuum data but no discrete data for this E_in
      continuum_dist.erase( contin_ptr );
      contin_ptr = next_contin;
      ++next_contin;
    }
    else
    {
      // E_in is the same for both discrete and continuum data
      contin_ptr = next_contin;
      ++next_contin;
      ++disc_ptr; 
    }
  }
}
// ----------- MF14_list::widen_jumps -----------------
void MF14_list::widen_jumps( )
// widen the jump discontinuities
{
    Unimplemented("MF14_list::widen_jumps", "Write this code!");
}
// ----------- MF14_list::continuum -----------------
void MF14_list::continuum( int NR, int NP )
// read the mf15 file on continuum energy distributions
{
  if( !with_mf15 )
  {
    SevereError("MF14_list::continuum","no mf15 file!");
  }

  // read the interpolation regions
  inFile->get_regions( NR, continuum_dist.multiple.NBT,
    continuum_dist.multiple.INT );

  if( (NR > 1) || (continuum_dist.multiple.INT[0] != 2) )
  {
    Unimplemented("MF14_list::continuum","Implement interpolation in mf14 continuum");
  }
  continuum_dist.multiple.read_data( NP, *inFile );

  // read the mf15 file
  continuum_dist.read_data( contin_file );
  continuum_dist.set_interp();  // set the interpolation rules
  {
    Info("MF14_list::continuum","Closing file "+endf_file_name( 15, ENDL.T )+"...");
  }
  contin_file.close( );
}
// ----------- MF14_list::check_Legendre -----------------
void MF14_list::check_Legendre( bool halt_on_error )
// ensure that we have the same E_gamma values for all Legendre
// orders.
{
  MF12_base::iterator E_in_ptr[ order_ + 1 ];
  int L_order;
  bool two_d_done = false;
  // start at the first incident energy for each Legendre order
  for( L_order = 0; L_order <= order_; ++L_order )
  {
    E_in_ptr[ L_order ] = coefs_[ L_order ].begin();
  }
  // loop through the lists by incident energy
  for( ; ; )
  {
    // first check the incident energies
    double e_in = E_in_ptr[ 0 ]->E_in( );
    for( L_order = 1; L_order <= order_; ++L_order )
    {
      if( E_in_ptr[ L_order ]->E_in( ) != e_in )
      {
        string msg("Incompatible E_in\n"+
            pastenum("  For order 0, E_in = ",e_in)+
            pastenum("  for order ",L_order)+
            pastenum(" E_in = ",E_in_ptr[ L_order ]->E_in( )));
        if (halt_on_error) SevereError("MF14_list::check_Legendre",msg);
        else Warning("MF14_list::check_Legendre",msg);
      }
    }
    // start at the begining of each list
    bool one_d_done = false;
    dd_list::iterator L_ptr[ order_ + 1 ];
    for( L_order = 0; L_order <= order_; ++L_order )
    {
      L_ptr[ L_order ] = E_in_ptr[ L_order ]->begin( );
    }
    // now scan the lists
    for( ; ; )
    {
      for( L_order = 1; L_order <= order_; ++L_order )
      {
        double close_enough = ENDL_EPSILON( L_ptr[ 0 ]->x );
        if( abs(L_ptr[ L_order ]->x != L_ptr[ 0 ]->x) > close_enough )
        {
          string msg(
            pastenum("Incompatible gamma energy for E_in: ",e_in)+         
            pastenum("\n  For order 0, E_gamma = ",L_ptr[ 0 ]->x)+
            pastenum("  for order ",L_order)+
            pastenum(" E_gamma = ",L_ptr[ L_order ]->x));
          if (halt_on_error) SevereError("MF14_list::check_Legendre",msg);
          else Warning("MF14_list::check_Legendre",msg);
        }
      }
      // go to the next (E_gamma, probability) pair
      for( L_order = 0; L_order <= order_; ++L_order )
      {
        ++L_ptr[ L_order ];
        if( L_ptr[ L_order ] == E_in_ptr[ L_order ]->end( ) )
        {
          one_d_done = true;  // there is no more data
        }
      }
      if( one_d_done )
      {
        break;
      }
    }
    // did the lists finish for all Legendre orders?
    for( L_order = 0; L_order <= order_; ++L_order )
    {
      if( L_ptr[ L_order ] != E_in_ptr[ L_order ]->end( ) )
      {
        string msg(
            pastenum("For E_in: ",e_in)+
            pastenum(" extra link E_gamma: ",L_ptr[ L_order ]->x)+
            pastenum(" for order: ",L_order));
        if (halt_on_error) SevereError("MF14_list::check_Legendre",msg);
        else Warning("MF14_list::check_Legendre",msg);
      }
    }
    // go to the list for the next incident energy
    for( L_order = 0; L_order <= order_; ++L_order )
    {
      ++E_in_ptr[ L_order ];
      if( E_in_ptr[ L_order ] == coefs_[ L_order ].end( ) )
      {
        two_d_done = true;  // there is no more data
      }
    }
    if( two_d_done )
    {
      break;
    }
  }
  // did the lists finish for all Legendre orders?
  for( L_order = 0; L_order <= order_; ++L_order )
  {
    if( E_in_ptr[ L_order ] != coefs_[ L_order ].end( ) )
    {
        string msg(
            pastenum("\n  Extra list E_in: ",E_in_ptr[ L_order ]->E_in( ))+
            pastenum(" for order: ",L_order));
        if (halt_on_error) SevereError("MF14_list::check_Legendre",msg);
        else Warning("MF14_list::check_Legendre",msg);
    }
  }
}
// ----------- MF14_list::print -----------------
void MF14_list::print( )
// print the lists for debugging purposes
{
  for( int L_order = 0; L_order <= order_; ++L_order )
  {
    cout << "Legendre order: " << L_order << endl;
    coefs_[ L_order ].print( );
  }
}
// ----------- MF14_list::write_endl -----------------
void MF14_list::write_endl( )
// write an ENDL i=4 file.
{
  ENDL.set_I_number( 4 );

  fstream endl_file;
  string file_name = ENDL.file_name;

  if ( ENDL.new_file() )
  {
    endl_file.open(file_name.c_str(),ios::out);
    Info("MF14_list::write_endl","Opening ENDL file "+file_name);
  }
  else
  {
    endl_file.open(file_name.c_str(),ios::out|ios::app);
    Info("MF14_list::write_endl","Appending ENDL file "+file_name);
  }

  //Shove the header info in
  endl_file << ENDL.header_line_1 << endl;
  endl_file << ENDL.header_line_2 << endl;
  endl_file.setf(ios::scientific,ios::floatfield);

  // print the data
  for( int L_count = 0; L_count <= order_; ++L_count )
  {
    coefs_[ L_count ].out_data( 4, endl_file, 1.0*L_count );
  }

  //Put on end of file/section line into output file
  endl_file << ENDL.eof_line << endl;

  //Finally, we close the output file
  Info("MF14_list::write_endl","Closing ENDL file "+file_name);
  endl_file.close();
}

// ********* for class MF14_split ****************
// ----------- MF14_split::master -----------------
void MF14_split::master( double awr, int NK, mf12_file *InFile,
  mf14_file *angle_file )
// handle an mf12 multiplicity file (LO = 1) in the isotropic case
{
  // save the input data
  movers.AWR = awr;
  mf12_NK = NK;

  Warning("MF14_split::master", "Printing only the isotropic component");

  // Use InFile for all of the mf12 input
  inFile = InFile;
  statics.inFile = InFile;
  raw_move.inFile = InFile;
  raw_static.inFile = InFile;

  // initialize the Booleans
  with_move = false;
  with_static = false;
  with_mf15 = false;

  // get the energy distributions
  read_data( );
  // brach according to the type of data
  if( NK == 1 )
  {
    //Here, there is a single gamma in from the reaction (except continuum)
    //The i=9 file is created in the write_xxxx routines with a 
    //multiplicity of one (or a ratio of cross sections).
    if( with_mf15 )
    {
      // there are only continuum distributions
      Info("MF14_split::master","mf14 continuum only");
      write_continuum( );
    }
    else if( with_static )
    {
      // there is just 1 line (static)
      Info("MF14_split::master","mf14 1 static line");
      write_static( );
    }
    else
    {
      // there is just 1 line (moving)
      Info("MF14_split::master","mf14 1 moving line");
      write_mover( );
    }
  }
  else
  {
    if( with_mf15 )
    {
      if( with_move )
      {
        Info("MF14_split::master", "mf14 continuum plus moving lines");
        move_w_contin( );
      }
      else
      {
        // static lines plus continuum
        Info("MF14_split::master","mf14 static lines plus continuum");
        plus_contin( );
      }
    }
    else if( with_move )
    {
      // continuum represented by moving lines
      Info("MF14_split::master","mf14 static lines plus movers");
      do_movers( );
    }
    else
    {
      // only static lines
      Info("MF14_split::master","mf14 static lines only");
      just_statics( );
    }
  }
}
// ********* for class MF14_c55_list *************
// ----------- MF14_c55_list constructor ----------------
MF14_c55_list::MF14_c55_list()
{
}
// ----------- MF14_c55_list destructor ----------------
MF14_c55_list::~MF14_c55_list()
{
}
// ----------- MF14_c55_list::master -----------------
void MF14_c55_list::master( mf14_file& inFile, int LTT, int NK, int NI )
// Handle an anisotropic mf14 file; make a c55 file
// LTT: Legendre flag
// NK: number of anisotropic lines
// NI: total number of lines
{
  int num_to_do = NK - NI;
  double EG[ num_to_do ]; // the gamma energies
  int count;
  int start_at;

  ENDL.set_c_number(55);
  ENDL.set_yo(7);
  ENDL.set_x0(0.0);  // ENDL convention for Q of c55 gammas

  // skip the isotropic lines
  for( count = 0; count < NI; ++count )
  {
    double eg;
    inFile.discrete( &eg );
  }

  // the anisotropic data
  if( LTT == 1 )
  {
    Info("MF14_c55_list::master"," Doing mf14 Legendre for c55");
    // we have NK sets of Legendre data
    two_d_Legendre Legendre[ num_to_do ];
    for( count = 0; count < num_to_do; ++count )
    {
      EG[count] = make_Legendre( inFile,  Legendre[count] );
    }

    // write the data in order of increasing gamma energy
    start_at = num_to_do - 1;
    if ( EG[ start_at ] == 0.0 )
    {
      // this is continuum data
      ENDL.set_x1(0.0);
      ENDL.set_s_number(0);
      Legendre[ start_at ].write_endl(1);
      --start_at;
    }
    // write the discrete data
    ENDL.set_s_number(3);
    for( count = start_at; count >= 0; --count )
    {
      ENDL.set_x1( EG[ count ] );
      Legendre[ count ].write_endl(1); 
      ENDL.append = true;
    }     
  }
  else if( LTT == 2 )
  {
    // we have num_to_do sets of tabular data
    mf4_table tables[ num_to_do ];
    for( count = 0; count < num_to_do; ++count )
    {
      EG[count] = make_table( inFile, tables[count] );
    }

    // write the data in order of increasing gamma energy
    start_at = num_to_do - 1;
    if ( EG[ start_at ] == 0.0 )
    {
      // this is continuum data
      ENDL.set_x1(0.0);
      ENDL.set_s_number(0);
      tables[ start_at ].write_endl(1);
      --start_at;
    }
    // write the discrete data
    ENDL.set_s_number(3);
    for( count = start_at; count >= 0; --count )
    {
      ENDL.set_x1( EG[ count ] );
      tables[ count ].write_endl(1); 
      ENDL.append = true;
    }     
  }
  else
  {
    SevereError("MF14_list::master",pastenum("Bad LTT value ",LTT));
  }
}
// ----------- MF14_c55_list::make_Legendre -----------------
double MF14_c55_list::make_Legendre( mf14_file& inFile,
  two_d_Legendre& Legendre )
// make a Legendre list for one gamma.  Return the gamma energy
{
  double EG; // the gamma energy
  int NR;    // number of interpolation regions
  int NE;    // number of incident energies

  inFile.read_EG( &EG, &NR, &NE );
  inFile.get_regions( NR, Legendre.NBT, Legendre.INT );
  Legendre.expand_data( inFile, NE, 4 ); //read/expand mf4 data

  // change to MeV
  return EG * ENDL.eV2MeV;
}

// ----------- MF14_c55_list::make_table -----------------
double MF14_c55_list::make_table( mf14_file& inFile,
  mf4_table& table )
// make a table list for one gamma.  Return the gamma energy
{
  double EG; // the gamma energy
  int NR;    // number of interpolation regions
  int NE;    // number of incident energies

  inFile.read_EG( &EG, &NR, &NE );
  inFile.get_regions( NR, table.NBT, table.INT );
  table.read_data( inFile, NE ); //read data

  // change to MeV
  return EG * ENDL.eV2MeV;
}

