/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1892 $
 * $Date: 2006-09-07 10:29:17 -0700 (Thu, 07 Sep 2006) $
 * $Author: dbrown $
 * $Id: mf14classes.hpp 1892 2006-09-07 17:29:17Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// header for the MF=14 translation code

#ifndef MF14CLASSES
#define MF14CLASSES

#include "ENDF_file.hpp"
#include "mf12classes.hpp"
#include "mf15classes.hpp"

using namespace std;

// ----------- class MF14_list -----------------
//! Class for anisotropic gamma distributions given as Legendre expansions
class MF14_list
{
private:
  mf12_file *inFile;
  mf14_file *angle_file;   // angular distributions
  mf15_file contin_file;   // continuum energy distributions

  vector<MF12_base> coefs_;
  MF15_list continuum_dist;  // continuum distribution

  multiplicity multiple;   // the total gamma multiplicity

  double AWR;
  double mf14_EG;    // next anisotropic gamma
  int mf14_NR;       // number of mf14 interpolation regions
  int mf14_NE;       // number of mf14 incident energies
  int order_;        // maximum Legendre order
  int mf14_NI;       // number of isotropic gammas
  int mf14_NK;       // total number of mf14 gammas, including continuum
  int mf12_NK;       // total number of mf12 gammas, including continuum
  int num_aniso;     // number of anisotropic gammas
  int aniso_count;   // numer of anisotropic gammas read
  bool with_mf15;    // is there an mf15 continuum data file?

public:
  //! Default constructor
  MF14_list( ): inFile(0), angle_file(0), contin_file(), coefs_(0), 
    continuum_dist(), multiple(), AWR(0.0), mf14_EG(0.0), 
    mf14_NR(0), mf14_NE(0), order_(0), mf14_NI(0),
    mf14_NK(0), mf12_NK(0), num_aniso(0), aniso_count(0), with_mf15(false){}

  //! Constructor with given files and number of distributions
  // NI: number of isotropic distributions
  // NK: total number of gamma distributions
  // order: order of Legendre expansion
  MF14_list( int NI, int NK, int order, mf12_file *InFile,
    mf14_file *Angle_File );

  //! Default destructor
  ~MF14_list();

  //! Controls the processing of these data.
  // mf12_NK is the number of lines (including continuum).
  void master( int mf12_NK );

  //! Reads in gamma probabilities and their Legendre expansions
  void read_data( );

  //! Reads in data for one gamma line
  void one_line( double EG, double *prev_E_in, int NR, int NP,
    bool do_shift);

  //! Inserts an (EG, probability) pair into a list
  void insert_pair( dd_link& XYdata, MF12_base::iterator e_in_link );

  //! Reads Legendre data for one gamma from an mf14 file
  void read_Legendre( double *coef_buff, double **Coef_ptrs );

  //! Inserts the Legendre data for one gamma from an mf14 file into our lists
  // Coef_ptrs[ j ] points to the j-th m14 incident energy, and this is
  // followed by order_ Legendre coefficients
  void insert_Legendre( double e_in, dd_link& XYdata,
    double prev_E_in, double **Coef_ptrs );

  // strip any incident energy points that have 0 multiplicity from 
  // both the mulitplicity and outgoing angular distributions
  void strip_zero_multiplicity( );

  //! Widens the delta functions for the discrete spectrum
  void widen_delta( );

  //! Widens the jump discontinuities
  void widen_jumps( );

  //! Removes redundant data points
  void thinit( );

  //! Reads the mf15 file on continuum energy distributions
  void continuum( int NR, int NP );

  //! Combines the continuous spectrum with the discrete lines
  void add_lists( );

  //! Deletes any extra continuum incident energies.
  void extra_contin( );

  //! Ensures that we have the same list of E_gamma values for all Legendre orders.
  void check_Legendre( bool halt_on_error = true );

  //! Prints the lists for debugging purposes
  void print( );

  //! Sets the norms of the spectra according to the given multiplicities
  void set_norms( );

  //! Normalizes the probabilities to 1
  void renorm( );

  //! Writes an ENDL i=4 file.
  void write_endl( );

};

// ----------- class MF14_split -----------------
//! Class for anisotropic gamma distributions and discrete lines given as Legendre expansions
class MF14_split : public MF12_list
{
private:
  mf14_file *angle_file;   // angular distributions

public:
  //! Default constructor
  MF14_split( )
  {
  }

  //! Default destructor
  ~MF14_split()
    {
    }

  //! read and process the data
  // NI: number of isotropic distributions
  // NK: total number of gamma distributions
  void master( double awr, int NK, mf12_file *InFile, mf14_file *Angle_File );

};
// ----------- class MF14_c55_list -----------------
//! Class to handle anisotropic MF=14 file and write ENDL c=55 file
class MF14_c55_list
{
public:
  //! Default constructor
  MF14_c55_list();

  //! Default destructor
  ~MF14_c55_list();

  //! Controls the processing of the MF=14 data
  // LTT: Legendre flag
  // NK: total number of lines
  // NI: number of isotropic lines
  void master( mf14_file& inFile, int LTT, int NK, int NI );

  //! Makes a Legendre list for one gamma and returns the gamma energy
  double make_Legendre( mf14_file& inFile, two_d_Legendre& Legendre );

  //! Makes a table list for one gamma and returns the gamma energy
  double make_table( mf14_file& inFile, mf4_table& table );
};

#endif
