/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1748 $
 * $Date: 2006-03-14 12:18:15 -0800 (Tue, 14 Mar 2006) $
 * $Author: dbrown $
 * $Id: mf14table.cpp 1748 2006-03-14 20:18:15Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the classes for translating the MF=14 tabular data

#include <vector>

#include "endl_formats.hpp"
#include "messaging.hpp"
#include "mf14table.hpp"

extern ENDLClass ENDL;

// ********* for class MF14_table_raw *************
// ----------- MF14_table_raw::read_data ----------------
void MF14_table_raw::read_data( mf14_file& angle_file, int num_gamma )
{
  int NR;  // number of interpolation regions
  int NE;  // number of incident energies
  double EG;  // energy of the gamma

  mf4_table new_table;  // table to insert into the list
  MF14_table_raw::iterator this_data;

  // read the data for the lines
  for( int count = 0; count < num_gamma; ++count )
  {
    // get the gamma energy
    angle_file.read_EG( &EG, &NR, &NE );

    // prepend the next table and point to it
    insert( begin( ), new_table );
    this_data = begin( );

    // the gamma energy in MeV
    this_data->E_gamma( ) = EG * ENDL.eV2MeV;

    // read the interpolation regions
    angle_file.get_regions( NR, this_data->NBT, this_data->INT );

    // read the data for this line
    this_data->read_data( angle_file, NE );

    // set the interpolation rules between E_in values
    this_data->set_interp( );
  }
}
// ----------- MF14_table_raw::fill_E_in ----------------
void MF14_table_raw::fill_E_in( )
// There must be common E_in values
{
  list< double > E_in_list;
  list< double >::iterator E_in_ptr;
  MF14_table_raw::iterator this_gamma;
  mf4_table::iterator this_E;
  double E_in;

  // Make the list of all E_in values
  for( this_gamma = begin( ); this_gamma != end( ); ++this_gamma )
  {
    // scan the E_in values for this gamma
    for( this_E = this_gamma->begin( ); this_E != this_gamma->end( );
	 ++this_E )
    {
      E_in = this_E->E_in( );
      E_in_list.push_back( E_in );
    }
  }
  // sort the list
  E_in_list.sort( );

  // remove duplicates
  E_in_list.unique( );

  // For each gamma make common E_in values
  for( this_gamma = begin( ); this_gamma != end( ); ++this_gamma )
  {
    one_d_table new_table;    //  for new entries
    mf4_table::iterator new_table_ptr;
    mf4_table::iterator prev_E;

    // initial E_in value for this gamma
    this_E = this_gamma->begin( );
    E_in = this_E->E_in( );

    // Are there E_in_list values less than E_in?
    E_in_ptr = E_in_list.begin( );
    for( ; *E_in_ptr < E_in; ++E_in_ptr )
    {
      //      cout << "E_gamma: " << this_gamma->E_gamma( ) <<
      //	" prepend: " << *E_in_ptr << endl;

      if( E_in_ptr == E_in_list.end( ) )
      {
        SevereError( "MF14_table_raw::fill_E_in",
		     " passed the end of E_in_list" );
      }
      // insert a copy
      this_gamma->insert( this_E, new_table );
      new_table_ptr = this_E;
      --new_table_ptr;
      new_table_ptr->copy( *this_E );
      new_table_ptr->E_in( ) = *E_in_ptr;
    }

    // Do the rest of E_in_list
    for(  ; E_in_ptr != E_in_list.end( ); ++E_in_ptr )
    {
      E_in = *E_in_ptr;

      // We could be past the end
      if( this_E == this_gamma->end( ) )
      {
	//        cout << "E_gamma: " << this_gamma->E_gamma( ) <<
	//        	" append: " << E_in << endl;

        this_gamma->insert( this_E, new_table );
        new_table_ptr = this_E;
        --new_table_ptr;
        new_table_ptr->copy( *prev_E );
        new_table_ptr->E_in( ) = E_in;
      }
      // The energies could match
      else if( this_E->E_in( ) == E_in )
      {
        prev_E = this_E;
        ++this_E;
      }
      else
      {
	//        cout << "E_gamma: " << this_gamma->E_gamma( ) <<
	//        	" insert: " << E_in << endl;

        this_gamma->insert( this_E, new_table );
        new_table_ptr = this_E;
        --new_table_ptr;
        new_table_ptr->list_interp( E_in, *prev_E, *this_E );
	//        new_table_ptr->print( );
      }
    }
  }

  // for each value of E_in we need common mu values
  fill_mu( );
}
// ---------------- MF14_table_raw::fill_mu ----------------
void MF14_table_raw::fill_mu( )
// There must be common mu values for each E_in
{
  list< double > mu_list;   // the common mu values for one E_in
  double mu;
  vector< mf4_table::iterator > E_in_ptrs;  // one E_in pointer for each gamma
  MF14_table_raw::iterator this_gamma;
  one_d_table::iterator this_mu_p;

  // go through the gammas to set up E_in_ptrs
  for( this_gamma = begin( ); this_gamma != end( ); ++this_gamma )
  {
    mf4_table::iterator E_in_ptr = this_gamma->begin( );
    E_in_ptrs.push_back( E_in_ptr );
  }
  int num_gamma = E_in_ptrs.size( );
  int j;

  // go through the incident energies for the first gamma
  this_gamma = begin( );
  for(  ; E_in_ptrs[ 0 ] != this_gamma->end( ); )
  {
    // collect the mu values for this E_in
    for( j = 0; j < num_gamma; ++j )
    {
      for( this_mu_p = E_in_ptrs[ j ]->begin( );
	   this_mu_p != E_in_ptrs[ j ]->end( ); ++this_mu_p )
      {
        mu = this_mu_p->x;
        mu_list.push_back( mu );
      }
    }
    // sort and remove duplicates
    mu_list.sort( );
    mu_list.unique( );

    //    cout << "E_in: " << E_in_ptrs[ 0 ]->E_in( ) << endl;
    //    for( list< double >::iterator mu_list_ptr = mu_list.begin( );
    //	 mu_list_ptr != mu_list.end( ); ++mu_list_ptr )
    //    {
    //      cout << *mu_list_ptr << endl;
    //    }

    // For each E_in make common mu values
    for( j = 0; j < num_gamma; ++j )
    {
      list< double >::iterator mu_list_ptr = mu_list.begin( );
      dd_link new_pair;  //  for new (mu, p) entries
      one_d_table::iterator prev_mu_p;

      this_mu_p = E_in_ptrs[ j ]->begin( );
      mu = this_mu_p->x;

      // Are there mu_list values less than mu?
      mu_list_ptr = mu_list.begin( );
      for( ; *mu_list_ptr < mu; ++mu_list_ptr )
      {
	//        cout << "E_gamma: " << this_gamma->E_gamma( ) <<
	//	  " E_in: " << E_in_ptrs[ j ]->E_in( ) <<
	//          " prepend: " << *mu_list_ptr << endl;

        if( mu_list_ptr == mu_list.end( ) )
        {
          SevereError( "MF14_table_raw::fill_mu",
                       " passed the end of mu_list" );
        }
        // insert a zero
        new_pair.x = *mu_list_ptr;
        new_pair.y = 0.0;
        E_in_ptrs[ j ]->insert( this_mu_p, new_pair );
       }

      // Do the rest of mu_list
      for(  ; mu_list_ptr != mu_list.end( ); ++mu_list_ptr )
      {
        mu = *mu_list_ptr;

        // We could be past the end
        if( this_mu_p == E_in_ptrs[ j ]->end( ) )
        {
	  //          cout << "E_gamma: " << this_gamma->E_gamma( ) <<
	  //	          " E_in: " << E_in_ptrs[ j ]->E_in( ) <<
	  //                  " append: " << mu << endl;

          new_pair.x = *mu_list_ptr;
          new_pair.y = 0.0;
          E_in_ptrs[ j ]->insert( this_mu_p, new_pair );
        }
        // The energies could match
        else if( this_mu_p->mu( ) == mu )
        {
          prev_mu_p = this_mu_p;
          ++this_mu_p;
        }
        else
        {
	  //          cout << "E_gamma: " << this_gamma->E_gamma( ) <<
	  //         	  " E_in: " << E_in_ptrs[ j ]->E_in( ) <<
	  //                 " insert: " << mu << endl;

          new_pair.x = *mu_list_ptr;
          new_pair.y = E_in_ptrs[ j ]->linlin_interp( new_pair.x,
            *prev_mu_p, *this_mu_p );
          E_in_ptrs[ j ]->insert( this_mu_p, new_pair );
        }
      }
    }

    // start clean for the next E_in
    mu_list.erase( mu_list.begin( ), mu_list.end( ) );

    // increment the E_in pointers
    for( j = 0; j < num_gamma; ++j )
    {
      ++E_in_ptrs[ j ];
    }
  }
}

// ********* for class MF14_table *************
// ----------- MF14_table::master ----------------
void MF14_table::master( mf12_file& mult_file, mf14_file& angle_file,
  int num_mf12, int num_mf14 )
{
  // set up the mf12 input file
  line_mult.inFile = &mult_file; 

  // the first block is the total multiplicity
  int NR; // number of interpolation regions
  int NP; // number of (E, multiplicity) pairs
  mult_file.read_line2( &NR, &NP );
  mult_file.get_regions( NR, multiple.NBT, multiple.INT );
  multiple.read_data( NP, mult_file );

  if( ENDL.F == 13 )  //Need to convert cross section to multiplicity
  {
    multiple.divide_by_xs( );
  }
  multiple.write_endl( 9 );

  // read the multiplicities of the gammas
  read_mf12( mult_file, num_mf12 );

  // read the mf14 data
  endf_mf14.read_data( angle_file, num_mf14 );

  // make sure that there is raw mf14 data for every E_in value
  endf_mf14.fill_E_in( );

  // weight the raw mf14 data by their multiplicity
  mult_weight( );

  // put the data into ENDL order
  ENDL_order( );

  // convert the lines to delta-functions
  widen_deltas( );

  // create the ENDL i = 1 angular distributions
  make_i1( );

  // print the ENDL data
  write_endl( );
}
// ----------- MF14_table::read_mf12 ----------------
void MF14_table::read_mf12( mf12_file& mult_file, int num_mf12 )
// read the multiplicities of the gamma lines
{
  for( int count = 0; count < num_mf12; ++count )
  {
    double EG;     //  gamma energy
    double ES;     //  excitation level
    int LP;        // energy-dependent gamma?
    int LF;        // discrete-continuum flag
    int NR;        // number of interpolation regions
    int NP;        // number of (E, multiplicity) pairs

    mult_file.gamma_line1( &EG, &ES, &LP, &LF, &NR, &NP );
    // check the data
    if( ( LF != 2 ) || ( LP == 2 ) )
    {
      SevereError( "MF14_table::master", " strange mf12 data" );
    }
    line_mult.one_line( EG * ENDL.eV2MeV, NR, NP );
  }
}
// ----------- MF14_table::mult_weight ----------------
void MF14_table::mult_weight( )
// weight the raw mf14 data by their multiplicity
{
  // go through the gammas
  MF12_raw::iterator this_mult = line_mult.begin( );

  for( MF14_table_raw::iterator this_gamma = endf_mf14.begin( );
       this_gamma != endf_mf14.end( ); ++this_gamma, ++this_mult )
  {
    // do a check
    if( ( this_gamma->E_gamma( ) != this_mult->E_gamma( ) ) ||
	this_mult == line_mult.end( ) )
    {
      SevereError( "MF14_table::mult_weight", 
		   " gamma energy mismatch" );
    }

    // iterate over the incident energies
    for( mf4_table::iterator this_E = this_gamma->begin( );
	 this_E != this_gamma->end( ); ++this_E )
    {
      double mult = this_mult->evaluate( this_E->E_in( ) );
      // scale the probabilities
      *this_E *= mult;
    }
  }
}
// ----------- MF14_table::ENDL_order ----------------
void MF14_table::ENDL_order( )
// put the data into ENDL order
{
  two_d_ENDL_table ENDL_E_in;         // for inserting into the list
  MF14_table::iterator ENDL_E_in_ptr;  //  point to the new data
  mf4_table::iterator E_in_ptr;
  MF14_table_raw::iterator this_gamma;
  one_d_table::iterator this_mu_p;

  // set up the ENDL list structure
  // go through the incident energies for the first gamma
  this_gamma = endf_mf14.begin( );
  for( E_in_ptr = this_gamma->begin( ); E_in_ptr != this_gamma->end( );
       ++E_in_ptr )
  {
    // insert a new 2-d list and point to it
    insert( end( ), ENDL_E_in );
    ENDL_E_in_ptr = end( );
    --ENDL_E_in_ptr;
    ENDL_E_in_ptr->E_in( ) = E_in_ptr->E_in( );

    // set up the 1-d lists for each mu
    for( this_mu_p = E_in_ptr->begin( ); this_mu_p != E_in_ptr->end( );
	 ++this_mu_p )
    {
      one_d_table new_table;
      new_table.mu( ) = this_mu_p->x;
      ENDL_E_in_ptr->insert( ENDL_E_in_ptr->end( ), new_table );
      dd_link new_pair;  //  for new (mu, p) entries
    }
  }

  // fill in the ENDL list
  // go through the gammas
  for( this_gamma = endf_mf14.begin( ); this_gamma != endf_mf14.end( );
    ++this_gamma )
  {
    double E_gam = this_gamma->E_gamma( );

    // start with the first ENDL incident energy
    ENDL_E_in_ptr = begin( );

    // go through the ENDF incident energies
    for( E_in_ptr = this_gamma->begin( ), ENDL_E_in_ptr = begin( );
          E_in_ptr != this_gamma->end( ); ++E_in_ptr, ++ENDL_E_in_ptr )
    {
      dd_link new_pair;  //  for new (E_gamma, p) entries
      new_pair.x = E_gam;

      // go through the mu values
      two_d_ENDL_table::iterator ENDL_mu = ENDL_E_in_ptr->begin( );
      for( this_mu_p = E_in_ptr->begin( ); this_mu_p != E_in_ptr->end( );
           ++this_mu_p, ++ENDL_mu )
      {
        new_pair.y = this_mu_p->y;
        ENDL_mu->insert( ENDL_mu->end( ), new_pair );
      }
    }
  }
}
// ----------- MF14_table::widen_deltas ----------------
void MF14_table::widen_deltas( )
// convert the lines to delta-functions
{
  for( MF14_table::iterator ENDL_E_in_ptr = begin( );
       ENDL_E_in_ptr != end( ); ++ENDL_E_in_ptr )
  {
    for( two_d_ENDL_table::iterator mu_ptr = ENDL_E_in_ptr->begin( );
	 mu_ptr != ENDL_E_in_ptr->end( ); ++mu_ptr )
    {
      mu_ptr->widen_delta( );
    }
  }
}
// ----------- MF14_table::make_i1 ----------------
void MF14_table::make_i1( )
// create the ENDL i = 1 angular distributions
{
  for( MF14_table::iterator ENDL_E_in_ptr = begin( );
       ENDL_E_in_ptr != end( ); ++ENDL_E_in_ptr )
  {
    // add a link to the cosine list
    one_d_table new_cos;
    cosines.push_back( new_cos );

    // expand this cosine link
    two_d_list<one_d_table>::iterator cos_ptr = cosines.end( );
    --cos_ptr;
    ENDL_E_in_ptr->make_cos_link( ENDL_E_in_ptr->E_in( ), cos_ptr );
  }
}
