/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1785 $
 * $Date: 2006-03-22 16:03:58 -0800 (Wed, 22 Mar 2006) $
 * $Author: dbrown $
 * $Id: mf15classes.cpp 1785 2006-03-23 00:03:58Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the mf15 class for 2-d linked lists

#include "mf15classes.hpp"
#include "endl_formats.hpp"

extern ENDLClass ENDL;

// ********* for class MF15_list *************
// ----------- MF15_list::read_data -----------------
void MF15_list::read_data( mf15_file& inFile )
// This routine is used to read in ENDF/B-VI continuuous gamma
// distributions
{
  int ZA;
  double AWR;
  int LF;   // 1 for tabulated continuum
  int NC;   // number of file sections
  int NR;   // number of interpolation regions
  int NP;   // number of incident energies

  // read the first line
  inFile.read_line1(&ZA, &AWR, &NC);
  if(NC > 1)
  {
    Unimplemented("MF15_list::read_data","Implement multiple subsections for mf15");
  }

  // get the weights for this subsection
  inFile.weight_line(&LF, &NR, &NP);
  if(LF != 1)
  {
    SevereError("MF15_list::read_data",
        pastenum("Implement model LF: ",LF)+" in mf15");
  }
  
  // read the interpolation regions
  inFile.get_regions(NR, weights.NBT, weights.INT);
  
  // read and save the weights
  weights.read_data( NP, inFile, ENDL.Max_E_in );

  // how to interpolate with respect to incident energy
  inFile.E_in_interp(&NR, &NP);
  inFile.get_regions(NR, NBT, INT);

  // don't save the data if we are above the maximum incident energy
  bool done = false;

  // make the 2-d list
  for(int count = 0; count < NP; ++count)
  {
    one_e_dist( inFile, &done );
  }

  // record the types of interpolation between 1-d lists
  set_interp();

  if((NR > 1) || (INT[0] != 2))
  {
    table_interp( );
  }

  // did we exceed the maximum incident energy?
  two_d_iterator last_link = end( );
  --last_link;
  if( last_link->E_in( ) > ENDL.Max_E_in )
  {
    chop_highE( ENDL.Max_E_in );
  }
}

// ----------- MF15_list::one_e_dist -----------------
void MF15_list::one_e_dist( mf15_file& inFile, bool *done )
// This routine is used to read in ENDF/B-VI one continuuous gamma
// distribution.
// set done = true if E_in >= ENDL.Max_E_in
{
  string linebuff;
  string strbuff;

  // the head for this incident energy
  int NR;   // number of interpolation regions
  int NP;   // number of secondary energies
  double E_in;
  two_d_iterator e_in_link;

  inFile.get_E_in( &E_in, &NR, &NP );
  E_in *= ENDL.eV2MeV;  // change to MeV

  // start a new two_d_link for this incident energy
  one_d_table new_link;
  if( !(*done) )
  {
    insert(end(), new_link);
    e_in_link = end();
    --e_in_link;  // points to the new link
    e_in_link->E_in( ) = E_in;
    inFile.get_regions(NR, e_in_link->NBT, e_in_link->INT);
  }
  else
  {
    inFile.get_regions(NR, new_link.NBT, new_link.INT);
  }

  // read the data for this incident energy
  for( int iNP=0; iNP < NP; iNP++ ) //Loop over NP data pairs
  { 
    double E_out;
    double Prob;

    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 ) getline( inFile, linebuff); //Read in new line
    strbuff = linebuff.substr(iP*22, 22); //Grab the right pair
    read_dd(&strbuff, &E_out, &Prob );
    if ( E_out==-999.0 || Prob==-999.0 ) //read_PAIR sends -999.0 if problem
    {
      SevereError("MF15_list::one_e_dist","read_PAIR messed up!");
    }
    else if( !(*done) )
    {
      E_out *= ENDL.eV2MeV;
      dd_link XYdata(E_out, Prob);
      e_in_link->insert(e_in_link->end(), XYdata);
    }
  }
  if( !(*done) )
  {
    // expand any histogram data and widen any jumps
    e_in_link->expand_interp( ENDL.Max_E_out );
    e_in_link->widen_jumps();
    e_in_link->renorm();
    // we may be done
    double dE = ENDL_EPSILON( ENDL.Max_E_in );
    if( E_in > ENDL.Max_E_in - dE )
    {
      *done = true;
    }
  }
}
