/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1888 $
 * $Date: 2006-08-28 10:37:08 -0700 (Mon, 28 Aug 2006) $
 * $Author: dbrown $
 * $Id: mf4classes.cpp 1888 2006-08-28 17:37:08Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the classes used to translate the MF=4 data

#include "mf4classes.hpp"
#include "endl_formats.hpp"
#include "global_params.hpp"

extern ENDLClass ENDL;
extern GlobalParameterClass Global;

// ********* for class two_d_isotropic *************
// ----------- two_d_isotropic::expand_data -----------------
void two_d_isotropic::expand_data(int MT)
// Make an isotropic list
{
  double e_in;
  // make a new link for the threshold energy
  one_d_isotropic new_link;

  if(MT > 0)
  {
    // get the threshold energy
    e_in = ENDL.threshold;
  }
  else
  {
    // this list is for gamma production; threshold set later
    e_in = 0.0;
  }

  double min_e_in = Global.Value( "Min_n_energy" );
  if ( e_in < min_e_in ) e_in = min_e_in;

  // insert the 1-d list as a link
  insert(end(), new_link);
  two_d_isotropic::iterator this_ein = end();
  --this_ein;
  this_ein->initiate(e_in);

  // Do it again the maximum incident energy
  e_in = ENDL.Max_E_in;
  insert(end(), new_link);
  this_ein = end();
  --this_ein;
  this_ein->initiate(e_in);
}

// ********* for class two_d_Legendre *************
// ----------- two_d_Legendre::expand_data -----------------
void two_d_Legendre::expand_data(mf4_file& inFile, int NE, int MF)
// interpret the ENDF/B-VI Legendre data
// NE is the number of incident energies
{
  bool done = false;
  // handle each incident neutron energy
  for (int iNE=0; iNE < NE; iNE++ )
  {
    one_E_in( inFile, MF, &done );
  }
  // The Legendre coefficients may depend logarithmically on
  // the incident energy
  if ( ( INT.size() > 1 ) || ( INT[ 0 ] != 2 ) )
  {
    expand_interp();
  }
}
// ----------- two_d_Legendre::one_E_in -----------------
void two_d_Legendre::one_E_in( mf4_file& inFile, int MF, bool *done )
// interpret the ENDF/B-VI Legendre data for 1 incident energy
// return true if E_in >= ENDL.Max_E_in
{
  double ZERO;  // zero
  double E_1;   // incident neutron energy
  int iZERO;    // zero
  int nLC;      // number of Legendre Coefficients

  string linebuff;  // the input line
  string strbuff;   // a substring
  two_d_Legendre::iterator this_link;

  if ( MF == 6 )
  {
    int ND;
    int NA;
    int NW;   // number of Legendre coefficients (P_0 = 1 always)
    int NEP;
    
    getline( inFile, linebuff ); //read next line
    read_ddiiii(&linebuff, &ZERO, // zero 
        &E_1,             // incident energy
        &ND,              // number of discrete energies
        &NA,              // number of angular parameters (1)
        &NW,              // number of table entries
        &NEP);            // number of E_out values

    nLC=NW;
    if(ND > 0)
    {
      Unimplemented("two_d_Legendre::one_E_in",
        "Implement discrete energies for Legendre");
    }
  }
  else if ( MF==4 )
  {
    double dT;    // temperature
    int iLT;      // test for temperature dependence
    int NL;       // number of Legendre coefficients (P_0 = 1 always)
    
    getline( inFile, linebuff ); //read next line
    read_ddiiii(&linebuff, &dT,   // temperature
        &E_1,             // incident energy
        &iLT,             // test for temperature dependence
        &iZERO,           // zero
        &NL,              // number of Legendre Coefficients
        &iZERO);          // zero

    nLC=NL;
    if(dT != 0.0)
    {
      Warning("two_d_Legendre::one_E_in",
        pastenum("Non-zero temperature (",dT)+
        " eV) in Legendre data, zeroing it."
      );
    }
  }
  else
  {
	Unimplemented("two_d_Legendre::one_E_in",
      pastenum("No code written for Legendre in MF = ",MF));
  }

  // make a new link for this incident energy
  one_d_Legendre new_link;

  if( !(*done) )
  {
    // insert the 1-d list as a link
    insert(end(), new_link);
    this_link = end();
    --this_link;

    // change to MeV
    this_link->E_in() = E_1 * ENDL.eV2MeV;

    // set the Legendre coefficients
    this_link->coef.get_space( nLC + 1 );
    this_link->coef.at(0) = 1.0;   //  default c_0
  }

  // read the coefficients
  for ( int iNR=0; iNR < nLC; ++iNR )
  {
    double next_coef;
    int iR = iNR % 6;  //There are 6 numbers per line
    if ( iR == 0 ) getline( inFile, linebuff); //Read in new line
    if( !*(done) )
    {
      strbuff = linebuff.substr(iR*11, 11);
      read_d( &strbuff, &next_coef );
      this_link->coef.at( iNR + 1 ) = next_coef;
    }
  }

  // is this incident energy above our maximum?
  if( ( this_link->E_in() > ENDL.Max_E_in ) && !(*done) )
  {
    set_interp( );
    chop_highE( ENDL.Max_E_in );
    *done = true;
  }
  else if( !(*done) )
  {
    // expand the 1-d model
    this_link->expand();
    // we may be done
    double dE = ENDL_EPSILON( ENDL.Max_E_in );
    if( this_link->E_in() > ENDL.Max_E_in - dE )
    {
      *done = true;
    }
  }
}

// ----------- two_d_Legendre::expand_interp -----------------
void two_d_Legendre::expand_interp()
// expand possible dependence of the coefficients on the
// logarithm of the incident energy
{
  if( INT[0] == -7 )
  {
    SevereError("two_d_Legendre::expand_interp","expand_interp called twice");
  }

  // record the types of interpolation between 1-d lists
  set_interp();

  // construct the equiprobable bins
  for( two_d_Legendre::iterator this_link = begin();
    this_link != end(); ++this_link )
  {
    this_link->get_bins();
  }
  // fill in to get good linear interpolation
  thicken();

  // flag to ensure that we don't do this twice on the same data
  INT[0] = -7;
}

// ********* for class mf4_table *************
// ----------- mf4_table::master -----------------
void mf4_table::master( mf4_file& inFile )
{
  int NE;   // number of incident energy points
  int NR;   // number of interpolation regions

  // how much data
  inFile.read_line3( &NR, &NE );
  inFile.get_regions(NR, NBT, INT);

  read_data(inFile, NE ); //read in the data

  widen_jumps();

  // renormalize the probability distributions since, in the log-lin
  // to lin-lin interpolation, we may have lost accuracy
  renorm();

  if (ENDL.write_file) 
  {
    write_endl( 1 ); //write the ENDL file
    // do we have elastic scattering for a light target?
    if ( ( ENDL.T == 2 ) && ( ENDL.res_za <= 2004 ) )
    {
      ENDL.outgoing_particle = 10 + ENDL.za_to_yo( ENDL.res_za );
      mirror( );
      write_endl( 1 );
    }
  }
}
// ----------- mf4_table::read_data -----------------
void mf4_table::read_data(mf4_file& inFile, int NE)
// read in the ENDF/B-VI tabulated  data
// NE is the number of incident energies
{
  // handle each incident neutron energy
  for (int iNE=0; iNE < NE; iNE++ )
  {
    one_E_in(inFile);
  }
}
// ----------- mf4_table::one_E_in -----------------
void mf4_table::one_E_in(mf4_file& inFile)
//read in the tabulated ENDF/B-VI data at the current incident neutron energy
{
  int iZERO;    // zero
  
  string linebuff;  // the input line
  string strbuff;   // a substring

  double T;     // temperature
  double E_1;   // incident neutron energy
  int LT;       // test for temperature dependence
  int NR;       // number of regions
  int NP;       // number of data points
      
  getline( inFile, linebuff ); //read next line
  read_ddiiii(&linebuff, &T,    // temperature
	      &E_1,             // incident energy
	      &LT,              // test for temperature dependence
	      &iZERO,           // zero
	      &NR ,             // number of regions
	      &NP);             // number of data points
  
  if( T != 0.0)
  {
    SevereError("mf4_table::one_E_in",
      pastenum("Non-zero temperature for tabulated data ",T));
  }
  if( NR != 1 )
  {
    SevereError("mf4_table::one_E_in",
      "Unwritten code!  Expected single region ");
  }
  if( NP < 1 )
  {
    SevereError("mf4_table::one_E_in",
      pastenum("Negative number of data points ",NP));
  }

  // read the interpolation information for this E_in
  inFile.get_regions(NR, NBT, INT);

  // make a new link for this incident energy
  one_d_table new_link;

  // insert the 1-d list as a link
  insert(end(), new_link);
  mf4_table::iterator this_link = end();
  --this_link;

  // change to MeV
  this_link->E_in() = E_1 * ENDL.eV2MeV;

  // read the data for this E_in
  for( int iNP=0; iNP < NP; iNP++ ) //Loop over NP data pairs
  { 
    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 ) getline( inFile, linebuff); //Read in new line
    for( int i=0+iP*22; i<22+iP*22; i=i+22 ) //Grab the right pair
    strbuff = linebuff.substr(iP*22, 22); //Grab the right pair
    double X,Y; //Real variables we will put the data pair
    read_dd(&strbuff, &X, &Y );
    if ( X==-999. || Y==-999. ) //read_PAIR sends -999. if problem
    {
      SevereError("mf4_table::one_E_in","read_PAIR fucked up!");
    }
    else
    {
      dd_link XYdata(X, Y);
      this_link->insert(this_link->end(), XYdata);
    }
  }
  
}

// ********* for class mixed_mf4 *************
// ----------- mixed_mf4::master -----------------
void mixed_mf4::master( mf4_file& inFile )
// Handle the Legendre data at low incident energies and tables at high
{
  int NE;  // number of incident energies
  int NR;  // number of interpolation regions

  // read the interpolation regions for Legendre data
  inFile.read_line3( &NR, &NE ); 
  inFile.get_regions( NR, NBT, INT );

  // Legendre data
  expand_data( inFile, NE, ENDL.F ); //read/expand data to pointwise

  // read the interpolation regions for the table
  inFile.read_line3( &NR, &NE );
  inFile.get_regions(NR, table_data.NBT, table_data.INT);

  table_data.read_data( inFile, NE ); //read in the tabular data

  //the last Legendre incident energy may be the same as the first tabular one
  mixed_mf4::iterator last_Legendre = end( );
  --last_Legendre;
  double last_L_Ein = last_Legendre->E_in( );
  
  mf4_table::iterator first_table = table_data.begin( );
  double first_T_Ein = first_table->E_in( );

  // get the local energy epsilon (smallest dE can see in output file)
  double local_eps = ENDL_EPSILON( last_L_Ein );

  if( first_T_Ein < last_L_Ein + local_eps )
  {
    // how close is the previous Legendre E_in?
    mixed_mf4::iterator prev_link = last_Legendre;
    --prev_link;
    if( prev_link->E_in( ) <= last_L_Ein - 2*local_eps )
    {
      last_Legendre->E_in( ) -= local_eps;
    }
    else
    {
      // just get rid of this link
      erase( last_Legendre );
    }

    // how close is the next tabular E_in?
    mf4_table::iterator next_link = first_table;
    ++next_link;
    if( next_link->E_in( ) >= first_T_Ein + 2*local_eps )
    {
      first_table->E_in( ) += local_eps;
    }
    else
    {
      // just get rid of this link
      table_data.erase( first_table );
    }
  }
  
  unique();
  widen_jumps();
  table_data.widen_jumps();
  
  // renormalize the probability distributions since, in the log-lin
  // to lin-lin interpolation, we may have lost accuracy
  table_data.renorm();

  if ( ENDL.write_file )
  {
    fstream endl_file;
    // set up and print header for the ENDL file
    set_up_endl( 1, endl_file );

    // print the Legendre data
    out_data( 1, endl_file, 0.0 );

    // print the table
    table_data.out_data( 1, endl_file, 0.0 );

    // end the file and close it
    close_file( endl_file );

    // do we have elastic scattering for a light target?
    if ( ( ENDL.T == 2 ) && ( ENDL.res_za <= 2004 ) )
    {
      Unimplemented("mixed_mf4::master", " mixed_mf4 data for light targets");
    }
  }
}

// ********* for class gen_mf4 *************
// ----------- gen_mf4::master -----------------
void gen_mf4::master( mf6_file& inFile )
// interpret the ENDF/B-VI data
{
  int NR;  // number of interpolation regions
  int NE;  // number of incident energies
  inFile.mf4_line1(&NR, &NE);
  inFile.get_regions(NR, NBT, INT);
  if ( (NR > 1) || (INT[0] != 2) )
  {
    Unimplemented("gen_mf4::master",
        "Implement interpolation between incident energies for gen_mf4");
  }

  // loop over the incident energies
  int LANG;  // type of data
  int NW;    // number of data entries
  int NL;    // Legendre order or number of cosines
  double e_in; // incident energy
  for ( int count = 0; count < NE; ++count)
  {
    inFile.mf4_line2(&e_in, &LANG, &NW, &NL);
    if ( LANG == 0 )
    {
      // Legendre data
      one_legendre( inFile, e_in, NL );
    }
    else if ( LANG == 12 )
    {
      // linear-linear tabular data
      one_table( inFile, e_in, NL );
    }
    else if ( LANG == 14 )
    {
      // linear-log tabular data
      one_table( inFile, e_in, NL );

      // construct NBT and INT vectors for lin-log interpolation
      gen_mf4::iterator this_link = end();
      --this_link;
      this_link->NBT.push_back( NL );
      this_link->INT.push_back( 4 );
      this_link->expand_interp( ENDL.Max_E_out );
    }
    else
    {
    Unimplemented("gen_mf4::master",pastenum(" Implement LANG = ",LANG));
    }
  }
  write_endl(1);
}
// ----------- gen_mf4::one_legendre -----------------
void gen_mf4::one_legendre( mf6_file& inFile, double e_in, int NL )
// interpret Legendre data for 1 incident energy
{
  // make a 1-d Legendre list
  one_d_Legendre Legendre;

  // set the Legendre coefficients
  string linebuff;  // the input line
  string strbuff;   // a substring
  Legendre.coef.get_space( NL + 1 );
  Legendre.coef.at( 0 ) = 1.0;
  for ( int iNR=0; iNR < NL; ++iNR )
  {
    double next_coef;
    int iR = iNR % 6;  //There are 6 numbers per line
    if ( iR == 0 ) getline( inFile, linebuff); //Read in new line
    strbuff = linebuff.substr(iR*11, 11);
    read_d( &strbuff, &next_coef );
    Legendre.coef.at( iNR + 1 ) = next_coef;
  }
  // expand the model
  Legendre.expand();

  // *** copy the Legendre list into a table ***
  // make a new link for this incident energy
  one_d_table new_link;

  // insert the 1-d list as a link
  insert(end(), new_link);
  gen_mf4::iterator this_link = end();
  --this_link;

  // change to MeV
  this_link->E_in() = e_in * ENDL.eV2MeV;

  // go through the Legendre list
  dd_link XYdata;
  for( one_d_Legendre::iterator from_link = Legendre.begin();
    from_link != Legendre.end(); ++from_link )
  {
    XYdata.x = from_link->x;
    XYdata.y = from_link->y;
    this_link->insert(this_link->end(), XYdata);
  }
}
// ----------- gen_mf4::one_table -----------------
void gen_mf4::one_table( mf6_file& inFile, double e_in, int NL )
// copy a table for 1 incident energy
{
  // make a new link for this incident energy
  one_d_table new_link;

  // insert the 1-d list as a link
  insert(end(), new_link);
  gen_mf4::iterator this_link = end();
  --this_link;

  // change to MeV
  this_link->E_in() = e_in * ENDL.eV2MeV;

  // go through the data
  string linebuff;  // the input line
  string strbuff;   // a substring
  dd_link XYdata;
  for( int count = 0; count < NL; ++count )
  {
    int iP=count % 3;  //There are 3 pairs per line
    if ( iP == 0 ) getline( inFile, linebuff); //Read in new line
    strbuff = linebuff.substr(iP*22, 22); //Grab the right pair
    read_dd(&strbuff, &XYdata.x, &XYdata.y);
    this_link->insert(this_link->end(), XYdata);
  }
}
