/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1882 $
 * $Date: 2006-05-24 15:20:29 -0700 (Wed, 24 May 2006) $
 * $Author: dbrown $
 * $Id: mf5classes.cpp 1882 2006-05-24 22:20:29Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the classes used to translate the MF=5 data

#include "mf5classes.hpp"
#include "global_params.hpp"

extern ENDLClass ENDL;
extern GlobalParameterClass Global;

// ********* for class mf5_table *************
// ----------- mf5_table::read_data -----------------
void mf5_table::read_data( mf5_file& inFile, int NR, int NP )
// read in the ENDF/B-VI tabulated  data
// NR is the number of interpolation regions
// NP is the number of (energy, weight) pairs
{
  duplicate_Ein = false;

  if( ENDL.S == 7 )
  {
    // there is no weight information
  }
  else
  {
    // read the weights for this model
    inFile.get_regions( NR, _weight.NBT, _weight.INT );
    // read the weight with energy in MeV
    _weight.read_data( inFile, NP, ENDL.eV2MeV, 1.0 );
    _weight.expand_interp( ENDL.Max_E_in );
  }

  int iZERO, NE;
  double ZERO;

  inFile.read_line2( &ZERO, &iZERO, &NR, &NE );

  // read the interpolation information
  inFile.get_regions(NR, NBT, INT);
  // read the data
  for (int iNE=0; iNE < NE; iNE++ )
  {
    one_E_in(inFile);
  }
  // record the types of interpolation between 1-d lists
  set_interp();
  if((NR > 1) || (INT[0] != 2))
  {
    table_interp( );
    // remove any extra inserted points
    thinit( );
  }
}

// ----------- mf5_table::one_E_in -----------------
void mf5_table::one_E_in(mf5_file& inFile)
//read in the tabulated ENDF/B-VI data at the current incident neutron energy
{
  double ZERO;  // zero
  int iZERO;    // zero
  
  string linebuff;  // the input line
  string strbuff;   // a substring

  double E_1;   // incident neutron energy
  int NR;       // number of regions
  int NP;       // number of data points
      
  getline( inFile, linebuff ); //read next line
  read_ddiiii(&linebuff, &ZERO, // zero
	      &E_1,             // incident energy
	      &iZERO,           // zero
	      &iZERO,           // zero
	      &NR ,             // number of regions
	      &NP);             // number of data points
  
  if( NP < 1 )
    {
      SevereError("mf5_table::one_E_in",
        pastenum("Negative number of data points ",NP)+"in mf5_table class.");
    }

  // make a new link for this incident energy
  //  one_d_distrib new_link(NONE);
  one_d_table new_link;

  // insert the 1-d list as a link
  insert(end(), new_link);
  two_d_iterator this_link = end();
  --this_link;

  // change to MeV
  this_link->E_in() = E_1 * ENDL.eV2MeV;

  // read the interpolation information for this E_in
  inFile.get_regions(NR, this_link->NBT, this_link->INT);

  // read the data for this E_in; check the order of the energies
  double prev_x = -1.0;
  for( int iNP=0; iNP < NP; iNP++ ) //Loop over NP data pairs
  { 
    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 )
    {
      getline( inFile, linebuff); //Read in new line
    }
    strbuff = linebuff.substr(iP*22, 22); //Grab the right pair
    double X,Y; //Real variables we will put the data pair
    read_dd(&strbuff, &X, &Y );
    if ( X==-999. || Y==-999. ) //read_PAIR sends -999. if problem
    {
      SevereError("mf5_table::one_E_in","read_PAIR fucked up!");
    }
    else
    {
      X *= ENDL.eV2MeV;
      Y /= ENDL.eV2MeV;
      if( X < prev_x ){
        SevereError( "mf5_table::one_E_in",
	  	   pastenum( "energy out of order: ", X ) );
      }
      else if( X == prev_x ){
        duplicate_Ein = true;
        Warning( "mf5_table::one_E_in",
	  	    pastenum( "duplicate outgoing energy: ", prev_x ) );
      }
      dd_link XYdata( X, Y );
      this_link->insert(this_link->end(), XYdata);
      prev_x = X;
    }
  }
  // expand the interpolation for this incident energy
  this_link->expand_interp( ENDL.Max_E_out );
  if (this_link->needs_widening())  this_link->widen_jumps( 0.0 );
  this_link->renorm( );
  if( ENDL.S != 7 )
  {
    this_link->weight = _weight.evaluate( this_link->E_in() );
  }
}
// ----------- mf5_table::table_interp -----------------
void mf5_table::table_interp( )
// Interpolate between tables according to the INT array
// This coding is based on two_d_Legendre::expand_interp
{
  if(INT[0] == -7)
  {
    SevereError("mf5_table::table_interp","table_interp called twice");
  }

  // construct the equiprobable bins
  for( mf5_table::iterator this_link = begin();
    this_link != end(); ++this_link )
  {
    this_link->get_bins();
  }
  // fill in to get good linear interpolation
  thicken();

  // flag to ensure that we don't do this twice on the same data
  INT[0] = -7;
}

// ********* for class two_d_evap *************
// ----------- two_d_evap::two_d_evap ------
two_d_evap::two_d_evap()
// constructor
{
  UU = -10.0;  //  negative to ensure error message if not set
  Max_energy = ENDL.Max_E_in;
}
// ----------- two_d_evap::two_d_evap ------
two_d_evap::two_d_evap(double U)
// constructor
{
  UU = U*ENDL.eV2MeV;  // convert to MeV
  Max_energy = ENDL.Max_E_in;
}
// ----------- two_d_evap::~two_d_evap ------
two_d_evap::~two_d_evap()
// destructor
{
}
// ----------- two_d_evap::make_list ------
void two_d_evap::make_list()
// expand the data into a list
{
  const double EPS = Global.Value( "mf5_tol" );  // tolerance for mf5 data

  dd_list::iterator theta_ptr = _theta.begin();
  bool pass_maxE = false;  // does the data exceed that maximum energy?

  // ignore this data if the threshold is above Max_energy
  if (theta_ptr->x > Max_energy)
  {
    ENDL.write_file = false;
  }
  else
  {
    // make a 1-d list for each (incident energy, theta) pair
    for( ; ( (theta_ptr!= _theta.end()) && !pass_maxE );
       ++theta_ptr)
    {
      // check against the maximum energy
      if(theta_ptr->x > Max_energy)
      {
        pass_maxE = true;
      }
      // insert and expand the model
      insert(end(), theta_ptr->x, theta_ptr->y);

      // we may be done
      if(theta_ptr->x > (1 - EPS)*Max_energy)
      {
        break;
      }
    }
    // record the types of interpolation between 1-d lists
    set_interp();

    // do we want to replace the last list with one at Max_energy?
    if( pass_maxE )
    {
      chop_highE( Max_energy );
    }
    // fill in to get good linear interpolation
    thicken();

    // normalize the probability
    renorm();
  }
}
// ----------- two_d_evap::read_data -----------------
void two_d_evap::read_data(mf5_file& inFile, int NR, int NP)
// NR is the number of interpolation regions
// NP is the number of (energy, weight) pairs
{
  duplicate_Ein = false;

  string linebuff;
  string strbuff;
  dd_link XYdata;  // for links in our lists

  // read the interpolation information
  inFile.get_regions(NR, _weight.NBT, _weight.INT);
  // read the weight with energy in MeV
  _weight.read_data( inFile, NP, ENDL.eV2MeV, 1.0 );
  _weight.expand_interp( ENDL.Max_E_in );

  // for the theta data
  double zero;
  int izero;
  int nr, np;
  inFile.read_line2(&zero, &izero, &nr, &np);

  // read the interpolation information
  inFile.get_regions(nr, NBT, INT);
  // read the parameters for this model
  // and check for duplication of energies
  double prev_x = -1.0;
  for( int iNP=0; iNP < np; iNP++ ) //Loop over NP data pairs
  { 
    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 )
    {
      getline( inFile, linebuff); //Read in new line
    }
    strbuff = linebuff.substr(iP*22, 22); //Grab the right pair
    read_dd(&strbuff, &XYdata.x, &XYdata.y );
    // convert to MeV
    XYdata.x *= ENDL.eV2MeV;
    XYdata.y *= ENDL.eV2MeV;
    if( XYdata.x < prev_x ){
      SevereError( "two_d_evap::read_data",
		   pastenum( "energy out of order: ", XYdata.x ) );
    }
    else if( XYdata.x == prev_x ){
      duplicate_Ein = true;
      Warning( "two_d_evap::read_data",
		   pastenum( "duplicate energy: ", prev_x ) );
    }
    _theta.insert(_theta.end(), XYdata);
    prev_x = XYdata.x;
  }
}
// ----------- two_d_evap::insert -----------------
void two_d_evap::insert(two_d_iterator where, double e_in, double Theta)
// insert a distribution for this e_in
{
  one_d_evap new_link;

  // insert the 1-d list as a link
  two_d_list<one_d_evap>::insert(where, new_link);

  // expand the 1-d model
  two_d_iterator this_link = where;
  --this_link;
  this_link->weight = _weight.evaluate( e_in );
  this_link->initiate(e_in, Theta, UU);
}
// ********* for class two_d_Maxwell *************
// ----------- two_d_Maxwell::two_d_Maxwell ------
two_d_Maxwell::two_d_Maxwell()
// constructor
{
  UU = -10.0;  //  negative to ensure error message if not set
  Max_energy = ENDL.Max_E_in;
}
// ----------- two_d_Maxwell::two_d_Maxwell ------
two_d_Maxwell::two_d_Maxwell(double U)
// constructor
{
  UU = U*ENDL.eV2MeV;  // convert to MeV
  Max_energy = ENDL.Max_E_in;
}
// ----------- two_d_Maxwell::~two_d_Maxwell ------
two_d_Maxwell::~two_d_Maxwell()
// destructor
{
}
// ----------- two_d_Maxwell::make_list ------
void two_d_Maxwell::make_list()
// expand the data into a list
{
  const double EPS = Global.Value( "mf5_tol" );  // tolerance for mf5 data

  dd_list::iterator theta_ptr = _theta.begin();
  bool pass_maxE = false;  // does the data exceed that maximum energy?

  // ignore this data if the threshold is above Max_energy
  if (theta_ptr->x > Max_energy)
  {
    ENDL.write_file = false;
  }
  else
  {
    // make a 1-d list for each (incident energy, theta) pair
    for( ; ( (theta_ptr!= _theta.end()) && !pass_maxE );
       ++theta_ptr)
    {
      // check against the maximum energy
      if(theta_ptr->x > Max_energy)
      {
        pass_maxE = true;
      }
      // insert and expand the model
      insert(end(), theta_ptr->x, theta_ptr->y);

      // we may be done
      if(theta_ptr->x > (1 - EPS)*Max_energy)
      {
        break;
      }
    }
    // record the types of interpolation between 1-d lists
    set_interp();

    // do we want to replace the last list with one at Max_energy?
    if( pass_maxE )
    {
      chop_highE( Max_energy );
    }
    // fill in to get good linear interpolation
    thicken();

    // normalize the probability
    renorm();
  }
}
// ----------- two_d_Maxwell::read_data -----------------
void two_d_Maxwell::read_data(mf5_file& inFile, int NR, int NP)
// NR is the number of interpolation regions
// NP is the number of (energy, weight) pairs
{
  duplicate_Ein = false;

  string linebuff;
  string strbuff;
  dd_link XYdata;  // for links in our lists

  // read the interpolation information
  inFile.get_regions(NR, _weight.NBT, _weight.INT);
  // read the weights for this model with energy in MeV
  _weight.read_data( inFile, NP, ENDL.eV2MeV, 1.0 );
  _weight.expand_interp( ENDL.Max_E_in );

  // for the theta data
  double zero;
  int izero;
  int nr, np;
  inFile.read_line2(&zero, &izero, &nr, &np);
  double prev_x = -1.0;
  // read the interpolation information
  inFile.get_regions(nr, NBT, INT);
  // read the parameters for this model

  for( int iNP=0; iNP < np; iNP++ ) //Loop over NP data pairs
  { 
    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 )
    {
      getline( inFile, linebuff); //Read in new line
    }
    strbuff = linebuff.substr(iP*22, 22); //Grab the right pair
    read_dd(&strbuff, &XYdata.x, &XYdata.y );
    // convert to MeV
    XYdata.x *= ENDL.eV2MeV;
    XYdata.y *= ENDL.eV2MeV;
    if( XYdata.x < prev_x ){
      SevereError( "two_d_Maxwell::read_data",
		   pastenum( "energy out of order: ", XYdata.x ) );
    }
    else if( XYdata.x == prev_x ){
      duplicate_Ein = true;
      Warning( "two_d_Maxwell::read_data",
		   pastenum( "duplicate energy: ", prev_x ) );
    }
    _theta.insert(_theta.end(), XYdata);
    prev_x = XYdata.x;
  }
}
// ----------- two_d_Maxwell::insert -----------------
void two_d_Maxwell::insert(two_d_iterator where, double e_in, double Theta)
// insert a distribution for this e_in
{
  one_d_Maxwell new_link;

  // insert the 1-d list as a link
  two_d_list<one_d_Maxwell>::insert(where, new_link);

  // expand the 1-d model
  two_d_iterator this_link = where;
  --this_link;
  this_link->initiate(e_in, Theta, UU);
  this_link->weight = _weight.evaluate( this_link->E_in() );
}
// ********* for class two_d_Watt *************
// ----------- two_d_Watt::two_d_Watt ------
two_d_Watt::two_d_Watt()
// constructor
{
  UU = -10.0;  //  negative to ensure error message if not set
  Max_energy = ENDL.Max_E_in;
}
// ----------- two_d_Watt::two_d_Watt ------
two_d_Watt::two_d_Watt(double U)
// constructor
{
  UU = U*ENDL.eV2MeV;  // convert to MeV
  Max_energy = ENDL.Max_E_in;
}
// ----------- two_d_Watt::~two_d_Watt ------
two_d_Watt::~two_d_Watt()
// destructor
{
}
// ----------- two_d_Watt::make_list ------
void two_d_Watt::make_list()
// expand the data into a list
{
  const double EPS = Global.Value( "mf5_tol" );  // tolerance for mf5 data

  dd_list::iterator a_ptr = Watt_a.begin();
  dd_list::iterator b_ptr = Watt_b.begin();
  bool pass_maxE = false;  // does the data exceed that maximum energy?

  // ignore this data if the threshold is above Max_energy
  if (a_ptr->x > Max_energy)
  {
    ENDL.write_file = false;
  }
  else
  {
    // Set up a 1-d list for each (incident energy, a) pair
    for( ; ( (a_ptr != Watt_a.end()) && !pass_maxE ); ++a_ptr )
    {
      // check against the maximum energy
      if(a_ptr->x > Max_energy)
      {
        pass_maxE = true;
      }
      // insert a link with these E_in and a values
      insert( end(), a_ptr->x, a_ptr->y );

      // we may be done
      if(a_ptr->x > (1 - EPS)*Max_energy)
      {
        break;
      }
    }
    // record the types of interpolation for a between 1-d lists
    set_interp();

    // set the b values and expand the lists
    set_b();

    // do we want to replace the last list with one at Max_energy?
    if( pass_maxE )
    {
      chop_highE( Max_energy );
    }
    // fill in to get good linear interpolation
    thicken();

    // normalize the probability
    renorm();
  }
}
// ----------- two_d_Watt::read_data -----------------
void two_d_Watt::read_data(mf5_file& inFile, int NR, int NP)
// NR is the number of interpolation regions
// NP is the number of (energy, weight) pairs
{
  duplicate_Ein = false;

  string linebuff;
  string strbuff;
  dd_link XYdata;  // for links in our lists

  // read the interpolation information
  inFile.get_regions(NR, _weight.NBT, _weight.INT);

  // read the weights for this model with energy in MeV
  _weight.read_data( inFile, NP, ENDL.eV2MeV, 1.0 );
  _weight.expand_interp( ENDL.Max_E_in );

  // for the Watt_a data
  double zero;
  int izero;
  int nr, np;
  inFile.read_line2(&zero, &izero, &nr, &np);

  // read the interpolation information
  inFile.get_regions(nr, NBT, INT);
  // read the parameters for this model
  // and check for duplication of energies
  double prev_x = -1.0;
  for( int iNP=0; iNP < np; iNP++ ) //Loop over NP data pairs
  { 
    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 )
    {
      getline( inFile, linebuff); //Read in new line
    }
    strbuff = linebuff.substr(iP*22, 22); //Grab the right pair
    read_dd(&strbuff, &XYdata.x, &XYdata.y );
    // convert to MeV
    XYdata.x *= ENDL.eV2MeV;
    XYdata.y *= ENDL.eV2MeV;
    if( XYdata.x < prev_x ){
      SevereError( "two_d_Watt::read_data",
		   pastenum( "energy out of order: ", XYdata.x ) );
    }
    else if( XYdata.x == prev_x ){
      duplicate_Ein = true;
      Warning( "two_d_Watt::read_data",
		   pastenum( "duplicate energy: ", prev_x ) );
    }
    Watt_a.insert(Watt_a.end(), XYdata);
    prev_x = XYdata.x;
  }

  // for the Watt_b data
  inFile.read_line2(&zero, &izero, &nr, &np);

  // read the interpolation information
  inFile.get_regions(nr, Watt_b.NBT, Watt_b.INT);
  // read the parameters for this model

  for( int iNP=0; iNP < np; iNP++ ) //Loop over NP data pairs
  { 
    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 )
    {
      getline( inFile, linebuff); //Read in new line
    }
    strbuff = linebuff.substr(iP*22, 22); //Grab the right pair
    read_dd(&strbuff, &XYdata.x, &XYdata.y );
    // convert to MeV
    XYdata.x *= ENDL.eV2MeV;
    XYdata.y /= ENDL.eV2MeV;
    Watt_b.insert(Watt_b.end(), XYdata);
  }
  // ensure that the Watt_a and Watt_b lists have the same E_in values
  fill_in_lists(Watt_a, Watt_b);
}
// ----------- two_d_Watt::insert -----------------
void two_d_Watt::insert( two_d_iterator where, double e_in, double a )
// insert a distribution for this e_in
{
  one_d_Watt new_link;

  // insert the 1-d list as a link
  two_d_list<one_d_Watt>::insert(where, new_link);

  // set the U and a parameters for this e_in
  two_d_iterator this_link = where;
  --this_link;
  this_link->set_EUa( e_in, UU, a );
  this_link->weight = _weight.evaluate( this_link->E_in() );
}
// ----------- two_d_Watt::set_b -----------------
void two_d_Watt::set_b( )
// set the b values and expand the lists
{
  const double EPS = Global.Value( "mf5_tol" );  // tolerance for mf5 data
  dd_list::iterator a_ptr = Watt_a.begin();
  dd_list::iterator b_ptr = Watt_b.begin();

  double E_a = a_ptr->x;
  double E_b = b_ptr->x;
  if( abs( E_a - E_b ) > EPS*E_a )
  {
    SevereError("two_d_Watt::set_b",
        "Watt a and b lists should start at the same energies.");
  }

  int n_NBT = 0;  // for counting the NBT and INT arrays
  // set up the first list
  two_d_Watt::iterator list_ptr = begin();
  list_ptr->same_b( b_ptr->y, Watt_b.INT[ n_NBT ] );

  // the next a and b terms
  ++a_ptr;
  ++b_ptr;
  E_a = a_ptr->x;
  E_b = b_ptr->x;
  int b_count = 1;  // count the b entries

  // loop through the a and b lists
  for( ++list_ptr; list_ptr != end(); )
  {
    // we may have the same energies
    if( abs( E_a - E_b ) < EPS*E_a )
    {
      list_ptr->same_b( b_ptr->y, Watt_b.INT[ n_NBT ] );

      // the next a and b terms
      ++a_ptr;
      ++b_ptr;
      E_a = a_ptr->x;
      E_b = b_ptr->x;
      ++list_ptr;  // next list
      ++b_count;
      if( b_count >= Watt_b.NBT[ n_NBT ] )
      {
        ++n_NBT;
      }
    }
    else if( E_b < E_a )
    {
      // set up a new 1-d list at energy E_b
      one_d_Watt new_link;

      // insert the 1-d list as a link
      two_d_list<one_d_Watt>::insert(list_ptr, new_link);
      two_d_iterator this_link = list_ptr;
      --this_link;  // points to the new list

      // interpolate to get the intermediate a and make a new list
      dd_list::iterator prev_a = a_ptr;
      --prev_a;
      double a_value = Watt_a.interp( E_b, *prev_a, *a_ptr,
        list_ptr->interp_type );
      this_link->weight = _weight.evaluate( this_link->E_in() );
      this_link->intermediate_b( E_b, UU, a_value,
        list_ptr->interp_type, b_ptr->y, Watt_b.INT[ n_NBT ] );

      // the next b term
      ++b_ptr;
      E_b = b_ptr->x;
      ++b_count;
      if( b_count >= Watt_b.NBT[ n_NBT ] )
      {
        ++n_NBT;
      }
    }
    else // E_a < E_b
    {
      // interpolate to get the intermediate b
      dd_list::iterator prev_b = b_ptr;
      --prev_b;
      double b_value = Watt_b.interp( E_a, *prev_b, *b_ptr,
        Watt_b.INT[ n_NBT ] );
      list_ptr->same_b( b_value, Watt_b.INT[ n_NBT ] );

      // the next a term
      ++a_ptr;
      E_a = a_ptr->x;
      ++list_ptr;  // next list
    }
  }
}
// ********* for class two_d_Madland *************
// ----------- two_d_Madland::two_d_Madland ------
two_d_Madland::two_d_Madland()
// constructor
{
  EFL_ = -10.0;  //  negative to ensure error message if not set
  EFH_ = -10.0;
  Max_energy = ENDL.Max_E_in;
}
// ----------- two_d_Madland::~two_d_Madland ------
two_d_Madland::~two_d_Madland()
// destructor
{
}
// ----------- two_d_Madland::make_list ------
void two_d_Madland::make_list()
// expand the data into a list
{
  const double EPS = Global.Value( "mf5_tol" );  // tolerance for mf5 data

  dd_list::iterator TM_ptr = Madland_TM.begin();
  bool pass_maxE = false;  // does the data exceed that maximum energy?

  // ignore this data if the threshold is above Max_energy
  if (TM_ptr->x > Max_energy)
  {
    ENDL.write_file = false;
  }
  else
  {
    // Make a 1-d list for each (incident energy, TM) pair.
    for( ; ( (TM_ptr != Madland_TM.end()) && !pass_maxE );
       ++TM_ptr)
    {
      // check against the maximum energy
      if(TM_ptr->x > Max_energy)
      {
        pass_maxE = true;
      }
      // insert and expand the model
      insert(end(), TM_ptr->x, TM_ptr->y);

      // we may be done
      if(TM_ptr->x > (1 - EPS)*Max_energy)
      {
        break;
      }
    }
  }
  // record the types of interpolation between 1-d lists
  set_interp();

  // do we want to replace the last list with one at Max_energy?
  if( pass_maxE )
  {
    chop_highE( Max_energy );
  }
  // fill in to get good linear interpolation
  thicken();
}
// ----------- two_d_Madland::read_data -----------------
void two_d_Madland::read_data(mf5_file& inFile, int NR, int NP)
// NR is the number of interpolation regions
// NP is the number of (energy, weight) pairs
{
  duplicate_Ein = false;

  string linebuff;
  string strbuff;
  dd_link XYdata;  // for links in our lists

  // read the interpolation information
  inFile.get_regions(NR, _weight.NBT, _weight.INT);

  // read the weights for this model with energy in MeV
  _weight.read_data( inFile, NP, ENDL.eV2MeV, 1.0 );
  _weight.expand_interp( ENDL.Max_E_in );

  // for the Madland_TM data
  int nr, ne;
  double EFL, EFH;  // average energies of fission fragments
  inFile.read_Madland( &EFL, &EFH, &nr, &ne);
  set_EF( EFL, EFH );  // save them as MeV

  // read the interpolation information
  inFile.get_regions(nr, NBT, INT);
  // read the parameters for this model
  // and check for duplication of energies
  double prev_x = -1.0;
  for( int iNE=0; iNE < ne; iNE++ ) //Loop over NE data pairs
  { 
    int iP=iNE % 3;  //There are 3 pairs per line
    if ( iP == 0 )
    {
      getline( inFile, linebuff); //Read in new line
    }
    strbuff = linebuff.substr(iP*22, 22); //Grab the right pair
    read_dd(&strbuff, &XYdata.x, &XYdata.y );
    // convert to MeV
    XYdata.x *= ENDL.eV2MeV;
    XYdata.y *= ENDL.eV2MeV;
    if( XYdata.x < prev_x ){
      SevereError( "two_d_Madland::read_data",
		   pastenum( "energy out of order: ", XYdata.x ) );
    }
    else if( XYdata.x == prev_x ){
      duplicate_Ein = true;
      Warning( "two_d_Madland::read_data",
		   pastenum( "duplicate energy: ", prev_x ) );
    }
    Madland_TM.insert(Madland_TM.end(), XYdata);
    prev_x = XYdata.x;
  }
}
// ----------- two_d_Madland::insert -----------------
void two_d_Madland::insert(two_d_iterator where, double e_in, double TM)
// insert a distribution for this e_in
{
  one_d_Madland new_link;

  // insert the 1-d list as a link
  two_d_list<one_d_Madland>::insert(where, new_link);

  // expand the 1-d model
  two_d_iterator this_link = where;
  --this_link;
  this_link->initiate( e_in, EFL_, EFH_, TM );
  this_link->weight = _weight.evaluate( this_link->E_in() );
}
// ********* for class add_them_up *************
// ----------- add_them_up::constructor-----------
add_them_up::  add_them_up( int NK )
{
  NK_ = NK;

  // set up space for the models
  table_vec = new mf5_table[ NK ];
  evap_vec = new two_d_evap[ NK ];
  Maxwell_vec = new two_d_Maxwell[ NK ];
  Watt_vec = new two_d_Watt[ NK ];
  Madland_vec = new two_d_Madland[ NK ];

  num_table = 0;
  num_evap = 0;
  num_Maxwell = 0;
  num_Watt = 0;
  num_Madland = 0;

  // debug_ = true;
  debug_ = false;
}
// ----------- add_them_up::destructor-----------
add_them_up::~add_them_up( )
{
  if( num_table > 0 )
  {
    delete [] table_vec;
  }
  if( num_evap > 0 )
  {
    delete [] evap_vec;
  }
  if( num_Maxwell > 0 )
  {
    delete [] Maxwell_vec;
  }
  if( num_Watt > 0 )
  {
    delete [] Watt_vec;
  }
  if( num_Madland > 0 )
  {
    delete [] Madland_vec;
  }
}
// ----------- add_them_up::manager-----------
// manage NK output models
void add_them_up::manager( mf5_file& inFile )
{
  // expand the models
  for( int j = 0; j < NK_; ++j )
  {
    one_model( inFile );
  }

  // ensure that all lists have common incident energies
  common_Ein( );

  // do a weighted add
  wt_add( );

  // check for total weight != 1
  check_total_wt( );

  // widen any post-add jumps
  widen_jumps( );

  // make sure that the probabilities are normalized
  renorm( );
  if (ENDL.write_file)
  {
    write_endl( 4 ); //write the ENDL file
  }
}
// ----------- add_them_up::one_model -----------
void add_them_up::one_model( mf5_file& inFile )
// Handles one output model
{
  double U;
  int LF, NR, NP;

  // read the first line for the model
  inFile.read_line2(&U, &LF, &NR, &NP);

  if ( LF == 1  ) //Arbitrary tabulated function
  {
    // data for an mf5_table
    table_vec[num_table].read_data( inFile, NR, NP );  //read in the data
    if( table_vec[num_table].duplicate_Ein )
        table_vec[num_table].widen_jumps( );
    // Insert 1-d lists where the weight has jumps
    table_vec[num_table].weight_jumps();
    ++num_table;
  }
  else if ( LF == 5  ) //General evaporation Spectrum
  {
    Warning("add_them_up::one_model","Need new code for LF = 5");
    return;
  }
  else if ( LF == 7  )  // Simple fission spectrum (Maxwellian)
  {
    Maxwell_vec[num_Maxwell].set_U( U );
    Maxwell_vec[num_Maxwell].read_data( inFile, NR, NP );    //read in the data
    Maxwell_vec[num_Maxwell].make_list();   // expand the data
    if( Maxwell_vec[num_Maxwell].duplicate_Ein )
        Maxwell_vec[num_Maxwell].widen_jumps( );
    // Insert 1-d lists where the weight has jumps
    Maxwell_vec[num_Maxwell].weight_jumps();
    ++num_Maxwell;
  }
  else if ( LF == 9  )  // Evaporation spectrum
  {
    evap_vec[num_evap].set_U( U );
    evap_vec[num_evap].read_data( inFile, NR, NP );    //read in the data
    evap_vec[num_evap].make_list();   // expand the data
    if( evap_vec[num_evap].duplicate_Ein )
        evap_vec[num_evap].widen_jumps( );
    // Insert 1-d lists where the weight has jumps
    evap_vec[num_evap].weight_jumps();
    ++num_evap;
  }
  else if ( LF == 11 )  //Energy-Dependent Watt spectrum
  {
    Watt_vec[num_Watt].set_U( U );
    Watt_vec[num_Watt].read_data( inFile, NR, NP );    //read in the data
    Watt_vec[num_Watt].make_list();   // expand the data
    if( Watt_vec[num_Watt].duplicate_Ein )
        Watt_vec[num_Watt].widen_jumps( );
    // Insert 1-d lists where the weight has jumps
    Watt_vec[num_Watt].weight_jumps();
    ++num_Watt;
  }
  else if ( LF == 12 ) //Energy dep fission neutron spectrum (Madland & Nix)
  {
    Madland_vec[num_Madland].read_data( inFile, NR, NP );    //read in the data
    Madland_vec[num_Madland].make_list();   // expand the data
    if( Madland_vec[num_Madland].duplicate_Ein )
        Madland_vec[num_Madland].widen_jumps( );
    // Insert 1-d lists where the weight has jumps
    Madland_vec[num_Madland].weight_jumps();
    ++num_Madland;
  }
  else
  {
    SevereError("add_them_up::one_model",
      pastenum("There is no model for LF = ",LF)+" for MF = 5 data.");
  }
}
// ----------- add_them_up::common_Ein -----------
void add_them_up::common_Ein( )
// ensure that all lists have common incident energies
{
  int j;
  list< double > E_incident;

  // first collect all of the incident energies
  for( j = 0; j < num_table; ++j )
  {
    table_vec[j].collect_Ein( E_incident );
  }
  for( j = 0; j < num_evap; ++j )
  {
    evap_vec[j].collect_Ein( E_incident );
  }
  for( j = 0; j < num_Maxwell; ++j )
  {
    Maxwell_vec[j].collect_Ein( E_incident );
  }
  for( j = 0; j < num_Watt; ++j )
  {
    Watt_vec[j].collect_Ein( E_incident );
  }
  for( j = 0; j < num_Madland; ++j )
  {
    Madland_vec[j].collect_Ein( E_incident );
  }

  // sort and delete duplicates
  E_incident.sort( );
  E_incident.unique( );

  // then do the fills
  for( j = 0; j < num_table; ++j )
  {
    table_vec[j].fill_in_list( E_incident );
  }
  for( j = 0; j < num_evap; ++j )
  {
    evap_vec[j].fill_in_list( E_incident );
  }
  for( j = 0; j < num_Maxwell; ++j )
  {
    Maxwell_vec[j].fill_in_list( E_incident );
  }
  for( j = 0; j < num_Watt; ++j )
  {
    Watt_vec[j].fill_in_list( E_incident );
  }
  for( j = 0; j < num_Madland; ++j )
  {
    Madland_vec[j].fill_in_list( E_incident );
  }
}
// ----------- add_them_up::wt_add -----------
void add_them_up::wt_add( )
// Adds the next list to the weighted sum
{
  int j;
  bool started = false;

   // add up the lists
  for( j = 0; j < num_table; ++j )
  {
    if( started )
    {
      sum_lists( *this, table_vec[j] );
    }
    else
    {
      copy_data( table_vec[j] );
      started = true;
      use_weight( );
    }
  }

  for( j = 0; j < num_evap; ++j )
  {
    if( started )
    {
      sum_lists( *this, evap_vec[j] );
    }
    else
    {
      copy_data( evap_vec[j] );
      started = true;
      use_weight( );
    }
  }

  for( j = 0; j < num_Maxwell; ++j )
  {
    if( started )
    {
      sum_lists( *this, Maxwell_vec[j] );
    }
    else
    {
      copy_data( Maxwell_vec[j] );
      started = true;
      use_weight( );
    }
  }

  for( j = 0; j < num_Watt; ++j )
  {
    if( started )
    {
      sum_lists( *this, Watt_vec[j] );
    }
    else
    {
      copy_data( Watt_vec[j] );
      started = true;
      use_weight( );
    }
  }

  for( j = 0; j < num_Madland; ++j )
  {
    if( started )
    {
      sum_lists( *this, Madland_vec[j] );
    }
    else
    {
      copy_data( Madland_vec[j] );
      started = true;
      use_weight( );
    }
  }
}
// ----------- add_them_up::check_total_wt -----------
void add_them_up::check_total_wt( )
// Check the total weight; it should be 1
{
  const double EPS = Global.Value( "mf5_weight_tol" );
  for( mf5_table::iterator this_list = begin( );
       this_list != end( ); ++this_list )
  {
    if( abs( this_list->weight - 1.0 ) > EPS )
    {
      Warning( "add_them_up::check_total_wt",
	       pastenum( pastenum( "bad total weight for E_in ",
			 this_list->E_in( ) )+": ",this_list->weight) );
    }
  }
}

// ********* for class fission_delay *************
// ----------- fission_delay::manager ------
void fission_delay::manager(mf1_file& inFile)
// manage everything for delayed neutrons
{
  // get the time constants and the total multiplicity
  read_mf1(inFile);

  // the data for the channels is in an mf5 file
  mf5_file mf5_in;
  mf5_in.open(5, 455);
  ENDL.set_c_number(15);
  ENDL.set_s_number(7);

  if( !mf5_in )
  {
    Info("fission_delay::manager",
        "No mf05mt455 file; writing total delayed neutron multiplicity");
    ENDL.set_yo(1);
    total_mult.widen_jumps();
    if ( ENDL.write_file ) total_mult.write_endl( 7 );
    return;
  }

  // read the first line of the file
  int ZA;
  double AWR;
  int NK;  // number of channels
  mf5_in.read_line1(&ZA, &AWR, &NK);
  if(NK != time_consts.size())
  {
    SevereError("fission_delay::manager","Wrong count for delayed neutron decay channels");
  }

  // loop over the decay channels
  ENDL.append = false;  // make new output files on the first pass
  for(int count = 0; count < NK; ++count)
  {
    // read the multiplicity and get the model
    int LF = one_mult(mf5_in, count);
    if( LF == 5 )
    {
      // general evaporation model
      one_gen_evap( mf5_in, count );
    }
    else if( LF == 1 )
    {
      // probability table
      one_table( mf5_in, count );
    }
    else
    {
      Unimplemented("fission_delay::manager",
        pastenum(" Implement model LF = ", LF) );
    }
    ENDL.append = true;  // now just append to the output files
  }
  Info("fission_delay::manager","Closing file "+endf_file_name(5, 455)+"...");
  mf5_in.close();
  ENDL.append = false;  // go back to making new output files
}
// ----------- fission_delay::read_mf1 -----------------
void fission_delay::read_mf1(mf1_file& inFile)
// get the time constants and the total multiplicity
{
  int NC;  // number of time constants
  inFile.delay_head(&NC);
  time_consts.reserve(NC);
  time_consts.resize(NC);

  string linebuff;
  string strbuff;

  for( int iNC=0; iNC < NC; iNC++ ) //Loop over NC data items
  {
    int iC=iNC % 6;  //There are 6 items per line
    if ( iC == 0 )
    {
      getline( inFile, linebuff); //Read in new line
    }
    strbuff = linebuff.substr(iC*11, 11); //Grab the right entry
    read_d(&strbuff, &time_consts[iNC] );
  }

  // for the total multiplicity
  int NR;
  int NP;
  inFile.mult_head(&NR, &NP);

  // read the interpolation information
  inFile.get_regions(NR, total_mult.NBT, total_mult.INT);
  if((NR > 1) || (total_mult.INT[0] != 2))
  {
    Unimplemented("fission_delay::read_mf1",
        "Implement interpolation for delayed fission neutrons");
  }

  // read the multiplicities
  for( int iNP=0; iNP < NP; iNP++ ) //Loop over NP data pairs
  { 
    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 )
    {
      getline( inFile, linebuff); //Read in new line
    }
    strbuff = linebuff.substr(iP*22, 22); //Grab the right pair
    dd_link XYdata;
    read_dd(&strbuff, &XYdata.x, &XYdata.y );
    // convert to MeV
    XYdata.x *= ENDL.eV2MeV;
    total_mult.insert(total_mult.end(), XYdata);
  }
}
// ----------- fission_delay::one_mult -----------------
int fission_delay::one_mult(mf5_file& inFile, int j)
// get the multiplicities for j-th channel and return the model
{
  double U;  // not used
  int LF;    // mf5 model
  int NR;    // number of interpolation regions
  int NP;    // number of energy-multiplicity pairs
  inFile.read_line2(&U, &LF, &NR, &NP);
 
  multiplicity partial_mult; // multiplicity for one channel
  dd_link XYdata;

  inFile.get_regions(NR, partial_mult.NBT, partial_mult.INT);
  string linebuff;
  string strbuff;

  for( int iNP=0; iNP < NP; iNP++ ) //Loop over NP data pairs
  {
    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 )
    {
      getline( inFile, linebuff); //Read in new line
    }
    strbuff = linebuff.substr(iP*22, 22); //Grab the right pair
    read_dd(&strbuff, &XYdata.x, &XYdata.y );
    // convert to MeV
    XYdata.x *= ENDL.eV2MeV;
    partial_mult.insert(partial_mult.end(), XYdata);
  }
  partial_mult *= total_mult;  // scale by the total multiplicity
  ENDL.set_c_number(15);
  ENDL.set_yo(1);
  ENDL.set_s_number(7);
  ENDL.set_x1(time_consts[j]);
  if ( ENDL.write_file ) partial_mult.write_endl( 7 );

  return LF;
}
// ----------- fission_delay::one_gen_evap -----------------
void fission_delay::one_gen_evap(mf5_file& inFile, int count)
// get the energy distribution for the j-th channel
// for the general evaporation model
{
  two_d_table e_dist;
  one_d_table one_E_in;
  dd_link XYdata;

  // start an empty list
  e_dist.insert(e_dist.end(), one_E_in);
  two_d_table::iterator dist_ptr = e_dist.end();
  --dist_ptr;

  double U;  // not used
  int LF;    // not used
  int NR;    // number of interpolation regions
  int N_theta;    // number of energy-theta pairs

  inFile.read_line2(&U, &LF, &NR, &N_theta);
  inFile.get_regions(NR, theta.NBT, theta.INT);
  string linebuff;
  string strbuff;

  for( int iNP=0; iNP < N_theta; iNP++ ) //Loop over N_theta data pairs
  {
    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 )
    {
      getline( inFile, linebuff); //Read in new line
    }
    strbuff = linebuff.substr(iP*22, 22); //Grab the right pair
    read_dd(&strbuff, &XYdata.x, &XYdata.y );
    // convert to MeV
    XYdata.x *= ENDL.eV2MeV;
    theta.insert(theta.end(), XYdata);
  }

  // Read in the values of g(E'/theta) for the energy distribution
  int NP;  // number of (x, g(x)) pairs
  inFile.read_line2(&U, &LF, &NR, &NP);
  inFile.get_regions(NR, dist_ptr->NBT, dist_ptr->INT);

  for( int iNP=0; iNP < NP; iNP++ ) //Loop over NP data pairs
  {
    int iP=iNP % 3;  //There are 3 pairs per line
    if ( iP == 0 )
    {
      getline( inFile, linebuff); //Read in new line
    }
    strbuff = linebuff.substr(iP*22, 22); //Grab the right pair
    read_dd(&strbuff, &XYdata.x, &XYdata.y );
    // convert to MeV
    XYdata.x *= ENDL.eV2MeV;
    dist_ptr->insert(dist_ptr->end(), XYdata);
  }
  // expand to lin-lin interpolation
  dist_ptr->expand_interp( ENDL.Max_E_out );

  // fill out the 2-d list
  gen_mf5_expand( N_theta, e_dist );

  // print it
  ENDL.set_yo(1);
  ENDL.set_x1(time_consts[count]);
  if ( ENDL.write_file ) e_dist.write_endl( 4 );
}
// ----------- fission_delay::gen_mf5_expand -----------------
void fission_delay::gen_mf5_expand(int N_theta, two_d_table& e_dist)
// form the 2-list of energy distributions for general evaporation model
{
  one_d_table one_E_in;
  two_d_table::iterator g_list = e_dist.end();
  --g_list;
  two_d_table::iterator next_list = g_list;
  dd_link XYdata;
  dd_list::iterator g_list_ptr;
  int count;

  // append N_theta - 1 more one_d_distrib links to the list
  for(count = 1; count < N_theta; ++count)
  {
    e_dist.insert(e_dist.end(), one_E_in);
    ++next_list;  // pointer to the new list
    // copy the first list
    for(g_list_ptr = g_list->begin();
      g_list_ptr != g_list->end(); ++g_list_ptr)
    {
      XYdata.x = g_list_ptr->x;
      XYdata.y = g_list_ptr->y;
      next_list->insert(next_list->end(), XYdata);
    }
  }

  // set the incident energy and scale E' = theta*x
  dd_list::iterator theta_ptr = theta.begin();
  for(g_list = e_dist.begin(); g_list != e_dist.end();
    ++g_list, ++theta_ptr)
  {
    g_list->E_in() = theta_ptr->x;
    double scale_E = theta_ptr->y;
    // copy the first list
    for(dd_list::iterator g_list_ptr = g_list->begin();
      g_list_ptr != g_list->end(); ++g_list_ptr)
    {
      XYdata.x *= scale_E;
    }
    // normalize
    g_list->renorm();
  }
}
// ----------- fission_delay::one_table -----------------
void fission_delay::one_table( mf5_file& inFile, int count )
// get the energy distribution for the j-th channel
// for an mf5 table
{
  mf5_table dist_table;

  // zeros because there are no weights
  dist_table.read_data( inFile, 0, 0 );

  // print it
  ENDL.set_yo(1);
  ENDL.set_x1(time_consts[count]);
  if ( ENDL.write_file ) dist_table.write_endl( 4 );
}
