/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1861 $
 * $Date: 2006-04-28 11:51:26 -0700 (Fri, 28 Apr 2006) $
 * $Author: hedstrom $
 * $Id: mf5classes.hpp 1861 2006-04-28 18:51:26Z hedstrom $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

//header for the classes used in translating the MF=5 files

#ifndef MF5CLASSES
#define MF5CLASSES

#include "list_2d.hpp"
#include "ENDF_file.hpp"
#include "multiplicity.hpp"

using namespace std;

// ----------- class mf5_table -----------------
// derive this class from two_d_table
//! Class to handle tabular data
class mf5_table : public two_d_table
{
private:

public:
  bool duplicate_Ein;

  //! Reads in all the data
  void read_data( mf5_file& inFile, int NR, int NP );

  //! Reads in the data for a particular incident neutron energy
  void one_E_in( mf5_file& inFile );

  //! Interpolates between tables according to the INT array
  void table_interp( );

  //! Empty function
  inline void set_U( double U )
  {
  }

  //! Empty function
  inline void make_list( )
  {
  }

};

// ----------- class two_d_evap -----------------
// derive this class for the evaporation spectrum from two_d_list
//! Class for handling the general evaporation spectrum
class two_d_evap : public two_d_list< one_d_evap >
{
private:
  dd_list _theta;
  double UU;
  double Max_energy;

public:
  bool duplicate_Ein;

  //! Default constructor
  two_d_evap();
  //! Constructor with specific U parameter
  two_d_evap(double U);
  //! Default destructor
  ~two_d_evap();

  //! Sets the U parameter
  inline void set_U( double U )
  {
    UU = U*ENDL.eV2MeV;  // convert to MeV;
  }

  //! Inserts a distribution for a given incident neutron energy
  void insert(two_d_iterator where, double e_in, double Theta);

  //! Reads all the data
  void read_data(mf5_file& inFile, int NR, int NP);

  //! Expands the data into a list
  void make_list();
};

// ----------- class two_d_Maxwell -----------------
// derive this class for the Maxwell spectrum from two_d_list
//! Class to handle the simple fission spectrum (Maxwellian)
class two_d_Maxwell : public two_d_list< one_d_Maxwell >
{
private:
  dd_list _theta;
  double UU;
  double Max_energy;

public:
  bool duplicate_Ein;

  //! Default constructor
  two_d_Maxwell();
  //! Constructor with specific U parameter
  two_d_Maxwell(double U);
  //! Default destructor
  ~two_d_Maxwell();

  //! Sets the U parameter
  inline void set_U( double U )
  {
    UU = U*ENDL.eV2MeV;  // convert to MeV;
  }

  //! Inserts a distribution for the given incident neutron energy
  void insert(two_d_iterator where, double e_in, double Theta);

  //! Read in all the data
  void read_data(mf5_file& inFile, int NR, int NP);

  //! Expands the data into a list
  void make_list();
};

// ----------- class two_d_Watt -----------------
// derive this class for the Watt spectrum from two_d_list
//! Class to handle the energy dependent Watt spectrum
class two_d_Watt : public two_d_list< one_d_Watt >
{
private:
  dd_list Watt_a;   // Watt a parameters
  dd_list Watt_b;   // Watt b parameters
  double UU;
  double Max_energy;

public:
  bool duplicate_Ein;

  //! Default constructor
  two_d_Watt();
  //! Constructor with specifi U parameter
  two_d_Watt(double U);
  //! Default destructor
  ~two_d_Watt();

  //! Sets the U parameter
  inline void set_U( double U )
  {
    UU = U*ENDL.eV2MeV;  // convert to MeV;
  }

  //! Inserts a distribution for the given incident neutron energy
  void insert( two_d_iterator where, double e_in, double a );

  //! Reads in all the data
  void read_data(mf5_file& inFile, int NR, int NP);

  //! Expands the data into a list
  void make_list();

  //! Sets the b values and expand the lists
  void set_b( );
};

// ----------- class two_d_Madland -----------------
// derive this class for the Madland spectrum from two_d_list
//! Class to handle the energy dependent fission spectrum of Madland and Nix
class two_d_Madland : public two_d_list< one_d_Madland >
{
private:
  dd_list Madland_TM;   // Madland TM parameters
  double EFL_;
  double EFH_;
  double Max_energy;

public:
  bool duplicate_Ein;

  //! Default constructor
  two_d_Madland();
  //! Default destructor
  ~two_d_Madland();

  //! Sets the parameters
  inline void set_EF( double EFL, double EFH )
  {
    EFL_ = EFL*ENDL.eV2MeV;  // convert to MeV;
    EFH_ = EFH*ENDL.eV2MeV;  // convert to MeV;
  }

  //! Empty function
  inline void set_U( double U ) {}

  //! Inserts a distribution for the given incident neutron energy
  void insert(two_d_iterator where, double e_in, double TM);

  //! Reads all the data
  void read_data(mf5_file& inFile, int NR, int NP);

  //! Expands the data into a list
  void make_list();
};

// ************* class add_them_up ********************
//! Class to manage the addition of weighted models
class add_them_up : public mf5_table
{
private:
  // arrays for the possible types of distributions
  mf5_table* table_vec;  // array of mf5_tables
  two_d_evap* evap_vec;  // array of evaporation models
  two_d_Maxwell* Maxwell_vec;  // array of Maxwell models
  two_d_Watt* Watt_vec;   // array of Watt models
  two_d_Madland* Madland_vec;  // array of Madland models

  int NK_;  // how many models

  // how many of each type of model
  int num_table;
  int num_evap;
  int num_Maxwell;
  int num_Watt;
  int num_Madland;

  bool debug_; // do we test multiple models?

  //! Checks that the total weight is 1
  void check_total_wt( );

public:
  //! Constructor for NK sets of data
  add_them_up( int NK );

  //! Destructor
  ~add_them_up( );

  //! Manages all of the output models
  void manager( mf5_file& inFile );

  //! Handles one output model
  void one_model( mf5_file& inFile );

  //! Ensures that all lists have common incident energies
  void common_Ein( );

  //! Adds the lists to get the weighted sum
  void wt_add( );
};

// ----------- class fission_delay -----------------
// This class is used with MT = 455 and LNU = 2.
//! Class that handles multiplicites and energy distributions for delayed fission neutrons
class fission_delay
{
private:
  one_d_table theta;
  vector<double> time_consts;
  multiplicity total_mult;  // total multiplicity

public:
  //! Default constructor
  inline fission_delay()
    {}
  //! Default destructor
  ~fission_delay()
    {}

  //! Manages the class
  void manager(mf1_file& inFile);

  //! Gets the time constants and the total multiplicity
  void read_mf1(mf1_file& inFile);

  //! Gets the multiplicities for the j-th channel and returns the model
  int one_mult(mf5_file& inFile, int j);

  //! Gets the energy distribution for the j-th channel for the general evaporation model
  void one_gen_evap(mf5_file& inFile, int count);

  //! Forms the 2d list of energy distributions for general evaporation model
  void gen_mf5_expand(int N_theta, two_d_table& e_dist);

  //! Gets the energy distribution for the j-th channel for an mf5 table
  void one_table( mf5_file& inFile, int count );
};

#endif
