/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1855 $
 * $Date: 2006-04-21 16:12:38 -0700 (Fri, 21 Apr 2006) $
 * $Author: hedstrom $
 * $Id: mf6classes.cpp 1855 2006-04-21 23:12:38Z hedstrom $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the class mf6_class

#include <vector>
using namespace std;
#include "mf6classes.hpp"
#include "record_types.hpp"
#include "global_params.hpp"
#include "endl_precision.hpp"

extern ENDLClass ENDL;
extern GlobalParameterClass Global;

// ********* for class mf6_one_d *************
// ----------- mf6_one_d::mf6_one_d -----------------
// constructor
mf6_one_d::mf6_one_d( )
{
}

// ----------- mf6_one_d::one_E_row -----------------
void mf6_one_d::one_E_row( double min_E, double max_E )
// for a given mu make new links for secondary energies
// min_E <= E <= max_E
{
  double E_out;
  double Prob;
  dd_link e_out_link;
  const int Num = 2;  // to ensure that we start with one internal links

  // fudge the limits to ensure that we stay inside the range
  double first_E = min_E*1.00001;
  double dE = (max_E*0.99999 - first_E)/Num;

  for(int count = 0; count <= Num; ++count)
  {
    E_out = first_E + count*dE;
    Prob = f( E_out );
    e_out_link = dd_link( E_out, Prob );
    insert( end( ), e_out_link );
  }

  // fill in intermediate links
  thicken();

  // If we are checking interpolation using the equiprobable
  // bins, get the bins.
  if(bin_interp)
  {
    get_bins();
  }
}

// ********* for class one_d_phase *************
// ----------- one_d_phase::f -------------------
double one_d_phase::f( double E )
// the phase-space distribution function
{
  Pair lab_Emu; // (energy, cosine) in lab coordinates
  Pair cm_Emu;  // (energy, cosine) in center-of-mass coordinates
  double Prob;  // return value

  lab_Emu.mu = mu( );
  lab_Emu.E = E;

  // in center-of-mass coordinates
  cm_Emu = map_->lab_to_cm( lab_Emu );

  double E_diff = Ei_max - cm_Emu.E;

  if ( E_diff <= 0.0 )
  {
    Warning("one_d_phase::f",
        pastenum("attempt to take logarithm of ",E_diff));
    return 0.0;
  }

  // this probability includes the Jacobian: sqrt(lab_Emu.E/cm_Emu.E)
  Prob = sqrt( lab_Emu.E )*exp( exponent * log( E_diff ) );

  return Prob;
}

// ********* for class two_d_phase *************
// ----------- two_d_phase::new_one_d( ) ---------
two_d_phase::iterator two_d_phase::new_one_d( two_d_phase::iterator where )
{
  one_d_phase new_link;
  insert( where, new_link );

  // point to the new link
  two_d_phase::iterator link_ptr = where;
  --link_ptr;

  // set up the parameters
  link_ptr->Ei_max = Ei_max_;
  link_ptr->exponent = exponent_;
  link_ptr->copy_map( &map );

  return link_ptr;
}

// ********* for class phase_space *************
// ----------- phase_space::phase_space ------
phase_space::phase_space( int projectile, int target, int ZAP )
{
  a.set( projectile );
  A.set( target );
  b.set( ZAP );
  C.set( projectile + target );
  Q = ENDL.QI;
}

// ----------- phase_space::expand_data ------
void phase_space::expand_data( mf6_file& inFile, multiplicity& multiple)
{
  double dE;
  double e_in;
  int count;

  // the multiplicity ought to be constant
  if( ( multiple.INT.size() > 1 ) || ( multiple.INT[0] > 2 ) )
  {
    SevereError("phase_space::expand_data","Strange multiplicity for phase_space class");
  }

  // read the data line
  double APSX;  // the total mass involved
  int NPSX;      // for the exponent
  inFile.phase_space( &APSX, &NPSX );
  mass_ratio = ( APSX - b.AWR ) / APSX;
  exponent__ = 1.5 * NPSX - 4;

  // our maximum incident energy
  double Max_energy = ENDL.Max_E_in;
  if( ENDL.threshold >= Max_energy )
  {
    ENDL.write_file = false;
    return;
  }

  // find the transition from forward to omnidirectional emission
  double E_transition = find_E_trans( );
  // make lists for various incident energies
  int Num_E_in = static_cast<int>( Global.Value( "num_E_in_3d" ) );
  if( ( E_transition <= 0.0 ) || ( E_transition >= Max_energy ) )
  {
    dE = ( Max_energy - ENDL.threshold ) / Num_E_in;
    for ( count = 0; count <= Num_E_in; ++ count )
    {
      e_in = ENDL.threshold + count * dE;
      one_E_in( e_in );
    }
  }
  else
  {
    double local_eps = ENDL_EPSILON( E_transition );
    dE = ( E_transition - local_eps - ENDL.threshold ) / Num_E_in;
    for ( count = 0; count <= Num_E_in; ++ count )
    {
      e_in = ENDL.threshold + count * dE;
      one_E_in( e_in );
    }
    dE = ( Max_energy - E_transition - local_eps ) / Num_E_in;
    for ( count = 0; count <= Num_E_in; ++ count )
    {
      e_in = E_transition + local_eps + count * dE;
      one_E_in( e_in );
    }
  }
}

// ----------- phase_space::one_E_in ------
void phase_space::one_E_in( double e_in )
{
  two_d_phase e_in_link;

  e_in_link.E_in() = e_in;
  insert( end(), e_in_link );

  // make a pointer to the new 3-d link
  phase_space::iterator this_link = end();
  --this_link;

  // Expand this data into a 3-d link
  expand_E_in( this_link );

  // add a link to the cosine list
  one_d_table new_cos;
  cosines.insert( cosines.end(), new_cos );

  // expand this cosine link
  two_d_list<one_d_table>::iterator cos_ptr = cosines.end();
  --cos_ptr;
  this_link->make_cos_link( this_link->E_in(), cos_ptr );
}

// ----------- phase_space::expand_E_in -----------------
void phase_space::expand_E_in( phase_space::iterator e_in_link )
{
  // maximum center-of-mass energy for the emitted particle
  e_in_link->Ei_max_ = mass_ratio *
    ( Q + e_in_link->E_in()*( A.AWR/(A.AWR + a.AWR) ) );
  if( e_in_link->Ei_max_ <= 0.0 )
  {
    Warning("phase_space::expand_E_in",
         pastenum(" negative Ei_max_: ",e_in_link->Ei_max_)+
         pastenum(" for E_in: ", e_in_link->E_in()));
    // to increase the incident energy at or below threshold
    double EPSILON = Global.Value( "mf5_shift" );
    e_in_link->E_in() = -Q * ( 1 + EPSILON ) * (A.AWR + a.AWR) / A.AWR;
    e_in_link->Ei_max_ = mass_ratio *
      ( Q + e_in_link->E_in()*( A.AWR/(A.AWR + a.AWR) ) );
  }

  // to map between lab and center-of-mass coordinates
  e_in_link->map.set_map( &A, &a, &b, e_in_link->E_in() );
  e_in_link->exponent_ = exponent__;

  // Identify the geometry of the (E', mu) region in lab coordinates.
  // Note that the zero minimal energy makes 
  //   e_in_link->geom_.slowest == Forward.
  e_in_link->get_geom( 0.0, e_in_link->Ei_max_ );
  if( e_in_link->geom_.fastest == Both_ways )
  {
    e_in_link->full_range( e_in_link->Ei_max_ );
  }
  else if( e_in_link->geom_.fastest == Transition )
  {
    e_in_link->full_trans( e_in_link->Ei_max_ );
  }
  else if( e_in_link->geom_.fastest == Forward )
  {
    e_in_link->full_forward( 0.0, e_in_link->Ei_max_ );
  }
  else
  {
    SevereError("phase_space::expand_E_in","Strange geometry");
  }
}
// ----------- phase_space::find_E_trans -----------------
double phase_space::find_E_trans( )
{
  // calculate the incident energy at which we first have zero
  // outgoing energy in the lab frame
  double numerator = mass_ratio*Q;
  double denominator =
    ( (a.AWR*b.AWR) / (A.AWR + a.AWR) - mass_ratio*A.AWR ) /
    (A.AWR + a.AWR);

  double E_transition = numerator / denominator;
  return E_transition;
}

// ********* for class two_d_ENDL_table *************
// ----------- two_d_ENDL_table::new_one_d( ) ---------
two_d_ENDL_table::iterator 
  two_d_ENDL_table::new_one_d( two_d_ENDL_table::iterator where )
{
  one_d_table new_link;
  insert( where, new_link );

  // point to the new link
  two_d_ENDL_table::iterator link_ptr = where;
  --link_ptr;

  return link_ptr;
}

// ********* for class ENDL_table *************
// ----------- ENDL_table::ENDL_table ------
ENDL_table::ENDL_table( )
// constructor
{
}

// ----------- ENDL_table::read_data ---------
void ENDL_table::read_data( mf6_file& inFile, int ZAP,
  multiplicity& multiple )
// Reads in all the data
{
  // the multiplicity ought to be constant
  if( ( multiple.INT.size() > 1 ) || ( multiple.INT[0] > 2 ) )
  {
    SevereError("ENDL_table::read_data","Strange multiplicity for ENDL_table class");
  }

  // read the number of incident energies
  int NR; // number of interpolation regions
  int NE; // number of incident energies
  inFile.ENDL( &NR, &NE );
  inFile.get_regions( NR, NBT, INT );
  // loop over the incident energies
  for ( int count = 0; count < NE; ++count)
  {
    one_E_in( inFile );
  }
}

// ----------- ENDL_table::one_E_in ---------
void ENDL_table::one_E_in( mf6_file& inFile )
// Reads the data for one incident energy
{
  string linebuff;
  string strbuff;

  // read the incident energy and the number of cosines
  double e_in;
  int NR; // number of interpolation regions
  int NMU; // number of cosines
  inFile.ENDL_head( &e_in, &NR, &NMU );
  two_d_ENDL_table new_e_in_link;
  insert( end(), new_e_in_link );
  // make a pointer to the new 3-d link
  ENDL_table::iterator e_in_link = end();
  --e_in_link;
  e_in_link->E_in() = e_in*ENDL.eV2MeV; // convert to MeV

  inFile.get_regions( NR, e_in_link->NBT, e_in_link->INT );
  // loop over the cosines
  for ( int count = 0; count < NMU; ++count )
  {
    // mu corresponds to a new 2d link in the list for this E_in
    double this_mu;
    int NEP; // number of (E', probability) pairs
    inFile.ENDL_head( &this_mu, &NR, &NEP );
    two_d_ENDL_table::iterator mu_ptr =
      e_in_link->new_one_d( e_in_link->end( ) );
    mu_ptr->mu( ) = this_mu;
    inFile.get_regions( NR, mu_ptr->NBT, mu_ptr->INT );

    // make a 1-d list of (E', probability) values
    for( int iNP=0; iNP < NEP; ++iNP ) //Loop over NEP data pairs
    { 
      int iP=iNP % 3;  //There are 3 pairs per line
      if ( iP == 0 )
      {
        getline( inFile, linebuff); //Read in new line
      }
      strbuff = linebuff.substr( iP*22, 22 );  // grab the right pair
      double X, Y; //Real variables we will put the data pair
      read_dd(&strbuff, &X, &Y );
      if ( X==-999. || Y==-999. ) //read_PAIR sends -999. if problem
      {
        SevereError("ENDL_table::one_E_in","read_PAIR fucked up!");
      }
      else
      {
        dd_link XYdata( X*ENDL.eV2MeV, Y );
        mu_ptr->insert( mu_ptr->end(), XYdata );
      }
    }
  }
  // add a link to the cosine list
  one_d_table new_cos;
  cosines.insert( cosines.end(), new_cos );

  // expand this cosine link
  two_d_list<one_d_table>::iterator cos_ptr = cosines.end();
  --cos_ptr;
  e_in_link->make_cos_link( e_in_link->E_in(), cos_ptr );
}

// *********** class Legendre_item ***********
// ----------- Legendre_item::Legendre_item ---------
Legendre_item::Legendre_item( )
{
  // default constructor
}
// ----------- Legendre_item::read_coefs ---------
void Legendre_item::read_coefs( int i_start, string& linebuff, int order,
  mf6_file& inFile, bool *E_out_done )
// Reads the Legendre coefficients for one outgoing energy.
// There are 6 entries per line, but the reading may start at any entry.
// *E_out_done = true if E_out > ENDL.max_E_out
{
  if( !(*E_out_done) )
  {
    // Allocate space for the coefficients
    coefs.get_space( order + 1 );
  }

  string strbuff;

  // read the outgoing energy
  if( i_start == 0 )
  {
    getline( inFile, linebuff );
  }
  strbuff = linebuff.substr( 11*i_start, 11 );  // grab the first entry
  read_d( &strbuff, &E_out );
  E_out *= ENDL.eV2MeV; // convert to MeV

  // get the coefficients
  int count = 0;
  for(  ; count <= order; ++count )
  {
    // There are 6 entries per line, and we already read E_out.
    int iP = ( i_start + count + 1 ) % 6;
    if ( iP == 0 )
    {
      getline( inFile, linebuff); //Read in new line
    }
    if( !(*E_out_done) )
    {
      // grab the proper entry
      strbuff = linebuff.substr( 11*iP, 11 );
      read_d( &strbuff, &coefs[ count ] );
    }
  }
  double dE = ENDL_EPSILON( ENDL.Max_E_out );
  if( E_out > ENDL.Max_E_out - dE )
  {
    *E_out_done = true;
  }
}
// ----------- Legendre_item::copy_coef ---------
void Legendre_item::copy_coef( Legendre_item& copy_from )
// Copy the Legendre coefficients
{
  // Allocate space for the coefficients
  int num_coef = copy_from.coefs.size( );
  coefs.get_space( num_coef );

  for( int count = 0; count < num_coef; ++count )
  {
    coefs[ count ] = copy_from.coefs[ count ];
  }
}
// ----------- Legendre_item::interpolate ---------
void Legendre_item::interpolate( Legendre_item& prev_link,
  Legendre_item& next_link )
// interpolate the coefficients between next_link and the previous one
{
  double E_diff = next_link.E_out - prev_link.E_out;

  // check for division by 0
  if(E_diff == 0.0)
  {
    SevereError("Legendre_item::interpolate","division by 0");
  }

  double alpha = (E_out - prev_link.E_out)/E_diff;

  // check for extrapolation
  if((alpha < 0.0) || (alpha > 1.0))
  {
    SevereError("Legendre_item::interpolate", "attempt to extrapolate");
  }

  for( int j = 0; j < coefs.size( ); ++j )
  {
    coefs[ j ] += alpha*next_link.coefs[ j ] +
      (1.0 - alpha)*prev_link.coefs[ j ];
  }
}

// *********** class Legendre_list ***********
// ----------- Legendre_list::read_data ---------
void Legendre_list::read_data( int LEP, mf6_file& inFile, bool *done )
// Reads the Legendre data for one incident energy
// *done = true if E_in > ENDL.Max_E_in
{
  // we may need a list of gamma lines
  one_d_table gamma_lines;

  int num_E_out;
  int ND;  // number of discrete lines
  // How much data?
  inFile.Legendre_E_in( &E_in, &ND, &order, &num_E_out );

  E_in *= ENDL.eV2MeV; // convert to MeV

  // read the data
  string linebuff;
  string strbuff;

  // start with the lines
  for( int count = 0; count < ND; ++count )
  {
    int iP = count % 3;  //There are 3 pairs per line
    if ( iP == 0 )
    {
      getline( inFile, linebuff ); //Read in new line
    }
    strbuff = linebuff.substr( iP*22, 22 );  // grab the right pair
    double X, Y; //Real variables we will put the data pair
    read_dd(&strbuff, &X, &Y );
    // to MeV
    X *= ENDL.eV2MeV;
    dd_link XYdata( X, Y );

    // insert the line if it is within our range
    double dE = ENDL_EPSILON( ENDL.Max_E_out );
    if( ( Y > 0.0 ) && ( X <= ENDL.Max_E_out + dE ) )
    {
      // find where to insert the line
      one_d_table::iterator next_line = gamma_lines.find_next( X );
      gamma_lines.insert( next_line, XYdata );
    }
  }

  // now read the Legendre coefficients for the continuum
  int i_start = 2*ND % 6;
  bool E_out_done = false;
  for( int count = 0; count < num_E_out - ND; ++count )
  {
    Legendre_item new_E_out;
    if( !(*done) )
    {
      insert( end( ), new_E_out );
      Legendre_list::iterator E_out_data = end( );
      --E_out_data;  // now points to the new list
      E_out_data->read_coefs( i_start, linebuff, order, inFile,
        &E_out_done );
    }
    else
    {
      new_E_out.read_coefs( i_start, linebuff, order, inFile,
        &E_out_done );
    }
    i_start += order + 2;
    i_start = i_start % 6;
  }

  // expand the interpolation
  expand_interp( LEP );

  // are there lines?
  if( gamma_lines.size( ) > 0 )
  {
    gamma_lines.widen_delta( );

    // are there Legendre coefficients for the continuum?
    if( empty( ) )
    {
      // copy the lines
      only_lines( gamma_lines );
    }
    else
    {
      // add the lines
      add_lines( gamma_lines );
    }
  }
  // is this incident energy too big?
  double dE = ENDL_EPSILON( ENDL.Max_E_in );
  if( E_in > ENDL.Max_E_in - dE )
  {
    *done = true;
  }
}
// ----------- Legendre_list::only_lines ---------
void Legendre_list::only_lines( one_d_table& gamma_lines )
// we have only gamma lines---no continuum
{
  for( one_d_table::iterator line_ptr = gamma_lines.begin( );
       line_ptr != gamma_lines.end( ); ++line_ptr )
  {
    Legendre_item new_E_out;
    insert( end( ), new_E_out );
    Legendre_list::iterator E_out_data = end( );
    --E_out_data;  // now points to the new list
    E_out_data->E_out = line_ptr->x;
    E_out_data->coefs.get_space( 1 );
    E_out_data->coefs[ 0 ] = line_ptr->y;
  }
}
// ----------- Legendre_list::add_lines ---------
void Legendre_list::add_lines( one_d_table& gamma_lines )
// we have gamma lines and continuum
{
  // pad the start with zeros if necessary
  one_d_table::iterator line_ptr = gamma_lines.begin( );
  Legendre_list::iterator continuum = begin( );

  double Eout_line = line_ptr->x;
  double Eout_continuum = continuum->E_out;

  if( Eout_line < Eout_continuum )
  {
    pad_head( Eout_line );
  }
  else if( Eout_continuum < Eout_line )
  {
    gamma_lines.pad_head( Eout_continuum );
  }

  // pad the tail with zeros if necessary
  line_ptr = gamma_lines.end( );
  continuum = end( );
  --line_ptr;
  --continuum;

  Eout_line = line_ptr->x;
  Eout_continuum = continuum->E_out;

  if( Eout_line > Eout_continuum )
  {
    pad_tail( Eout_line );
  }
  else if( Eout_continuum > Eout_line )
  {
    gamma_lines.pad_tail( Eout_continuum );
  }

  // in order to add them, first make a list of the common E_out values
  list< double > E_out_list;
  gamma_lines.collect_E_in( E_out_list );

  for( Legendre_list::iterator continuum = begin( );
       continuum != end( ); ++continuum )
  {
    double E_cont = continuum->E_out;
    E_out_list.push_back( E_cont );
  }

  E_out_list.sort( );
  E_out_list.unique( );
  // fill the line list with values from E_out_list
  gamma_lines.fill_with( E_out_list );

  // fill the continuum list with values from E_out_list
  fill_with( E_out_list );

  // add the gamma data
  for( line_ptr = gamma_lines.begin( ), continuum = begin( );
       (  line_ptr != gamma_lines.end( ) ) && ( continuum != end( ) );
       ++line_ptr, ++continuum )
  {
    double E_gamma = line_ptr->x;
    double E_cont = continuum->E_out;

    if( E_gamma != E_cont )
    {
      SevereError( "Legendre_list::add_lines",
		   "energies don't match" );
    }
    // add the line data
    continuum->coefs[ 0 ] += line_ptr->y;
  }
}
// ----------- Legendre_list::pad_head ---------
void Legendre_list::pad_head( double E0 )
// insert zero links from E0 to the head
{
  Legendre_item new_link;  // for creating new links
  Legendre_list::iterator E_out_data;

  double first_E = begin( )->E_out;
  // smallest dE can see in output file
  double dE = ENDL_EPSILON( first_E );

  // insert a jump if needed
  if( first_E - E0 > dE )
  {
    bool non_zero = false;
    for( int j = 0; j <= order; ++j )
    {
      if( begin( )->coefs[ j ] != 0.0 )
      {
        non_zero = true;
	break;
      }
    }
    if( non_zero )
    {
      new_link.E_out = first_E - dE;
      push_front( new_link );
      E_out_data = begin( );
      E_out_data->coefs.get_space( order + 1 );
      E_out_data->coefs.set_zero( );
    }
  }

  // insert a new head
  new_link.E_out = E0;
  push_front( new_link );
  E_out_data = begin( );
  E_out_data->coefs.get_space( order + 1 );
  E_out_data->coefs.set_zero( );
}
// ----------- Legendre_list::pad_tail ---------
void Legendre_list::pad_tail( double E0 )
// insert zero links from E0 to the tail
{
  Legendre_item new_link;  // for creating new links

  Legendre_list::iterator E_out_data = end( );
  --E_out_data;
  double last_E = E_out_data->E_out;
  // smallest dE can see in output file
  double dE = ENDL_EPSILON( last_E );

  // insert a jump if needed
  if( E0 - last_E > dE )
  {
    bool non_zero = false;
    for( int j = 0; j <= order; ++j )
    {
      if( E_out_data->coefs[ j ] != 0.0 )
      {
        non_zero = true;
	break;
      }
    }
    if( non_zero )
    {
      new_link.E_out = last_E + dE;
      push_back( new_link );
      E_out_data = end( );
      --E_out_data;
      E_out_data->coefs.get_space( order + 1 );
      E_out_data->coefs.set_zero( );
    }
  }

  // insert a new tail
  new_link.E_out = E0;
  push_back( new_link );
  E_out_data = end( );
  --E_out_data;
  E_out_data->coefs.get_space( order + 1 );
  E_out_data->coefs.set_zero( );
}
// ----------- Legendre_list::fill_with ------------
void Legendre_list::fill_with( list< double >& E_out_list )
// make sure that there are links for each energy in E_out_list
{
  Legendre_list::iterator this_link = begin( );

  // go through the E_out_list
  for( list< double >::iterator E_out_ptr = E_out_list.begin( ); 
       E_out_ptr != E_out_list.end( ); ++E_out_ptr )
  {
    double E_out = *E_out_ptr;

    // if there is a node here, just go on to the next
    if( this_link->E_out == E_out )
    {
      ++this_link;
    }
    else
    {
      // interpolate and make a new link
      Legendre_item new_link;  // for creating new links
      new_link.E_out = E_out;
      Legendre_list::iterator prev_link = this_link;
      --prev_link;
      insert( this_link, new_link );
      Legendre_list::iterator new_ptr = this_link;
      --new_ptr;
      new_ptr->coefs.get_space( order + 1 );
      new_ptr->interpolate( *prev_link, *this_link );
    }
  }
}

// ----------- Legendre_list::get_norm ---------
double Legendre_list::get_norm( )
// Get the norm with respect to E_out
{
  // point to the zeroth Legendre_item
  Legendre_list::iterator data_ptr = begin( );

  if (size()==0) SevereError("Legendre_list::get_norm","Empty list!");

  double prev_E_out = data_ptr->E_out;
  double prev_P = data_ptr->coefs.at( 0 );
  ++data_ptr;
  double sum = 0.0;

  // loop through the other Legendre_items
  for(  ; data_ptr != end( ); ++data_ptr )
  {
    double next_E_out = data_ptr->E_out;
    double next_P = data_ptr->coefs.at( 0 );

    // accumulate the probability
    sum += 0.5*( next_P + prev_P )*( next_E_out - prev_E_out );
    // save the values for the next interval
    prev_E_out = next_E_out;
    prev_P = next_P;
  }
  return sum;
}
// ----------- Legendre_list::renorm ---------
void Legendre_list::renorm( )
// Set the norm with respect to E_out to 1.
{
  double norm = get_norm( );
  // point to the first Legendre_item
  Legendre_list::iterator data_ptr = begin( );

  // error check
  if( norm == 0.0 )
  {
    if (size()==0) SevereError("Legendre_list::renorm","Empty list!");
    Warning("Legendre_list::renorm",
         pastenum("zero norm in Legendre_list::renorm routine for E_in: ",E_in));
    // fudge something
    data_ptr->coefs[ 0 ] = 1.0;
  }
  else
  {
    // scale the data
    for(  ;  data_ptr != end( ); ++data_ptr )
    {
      for( int i_coef = 0; i_coef < data_ptr->coefs.size( );
          ++i_coef )
      {
        data_ptr->coefs[ i_coef ] /= norm;
      }
    }
  }
}

// ----------- Legendre_list::widen_jumps ---------
//! shift (near) duplicate x-values in the list
void Legendre_list::widen_jumps( )
{

  // scale dE by the smoother used for histograms
  double local_eps, jump_width;

  typedef Legendre_list::iterator list_iterator;

  // container to hold links to elements we need to widen
  list_iterator cluster_start = begin();
  list_iterator cluster_end = begin();
  int cluster_size=1;

  for( list_iterator this_link = begin(); this_link != end( ); ++this_link )
  {
    // get the local energy epsilon (smallest dE can see in output file)
    local_eps = ENDL_EPSILON( this_link->E_out );
    jump_width = ENDL_JUMP_WIDTH( this_link->E_out );

    // this_link is too close to the cluster_end, so add it to the cluster
    if( abs( this_link->E_out - cluster_end->E_out ) <= local_eps )
    {
      cluster_end = this_link;
      ++cluster_size;
    }
    // this_link is far from the cluster_end, so let's process the cluster
    else 
    {
      if ( cluster_size > 1 ) 
        widen_cluster( cluster_start, cluster_end, cluster_size, jump_width );
      // clear the clusters
      cluster_start = ++cluster_end;
      cluster_size  = 1;
    }
  }
  // check the last cluster since doesn't get processed otherwise
  if ( cluster_size > 1 ) 
    widen_cluster( cluster_start, cluster_end, cluster_size, jump_width );
}

// -------------- Legendre_list::widen_clusters ------------------
//! used by widen_jumps to do the actual widening between the cluster_start and cluster_end iterator
void Legendre_list::widen_cluster( Legendre_list::iterator cluster_start, 
  Legendre_list::iterator cluster_end, int cluster_size, double jump_width )
{
  if( ( cluster_size == 1 ) || ( cluster_start==cluster_end ) ) return;
  if ( cluster_size>4 ) 
  {
    Warning("Legendre_list::widen_cluster",pastenum("Found ",cluster_size)+
      pastenum(" links in a cluster, check ENDL_EPSILON and ENDL_JUMP_WIDTH values or the data is broken at E_out=",
      cluster_start->E_out)); 
  }

  if ( cluster_start->E_out <= jump_width )
  {
    Legendre_list::iterator cluster_middle = cluster_start; 
    for (int ijump = 1; ijump<=cluster_size; ++ijump)
    {
      ++cluster_middle; 
      cluster_middle->E_out += static_cast<double>(ijump)*jump_width;
    }
  }
  else
  {
    switch (cluster_size)
    {
      case 2:
        cluster_start->E_out -= 0.5*jump_width;
        cluster_end->E_out   += 0.5*jump_width;
        break;
      case 3:
        cluster_start->E_out -= jump_width;
        cluster_end->E_out   += jump_width;
        break;
      case 4:
        {
        Legendre_list::iterator cluster_left = cluster_start;
        Legendre_list::iterator cluster_right = cluster_end;
        ++cluster_left;--cluster_right;
        cluster_start->E_out -= 1.5*jump_width;
        cluster_left->E_out  -= 0.5*jump_width;
        cluster_right->E_out += 0.5*jump_width;
        cluster_end->E_out   += 1.5*jump_width;
        }
        break;
      default:
        Warning("Legendre_list::widen_cluster",pastenum("Found ",cluster_size)+
        pastenum(" links in a cluster, check ENDL_EPSILON and ENDL_JUMP_WIDTH values or the data is broken at E_out=",
        cluster_start->E_out)); 
        return;
    }
  }
}

// ----------- Legendre_list::expand_interp ---------
void Legendre_list::expand_interp( int LEP )
// Expand the interpolation with respect to energy of 
// the outgoing particle
{

  if ( LEP == 2 ) // lin-lin
  {
  }
  else if ( LEP == 1 ) // histogram
  {
    Legendre_list::iterator data_ptr = begin( );

    // We could be finished already
    if( data_ptr == end( ) )
    {
      return;
    }

    // loop through the list
    Legendre_list::iterator prev_ptr = begin( );
    ++data_ptr;
    Legendre_list::iterator next_ptr = data_ptr;
    ++next_ptr;

    for(  ; data_ptr != end( );
        prev_ptr = data_ptr, data_ptr = next_ptr, ++next_ptr )
    {
      Legendre_item new_link;
      insert( data_ptr, new_link );
      Legendre_list::iterator new_ptr = data_ptr;
      --new_ptr;  //  points to the new link

      // how wide the jump
      double dE_L = data_ptr->E_out - prev_ptr->E_out;
      double dE_R = ( next_ptr == end( ) ) ? data_ptr->E_out :
        next_ptr->E_out - data_ptr->E_out;
      double dE = (dE_L < dE_R) ? dE_L : dE_R;

      if(dE <= 0.0)
      {
        SevereError("Legendre_list::expand_interp", "bad histogram data");
      }

      // set smoothing dE to be smallest dE we can see in output file
      dE = ENDL_EPSILON( data_ptr->E_out );

      new_ptr->E_out = data_ptr->E_out - dE;
      data_ptr->E_out += dE;
      new_ptr->copy_coef( *prev_ptr );  // copy the Legendre coeficients
    }
  }
  else
  {
    SevereError("Legendre_list::expand_interp",
        pastenum("Implement interpolation type: ",LEP));
  }
}
// ----------- Legendre_list::unit_base ---------
void Legendre_list::unit_base( )
// map to unit base, 0 <= E' <= 1
{
  least_E = begin( )->E_out;
  Legendre_list::iterator list_ptr = end( );
  --list_ptr;
  most_E = list_ptr->E_out;
  double scale = most_E - least_E;
  if( scale <= 0.0 )
  {
    SevereError( "Legendre_list::unit_base",
		 "invalid scale factor");
  }

  for( list_ptr = begin( ); list_ptr != end( ); ++list_ptr )
  {
    list_ptr->E_out = ( list_ptr->E_out - least_E )/scale;
    list_ptr->coefs *= scale;
  }
}
// ----------- Legendre_list::rescale ---------
void Legendre_list::rescale( )
// undo the unit base transfoprmation
{
  double scale = most_E - least_E;
  if( scale <= 0.0 )
  {
    SevereError( "Legendre_list::rescale",
		 "invalid scale factor");
  }

  for( Legendre_list::iterator list_ptr = begin( );
    list_ptr != end( ); ++list_ptr )
  {
    list_ptr->E_out = least_E + scale*list_ptr->E_out;
    list_ptr->coefs *= 1.0/scale;
  }
}
// ----------- Legendre_list::copy ---------
void Legendre_list::copy( Legendre_list& to_copy )
// copy the Legendre coefficients
{
  order = to_copy.order;
  interp_type = to_copy.interp_type;
  Legendre_item new_link;

  for( Legendre_list::iterator copy_ptr = to_copy.begin( );
    copy_ptr != to_copy.end( ); ++copy_ptr )
  {
    insert( end( ), new_link );
    Legendre_list::iterator list_ptr = end( );
    --list_ptr;  // now points to the new link

    list_ptr->E_out = copy_ptr->E_out;
    list_ptr->copy_coef( *copy_ptr );
  }
}
// ----------- Legendre_list::operator*= ---------
Legendre_list& Legendre_list::operator*=( double alpha )
// scale the Legendre coefficients by alpha
{
  for( Legendre_list::iterator list_ptr = begin( );
    list_ptr != end( ); ++list_ptr )
  {
    list_ptr->coefs *= alpha;
  }
  return *this;
}
// ----------- Legendre_list::operator+= ---------
Legendre_list& Legendre_list::operator+=( Legendre_list& to_add )
// scale the Legendre coefficients by alpha
{
  Legendre_list::iterator add_ptr = to_add.begin( );
  for( Legendre_list::iterator list_ptr = begin( );
    list_ptr != end( ); ++list_ptr, ++add_ptr )
  {
    if( ( list_ptr->E_out != add_ptr->E_out ) ||
	(add_ptr == to_add.end( ) ) )
    {
      SevereError( "Legendre_list::operator+=",
		   "summands don't agree" );
    }
    list_ptr->coefs += add_ptr->coefs;
  }
  return *this;
}
// ----------- Legendre_list::chop_E_out ---------
void Legendre_list::chop_E_out( double max_E_out )
// deletes entries with E_out > max_E_out
{
  // I do not expect that all of the entries will have energy too big
  double dE = ENDL_EPSILON( max_E_out );
  if( begin( )->E_out > max_E_out + dE )
  {
    Warning( "Legendre_list::chop_E_out",
	     "all outgoing energies are too big" );
  }

  Legendre_list::iterator last_ptr = end( );
  --last_ptr;
  if( last_ptr->E_out > max_E_out + dE )
  {
    // loop through the incident-energy lists
    for( Legendre_list::iterator list_ptr = begin( );
         list_ptr != end( ); ++list_ptr )
    {
      if( list_ptr->E_out > max_E_out + dE )
      {
        // the previous list may be at max_E_out
        Legendre_list::iterator prev_ptr = list_ptr;
        --prev_ptr;
        if( prev_ptr->E_out > max_E_out - dE )
        {
	  // are there any others with big enough E_out?
          Legendre_list::iterator prev_prev_ptr = prev_ptr;
          --prev_prev_ptr;
          for( ; prev_prev_ptr->E_out > max_E_out - dE; --prev_prev_ptr )
	  {
            list_ptr = prev_ptr;
            prev_ptr = prev_prev_ptr;
	  }
	  prev_ptr->E_out = max_E_out;
          erase( list_ptr, end( ) );
	}
        else
	{
          // interpolate and make a new link
          Legendre_item new_link;  // for creating new links
          insert( list_ptr, new_link );
          Legendre_list::iterator new_ptr = list_ptr;
          --new_ptr;
          new_ptr->E_out = max_E_out;
          new_ptr->coefs.get_space( order + 1 );
          new_ptr->interpolate( *prev_ptr, *list_ptr );
          erase( list_ptr, end( ) );
	}
        break;
      }
    }
  }
}
// ----------- Legendre_list::print ---------
void Legendre_list::print( )
// print for debugging
{
  cout.setf(ios::scientific,ios::floatfield);
  for( Legendre_list::iterator list_ptr = begin( );
    list_ptr != end( ); ++list_ptr )
  {
    cout << ENDL.data( list_ptr->E_out ) << ":  ";
    list_ptr->coefs.print( );
  }
}

// *********** class mf6_Legendre_raw ***********
// ----------- mf6_Legendre_raw::read_data ---------
void mf6_Legendre_raw::read_data( int NR, int num_E_in,
  int LEP, mf6_file& inFile )
// Reads the Legendre data for all incident energies
{
  bool done = false;

  // read the interpolation regions
  inFile.get_regions( NR, NBT, INT );

  // convert the data into a linked list
  for( int count = 0; count < num_E_in; ++count )
  {
    Legendre_list one_E_in;
    if( !done )
    {
      insert( end( ), one_E_in );
      mf6_Legendre_raw::iterator new_ptr = end( );
      --new_ptr;  // now points to the new link
      new_ptr->read_data( LEP, inFile, &done );
    }
    else
    {
      one_E_in.read_data( LEP, inFile, &done );
    }
  }
}
// ----------- mf6_Legendre_raw::get_max_order ---------
int mf6_Legendre_raw::get_max_order( )
// Gets the maximum Legendre order
{
  int max_order = -1;
  // loop through the incident-energy lists
  for( mf6_Legendre_raw::iterator list_ptr = begin( );
       list_ptr != end( ); ++list_ptr )
  {
    if( list_ptr->order > max_order )
    {
      max_order = list_ptr->order;
    }
  }
  return max_order;
}
// ----------- mf6_Legendre_raw::expand_interp ---------
void mf6_Legendre_raw::expand_interp( )
// Expand the interpolation for the ENDF lists with respect to E_in
{
  // Set the interpolation with respect to E_in
  int count = 0;
  int INT_index = 0;
  if( INT[ INT_index ] == 12 )
  {
    Info("mf6_Legendre_raw::expand_interp", "type 12 interpolation used");
  }
  else if( ( INT[ INT_index ] != 2 ) && ( INT[ INT_index ] != 22 ) )
  {
    Unimplemented("mf6_Legendre_raw::expand_interp",
        pastenum("Write code to expand interpolation ",INT[ INT_index ]));
  }

  // loop through the list
  for( mf6_Legendre_raw::iterator this_link = begin( );
      this_link != end( ); ++this_link, ++count )
  {
    // do we change interpolation types?
    if ( count >= NBT[ INT_index ] )
    {
      ++INT_index;
      if ( INT_index >= INT.size( ) )
      {
        SevereError("mf6_Legendre_raw::expand_interp","INT_index overflow");
      }
      if( ( INT[ INT_index ] != 2 ) && ( INT[ INT_index ] != 22 ) )
      {
        Unimplemented("mf6_Legendre_raw::expand_interp",
            pastenum("write code to expand interpolation ",INT[ INT_index ]));
      }
    }
    this_link->interp_type = INT[ INT_index ];
  }

  // Write code here to expand the interpolation with respect to E_in

}


// ----------- mf6_Legendre_raw::widen_jumps ---------
void mf6_Legendre_raw::widen_jumps( )
{
  // loop through the incident-energy lists
  for( mf6_Legendre_raw::iterator list_ptr = begin( );
       list_ptr != end( ); ++list_ptr )
  {
    list_ptr->widen_jumps( );
  }
}


// ----------- mf6_Legendre_raw::renorm ---------
void mf6_Legendre_raw::renorm( )
// Normalizes all of the lists
{
  // loop through the incident-energy lists
  for( mf6_Legendre_raw::iterator list_ptr = begin( );
       list_ptr != end( ); ++list_ptr )
  {
    list_ptr->renorm( );
  }
}
// ----------- mf6_Legendre_raw::chop_E_in ---------
void mf6_Legendre_raw::chop_E_in( double max_E_in )
// deletes entries with E_in > max_E_in
{
  double dE = ENDL_EPSILON( max_E_in );
  // All of the data may have E_in too big
  if( begin( )->E_in > max_E_in + dE )
  {
    ENDL.write_file = false;
    return;
  }

  mf6_Legendre_raw::iterator last_ptr = end( );
  --last_ptr;
  if( last_ptr->E_in > max_E_in + dE )
  {
    mf6_Legendre_raw::iterator prev_ptr = last_ptr;
    --prev_ptr;
    interpolate( max_E_in, prev_ptr, last_ptr );
    erase( last_ptr );
  }
}
// ----------- mf6_Legendre_raw::interpolate ---------
void mf6_Legendre_raw::interpolate( double E_in,
  mf6_Legendre_raw::iterator prev_ptr,
  mf6_Legendre_raw::iterator next_ptr )
// Insert a new interpolated link
{
  Legendre_list new_list;
  insert( next_ptr, new_list );
  mf6_Legendre_raw::iterator new_ptr = next_ptr;
  --new_ptr; // now points to the new link

  // copy the prev_ptr data
  new_ptr->copy( *prev_ptr );
  new_ptr->E_in = E_in;
  // map to unit base
  new_ptr->unit_base( );
  next_ptr->unit_base( );

  // collect all of the outgoing energies
  list< double > E_outgoing;
  Legendre_list::iterator list_ptr;
  for( list_ptr = new_ptr->begin( ); 
      list_ptr != new_ptr->end( ); ++list_ptr )
  {
    E_outgoing.push_back( list_ptr->E_out );
  }
  for( list_ptr = next_ptr->begin( ); 
      list_ptr != next_ptr->end( ); ++list_ptr )
  {
    E_outgoing.push_back( list_ptr->E_out );
  }
  // sort the outgoing energies and remove duplicates
  E_outgoing.sort( );
  E_outgoing.unique( );
  // enforce common outgoing energies
  new_ptr->fill_with( E_outgoing );
  next_ptr->fill_with( E_outgoing );

  // take the average
  double alpha = ( E_in - prev_ptr->E_in ) /
    ( next_ptr->E_in - prev_ptr->E_in );
  *new_ptr *= 1.0 - alpha;
  *next_ptr *= alpha;
  *new_ptr += *next_ptr;
  new_ptr->least_E = ( 1.0 - alpha )*new_ptr->least_E +
    alpha*next_ptr->least_E;
  new_ptr->most_E = ( 1.0 - alpha )*new_ptr->most_E +
    alpha*next_ptr->most_E;
  // undo unit base
  new_ptr->rescale( );
}
// ----------- mf6_Legendre_raw::chop_E_out ---------
void mf6_Legendre_raw::chop_E_out( double max_E_out )
// deletes entries with E_out > max_E_out
{
  for( mf6_Legendre_raw::iterator list_ptr = begin( );
       list_ptr != end( ); ++list_ptr )
  {
    list_ptr->chop_E_out( max_E_out );
  }
}
// ----------- mf6_Legendre_raw::print ---------
void mf6_Legendre_raw::print( )
// print for debugging
{
  for( mf6_Legendre_raw::iterator list_ptr = begin( );
    list_ptr != end( ); ++list_ptr )
  {
    cout << ENDL.data( list_ptr->E_in ) << ":\n";
    list_ptr->print( );
  }
}
// *********** class mf6_one_order ***********
// ----------- mf6_one_order::null_coefs ---------
bool mf6_one_order::null_coefs( )
// returns true if all entries are zero
{
  for( mf6_one_order::iterator this_list = begin( );
       this_list != end( ); ++this_list)
  {
    double Norm = this_list->get_norm( );
    if( Norm != 0.0 )
    {
      return false;
    }
  }
  // they all were zero
  return true;
}

// *********** class mf6_Legendre ***********
// ----------- mf6_Legendre::master ---------
void mf6_Legendre::master( mf6_file& inFile, int LEP,
  int NR, int NE, bool res_copy )
// Handles all the data
{
  // Read all of the data
  ENDF_data.read_data( NR, NE, LEP, inFile );
  // delete the high incident energies
  ENDF_data.chop_E_in( ENDL.Max_E_in );
  // delete the high outgoing energies
  ENDF_data.chop_E_out( ENDL.Max_E_out );

  order = -1;  // default
  if( !ENDL.write_file )
  {
    return;  // this is data for the residual nucleus
  }
  order = ENDF_data.get_max_order( );

  // expand the interpolation with respect to E_in
  ENDF_data.expand_interp( );
  ENDF_data.widen_jumps( );  
  ENDF_data.renorm( );

  // convert to ENDL format
  reformat( );

  // remove redunant zeros
  reduce_order( );

  // print the ENDL file
  write_endl( );

  // Do we need to write a copy for the residual as particle?
  if( res_copy )
  {
    ENDL.outgoing_particle += 10;
    write_endl( );
  }
}
// ----------- mf6_Legendre::reformat ---------
void mf6_Legendre::reformat( )
// copy ENDF_data into ENDL_data
{
  mf6_Legendre_raw::iterator ENDF_ptr;

  // pointers to the ENDL lists
  mf6_one_order::iterator ENDL_E_in[ order + 1 ];

  // set up space for ENDL_data
  ENDL_data.resize( order + 1 );

  // go through the ENDF data
  for( mf6_Legendre_raw::iterator ENDF_ptr = ENDF_data.begin( );
       ENDF_ptr != ENDF_data.end( ); ++ENDF_ptr )
  {
    // make a new 1-d list for each Legendre order
    for( int count = 0; count <= order; ++count )
    {
      one_d_table new_E_out;
      ENDL_data[ count ].insert( ENDL_data[ count ].end( ),
        new_E_out );
      ENDL_E_in[ count ] = ENDL_data[ count ].end( );
      --ENDL_E_in[ count ];
      ENDL_E_in[ count ]->E_in( ) = ENDF_ptr->E_in;
    }

    // now loop through the ENDF E_out links
    for( Legendre_list::iterator ENDF_Eout_ptr = ENDF_ptr->begin( );
         ENDF_Eout_ptr != ENDF_ptr->end( ); ++ENDF_Eout_ptr )
    {
      dd_link XYdata;
      XYdata.x = ENDF_Eout_ptr->E_out;
      // loop through the Legendre orders
      for( int ord_count = 0; ord_count <= ENDF_ptr->order;
        ++ord_count )
      {
        XYdata.y = ENDF_Eout_ptr->coefs[ ord_count ];
        ENDL_E_in[ ord_count ]->insert( ENDL_E_in[ ord_count ]->end( ),
          XYdata );
      }
      // we may need to fill out the list with zeros
      XYdata.y = 0.0;
      for( int ord_count = ENDF_ptr->order + 1; ord_count <= order;
        ++ord_count )
      {
        ENDL_E_in[ ord_count ]->insert( ENDL_E_in[ ord_count ]->end( ),
          XYdata );
      }
    }
  }
}
// ----------- mf6_Legendre::reduce_order ---------
void mf6_Legendre::reduce_order( )
// Reduces the order if high-order Legendre coefficients are all zero
{
  int L_count;

  // print the data
  for( L_count = order; L_count > 0; --L_count )
  {
    if( !ENDL_data[ L_count ].null_coefs( ) )
    {
      // we found something nonzero at this order
      break;
    }
  }
  // reset the order
  order = L_count;
}
// ----------- mf6_Legendre::write_endl ---------
void mf6_Legendre::write_endl( )
// print the data in ENDL format
{
  ENDL.set_I_number( 4 );

  fstream endl_file;
  string file_name = ENDL.file_name;

  if ( ENDL.new_file() )
  {
    endl_file.open(file_name.c_str(),ios::out);
    Info("mf6_Legendre::write_endl","Opening ENDL file "+file_name);
  }
  else
  {
    endl_file.open(file_name.c_str(),ios::out|ios::app);
    Info("mf6_Legendre::write_endl","Appending ENDL file "+file_name);
  }

  //Shove the header info in
  endl_file << ENDL.header_line_1 << endl;
  endl_file << ENDL.header_line_2 << endl;
  endl_file.setf(ios::scientific,ios::floatfield);

  // print the data
  for( int L_count = 0; L_count <= order; ++L_count )
  {
    ENDL_data[ L_count ].out_data( 4, endl_file, 1.0*L_count );
  }

  //Put on end of file/section line into output file
  endl_file << ENDL.eof_line << endl;

  //Finally, we close the output file
  Info("mf6_Legendre::write_endl","Closing ENDL file "+file_name);
  endl_file.close();
}
