/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1838 $
 * $Date: 2006-04-11 16:20:40 -0700 (Tue, 11 Apr 2006) $
 * $Author: hedstrom $
 * $Id: mf6classes.hpp 1838 2006-04-11 23:20:40Z hedstrom $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// header for the classes used in translating the MF=6 files

#ifndef MF6CLASSES
#define MF6CLASSES

#include <cmath>

#include "list_3d.hpp"
#include "mappings.hpp"
#include "nuclei.hpp"
#include "ENDF_file.hpp"
#include "math_util.hpp"  // for the Param class
#include "multiplicity.hpp"

// ----------- class mf6_one_d -----------------
//! This class is for the 1-dimensional lists of (energy, probability density) used for double-differential data.
class mf6_one_d : public distrib_base
{
public:
  // constructor
  mf6_one_d();

  // destructor
  ~mf6_one_d() {}

  // to map between center-of-mass and laboratory coordinates
  mappings *map_;

  inline void copy_map( mappings *Map )
  {
    map_ = Map;
  }

  // for a given mu make Num+1 new links for secondary
  // energies min_E <= E <= max_E
  //! Generates new links in outgoing particle energy for a given angle cosine
  void one_E_row( double min_E, double max_E );

};

// ----------- class one_d_phase -----------------
//! One (E', probability) list for the N-body phase-space distribution
class one_d_phase : public mf6_one_d
{
public:
  double Ei_max;
  double exponent;

  //! The phase-space probability distribution
  double f( double E );
};

// ----------- class two_d_phase -----------------
//! Class for one incident energy for the N-body phase-space distribution
class two_d_phase : public three_d_link< one_d_phase >
{
public:
  double Ei_max_;
  double exponent_;

  // ********* implement virtual functions *******************
  //! Append a new one_d_phase and set its parameters
  two_d_phase::iterator new_one_d( two_d_phase::iterator where );
  // ********** end of virtual functions *********************
};

// ----------- class phase_space -----------------
//! Class for the N-body phase-space distribution
class phase_space : public three_d_list< two_d_phase >
{
private:
  double mass_ratio;  // (total_mass - particle_mass)/total_mass
  double Q;           // energy of the reaction
  double exponent__;  // 1.5*n - 4

  //! Handles one incident neutron energy
  void one_E_in( double e_in );

  //! Finds the transition from forward to omnidirectional emission
  double find_E_trans( );

public:
  //! Constructor identifies the participating particles in the reaction
  phase_space( int projectile, int target, int ZAP );
  ~phase_space() {}

  //! Reads the data and expands it into a list
  void expand_data( mf6_file& inFile, multiplicity& multiple );

  //! Expand data for a single incident neutron energy
  void expand_E_in( phase_space::iterator e_in_link );

};

// ----------- class two_d_ENDL_table -----------------
//! Class for one incident energy for the LAW=7 double differential table
class two_d_ENDL_table : public three_d_link< one_d_table >
{
public:

  // ********* implement virtual functions *******************
  //! Append a new one_d_ENDL_tablep
  two_d_ENDL_table::iterator
    new_one_d( two_d_ENDL_table::iterator where );
  // ********** end of virtual functions *********************
};

// ----------- class ENDL_table -----------------
//! Class for the LAW=7 double differential table
class ENDL_table : public three_d_list< two_d_ENDL_table >
{
private:
  //! There is no ENDL_table probability function
  inline double get_f(Pair& E_mu)
  {
    return 0.0;
  }

public:
  //! Default constructor
  ENDL_table( );
  //! Default destructor
  ~ENDL_table( ) {}

  //! Reads all the data
  void read_data( mf6_file& inFile, int ZAP,
    multiplicity& multiple );

  //! Reads the data for one incident energy
  void one_E_in( mf6_file& inFile );

};

// ----------- class Legendre_item -----------------
//! Class holding Legendre coefficients for the mf6, LAW=1, LANG=1 Legendre expansions
class Legendre_item
{
public:
  //! Default constructor
  Legendre_item( );

  //! Default destructor
  ~Legendre_item( ) {}

  //! Energy of the outgoing particle
  double E_out;

  //! The Legendre coefficients
  Param coefs;

  //! Reads a set of data
  //! Sets *E_out_done = true if the outgoing energy is too big
  void read_coefs( int i_start, string& linebuff, int order,
    mf6_file& inFile, bool *E_out_done );

  //! Copies the Legendre coefficients
  void copy_coef( Legendre_item& copy_from );

  //! Interpolates the coefficients between next_link and the previous one
  void interpolate( Legendre_item& prev_item,
		   Legendre_item& next_item );
};

// ----------- class Legendre_list -----------------
//! Class holding an array of Legendre coefficients for the mf6, LAW=1, LANG=1 Legendre expansions
class Legendre_list : public list< Legendre_item >
{
private:
  // insert zero links from E0 to the head
  void pad_head( double E0 );

  // append zero links up to E0
  void pad_tail( double E0 );

public:
  //! Energy of the incident particle
  double E_in;

  //! The Legendre order
  int order;

 // type of interpolation between this list and the previous one
  int interp_type;

  // for the unit_base transformation
  double least_E;
  double most_E;

  //! Default destructor
  inline ~Legendre_list( ) {}

  //! Reads the data
  //! Sets *done = true if the incident energy is too big
  void read_data( int LEP, mf6_file& inFile, bool *done );

  //! We have only gamma lines---no continuum
  void only_lines( one_d_table& gamma_lines );

  //! We have gamma lines and continuum
  void add_lines( one_d_table& gamma_lines );

  //! Gets the norm with respect to E_out.
  double get_norm( );

  //! Removes jump discontinuities
  void widen_jumps( );

  //! Utility code used by widen_jumps
  void widen_cluster( Legendre_list::iterator cluster_start, 
    Legendre_list::iterator cluster_end, int cluster_size, double jump_width );

  //! Sets the norm with respect to E_out to 1.
  void renorm( );

  //! Expands the interpolation with respect to energy of  the outgoing particle
  void expand_interp( int LEP );

  // make sure that there are links for each energy in E_out_list
  void fill_with( list< double >& E_out_list );

  //! deletes entries with E_out > max_E_out
  void chop_E_out( double max_E_out );

  //! Transforms to unit base 0 <= E' <= 1
  void unit_base( );

  //! Undoes the unit-base transformation
  void rescale( );

  //! operator*=, used in interpolation
  Legendre_list& operator*=( double alpha );

  //! operator+=, used in interpolation
  Legendre_list& operator+=( Legendre_list& to_add );

  //! Copies the Legendre coefficients
  void copy( Legendre_list& to_add );

  // print for debugging
  void print( );
};

// ----------- class mf6_Legendre_raw -----------------
//! Class holding all of the mf6 Legendre data in ENDF format
class mf6_Legendre_raw : public list<Legendre_list>
{
public:
  // The interpolation information
  vector<int> NBT;
  vector<int> INT;

  //! Default destructor
  inline ~mf6_Legendre_raw( ) {}

  //! Reads the data
  void read_data( int NR, int num_E_in, int LEP, mf6_file& inFile );

  //! Gets the maximum Legendre order
  int get_max_order( );

  //! Normalizes all of the lists
  void renorm( );

  //! Removes jump discontinuities
  void widen_jumps( );

  //! Expands the interpolation for the ENDF lists
  void expand_interp( );

  //! deletes entries with E_in > max_E_in
  void chop_E_in( double max_E_in );

  //! Inserts a new interpolated link
  void interpolate( double E_in,
    mf6_Legendre_raw::iterator prev_ptr,
    mf6_Legendre_raw::iterator next_ptr );

  //! deletes entries with E_out > max_E_out
  void chop_E_out( double max_E_out );

  //! For debugging
  void print( );
};

// ----------- class mf6_one_order -----------------
//! Class for one the Legendre order for mf6, LAW=1, LANG=1
class mf6_one_order : public two_d_table
{
public:
  //! Checks whether all coefficients are zero at this order
  bool null_coefs( );
};

// ----------- class mf6_Legendre -----------------
//! Class for the mf6, LAW=1, LANG=1 Legendre expansions
class mf6_Legendre
{
private:
  // the multiplicity (used mostly for gammas)
  multiplicity multiple;

  // a class for all of the ENDF/B-VI data
  mf6_Legendre_raw ENDF_data;

  // an stl::vector to hold the data in ENDL order
  vector<mf6_one_order> ENDL_data;

  int order; // Legendre order

public:
  //! Default constructor
  mf6_Legendre( ) {}

  //! Default destructor
  ~mf6_Legendre( ) {}

  //! Handles all the data
  void master( mf6_file& inFile, int LEP, int NR, int NE, bool res_copy );

  //! Converts to ENDL format
  void reformat( );

  //! Reduces the order if high-order Legendre coefficients are all zero
  void reduce_order( );

  //! Prints the ENDL file
  void write_endl( );
};


#endif
