/*
 * ******** fete: From ENDF To ENDL *********
 * 
 * Copyright (c) 2006, The Regents of the University of California. 
 * All rights reserved.
 * 
 * Produced at the Lawrence Livermore National Laboratory. 
 * Written by David A. Brown, Gerry Hedstrom, Tony Hill
 * 
 * This file is part of fete v1.0  (UCRL-CODE-218718)
 * 
 * Please read the COPYING file for "Our Notice and GNU General 
 * Public License" in the root of this software distribution.  
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License (as published by 
 * the Free Software Foundation) version 2, dated June 1991. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms 
 * and conditions of the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * $Revision: 1866 $
 * $Date: 2006-05-15 09:12:12 -0700 (Mon, 15 May 2006) $
 * $Author: dbrown $
 * $Id: multiplicity.cpp 1866 2006-05-15 16:12:12Z dbrown $
 * 
 * ******** fete: From ENDF To ENDL *********
 */

// implementation of the one_d_distrib class

#include <cmath>   // math library

#include "multiplicity.hpp"
#include "endl_formats.hpp"
#include "endl_precision.hpp"

extern ENDLClass ENDL;

// *********** multiplicity class ***********************


// ----------- multiplicity::read_data ----------
// read in and expand 2-column data
void multiplicity::read_data( int NP, ENDF_file& inFile )
{
  one_d_table::read_data( NP, inFile, ENDL.Max_E_in );
  if( ENDL.write_file )
  {
    expand_interp( ENDL.Max_E_in );
  }
}

// ----------- multiplicity::get_mult ----------
//! Returns the constant multiplicity as may be used in an (n, Xn) reaction
int multiplicity::get_mult( )
{
  multiplicity::iterator left_ptr = begin( );
  multiplicity::iterator right_ptr = end( );
  --right_ptr;
  double dm = ENDL_EPSILON( 1.0 );

  if( abs( left_ptr->y - right_ptr->y ) > dm )
  {
    Warning( "multiplicity::get_mult", pastenum( 
      pastenum( pastenum("Multiplicity not constant: ",left_ptr->y)+" vs. ", right_ptr->y )+" at E = ",left_ptr->x) );
  }
  return static_cast<int>( dm + 0.5*( left_ptr->y + right_ptr->y ) );
}

// ----------- multiplicity::check_const_mult ----------
//! Checks the multiplicity to make sure it is constant, as it should be
bool multiplicity::check_const_mult( )
{
  return check_const_mult( multiplicity::begin()->y );
}

// ----------- multiplicity::check_const_mult ----------
//! Checks the multiplicity to make sure it is constant, as it should be
bool multiplicity::check_const_mult( double first_mult )
{
  double dm = ENDL_EPSILON( 1.0 );

  for( multiplicity::iterator itr = multiplicity::begin(); itr != multiplicity::end(); ++itr )
  {
    if( abs( itr->y - first_mult ) > dm ) return false;
  }
  return true;
}

// ----------- multiplicity::set_min_mult ----------
//! Sets a floor value on the multiplicity for those times where 
//! it is not constant, but a minimum value is requires
void multiplicity::set_min_mult( double min_mult ){
  for( multiplicity::iterator itr = multiplicity::begin(); itr != multiplicity::end(); ++itr )
  {
    if( itr->y < min_mult )
    {
        Warning("multiplicity::set_min_mult",
            pastenum("Multiplicity at E = ",itr->x)+
            pastenum("MeV is ",itr->y)+
            pastenum(".  Resetting to minimum allowed value of ",min_mult));
        itr->y = min_mult;
    }
  }
}

// ----------- multiplicity::set_max_mult ----------
//! Sets a ceiling value on the multiplicity for those times where 
//! it is not constant, but a minimum value is requires
void multiplicity::set_max_mult( double max_mult ){
  for( multiplicity::iterator itr = multiplicity::begin(); itr != multiplicity::end(); ++itr )
  {
    if( itr->y > max_mult )
    {
        Warning("multiplicity::set_min_mult",
            pastenum("Multiplicity at E = ",itr->x)+
            pastenum("MeV is ",itr->y)+
            pastenum(".  Resetting to maximum allowed value of ",max_mult));
        itr->y = max_mult;
    }
  }
}

// ----------- multiplicity::get_mult ----------
//! Sets the multiplicity to a constant, 
//! if we know it from another method (say the reaction MT)
void multiplicity::set_mult( int mult )
{
  for( multiplicity::iterator itr = multiplicity::begin(); itr != multiplicity::end(); ++itr )
  {
    itr->y=static_cast<double>(mult);
  }
}

// ----------- multiplicity::scale_by_xs ----------
//! Scale photon multiplicity by the cross section to get photon production
void multiplicity::scale_by_xs( )
{
  mf3_file xs_File;   //Instantiate the input file
  xs_File.open(3, ENDL.T);
  one_d_table xs;  // Instantiate the cross section object

  // get the target ZA and mass
  int ZA;
  double AWR;
  xs_File.read_line1(&ZA, &AWR);

  //read second line
  double QM, QI;
  int LR, NR, NP;
  xs_File.read_line2(
	  &QM,    // mass difference Q value
	  &QI,    // reaction Q value for the lowest state
	  &LR,    // complex or "breakup" flag
	  &NR,    // number of INT regions 
	  &NP     // number of data points
  );
  xs_File.get_regions( NR, xs.NBT, xs.INT );
  xs.read_data( NP, xs_File, ENDL.Max_E_in );
  xs_File.close();
  
  *this *= xs;
  
  thinit();
}
// ----------- multiplicity::divide_by_xs ----------
//! Divide photon production by the cross section to get photon multiplicity
void multiplicity::divide_by_xs()
{
  mf3_file xs_File;   //Instantiate the input file
  xs_File.open(3, ENDL.T);
  one_d_table xs;  // Instantiate the cross section object

  // get the target ZA and mass
  int ZA;
  double AWR;
  xs_File.read_line1(&ZA, &AWR);

  //read second line
  double QM, QI;
  int LR, NR, NP;
  xs_File.read_line2(
	  &QM,   // mass difference Q value
	  &QI,   // reaction Q value for the lowest state
	  &LR,   // complex or "breakup" flag
	  &NR,   // number of INT regions 
	  &NP    // number of data points
  );    
  xs_File.get_regions(NR, xs.NBT, xs.INT);
  xs.read_data( NP, xs_File, ENDL.Max_E_in );
  xs_File.close();

  *this /= xs;

}

// ----------- multiplicity::check_count ----------
void multiplicity::check_count( )
{
  multiplicity::iterator this_link;

  for( this_link = begin( ); this_link != end( );
       ++this_link )
  {
    if( this_link->y > 1.0 )
    {
      Warning( "multiplicity::check_count",
	       " the multiplicity of a single gamma exceeds 1" );
      break;
    }
  }
}

// ----------- multiplicity::list_interp ----------
void multiplicity::list_interp( double e_in, multiplicity& left_list,
  multiplicity& right_list )
// Make a null list to be used for discrete gamma multiplicities.
{
  if( !empty() )
  {
    SevereError("multiplicity::list_interp",
        "You are trying to fill a multiplicity which has already been made");
  }

  // get the first incident energy
  multiplicity::iterator left_ptr = left_list.begin( );
  multiplicity::iterator right_ptr = right_list.begin( );

  double left_e_in = left_ptr->x;
  double right_e_in = right_ptr->x;

  dd_link XYdata;
  XYdata.x = ( left_e_in < right_e_in ) ? left_e_in : right_e_in;
  XYdata.y = 0.0;
  insert( end( ), XYdata );

  // get the last incident energy
  left_ptr = left_list.end( );
  --left_ptr;
  right_ptr = right_list.end( );
  --right_ptr;

  left_e_in = left_ptr->x;
  right_e_in = right_ptr->x;

  XYdata.x = ( left_e_in < right_e_in ) ? left_e_in : right_e_in;
  insert( end( ), XYdata );
}
